From Coq Require Import Arith List Relations.
Import ListNotations.
Require Import
        Coq.Logic.FunctionalExtensionality.

From C4.Util Require Import
  FunOverride
  LibTactics.

From C4.Core Require Import
  Program
  ProgramFacts
  ProgramOps
  ProgramOpsFacts
  Object
  Simulates
  SimulatesFacts
  Threads
  ThreadsFacts
  Linearizability.

Set Implicit Arguments.
Set Maximal Implicit Insertion.

Definition pos := { n : nat | 0 < n }.

Inductive ArrayMethod M: Type -> Type :=
| arrayCall {R}: nat -> M R -> ArrayMethod M R
| getSize: ArrayMethod M nat.
Arguments arrayCall {M} {R} _ _.
Arguments getSize {M}.

(*
  (*
     To support resizing, we add a back up array and support call on it.
     moveBack copies the current array to the back array.
     removeBack removes the back array.
     resize makes a new current array of the given size and initial value.
   *)
  | moveBack: ArrayMethod M unit
  | resize: nat -> ArrayMethod M unit
  | arrayBackCall {R}: nat -> M R -> ArrayMethod M R
  | removeBack: ArrayMethod M unit.
*)
(* Arguments resize {M} _. *)


Section AbsArray.

  Context {I: Type -> Type}.
  Context (spec: Spec I).
  (* Variable initObjState: obj.(State). *)

  Record ArrayState := {
    arrayMapState: nat -> spec.(State);
    arraySizeState: { n: nat | 0 < n }
  }.
  Definition arrayState := Build_ArrayState.
  Canonical Structure Default_ArrayState m s := Build_ArrayState m s.

  Inductive ArraySem:
    forall {R}, ArrayMethod I R -> ArrayState -> R -> ArrayState -> Prop :=

  | succArrayCall i R (m: I R) st r o':
      i < proj1_sig (arraySizeState st) ->
      spec.(EvalInstruction) m (arrayMapState st i) r o' ->
      ArraySem
        (arrayCall i m)
        st
        r
        (arrayState (noverride (arrayMapState st) i o') (arraySizeState st))

    (* Out-of-bounds call is undefined behavior: any result may be returned *)
  | failArrayCall i R (m: I R) st v:
      proj1_sig (arraySizeState st) <= i ->
      ArraySem
        (arrayCall i m) st
        v st

  | getSizeCall st:
      ArraySem
        getSize st
        (proj1_sig (arraySizeState st)) st.

  Definition absArray (len : { n : nat | 0 < n }) : Spec (ArrayMethod I) :=
    Build_Spec _
      ArrayState
      (@ArraySem)
      (arrayState (fun _ => spec.(initialState)) len).

End AbsArray.
Arguments arrayMapState {I} {spec} _ _.
Arguments arraySizeState {I} {spec} _.

Section ArrayImpl.

  Context {M: Type -> Type}.
  Context (impl: Object M).
  Context (s : impl.(objBase).(State)).
  Context (len : { n | n > 0 }).

  Definition arrayLift {M R} (p: Program M R) i: Program (ArrayMethod M) R :=
    mapProgram
      (fun _ bm => arrayCall i bm)
      p.

  (* Todo: Refactor to inversion lemma. *)
  Lemma arrayLiftEqRet:
    forall I R (p : Program I R) i r,
      arrayLift p i = Return r ->
      p = Return r.

  Proof.
    introv H1.
    destruct p; progFrobSimplIn H1; inverts H1.
    reflexivity.
  Qed.

  Lemma arrayLiftEqNoop:
    forall I R (p: Program I R) p' i,
      arrayLift p i = Noop p' ->
      exists p'', p = Noop p'' /\ p' = arrayLift p'' i.

  Proof.
    introv H1.
    destruct p; progFrobSimplIn H1; inverts H1.
    eauto.
  Qed.

  Lemma arrayLiftEqBind:
    forall I R S (p: Program I S) (e : _ R) (f : R -> Program _ S) i,
      arrayLift p i = Bind e f ->
      exists e' f', p = Bind e' f' /\ e = arrayCall i  e' /\ forall r, f r = arrayLift (f' r) i.
  Proof.
    introv H1.
    destruct p; progFrobSimplIn H1; inverts H1.
    eauto.
  Qed.

  Lemma EvalProgramArrayLift:
    forall R M (o: Spec M) i (s1a: nat -> o.(State)) s1s s2ai p r,
      i < proj1_sig s1s ->
      EvalProgram
        o R
        p
        (s1a i) r s2ai ->
      EvalProgram
        (absArray o len) R
        (arrayLift p i)
        {| arrayMapState := s1a; arraySizeState := s1s |} r
        {| arrayMapState := noverride s1a i s2ai; arraySizeState := s1s |}.

  Proof.

    introv H1 H2.

    replace s1a with (noverride s1a i (s1a i)) at 1. Focus 2.
    apply functional_extensionality. intros. destruct_override; auto.

    remember (s1a i) as s1ai eqn:E1. clear E1.
    gen R p s1ai r s2ai.

    eapply EvalProgramInd.
    - intros.
      progFrobSimpl.
      eapply ProgramStepsRefl.
    - introv H3 H4. progFrobSimpl.
      eapply ProgramStepStepsImplSteps.
      econstructor. constructor. simpl; assumption. simpl. simpl_override. eassumption.
      simpl. applys_eq H4 3. unfold arrayState. f_equal.
      apply functional_extensionality. intros. destruct_override; auto.
    - introv H3. progFrobSimpl.
      eapply ProgramStepStepsImplSteps.
      econstructor. eassumption.
  Qed.


(*    * A parametric object of array. *)
(*    * Given an object, returns the object of an array of that object. *)
(*    *)

  Definition arrayImpl: Object (ArrayMethod M) := {|
    ObjInstruction :=
      ArrayMethod (ObjInstruction impl);

    objBase :=
      absArray (objBase impl) len;

    objProgram :=
      (* implProg *)
      fun R (m: ArrayMethod M R) =>
        match m with
        | arrayCall i im => arrayLift (objProgram impl _ im) i
        | getSize =>
          s <- getSize;
          Return s
        end
  |}.

End ArrayImpl.

Record ArrayStateEq {M1 M2} (o1 : Spec M1) (o2 : Spec M2) (RR : State o1 -> State o2 -> Prop)
    (s1 : ArrayState o1) (s2 : ArrayState o2)
  : Prop :=
  { array_size_eq : arraySizeState s1 = arraySizeState s2
  ; array_map_eq : forall i, i < proj1_sig (arraySizeState s1) ->
      RR (arrayMapState s1 i) (arrayMapState s2 i)
  }.

Lemma initArrayStateEq {M1 M2} (o1 : Spec M1) (o2 : Spec M2) (RR : State o1 -> State o2 -> Prop)
    (s1 : State o1) (s2 : State o2)
  : RR s1 s2 -> forall n, ArrayStateEq RR (arrayState o1 (fun _ => s1) n) (arrayState o2 (fun _ => s2) n).
Proof.
  intros SIM n. constructor; auto.
Qed.

Lemma noverride_eq {V} (f : nat -> V) (n : nat) (v : V) : noverride f n v n = v.
Proof.
  unfold noverride.
  apply override_new_val.
Qed.

Lemma noverride_neq {V} (f : nat -> V) ( n : nat) (v : V) (i : nat)
  : i <> n -> noverride f n v i = f i.
Proof.
  unfold noverride. intros. apply override_old_val. auto.
Qed.

Lemma evalArrayLift {M} (o1 : Spec M) len
  : forall s1 (n : nat) R (p : Program M R) (r : R) s1',
    n < proj1_sig (arraySizeState s1) ->
    EvalProgram (absArray o1 len) R (arrayLift p n) s1 r s1' ->
    arraySizeState s1 = arraySizeState s1' /\
    exists ss1',
      EvalProgram o1 R p (arrayMapState s1 n) r ss1' /\
      arrayMapState s1' n = ss1' /\
      (forall i, i <> n -> arrayMapState s1' i = arrayMapState s1 i).
Proof.
  intros * Hn H.
  remember (arrayLift p n) as q eqn:Eq.
  revert p Eq.
  induction H using EvalProgramInd; intros.
  - symmetry in Eq; apply arrayLiftEqRet in Eq; subst.
    split; [ reflexivity | ].
    exists (arrayMapState s n).
    split; [ | split ].
    + apply EvalProgramReturn.
    + reflexivity.
    + reflexivity.
  - symmetry in Eq; apply arrayLiftEqBind in Eq; destruct Eq as [e' [f' [Hp [Hm Hf]]]]; subst.
    inversion H; injpair_; subst; cbn in *. 2:{ contradict Hn; apply le_not_lt; auto. }
    specialize (IHEvalProgram Hn _ (Hf _)).
    destruct IHEvalProgram as [SZ [ss1' [EVAL [Hss1' Hrest]]]].
     revert Hss1'; subst; intros Hss1'.
    split; [ auto |].
    exists ss1'.
    split; [| split ].
    + eapply EvalProgramBind; try eassumption. cbn in EVAL.
      rewrite noverride_eq in EVAL.
      assumption.
    + auto.
    + intros i Hi. rewrite (Hrest i Hi).
      cbn.
      rewrite noverride_neq; auto.
  - symmetry in Eq; apply arrayLiftEqNoop in Eq. destruct Eq as [p2 [Hp2 Hp]]. subst.
    specialize (IHEvalProgram Hn _ eq_refl).
    destruct IHEvalProgram as [SZ [ss1' [EVAL ?]]].
    split; [ auto |].
    exists ss1'; split; auto.
    apply EvalProgramNoop; assumption.
Qed.

Lemma evalArrayLiftFail {M} (o1 : Spec M) len
  : forall s1 (n : nat) R (p : Program M R) (r : R) s1',
    ~ n < proj1_sig (arraySizeState s1) ->
    EvalProgram (absArray o1 len) R (arrayLift p n) s1 r s1' ->
    s1' = s1.
Proof.
  intros * Hn H.
  remember (arrayLift p n) as q eqn:Eq.
  revert p Eq.
  induction H using EvalProgramInd; intros.
  - auto.
  - symmetry in Eq; apply arrayLiftEqBind in Eq; destruct Eq as [e' [f' [Hp [Hm Hf]]]]; subst.
    inversion H; injpair_; subst; cbn in *. 1:{ contradict Hn; auto. }
    eauto.
  - symmetry in Eq; apply arrayLiftEqNoop in Eq. destruct Eq as [p2 [Hp2 Hp]]. subst.
    specialize (IHEvalProgram Hn _ eq_refl).
    auto.
Qed.

Lemma clos_refl_trans__norefl {A} (R : relation A) x z
  : x <> z ->
    clos_refl_trans_1n A R x z ->
    exists y, R x y /\ clos_refl_trans_1n A R y z.
Proof.
  intros ? w. destruct w; [ contradiction | eauto ].
Qed.

Lemma EvalProgramBindInv {M} (spec : Spec M) Q (e : M Q) R (f : Q -> Program M R)
   s1 s1' r
  : EvalProgram spec R (x <- e; f x) s1 r s1' ->
    exists (x : Q) s10, EvalInstruction spec e s1 x s10 /\
      EvalProgram spec R (f x) s10 r s1'.
Proof.
  intros H. apply clos_refl_trans__norefl in H; [ | discriminate].
  destruct H as [[] []].
  inversion H. injpair_. subst. eauto.
Qed.

Lemma EvalGetSize {M} (spec : Spec M) len (R : Type) (f : nat -> Program _ R) s1 r s1'
  : EvalProgram (absArray spec len) R (Bind getSize f) s1 r s1' ->
    EvalProgram (absArray spec len) R (f (proj1_sig (arraySizeState s1))) s1 r s1'.
Proof.
  intros H. apply EvalProgramBindInv in H. destruct H as [s [s1_ [EVALI EVAL]]].
  inversion EVALI. injpair. assumption.
Qed.

Lemma EvalReturn {M} (spec : Spec M) (R : Type) s r0 r s'
  : EvalProgram spec R (Return r0) s r s' ->
    r0 = r /\ s = s'.
Proof.
  intros H. do 2 red in H. remember (Return r, s') as Rr. destruct H.
  - inversion HeqRr; auto.
  - inversion H.
Qed.

Lemma EvalProgramSpin {M} (spec : Spec M) R s (r : R) s'
  : ~ EvalProgram spec R (cofix spin : Program M R := Noop spin) s r s'.
Proof.
  intros H. hnf in H.
  remember (Return r, s') as q'.
  remember ((cofix spin : Program M R := Noop spin), s) as q.
  induction H.
  - rewrite (frobProgramId (cofix spin : Program M R := Noop spin)) in Heqq.
    subst; discriminate.
  - rewrite (frobProgramId (cofix spin : Program M R := Noop spin)) in Heqq.
    destruct y; cbn in *.
    subst; inversion H.
    injpair_.
    cbn in *; subst.
    auto.
Qed.

Definition overrideArrayState M (spec : Spec M) (a : ArrayState spec) (i : nat) (s : State spec)
  : ArrayState spec :=
  {| arrayMapState := noverride (arrayMapState a) i s
  ;  arraySizeState := arraySizeState a
  |}.

Theorem implSeqRefObjImpliesArrayImplSeqRefAbsArray:
  forall {M}(impl: Object M) spec len initObjState initSpecState,
    Simulates_
      (coarsen impl) spec
      initObjState initSpecState ->
    Simulates_
      (coarsen (arrayImpl impl len)) (absArray spec len)
      (arrayState impl.(objBase) (fun _ => initObjState) len) (arrayState spec (fun _ => initSpecState) len).
Proof.
  intros M impl spec len sO sI SIM.
  apply ForSimulatesSimulates with
    (rel := @ArrayStateEq _ _ (objBase impl) spec (Simulates_ (coarsen impl) spec)).
  split; [ apply initArrayStateEq; auto | ].
  clear; intros.
  cbn in H0. destruct int.
  - destruct (lt_dec n (proj1_sig (arraySizeState s1))).
    + apply evalArrayLift in H0; auto.
      destruct H0 as [SZ [ss1' [EVAL [Hn Hi]]]].
      assert (SIM' : Simulates_ (coarsen impl) spec (arrayMapState s1 n) (arrayMapState s2 n)).
      { apply H; auto. }
      apply SIM' in EVAL.
      destruct EVAL as [ss2' EVAL_ SIM_].
      exists (overrideArrayState s2 n ss2').
      split.
      * constructor.
        { rewrite <- (array_size_eq H). auto. }
        { auto. }
      * constructor.
        { rewrite <- SZ. apply H. }
        { intros. cbn. destruct_override.
          { subst. auto. }
          { rewrite Hi; [ apply H; rewrite SZ; auto | auto]. }
        }
    + exists s2. apply evalArrayLiftFail in H0; auto.
      subst; split; auto.
      apply failArrayCall. apply Nat.nlt_ge. rewrite <- (array_size_eq H). auto.
  - exists s2.
    apply EvalGetSize in H0.
    apply EvalReturn in H0.
    destruct H0; subst.
    split.
    + rewrite (array_size_eq H); constructor.
    + assumption.
Qed.
