From Coq Require Import Arith
        Lia
        Setoid
        SetoidClass
        Morphisms
        FunctionalExtensionality.
From Coq Require String.

From C4.Util Require Import
     FunOverride
     MathUtil
     LibTactics.
From C4.Core Require Import
     Program
     ProgramFacts
     ProgramOps
     ProgramOpsFacts
     Object
     Pair
     SimulatesFacts
     Threads
     ThreadsFacts
     LinCoInd
     Linearizability
     LinearizabilityFacts.
From C4.App Require Import
     Lock
     Array
     Maps
     SHashMap
     Annotated
     AnnotatedFacts
     Dum
     SHashMap.
Require Import C4.App.Register. (* "Register" is a keyword *)

Import EMap.



Arguments arrayCall {M} {R} _ _.
Arguments getSize {M}.

Module CHashMapFun (E : HashableType).

  Definition Key := E.t.
  Definition eq_key_dec := E.eq_dec.
  Definition hash := E.hash.
  Definition bucketMapImpl := E.bucketMapImpl.

  Section CHashMapFun1.

  Context (Val : Type) (nLocks nBuckets : {n : nat | 0 < n}).

  Definition locks := serialObject (absArray (atomicLock) nLocks).
  Definition buckets := arrayImpl (bucketMapImpl Val) nBuckets.

  Definition base := pairObject locks buckets.

  Definition BaseState :=
    prod
      locks.(objBase).(State)
      buckets.(objBase).(State).

  Definition posMult (n1 n2: {n: nat | 0 < n}): {n: nat | 0 < n}.
    refine (exist _ ((proj1_sig n1)*(proj1_sig n2)) _).
      destruct n1. destruct n2. simpl. eapply Nat.mul_pos_pos; eassumption.
    Defined.

  Definition BaseMethod :=
    PairMethod
      (ArrayMethod LockMethod)
      (ArrayMethod (EMapMethod Key Val)).
  Definition callLocks {R} (m: ArrayMethod LockMethod R): BaseMethod R :=
    inl m.
  Definition callLocksArray (i: nat) {R} (m: LockMethod R): BaseMethod R :=
    inl (arrayCall i m).
  Definition callBuckets {R} (m: ArrayMethod (EMapMethod Key Val) R): BaseMethod R :=
    inr m.
  Definition callBucketsArray {R} (i: nat) (m: EMapMethod Key Val R): BaseMethod R :=
    inr (arrayCall i m).

  Definition impl {R} (m: EMapMethod Key Val R):
    Program BaseMethod R :=

    let k := keyArg m in
    ls <- callLocks getSize;
    let li := (hash k) mod ls in
    _ <- callLocksArray li doLock;
    bs <- callBuckets getSize;
    let bi := (hash k) mod bs in
    r <- callBucketsArray bi m;
    _ <- callLocksArray li doUnlock;
    Return r.

  Definition cHashMapImpl :=
    flattenObject
      base _
      (@impl).

  End CHashMapFun1.

  Module SHashMap := SHashMapFun E.

  Module CHashMapLinFun (SHashMapSeqRefArgs: SHashMap.SHashMapSeqRefParams).

    Section CHashMapLinFun1.

    Context (Val : Type) (nLocks nStripes : {n | n > 0}).
    Definition nBuckets := posMult nLocks nStripes.

    (* Annotating *)
    Definition Ann :=
      option nat.

    (*
    Definition initState (s: {n: nat | 0 < n}): base.(objBase).(State) :=
      (initArrayObjState initLockObjState s,
       initArrayObjState initBucketMapObjState s).
    *)

    Definition AnnBaseMethod :=
      AnnMethod
        Ann
        (PairMethod
           (ArrayMethod LockMethod)
           (ArrayMethod (EMapMethod Key Val))).
    Definition annCallLocks {R} (a: option nat) (m: ArrayMethod LockMethod R): AnnBaseMethod R :=
      annMethod a(inl m).
    Definition annCallLocksArray (a: option nat) (i: nat) {R} (m: LockMethod R): AnnBaseMethod R :=
      annMethod a (inl (arrayCall i m)).
    Definition annCallBuckets (a: option nat) {R} (m: ArrayMethod (EMapMethod Key Val) R): AnnBaseMethod R :=
      annMethod a (inr m).
    Definition annCallBucketsArray (a: option nat) {R} (i: nat) (m: EMapMethod Key Val R): AnnBaseMethod R :=
      annMethod a (inr (arrayCall i m)).

    Definition annImpl {R} (m: EMapMethod Key Val R):
      Program AnnBaseMethod R :=

      let k := keyArg m in    
      ls <- annCallLocks None getSize;
      let li := (hash k) mod ls in
      _ <- annCallLocksArray None li doLock;
      bs <- annCallBuckets (Some li) getSize;
      let bi := (hash k) mod bs in
      r <- annCallBucketsArray (Some li) bi m;
      _ <- annCallLocksArray (Some li) li doUnlock;
      Return r.

    Definition annCHashMapImpl :=
      flattenAnnObject
        (base Val nLocks nBuckets)
        (@annImpl).
    
    (* ------------------------------- *)

    (* 
     * Because the lock object is linearizable,
     * we can replace it with an abstract lock.
     *)
    Definition rLocks := absArray atomicLock nLocks.

    (* R for reduced or replaced. *)
    Definition rBase :=
      pairSpec
        (Annotated Ann rLocks)
        (Annotated Ann (pairSpec ((buckets Val nBuckets).(objBase)) dumObj)).


    Definition RBaseState :=
      prod
        rLocks.(State)
        (prod
           (buckets Val nBuckets).(objBase).(State)
           dumObj.(State)).

    Definition rBaseState
               (l: rLocks.(State))
               (r: (buckets Val nBuckets).(objBase).(State)):
      RBaseState :=
      (l, (r, dumState)).

    Definition rLocksState (rs: RBaseState): rLocks.(State) :=
      fst rs.
    Definition rBucketsState (rs : RBaseState) : (buckets Val nBuckets).(objBase).(State) :=
      fst (snd rs).

    Definition RBaseMethod :=
      PairMethod
        (AnnMethod Ann (ArrayMethod LockMethod))
        (AnnMethod Ann (PairMethod ((arrayImpl (bucketMapImpl Val) nBuckets).(ObjInstruction)) DumMethod)).

    Definition rCallLocks (a: Ann) {R} (m: ArrayMethod LockMethod R): RBaseMethod R :=
      inl (annMethod a m).
    Definition rCallLocksArray (a: Ann) (i: nat) {R} (m: LockMethod R): RBaseMethod R :=
      inl (annMethod a (arrayCall i m)).
    Definition rCallBaseBuckets (a: Ann) {R} (m: (arrayImpl (bucketMapImpl Val) nBuckets).(ObjInstruction) R): RBaseMethod R :=
      inr (annMethod a (inl m)).

    (* Calls on the lock remain unchanged but the calls on the buckets are expanded. *)
    Definition flattenSeqImpl {R} (p: Program AnnBaseMethod R): Program RBaseMethod R :=
      flattenAnnProgramRight
        (* (objProgram (dumNoopImpl (arrayImpl bucketMapImpl))) *)
        (dumNoopImplProg (objProgram (arrayImpl (bucketMapImpl Val) nBuckets)))
        (distPairAnnProg p).

    Definition bindFlattenSeqImpl
               {R} (ann: Ann) (p: Program _ R)
               {R'} (f: R -> Program AnnBaseMethod R'):
      Program RBaseMethod R' :=
      bindFlattenAnnProgramRight
        (objProgram (dumNoopImpl (arrayImpl (bucketMapImpl Val) nBuckets)))
        (mapProgram (fun _ m => inr m) (annProgram ann p))
        (fun r => distPairAnnProg (f r)).

    Definition cImpl :=
      object
        rBase
        (fun _ m => flattenSeqImpl (annImpl m)).

    (* The abstract map is the coarsened sequential hash map. *)
    Definition aImpl: Object (EMapMethod Key Val) :=
      coarsenImpl (SHashMap.hashMapImpl Val nBuckets).

    Inductive PDRel {R}:
      Program RBaseMethod R ->
      RBaseState ->
      Program (EMapMethod Key Val) R ->
      aImpl.(objBase).(State) ->
      (* SHashMap.buckets.(objBase).(State) -> *)
      Prop :=


    | PDRel1PN m sc sa:
        PDRel
          (Noop
             (ls <- rCallLocks None getSize;
              (Noop
                 (flattenSeqImpl
                    (let k := keyArg m in
                     let li := (hash k) mod ls in
                     _ <- annCallLocksArray None li doLock;
                     bs <- annCallBuckets (Some li) getSize;
                     let bi := (hash k) mod bs in
                     r <- annCallBucketsArray (Some li) bi m;
                     _ <- annCallLocksArray (Some li) li doUnlock;
                     Return r)))))
          sc
          (r <- m; Return r)
          sa

    (* Getting size. *)
    | PDRel1 m sc sa:
        PDRel
          (ls <- rCallLocks None getSize;
           (Noop
              (flattenSeqImpl
                 (let k := keyArg m in
                  let li := (hash k) mod ls in
                  _ <- annCallLocksArray None li doLock;
                  bs <- annCallBuckets (Some li) getSize;
                  let bi := (hash k) mod bs in
                  r <- annCallBucketsArray (Some li) bi m;
                  _ <- annCallLocksArray (Some li) li doUnlock;
                  Return r))))
          sc
          (r <- m; Return r)
          sa

    | PDRelNN m scla scls scb sa t:
        PDRel
          (Noop
             (flattenSeqImpl
                (let k := keyArg m in
                 let li := (hash k) mod (proj1_sig scls) in
                 _ <- annCallLocksArray None li doLock;
                 bs <- annCallBuckets (Some li) getSize;
                 let bi := (hash k) mod bs in
                 r <- annCallBucketsArray (Some li) bi m;
                 _ <- annCallLocksArray (Some li) li doUnlock;
                 Return r)))
              ({| arrayMapState := scla; arraySizeState := scls|}, (scb, t))
          (r <- m; Return r)
          sa

    | PDRel1''PN m scla scls scb sa t:
        PDRel
          (let k := keyArg m in
           let li := (hash k) mod (proj1_sig scls) in
           (Noop
             (_ <- rCallLocksArray None li doLock;
              (Noop
                 (flattenSeqImpl
                    (bs <- annCallBuckets (Some li) getSize;
                     let bi := (hash k) mod bs in
                     r <- annCallBucketsArray (Some li) bi m;
                     _ <- annCallLocksArray (Some li) li doUnlock;
                     Return r))))))
          ({| arrayMapState := scla; arraySizeState := scls|}, (scb, t))
          (r <- m; Return r)
          sa
        (* Aquiring the lock. *)
    | PDRel1'' m scla scls scb sa t:
        PDRel
          (let k := keyArg m in
           let li := (hash k) mod (proj1_sig scls) in
           _ <- rCallLocksArray None li doLock;
           (Noop
              (flattenSeqImpl
                 (bs <- annCallBuckets (Some li) getSize;
                  let bi := (hash k) mod bs in
                  r <- annCallBucketsArray (Some li) bi m;
                  _ <- annCallLocksArray (Some li) li doUnlock;
                  Return r))))
          ({| arrayMapState := scla; arraySizeState := scls|}, (scb, t))
          (r <- m; Return r)
          sa

    | PDRel1''N m (scla: nat -> atomicLock.(State)) scls scba scbs saa sas t:
        (let k := keyArg m in
         let li := (hash k) mod (proj1_sig scls) in
         scla li = true) ->
            (let k := keyArg m in
         scba ((hash k) mod (proj1_sig scbs)) =
         saa ((hash k) mod (proj1_sig scbs))) ->

        (let k := keyArg m in           forall h,
           h mod (proj1_sig scls) = (hash k) mod (proj1_sig scls) ->
           h mod (proj1_sig scbs) <> (hash k) mod (proj1_sig scbs) ->
           scba (h mod (proj1_sig scbs)) =
           saa (h mod (proj1_sig scbs))) ->
          PDRel
          (let k := keyArg m in
           let li := (hash k) mod (proj1_sig scls) in
           (Noop
              (flattenSeqImpl
                 (bs <- annCallBuckets (Some li) getSize;
                  let bi := (hash k) mod bs in
                  r <- annCallBucketsArray (Some li) bi m;
                  _ <- annCallLocksArray (Some li) li doUnlock;
                  Return r))))
          ({| arrayMapState := scla; arraySizeState := scls|}, ({| arrayMapState := scba; arraySizeState := scbs|}, t))
          (r <- m; Return r)
          (arrayState _ saa sas)
        (* Noop before getting size'. *)
    | PDRel1p' m (scla: nat -> atomicLock.(State)) scls scba scbs saa sas t:
        (let k := keyArg m in
         let li := (hash k) mod (proj1_sig scls) in
         scla li = true) ->

        (* The buckets whose lock is aquired are equal to the abstract buckets. *)
        (let k := keyArg m in           scba ((hash k) mod (proj1_sig scbs)) =
         saa ((hash k) mod (proj1_sig scbs))) ->

        (let k := keyArg m in           forall h,
            h mod (proj1_sig scls) = (hash k) mod (proj1_sig scls) ->
            h mod (proj1_sig scbs) <> (hash k) mod (proj1_sig scbs) ->
            scba (h mod (proj1_sig scbs)) =
            saa (h mod (proj1_sig scbs))) ->

        PDRel
          (let k := keyArg m in
           let li := (hash k) mod (proj1_sig scls) in
           (Noop
              (bs <- rCallBaseBuckets (Some li) getSize;
               (Noop
                  (flattenSeqImpl
                     (let bi := (hash k) mod bs in
                      r <- annCallBucketsArray (Some li) bi m;
                      _ <- annCallLocksArray (Some li) li doUnlock;
                      Return r))))))
          ({| arrayMapState := scla; arraySizeState := scls|}, ({| arrayMapState := scba; arraySizeState := scbs|}, t))
          (r <- m; Return r)
          (arrayState _ saa sas)

    (* Getting size'. *)
    | PDRel1' m (scla: nat -> atomicLock.(State)) scls scba scbs saa sas t:
        (let k := keyArg m in
         let li := (hash k) mod (proj1_sig scls) in
         scla li = true) ->

        (let k := keyArg m in           scba ((hash k) mod (proj1_sig scbs)) =
         saa ((hash k) mod (proj1_sig scbs))) ->

        (let k := keyArg m in
         forall h,
            h mod (proj1_sig scls) = (hash k) mod (proj1_sig scls) ->
            h mod (proj1_sig scbs) <> (hash k) mod (proj1_sig scbs) ->
            scba (h mod (proj1_sig scbs)) =
            saa (h mod (proj1_sig scbs))) ->

         PDRel
           (let k := keyArg m in
            let li := (hash k) mod (proj1_sig scls) in
            (bs <- rCallBaseBuckets (Some li) getSize;
             (Noop
                (flattenSeqImpl
                   (let bi := (hash k) mod bs in
                    r <- annCallBucketsArray (Some li) bi m;
                    _ <- annCallLocksArray (Some li) li doUnlock;
                    Return r)))))
            ({| arrayMapState := scla; arraySizeState := scls|}, ({| arrayMapState := scba; arraySizeState := scbs|}, t))
            (r <- m; Return r)
            (arrayState _ saa sas)

    | PDRel1'N m (scla: nat -> atomicLock.(State)) scls scba scbs saa sas t:
        (let k := keyArg m in
         let li := (hash k) mod (proj1_sig scls) in
         scla li = true) ->

        (let k := keyArg m in
         scba ((hash k) mod (proj1_sig scbs)) =
         saa ((hash k) mod (proj1_sig scbs))) ->

        (let k := keyArg m in
         forall h,
            h mod (proj1_sig scls) = (hash k) mod (proj1_sig scls) ->
            h mod (proj1_sig scbs) <> (hash k) mod (proj1_sig scbs) ->
            scba (h mod (proj1_sig scbs)) =
            saa (h mod (proj1_sig scbs))) ->

         PDRel
           (let k := keyArg m in
            let li := (hash k) mod (proj1_sig scls) in
            let bi := (hash k) mod (proj1_sig scbs) in
            (Noop
               (flattenSeqImpl
                  (r <- annCallBucketsArray (Some li) bi m;
                   _ <- annCallLocksArray (Some li) li doUnlock;
                   Return r))))
            ({| arrayMapState := scla; arraySizeState := scls|}, ({| arrayMapState := scba; arraySizeState := scbs|}, t))
            (r <- m; Return r)
            (arrayState _ saa sas)

    (* The Noop before the execution of the method object. *)
    | PDRel2 m (scla: nat -> atomicLock.(State)) scls scba scbs saa sas t:
        (let k := keyArg m in
         let li := (hash k) mod (proj1_sig scls) in
         scla li = true) ->

        (let k := keyArg m in
         scba ((hash k) mod (proj1_sig scbs)) =
         saa ((hash k) mod (proj1_sig scbs))) ->

        (let k := keyArg m in
         forall h,
            h mod (proj1_sig scls) = (hash k) mod (proj1_sig scls) ->
            h mod (proj1_sig scbs) <> (hash k) mod (proj1_sig scbs) ->
            scba (h mod (proj1_sig scbs)) =
            saa (h mod (proj1_sig scbs))) ->

        PDRel
          (let k := keyArg m in
           let li := (hash k) mod (proj1_sig scls) in
           let bi := (hash k) mod (proj1_sig scbs) in
           (Noop
              (bindFlattenSeqImpl
                 (Some li)
                 (* (objProgram (arrayImpl bucketMapImpl) _ (arrayCall bi m)) *)
                 (dumNoopProgram (objProgram (arrayImpl (bucketMapImpl Val) nBuckets) _ (arrayCall bi m)))
                 (fun r =>
                    _ <- annCallLocksArray (Some li) li doUnlock;
                    Return r))))
          ({| arrayMapState := scla; arraySizeState := scls|}, ({| arrayMapState := scba; arraySizeState := scbs|}, t))
          (r <- m; Return r)
          (arrayState _ saa sas)

    (* At the beginning or in the middle of the object of m. *)
    (* 
     * sc: state, concrete
     * sa: state, abstract
     * pc: program, concrete
     *)
    | PDRel2' m pc (scla: nat -> atomicLock.(State)) scls scba scbs saa sas t:
        (let k := keyArg m in
         let li := (hash k) mod (proj1_sig scls) in
         scla li = true) ->

        (let k := keyArg m in
         let li := (hash k) mod (proj1_sig scls) in
         let bi := (hash k) mod (proj1_sig scbs) in                 
         ProgramSteps
           (objBase (bucketMapImpl Val)) _
           (objProgram (bucketMapImpl Val) _ m)
           (saa bi)
           pc
           (scba bi)) ->

        (let k := keyArg m in        
         forall h,        
            h mod (proj1_sig scls) = (hash k) mod (proj1_sig scls) ->
            h mod (proj1_sig scbs) <> (hash k) mod (proj1_sig scbs) ->
            scba (h mod (proj1_sig scbs)) =
            saa (h mod (proj1_sig scbs))) ->
        
        PDRel
          (let k := keyArg m in
           let li := (hash k) mod (proj1_sig scls) in
           let bi := (hash k) mod (proj1_sig scbs) in
           (bindFlattenSeqImpl
              (Some li)
              (* (arrayLift pc bi) *)
              (dumNoopProgram (arrayLift pc bi))
              (fun r =>
                 _ <- annCallLocksArray (Some li) li doUnlock;
                 Return r)))
          ({| arrayMapState := scla; arraySizeState := scls|}, ({| arrayMapState := scba; arraySizeState := scbs|}, t))
          (r <- m; Return r)
          (arrayState _ saa sas)

    (* At the end of the object of m, before the unlock method call. *)
    | PDRel3 m r (scla: nat -> atomicLock.(State)) scls scba scbs saa sas t:
        (let k := keyArg m in
         let li := (hash k) mod (proj1_sig scls) in
         scla li = true) ->

        (let k := keyArg m in
         let li := (hash k) mod (proj1_sig scls) in
         let bi := (hash k) mod (proj1_sig scbs) in                 
         ProgramSteps
           (objBase (bucketMapImpl Val)) _
           (objProgram (bucketMapImpl Val) _ m)
           (saa bi)
           (Return r)
           (scba bi)) ->

        (let k := keyArg m in        
         forall h,        
            h mod (proj1_sig scls) = (hash k) mod (proj1_sig scls) ->
            h mod (proj1_sig scbs) <> (hash k) mod (proj1_sig scbs) ->
            scba (h mod (proj1_sig scbs)) =
            saa (h mod (proj1_sig scbs))) ->

        PDRel
          (let k := keyArg m in
           let li := (hash k) mod (proj1_sig scls) in
           (Noop 
              (_ <- rCallLocksArray (Some li) li doUnlock;
               Noop (Return r))))
          ({| arrayMapState := scla; arraySizeState := scls|}, ({| arrayMapState := scba; arraySizeState := scbs|}, t))
          (r <- m; Return r)
          (arrayState _ saa sas)

    (* Before the unlock method call. *)
    | PDRel4 m r (scla: nat -> atomicLock.(State)) scls scba scbs saa sas t:
        (let k := keyArg m in
         let li := (hash k) mod (proj1_sig scls) in
         scla li = true) ->

        (let k := keyArg m in
         let li := (hash k) mod (proj1_sig scls) in
         let bi := (hash k) mod (proj1_sig scbs) in                 
         ProgramSteps
           (objBase (bucketMapImpl Val)) _
           (objProgram (bucketMapImpl Val) _ m)
           (saa bi)
           (Return r)
           (scba bi)) ->

        (let k := keyArg m in        
         forall h,        
            h mod (proj1_sig scls) = (hash k) mod (proj1_sig scls) ->
            h mod (proj1_sig scbs) <> (hash k) mod (proj1_sig scbs) ->
            scba (h mod (proj1_sig scbs)) =
            saa (h mod (proj1_sig scbs))) ->

        PDRel
          (let k := keyArg m in
           let li := (hash k) mod (proj1_sig scls) in
           (_ <- rCallLocksArray (Some li) li doUnlock;
            Noop (Return r)))
          ({| arrayMapState := scla; arraySizeState := scls|}, ({| arrayMapState := scba; arraySizeState := scbs|}, t))
          (r <- m; Return r)
          (arrayState _ saa sas)

    (* At the end noop. *)
    | PDRel5' r s1 s2:
        PDRel
          (Noop (Return r)) s1
          (Return r) s2

    (* At the end. *)
    | PDRel5 r s1 s2:
        PDRel
          (Return r) s1
          (Return r) s2.

    Hint Constructors PDRel.

    Definition DRel
               (sc: RBaseState)
               (sa: aImpl.(objBase).(State)): Prop :=
               (* (s2: ObjState seqImpl): Prop := *)

      arraySizeState (rBucketsState sc) =
      arraySizeState sa

      /\ (proj1_sig (arraySizeState (rBucketsState sc)))mod
        (proj1_sig (arraySizeState (rLocksState sc))) = 0

      /\ forall h,
        arrayMapState
          (rLocksState sc)
          (h mod (proj1_sig (arraySizeState (rLocksState sc)))) =
        false ->

        arrayMapState
          (rBucketsState sc)
          (h mod (proj1_sig (arraySizeState (rBucketsState sc)))) =
        arrayMapState
          sa
          (h mod (proj1_sig (arraySizeState sa))).

    CoInductive outCritical {R}: nat -> Program RBaseMethod R -> Prop :=
      
    | BindOutCritical1 i R' (m: AnnMethod Ann (ArrayMethod LockMethod) R') f:
        annMethodAnn m <> Some i ->
        outCritical i (Bind (inl m) f)
    | BindOutCritical2 i R' (m: AnnMethod Ann _ R') f:
        annMethodAnn m <> Some i ->
        outCritical i (Bind (inr m) f)
    | ReturnOutCritical i r:
        outCritical i (Return r)
    | NoopOutCritical i p:
        outCritical i p ->
        outCritical i (Noop p).

    
    Definition PPRel {R R'}
               (p11: Program RBaseMethod R)
               (p21: Program (EMapMethod Key Val) R)
               (p12: Program RBaseMethod R')
               (p22: Program (EMapMethod Key Val) R'):
      Prop :=
      forall i,
        outCritical i p11 \/
        outCritical i p12.


    Lemma PPRelSym:
      forall R (p1: Program RBaseMethod R) p2
        R' (p1': Program RBaseMethod R') p2',

        PPRel p1 p2 p1' p2' ->
        PPRel p1' p2' p1 p2.

    Proof.
      introv H1.
      unfold PPRel in *.
      intros. specialize (H1 i). iauto.
    Qed.
    
    Lemma initRel:
      DRel (initialState rBase) (initialState (buckets Val nBuckets).(objBase)).
    Proof.
      intros. unfold DRel. splits.
      - simpl. reflexivity.
      - simpl. rewrite Nat.mul_comm. apply Nat.mod_mul. destruct nLocks; cbn; lia.
      - intros. simpl. reflexivity.
    Qed.  

    Import Coq.Strings.String.
    Ltac print_o b s :=
      match b with
      | true => idtac "!!!" s      
      | false => idtac
      end.

    Ltac print s :=
      print_o true s.

    Ltac mark :=
      print "Mark "%string.

    Ltac substLet v :=
      unfold v in *;
      clear v.

    Ltac substLetAll :=                       
      repeat (
          match goal with
          | [ x := _ |- _ ] =>
            substLet x
          end).    

    Ltac simplFlatten :=
      try unfold flattenSeqImpl;
      try unfold bindFlattenSeqImpl;
      try unfold bindFlattenAnnProgramRight;
      try unfold flattenAnnProgramRight;      
      try unfold bindFlattenProgramRight; 
      try unfold flattenProgramRight;
      match goal with
      | [ |- context [ flattenProgram _ ?p ] ] =>
        setoid_rewrite (frobProgramId (flattenProgram _ p)); simpl
      | [ |- context[ bindFlattenProgram _ _ ?p ] ] =>
        setoid_rewrite (frobProgramId (bindFlattenProgram _ _ p)); simpl
      end.

    Ltac simplFlattenIn H :=
      try unfold flattenSeqImpl in H;
      try unfold bindFlattenSeqImpl in H;
      try unfold bindFlattenAnnProgramRight in H;
      try unfold flattenAnnProgramRight in H;
      try unfold bindFlattenProgramRight in H;
      try unfold flattenProgramRight in H;
      match type of H with
      | context[ flattenProgram _ ?p ] =>
        setoid_rewrite (frobProgramId (flattenProgram _ p)) in H; simpl in H
      | context[ bindFlattenProgram _ _ ?p] =>
        setoid_rewrite (frobProgramId (bindFlattenProgram _ _ p)) in H; simpl in H
      end.

    Lemma noopRel:
      forall R (p1: Program RBaseMethod R) p2 s1 s2,
        PDRel (Noop p1) s1 p2 s2 ->
        (PDRel p1 s1 p2 s2
         /\ (forall R (p1t: Program RBaseMethod R) p2t,
               PDRel p1t s1 p2t s2 ->
               PPRel (Noop p1) p2 p1t p2t ->
               PPRel p1 p2 p1t p2t))
        \/ (exists p2',
              p2 = Noop p2'
              /\ PDRel p1 s1 p2' s2
              /\ (forall R (p1t: Program RBaseMethod R) p2t,
                    PDRel p1t s1 p2t s2 ->
                    PPRel (Noop p1) p2 p1t p2t ->
                    PPRel p1 p2' p1t p2t))
      \/ (exists p2',
          ProgramSteps aImpl.(objBase) _ p2 s2 p2' s2
          /\ PDRel p1 s1 p2' s2
          /\ (forall R (p1t: Program cImpl.(ObjInstruction) R) p2t,
                PDRel p1t s1 p2t s2 ->
                PPRel (Noop p1) p2 p1t p2t ->
                PPRel p1 p2' p1t p2t)).


    Proof.
      introv H1.
      inverts H1 as. (* sion H1; subst. *)

      - left. splits.
        + constructor.
        + introv H3 H4.
          unfold PPRel in *.
          intros. specialize (H4 i).
          destruct H4 as [H4 | H4].
          * inverts H4. auto.
          * auto.

      - left. splits.
        + simplFlatten. simplFlatten.
          setoid_rewrite bind_eq_eq at 1.
          { constructor; assumption. }
          { reflexivity. }
          unfold pointwise_relation. intros.
          match goal with
          | [ |- context [bindFlattenProgram _ _ (mapProgram ?p ?p')]] =>
            rewrite (frobProgramId (mapProgram p p')); simpl
          end.
          progFrobSimplAt 1.
          reflexivity.

        + introv H3 H4.
          unfold PPRel in *.
          intros. specialize (H4 i).
          destruct H4 as [H4 | H4].
          * inverts H4. auto.
          * auto.            

      - left. splits.
        + constructor.
        + introv H3 H4.
          unfold PPRel in *.
          intros. specialize (H4 i).
          destruct H4 as [H4 | H4].
          * inverts H4. auto.
          * auto.            

      - left. splits.
        + simplFlatten. simplFlatten.
          setoid_rewrite bind_eq_eq at 1.
          2: reflexivity.
          { constructor; assumption. }
          unfold pointwise_relation. intros.
          match goal with
          | [ |- context [bindFlattenProgram _ _ (mapProgram ?p ?p')]] =>
            rewrite (frobProgramId (mapProgram p p')); simpl
          end.
          progFrobSimplAt 1.
          reflexivity.

        + introv H4 H5.
          unfold PPRel in *.
          intros. specialize (H5 i).
          destruct H5 as [H5 | H5].
          * inverts H5. auto.
          * auto.            
            
      (* Noop before getting size'. *)      
      - introv H1 H2.
        left. splits.
        + econstructor.
          eassumption.
          simpl in *; assumption.
          assumption.
          
        + introv H4 H5.
          unfold PPRel in *.
          intros. specialize (H5 i).
          destruct H5 as [H5 | H5].
          * left. inverts H5 as H5. assumption.
          * auto.

      - introv H1 H2.
        left. splits.
        + simplFlatten.
          econstructor. eassumption.
          simpl in *; assumption.
          assumption.
          
        + introv H4 H5.
          unfold PPRel in *.
          intros. specialize (H5 i).
          destruct H5 as [H5 | H5].
          * left. inverts H5 as H5. assumption.
          * auto.
            
      (* Right after the lock call *)            
      - introv H1 H2.
        left. splits.
        + econstructor.
          eassumption.
          simpl in *.
          rewrite H2. constructor.
          assumption.
          
        + introv H4 H5.
          unfold PPRel in *.
          intros. specialize (H5 i).
          destruct H5 as [H5 | H5].
          * left. inverts H5 as H5. assumption.
          * auto.


      (* Noop as the result of flattening *) 
      - introv H1 H2 H2' H3.
        left.
        unfold bindFlattenSeqImpl in H3.
        unfold bindFlattenAnnProgramRight in H3.
        eapply bindFlattenProgramRightEqNoop in H3.
        destruct H3 as [(r & A1 & A2) | (p' & A1 & A2)].
        + apply mapProgramEqRet in A1.
          apply annProgramEqRet in A1.
          apply dumNoopProgramEqRet in A1.
          apply arrayLiftEqRet in A1.
          subst. splits.
          * simplFlatten. simplFlatten. 
            setoid_rewrite bind_eq_eq at 1.
            2: reflexivity.
            2:{ unfold pointwise_relation. intros.
            match goal with
            | [ |- context [mapProgram _ _]] =>
              rewrite (frobProgramId (mapProgram _ _)); simpl
            end.
            match goal with
            | [ |- context [mapProgram _ _]] =>
              rewrite (frobProgramId (mapProgram _ _)); simpl
            end.
            progFrobSimplAt 1.
            progFrobSimplAt 2.
            reflexivity.
            }
            { constructor; assumption. }

          * { introv H3 H4.
              progFrobSimplInAt H4 1. progFrobSimplInAt H4 2. progFrobSimplInAt H4 3.
              progFrobSimplInAt H4 4. progFrobSimplInAt H4 5.
              unfold PPRel in *.
              intros. specialize (H4 i).
              destruct H4 as [H4 | H4].
              * inverts H4 as H4. inverts H4 as H4. inverts H4 as H4. simpl in H4.
                constructor. simplFlatten. constructor. simplFlatten. constructor.
                assumption.
              * auto.
            }

        + eapply mapProgramEqNoop in A1. destruct A1 as (p'' & A11 & A12).
          eapply annProgramEqNoop in A11. destruct A11 as (p''' & A111 & A112).
          eapply dumNoopProgramEqNoop in A111. contradiction.

      - (* introv H1 H2. *)
        left. splits.
        + constructor; assumption.
        + introv _ H4.
          unfold PPRel in *.
          intros. specialize (H4 i).
          destruct H4 as [H4 | H4].
          * inverts H4. auto.
          * auto.            

      - (* introv H1 H2. *)
        left. splits.
        + constructor; assumption.
        + introv _ H4.
          unfold PPRel in *.
          intros. specialize (H4 i).
          destruct H4 as [H4 | H4].
          * inverts H4. auto.
          * auto.            
    Qed.


    Lemma invRel:
      forall R m sc sa,
        DRel sc sa ->
        PDRel (objProgram cImpl R m) sc
              (objProgram aImpl R m) sa
        /\ forall R' (pc: Program cImpl.(ObjInstruction) R') pa,
            PDRel pc sc pa sa ->
            PPRel (objProgram cImpl R m) (objProgram aImpl R m)
                  pc pa.

    Proof.
      introv H1.
      unfold cImpl. unfold aImpl. simpl. unfold annImpl.
      simplFlatten. simplFlatten.
      splits.
      + setoid_rewrite bind_eq_eq at 1.
        2: reflexivity.
        { constructor. }
        unfold pointwise_relation. intros.
        match goal with
        | [ |- context [bindFlattenProgram _ _ (mapProgram ?p ?p')]] =>
          rewrite (frobProgramId (mapProgram p p')); simpl
        end.
        progFrobSimplAt 1.
        reflexivity.

      + introv H2. unfold PPRel. left. constructor. constructor. simpl. intro. inverts H.
        
    Qed.


    Lemma retRel:
      forall R (r: R) sc pa sa,
        PDRel (Return r) sc pa sa ->
        DRel sc sa ->
        pa = Return r.

    Proof.
      introv H1 H2.
      (* inverts H1 as. *)
      inversion H1.
      - eapply bindFlattenProgramEqRet in H. contradiction.
      - reflexivity.
    Qed.

    Lemma OutCriticalHelper0:
      forall (i: nat) a R R' (m: PairMethod (ArrayMethod LockMethod) (ArrayMethod (EMapMethod Key Val)) R)
        (f: R -> Program AnnBaseMethod R'),
        a <> Some i ->
        (forall r,
          outCritical
          i
          (flattenSeqImpl (f r))) ->
        outCritical
          i
          (flattenSeqImpl
             (r <- annMethod a m; f r)).

    Proof.
      introv H1 H2.
      simplFlatten. constructor.
      destruct m.

      - progFrobSimpl. constructor. simpl. assumption.
      - destruct a0.
        + match goal with
          | [ |- context [bindFlattenProgram _ _ (pairImplProg ?p ?p' _ _)]] =>
            setoid_rewrite (frobProgramId (pairImplProg p p' _ _)); simpl
          end.
          generalize (objProgram (bucketMapImpl Val) R e).
          (* cofix CoInd'. *)
          intro p'. destruct p'.
          * progFrobSimpl. constructor. simpl. assumption.
          * progFrobSimpl. constructor. progFrobSimpl. specialize (H2 r). simplFlattenIn H2. eapply H2.
          * progFrobSimpl. constructor.
            simpl. assumption.
        + progFrobSimpl. constructor. simpl. assumption.
    Qed.
    
    Lemma OutCriticalHelper1:
      forall (i k: nat) a R R' (m: PairMethod (ArrayMethod LockMethod) (ArrayMethod (EMapMethod Key Val)) R') p,
        a <> Some i ->
        outCritical
          i
          (bindFlattenSeqImpl
             a
             (dumNoopProgram (arrayLift p k))
             (fun r : R =>
                _ <- annMethod a m; Return r)).

    Proof.
      introv H1.
      gen p. (* cofix CoInd. *)
      intro p; destruct p.
      - simplFlatten. constructor. simpl. assumption.
      - simplFlatten. constructor.
        (* Guarded. *) destruct m. (* Guarded. *)

        + progFrobSimpl. constructor. progFrobSimpl. constructor. simpl. assumption.
        + eapply OutCriticalHelper0. assumption.
          intro. simplFlatten. constructor.
          
      - simplFlatten. constructor. simpl. assumption.
    Qed.
    
    (* Helper lemmas for stepRel lemma. *)
    Lemma PDRelPres1:
      forall {R} (scla: nat -> atomicLock.(State)) scls scba scbs s2 h
        (p1: Program RBaseMethod R)
        (p2: Program (EMapMethod Key Val) R) t,

        scla h = false ->
        PDRel p1 ({| arrayMapState := scla; arraySizeState := scls |}, ({| arrayMapState := scba; arraySizeState := scbs |}, t))
              p2 s2 ->
        PDRel p1
              ({| arrayMapState := (noverride scla h true); arraySizeState := scls |}, ({| arrayMapState := scba; arraySizeState := scbs |}, t))
              p2 s2.

    Proof.
      introv H1 H2.      
      inverts H2; simpl in *.
      - constructor.
      - constructor.
      - constructor.
      - constructor.
      - constructor.
      - constructor; simpl.
        + destruct_override; auto. 
        + assumption.
        + assumption.
      - constructor; simpl.
        + destruct_override; auto. 
        + assumption.
        + assumption.
      - constructor; simpl.
        + destruct_override; auto. 
        + assumption.
        + assumption.
      - constructor; simpl.
        + destruct_override; auto. 
        + assumption.
        + assumption.
      - constructor; simpl.
        + destruct_override; auto. 
        + assumption.
        + assumption.
      - constructor; simpl.
        + destruct_override; auto. 
        + assumption.
        + assumption.
      - constructor; simpl.
        + destruct_override; auto. 
        + assumption.
        + assumption.          
      - constructor; simpl.
        + destruct_override; auto. 
        + assumption.          
        + assumption.
      - constructor.        
      - constructor.        
    Qed.    


    Lemma PPRelPres1:
      forall R s1la s1ls s1b s2 (m: EMapMethod Key Val R)
        R' (p1t : Program RBaseMethod R') p2t,
        s1la (hash (keyArg m) mod (proj1_sig s1ls)) = false ->
        PDRel p1t (arrayState atomicLock s1la s1ls, s1b) p2t s2 ->
        PPRel
          (Noop
             (flattenSeqImpl
                (bs <- annCallBuckets
                    (Some (hash (keyArg m) mod (proj1_sig s1ls)))
                    getSize;
                 r <- annCallBucketsArray
                   (Some (hash (keyArg m) mod (proj1_sig s1ls)))
                   (hash (keyArg m) mod bs) m;
                 _ <- annCallLocksArray
                   (Some (hash (keyArg m) mod (proj1_sig s1ls)))
                   (hash (keyArg m) mod (proj1_sig s1ls)) doUnlock; 
                 Return r)))
          (r <- m; Return r) 
          p1t p2t.

    Proof.  
      introv H1 H2.                
      unfold PPRel.
      inverts H2 as; simpl in *.
      + right. constructor. constructor. simpl. intro. inverts H.
      + right. constructor. simpl. intro. inverts H.
      + right. constructor. simplFlatten. constructor.
        simplFlatten. constructor.
        simpl. intro. inverts H.
      + right. constructor. constructor.
        simpl. intro. inverts H.
      + right. constructor.
        simpl. intro. inverts H.
      + introv H2 _ _.
        intro.
        destruct (eq_nat_dec i (hash (keyArg m) mod (proj1_sig s1ls))).
        destruct (eq_nat_dec i (hash (keyArg m0) mod (proj1_sig s1ls))).
        congruence.
        right. constructor. simplFlatten. constructor. simplFlatten. constructor. simpl. congruence.
        left. constructor. simplFlatten. constructor. simplFlatten. constructor. simpl. congruence.        
      + introv H2 _ _.
        intro.
        destruct (eq_nat_dec i (hash (keyArg m) mod (proj1_sig s1ls))).
        destruct (eq_nat_dec i (hash (keyArg m0) mod (proj1_sig s1ls))).
        congruence.
        right. constructor. constructor. simpl. congruence.
        left. constructor. simplFlatten. constructor. simplFlatten. constructor. simpl. congruence.
      + introv H2 _ _.
        intro.
        destruct (eq_nat_dec i (hash (keyArg m) mod (proj1_sig s1ls))).
        destruct (eq_nat_dec i (hash (keyArg m0) mod (proj1_sig s1ls))).
        congruence.
        right. constructor. simpl. congruence.
        left. constructor. simplFlatten. constructor. simplFlatten. constructor. simpl. congruence.
      + introv H2 _ _.
        intro.
        destruct (eq_nat_dec i (hash (keyArg m) mod (proj1_sig s1ls))).
        destruct (eq_nat_dec i (hash (keyArg m0) mod (proj1_sig s1ls))).
        congruence.
        right. constructor. simplFlatten. constructor. eapply OutCriticalHelper1. simpl. congruence.
        left. constructor. simplFlatten. constructor. simplFlatten. constructor. simpl. congruence.
      + introv H2 _ _.
        intro.
        destruct (eq_nat_dec i (hash (keyArg m) mod (proj1_sig s1ls))).
        destruct (eq_nat_dec i (hash (keyArg m0) mod (proj1_sig s1ls))).
        congruence.
        right. constructor. eapply OutCriticalHelper1. congruence.
        left. constructor. simplFlatten. constructor. simplFlatten. constructor. simpl. congruence.
      + introv H2 _ _.
        intro.
        destruct (eq_nat_dec i (hash (keyArg m) mod (proj1_sig s1ls))).
        destruct (eq_nat_dec i (hash (keyArg m0) mod (proj1_sig s1ls))).
        congruence.
        right. eapply OutCriticalHelper1. congruence.
        left. constructor. simplFlatten. constructor. simplFlatten. constructor. simpl. congruence.
      + introv H2 _ _.
        intro.
        destruct (eq_nat_dec i (hash (keyArg m) mod (proj1_sig s1ls))).
        destruct (eq_nat_dec i (hash (keyArg m0) mod (proj1_sig s1ls))).
        congruence.
        right. constructor. constructor. simpl. congruence.
        left. constructor. simplFlatten. constructor. simplFlatten. constructor. simpl. congruence.
      + introv H2 _ _.
        intro.
        destruct (eq_nat_dec i (hash (keyArg m) mod (proj1_sig s1ls))).
        destruct (eq_nat_dec i (hash (keyArg m0) mod (proj1_sig s1ls))).
        congruence.
        right. constructor. simpl. congruence.
        left. constructor. simplFlatten. constructor. simplFlatten. constructor. simpl. congruence.
      + intro. right. constructor. constructor.
      + intro. right. constructor.
    Qed.
        

    Lemma lockSizeBucketSizeModEq:
      forall (s1ls s1bs : {n : nat | 0 < n}) h h',
        proj1_sig s1bs mod proj1_sig s1ls = 0 ->
        h mod proj1_sig s1bs = h' mod proj1_sig s1bs ->
        h mod proj1_sig s1ls = h' mod proj1_sig s1ls.
    Proof.
      introv H1 H2.         
      applys DivDivModEq H1 H2.
      destruct s1ls. simpl. eapply not_eq_sym. eapply Nat.lt_neq. assumption.
      destruct s1bs. simpl. eapply not_eq_sym. eapply Nat.lt_neq. assumption.
    Qed.
    
    Lemma PDRelPres2:
      forall (R' R : Type) s1la
        (s1ls : {n : nat | 0 < n})
        (s1ba : nat -> State (objBase (bucketMapImpl Val)))
        (s1bs : {n : nat | 0 < n}) (s2 : ArrayState (objBase (bucketMapImpl Val)))
        (m0 : EMapMethod Key Val R) (i : ObjInstruction (bucketMapImpl Val) R')
        (p : R' -> Program (ObjInstruction (bucketMapImpl Val)) R)
        (o' : State (objBase (bucketMapImpl Val))) (R0 : Type)
        (p1t : Program RBaseMethod R0) (p2t : Program (EMapMethod Key Val) R0) t,
        PDRel p1t
              ({| arrayMapState := s1la; arraySizeState := s1ls |},
               ({| arrayMapState := s1ba; arraySizeState := s1bs |}, t))
              p2t s2 ->
        proj1_sig s1bs mod proj1_sig s1ls = 0 ->
        PPRel
          (bindFlattenSeqImpl
             (Some (hash (keyArg m0) mod proj1_sig s1ls))
             (dumNoopProgram (arrayLift (x <- i; p x) (hash (keyArg m0) mod proj1_sig s1bs)))
             (fun r : R =>
                _ <-
                  annCallLocksArray (Some (hash (keyArg m0) mod proj1_sig s1ls))
                  (hash (keyArg m0) mod proj1_sig s1ls) doUnlock; 
              Return r))
          (r <- m0; Return r)
          p1t p2t ->
        PDRel p1t
              ({| arrayMapState := s1la;
                  arraySizeState := s1ls |},
               ({| arrayMapState := noverride s1ba (hash (keyArg m0) mod proj1_sig s1bs) o';
                  arraySizeState :=  s1bs |}, t))
              p2t s2.
          
          
    Proof.
      introv H1 H2 H3.
      inverts H1 as; simpl in *.

      - constructor.
      - constructor.
      - constructor.
      - constructor.
      - constructor.
        
      - { introv H4 H5 H6.

          assert (H3': hash (keyArg m0) mod proj1_sig s1ls <> hash (keyArg m) mod proj1_sig s1ls).
          - intro H7.
            unfold PPRel in H3.
            specialize (H3 (hash (keyArg m0) mod proj1_sig s1ls)). destruct H3 as [H3 | H3].
            * simplFlattenIn H3. inverts H3. simpl in *. congruence.
            * simplFlattenIn H3. simplFlattenIn H3. inverts H3 as H3. inverts H3 as H3. inverts H3 as H3. simpl in H3. congruence.
              
          - clear H3.  
            constructor; simpl.
            + auto.
            + destruct_override.
              * eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * congruence.
            + introv H7 H8.
              destruct_override.
              * eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * auto.
        }
          

      - { introv H4 H5 H6.

          assert (H3': hash (keyArg m0) mod proj1_sig s1ls <> hash (keyArg m) mod proj1_sig s1ls).
          - intro H7.
            unfold PPRel in H3.
            specialize (H3 (hash (keyArg m0) mod proj1_sig s1ls)). destruct H3 as [H3 | H3].
            * simplFlattenIn H3. inverts H3. simpl in *. congruence.
            * inverts H3 as H3. inverts H3 as H3. simpl in H3. congruence.
              
          - clear H3.  
            constructor; simpl.
            + auto.
            + destruct_override.
              * eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * congruence.
            + introv H7 H8.
              destruct_override.
              * eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * auto.
        }

      - { introv H4 H5 H6.

          assert (H3': hash (keyArg m0) mod proj1_sig s1ls <> hash (keyArg m) mod proj1_sig s1ls).
          - intro H7.
            unfold PPRel in H3.
            specialize (H3 (hash (keyArg m0) mod proj1_sig s1ls)). destruct H3 as [H3 | H3].
            * simplFlattenIn H3. inverts H3. simpl in *. congruence.
            * inverts H3 as H3. simpl in H3. congruence.
            
          - clear H3.  
            constructor; simpl.
            + auto.
            + destruct_override.
              * eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * congruence.
            + introv H7 H8.
              destruct_override.
              * eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * auto.
        }

      - { introv H4 H5 H6.

          assert (H3': hash (keyArg m0) mod proj1_sig s1ls <> hash (keyArg m) mod proj1_sig s1ls).
          - intro H7.
            unfold PPRel in H3.
            specialize (H3 (hash (keyArg m0) mod proj1_sig s1ls)). destruct H3 as [H3 | H3].
            * simplFlattenIn H3. inverts H3. simpl in *. congruence.
            * { inverts H3 as H3. simplFlattenIn H3.
                inverts H3 as H3. simplFlattenIn H3.
                destruct (objProgram (bucketMapImpl Val) R0 m).
                - inverts H3 as H3.
                  simpl in *. congruence.
                - inverts H3 as H3. progFrobSimplIn H3.
                  inverts H3 as H3. progFrobSimplIn H3.
                  inverts H3 as H3.
                  simpl in *. congruence.
                  
                - (* inverts OS. *)
                  inverts H3 as H3.
                  simpl in *. congruence.                
              }
            
          - clear H3.  
            constructor; simpl.
            + auto.
            + destruct_override.
              * eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * congruence.
            + introv H7 H8.
              destruct_override.
              * eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * auto.
        }

      - { introv H4 H5 H6.

          assert (H3': hash (keyArg m0) mod proj1_sig s1ls <> hash (keyArg m) mod proj1_sig s1ls).
          - intro H7.
            unfold PPRel in H3.
            specialize (H3 (hash (keyArg m0) mod proj1_sig s1ls)). destruct H3 as [H3 | H3].
            * simplFlattenIn H3. inverts H3. simpl in *. congruence.
            * { inverts H3 as H3. simplFlattenIn H3.
                destruct (objProgram (bucketMapImpl Val) R0 m).
                - inverts H3 as H3.
                  simpl in *. congruence.
                - inverts H3 as H3. progFrobSimplIn H3.
                  inverts H3 as H3. progFrobSimplIn H3.
                  inverts H3 as H3.
                  simpl in *. congruence.
                  
                - (* inverts OS. *)
                  inverts H3 as H3.
                  simpl in *. congruence.                
              }
            
          - clear H3.  
            constructor; simpl.
            + auto.
            + destruct_override.
              * eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * congruence.
            + introv H7 H8.
              destruct_override.
              * eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * auto.
        }

                  
      - { introv H4 H5 H6.

          assert (H3': hash (keyArg m0) mod proj1_sig s1ls <> hash (keyArg m) mod proj1_sig s1ls).
          - intro H7.
            unfold PPRel in H3.
            specialize (H3 (hash (keyArg m0) mod proj1_sig s1ls)). destruct H3 as [H3 | H3].
            * simplFlattenIn H3. inverts H3. simpl in *. congruence.
            * { destruct pc; simplFlattenIn H3; inverts H3 as H3.
                - simpl in *. congruence.
                - progFrobSimplIn H3. inverts H3 as H3.
                  progFrobSimplIn H3. inverts H3 as H3.
                  simpl in *. congruence.
                - simpl in *. congruence.
              }
            
          - clear H3.  
            constructor; simpl.
            + auto.
            + destruct_override.
              * eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * congruence.
            + introv H7 H8.
              destruct_override.
              * eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * auto.
        }

      - { introv H4 H5 H6.

          assert (H3': hash (keyArg m0) mod proj1_sig s1ls <> hash (keyArg m) mod proj1_sig s1ls).
          - intro H7.
            unfold PPRel in H3.
            specialize (H3 (hash (keyArg m0) mod proj1_sig s1ls)). destruct H3 as [H3 | H3].
            * simplFlattenIn H3. inverts H3. simpl in *. congruence.
            * inverts H3 as H3. inverts H3 as H3. simpl in H3. congruence.
            
          - clear H3.  
            constructor; simpl.
            + auto.
            + destruct_override.
              * eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * congruence.
            + introv H7 H8.
              destruct_override.
              * eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * auto.
        }

      - { introv H4 H5 H6.

          assert (H3': hash (keyArg m0) mod proj1_sig s1ls <> hash (keyArg m) mod proj1_sig s1ls).
          - intro H7.
            unfold PPRel in H3.
            specialize (H3 (hash (keyArg m0) mod proj1_sig s1ls)). destruct H3 as [H3 | H3].
            * simplFlattenIn H3. inverts H3. simpl in *. congruence.
            * inverts H3 as H3. simpl in H3. congruence.
            
          - clear H3.  
            constructor; simpl.
            + auto.
            + destruct_override.
              * eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * congruence.
            + introv H7 H8.
              destruct_override.
              * eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * auto.
        }

      - constructor.
      - constructor.
    Qed.          



    
    Lemma PPRelPres2:
      forall (R' R : Type) (s1ls: { n: nat | 0 < n }) (s1bs: { n: nat | 0 < n })
        (r : R') (m : EMapMethod Key Val R) (i : ObjInstruction (bucketMapImpl Val) R')
                 (f : R' -> Program (ObjInstruction (bucketMapImpl Val)) R)
                 (R'' : Type) (p1t : Program RBaseMethod R'') (p2t : Program (EMapMethod Key Val) R''),
      
      PPRel
        (bindFlattenSeqImpl
           (Some (hash (keyArg m) mod proj1_sig s1ls))
           (dumNoopProgram
              (arrayLift (x <- i; f x)
                         (hash (keyArg m) mod proj1_sig s1bs)))
           (fun r : R =>
              _ <-
                annCallLocksArray
                (Some (hash (keyArg m) mod proj1_sig s1ls))
                (hash (keyArg m) mod proj1_sig s1ls) doUnlock; 
            Return r))
        (r <- m; Return r) p1t p2t ->

      PPRel
        (bindFlattenSeqImpl
           (Some (hash (keyArg m) mod proj1_sig s1ls))
           (dumNoopProgram
              (mapProgram
                 (fun (A : Type) (bm : ObjInstruction (bucketMapImpl Val) A) =>
                    arrayCall
                      (hash (keyArg m) mod proj1_sig s1bs)
                      bm)
                 (f r)))
           (fun r : R =>
              (_ <- annCallLocksArray
                 (Some (hash (keyArg m) mod proj1_sig s1ls))
                 (hash (keyArg m) mod proj1_sig s1ls)
                 doUnlock; 
               Return r)))
        (r <- m; Return r)
        p1t p2t.
    
    Proof.
      introv H1.
      unfold PPRel in *.
      intro i'. specialize (H1 i').
      destruct H1 as [H1 | H1].
      - simplFlattenIn H1. inverts H1. simpl in *. left. apply OutCriticalHelper1. assumption.
      - right. assumption.
    Qed.

    Lemma PPRelPres2':
      forall (R : Type) (s1ls: { n: nat | 0 < n }) (s1bs: { n: nat | 0 < n })
        (m : EMapMethod Key Val R)
        (p : Program (ObjInstruction (bucketMapImpl Val)) R)
        (R'' : Type) (p1t : Program RBaseMethod R'') (p2t : Program (EMapMethod Key Val) R''),
      
      PPRel
        (bindFlattenSeqImpl
           (Some (hash (keyArg m) mod proj1_sig s1ls))
           (dumNoopProgram
              (arrayLift (Noop p)
                         (hash (keyArg m) mod proj1_sig s1bs)))
           (fun r : R =>
              _ <-
                annCallLocksArray
                (Some (hash (keyArg m) mod proj1_sig s1ls))
                (hash (keyArg m) mod proj1_sig s1ls) doUnlock; 
            Return r))
        (r <- m; Return r) p1t p2t ->

      PPRel
        (bindFlattenSeqImpl
           (Some (hash (keyArg m) mod proj1_sig s1ls))
           (dumNoopProgram
              (mapProgram
                 (fun (A : Type) (bm : ObjInstruction (bucketMapImpl Val) A) =>
                    arrayCall
                      (hash (keyArg m) mod proj1_sig s1bs)
                      bm)
                 p))
           (fun r : R =>
              (_ <- annCallLocksArray
                 (Some (hash (keyArg m) mod proj1_sig s1ls))
                 (hash (keyArg m) mod proj1_sig s1ls)
                 doUnlock; 
               Return r)))
        (r <- m; Return r)
        p1t p2t.
    
    Proof.
      introv H1.
      unfold PPRel in *.
      intro i'. specialize (H1 i').
      destruct H1 as [H1 | H1].
      - simplFlattenIn H1. inverts H1. simpl in *. left. apply OutCriticalHelper1. assumption.
      - right. assumption.
    Qed.


    Lemma PDRelPres3:
      forall (R : Type) (scls : {n : nat | 0 < n})
        (scba : nat -> State (objBase (bucketMapImpl Val)))
        (scbs : {n : nat | 0 < n}) (t' : DumState) m
        (r0 : R) (scla : nat -> atomicLock.(State))
        (saa : nat -> State (objBase (bucketMapImpl Val)))
        (sas : {n : nat | 0 < n}) (o' : State atomicLock) 
        (R0 : Type) (p1t : Program RBaseMethod R0) (p2t : Program (EMapMethod Key Val) R0),
        PDRel
          p1t
          ({| arrayMapState := scla; arraySizeState := scls |},
           ({| arrayMapState := scba; arraySizeState := scbs |}, t'))
          p2t
          {| arrayMapState := saa; arraySizeState := sas |} ->
        proj1_sig scbs mod proj1_sig scls = 0 ->        
        PPRel
          (_ <- rCallLocksArray (Some (hash (keyArg m) mod proj1_sig scls)) (hash (keyArg m) mod proj1_sig scls) doUnlock; 
           Noop (Return r0))
          (r <- m; Return r)
          p1t p2t ->
        PDRel
          p1t
          ({| arrayMapState := noverride scla (hash (keyArg m) mod proj1_sig scls) o';
              arraySizeState := scls |},
           ({| arrayMapState := scba; arraySizeState := scbs |}, t')) p2t
          {| arrayMapState := noverride saa (hash (keyArg m) mod proj1_sig scbs) (scba (hash (keyArg m) mod proj1_sig scbs));
             arraySizeState := scbs |}.

    Proof.
      introv H1 H2 H3.
      inverts H1 as.
      
      + constructor.
      + constructor.
      + constructor.
      + constructor.
      + constructor.
      + { introv H4 H5 H6.
          assert (H6': hash (keyArg m) mod proj1_sig scls <> hash (keyArg m0) mod proj1_sig scls).
          - intro H7.
            unfold PPRel in H3.
            specialize (H3 (hash (keyArg m) mod proj1_sig scls)). destruct H3 as [H3 | H3].
            + inverts H3. simpl in *. congruence.
            + inverts H3 as H3. simplFlattenIn H3. inverts H3 as H3.
              simplFlattenIn H3. inverts H3 as H3. simpl in H3. congruence.

          - clear H3.
            constructor; simpl.
            + simpl_override. assumption.
            + destruct_override.
              * congruence.
              * congruence.
            + introv H7 H8.
              destruct_override.
              * congruence.
              * auto.
        }

      + { introv H4 H5 H6.
          assert (H6': hash (keyArg m) mod proj1_sig scls <> hash (keyArg m0) mod proj1_sig scls).
          - intro H7.
            unfold PPRel in H3.
            specialize (H3 (hash (keyArg m) mod proj1_sig scls)). destruct H3 as [H3 | H3].
            + inverts H3. simpl in *. congruence.
            + inverts H3 as H3. inverts H3 as H3. simpl in H3. congruence.

          - clear H3.  
            constructor; simpl.
            + simpl_override. assumption.
            + destruct_override.
              * congruence.
              * congruence.
            + introv H7 H8.
              destruct_override.
              * congruence.
              * auto.
        }

      + { introv H4 H5 H6.
          assert (H3': hash (keyArg m) mod proj1_sig scls <> hash (keyArg m0) mod proj1_sig scls).
          - intro H7.
            unfold PPRel in H3.
            specialize (H3 (hash (keyArg m) mod proj1_sig scls)). destruct H3 as [H3 | H3].
            + inverts H3. simpl in *. congruence.
            + inverts H3 as H3. simpl in H3. congruence.

          - clear H3.  
            constructor; simpl.
            + simpl_override. assumption.
            + destruct_override.
              * congruence.
              * congruence.
            + introv H7 H8.
              destruct_override.
              * congruence.
              * auto.
        }

      + { introv H4 H5 H6.
          assert (H3': hash (keyArg m) mod proj1_sig scls <> hash (keyArg m0) mod proj1_sig scls).
          - intro H7.
            unfold PPRel in H3.
            specialize (H3 (hash (keyArg m) mod proj1_sig scls)). destruct H3 as [H3 | H3].
            + inverts H3. simpl in *. congruence.
            + inverts H3 as H3. simplFlattenIn H3.
              inverts H3 as H3. simplFlattenIn H3.
              destruct (objProgram (bucketMapImpl Val) R0 m0).
              * inverts H3 as H3.
                simpl in *. congruence.
              * inverts H3 as H3. progFrobSimplIn H3.
                inverts H3 as H3. progFrobSimplIn H3.
                inverts H3 as H3.
                simpl in *. congruence. 
              * inverts H3 as H3.
                simpl in *. congruence.

          - clear H3.  
            constructor; simpl.
            + simpl_override. assumption.
            + destruct_override.
              * congruence.
              * congruence.
            + introv H7 H8.
              destruct_override.
              * congruence.
              * auto.                
        }


      + { introv H4 H5 H6.
          assert (H3': hash (keyArg m) mod proj1_sig scls <> hash (keyArg m0) mod proj1_sig scls).
          - intro H7.
            unfold PPRel in H3.
            specialize (H3 (hash (keyArg m) mod proj1_sig scls)). destruct H3 as [H3 | H3].
            + inverts H3. simpl in *. congruence.
            + inverts H3 as H3. simplFlattenIn H3.
              destruct (objProgram (bucketMapImpl Val) R0 m0).
              * inverts H3 as H3.
                simpl in *. congruence.
              * inverts H3 as H3. progFrobSimplIn H3.
                inverts H3 as H3. progFrobSimplIn H3.
                inverts H3 as H3.
                simpl in *. congruence. 
              * inverts H3 as H3.
                simpl in *. congruence.

          - clear H3.  
            constructor; simpl.
            + simpl_override. assumption.
            + destruct_override.
              * congruence.
              * congruence.
            + introv H7 H8.
              destruct_override.
              * congruence.
              * auto.                
        }

      + { introv H4 H5 H6.
          assert (H3': hash (keyArg m) mod proj1_sig scls <> hash (keyArg m0) mod proj1_sig scls).
          - intro H7.
            unfold PPRel in H3.
            specialize (H3 (hash (keyArg m) mod proj1_sig scls)). destruct H3 as [H3 | H3].
            + inverts H3. simpl in *. congruence.
            + { destruct pc; simplFlattenIn H3; inverts H3 as H3.
                - simpl in *. congruence.
                - progFrobSimplIn H3. inverts H3 as H3.
                  progFrobSimplIn H3. inverts H3 as H3.
                  simpl in *. congruence.
                - simpl in *. congruence.
              }
              
          - clear H3.  
            constructor; simpl.
            + simpl_override. assumption.
            + destruct_override.
              * exfalso. eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * congruence.
            + introv H7 H8.
              destruct_override.
              * congruence.
              * auto.                
        }

      + { introv H4 H5 H6.
          assert (H3': hash (keyArg m) mod proj1_sig scls <> hash (keyArg m0) mod proj1_sig scls).
          - intro H7.
            unfold PPRel in H3.
            specialize (H3 (hash (keyArg m) mod proj1_sig scls)). destruct H3 as [H3 | H3].
            + inverts H3. simpl in *. congruence.
            + inverts H3 as H3. inverts H3 as H3. simpl in H3. congruence.
              
          - clear H3.  
            constructor; simpl.
            + simpl_override. assumption.
            + destruct_override.
              * exfalso. eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * congruence.
            + introv H7 H8.
              destruct_override.
              * congruence.
              * auto.                
        }

      + { introv H4 H5 H6.
          assert (H3': hash (keyArg m) mod proj1_sig scls <> hash (keyArg m0) mod proj1_sig scls).
          - intro H7.
            unfold PPRel in H3.
            specialize (H3 (hash (keyArg m) mod proj1_sig scls)). destruct H3 as [H3 | H3].
            + inverts H3. simpl in *. congruence.
            + inverts H3 as H3. simpl in H3. congruence.
              
          - clear H3.  
            constructor; simpl.
            + simpl_override. assumption.
            + destruct_override.
              * exfalso. eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
              * congruence.
            + introv H7 H8.
              destruct_override.
              * congruence.
              * auto.                
        }
        
      + constructor.
      + constructor.

    Qed.


    Lemma PPRelPres3:
      forall (R : Type) (sc'ls : {n : nat | 0 < n}) (m : EMapMethod Key Val R) 
        (r : R) (R : Type) (p1t : Program RBaseMethod R)
        (p2t : Program (EMapMethod Key Val) R),
        PPRel
          (_ <- rCallLocksArray (Some (hash (keyArg m) mod proj1_sig sc'ls))
             (hash (keyArg m) mod proj1_sig sc'ls) doUnlock; 
           Noop (Return r))
          (r <- m; Return r) p1t p2t ->
        PPRel
          (Noop (Return r)) (Return r)
          p1t p2t.
    Proof.
      introv H1.
      unfold PPRel in *.
      intros.
      specialize (H1 i). destruct H1.
      - left. constructor. constructor.
      - right. assumption.
    Qed.


    Lemma stepRel:
      forall R' m R (f: R' -> Program RBaseMethod R) s1 p2 s2 s1' r,
        DRel s1 s2 ->
        PDRel (Bind m f) s1 p2 s2 ->
        rBase.(EvalInstruction) m s1 r s1' ->
        exists p2' s2',
          ProgramSteps (aImpl.(objBase)) _ p2 s2 p2' s2'
          /\ DRel s1' s2'                                      
          /\ PDRel (f r) s1' p2' s2'
          /\ (forall R (p1t: Program RBaseMethod R) p2t,
                PDRel p1t s1 p2t s2 ->
                PPRel (Bind m f) p2 p1t p2t (* s1 s2 *) ->
                PDRel p1t s1' p2t s2'
                /\ PPRel (f r) p2' p1t p2t (* s1' s2' *)).

    Proof.
      introv H1 H2 H3.
      inverts H2 as; subst.

      (* Getting the locks array size. *)
      (* 
       * Getting the size does not change the state. 
       * We take no action on the abstract. The conclusions are already available in the assumptions. 
       *)
      - { simpl in H3. destruct H3 as (H31 & H32). inverts H31 as H31.
          destruct s1 as ((s1la & s1ls) & ((s1ba & s1bs) & t)).
          destruct s1' as ((s1la' & s1ls') & ((s1ba' & s1bs') & t')).
          simpl in *. inverts H31.  inverts H32.
          
          
          eexists. eexists.
          simpl. splits.

          - eapply ProgramStepsRefl.

          - assumption.
            
          - constructor.
             
          - introv H4 H5.
            split.
            + assumption.
            + unfold PPRel in *.
              intros. specialize (H5 i).
              destruct H5 as [H5 | H5].
              * left. constructor. simplFlatten. constructor.
                match goal with
                | [ |- context [bindFlattenProgram _ _ (mapProgram ?p ?p')]] =>
                  rewrite (frobProgramId (mapProgram p p')); simpl
                end.
                simplFlatten.
                constructor.
                simpl. intro. inverts H.
              * right. assumption.
        }


      (* Aquiring the lock. *)
      - { simpl in H3. destruct H3 as (H31 & H32).
          destruct s1' as (sc'l & (sc'b & t')). destruct scb as (scba & scbs). simpl in *. inverts H32.
          destruct s2.
          inverts H31 as.

          - introv H311 H312.
            inverts H312 as H312. symmetry in H312.
          
            eexists. eexists.
            splits.

            + eapply ProgramStepsRefl.

            + unfold DRel in *. destruct H1 as (H11 & H12 & H13). simpl in *. splits.
              * assumption.      
              * assumption.
              * introv H4. destruct_override_in H4; eauto.

            + constructor; simpl in *.
              * simpl_override. reflexivity.
              * unfold DRel in H1. simpl in *. destruct H1 as (H11 & (H12 & H13)). subst arraySizeState.
                eapply H13. congruence.
              * unfold DRel in H1. simpl in *. destruct H1 as (H11 & (H12 & H13)). subst arraySizeState.
                introv H4 H5.
                eapply H13. congruence.
                
            + introv H4 H5.
              split.
              * eapply PDRelPres1; simpl; assumption.                
              * eapply PPRelPres1; eassumption.

          - introv H4. exfalso. simpl in *.
            eapply noOverflow. eassumption.
        }

      (* Getting the buckets array size. *)
      - { introv H4 H5.
          inverts H3 as. introv H31 H32. simpl in *.
          destruct H31 as (H311 & H312). 
          destruct s1' as (s1l' & ((s1'ba & a1'bs) & t')). simpl in *.
          inverts H311. subst.
          
          eexists. eexists.
          splits.

          - eapply ProgramStepsRefl.

          - assumption.
            
          - constructor; assumption.
             
          - introv H6 H7. splits.
            + assumption.
            + unfold PPRel in *.
              intros. specialize (H7 i).
              destruct H7 as [H7 | H7].
              * left. inverts H7 as H7. simpl in H7. constructor. eapply OutCriticalHelper0. assumption. 
                intros. simplFlatten. constructor. simplFlatten. constructor. assumption.
              * right. assumption.
        }

        
      (* A call on the seq impl. *)
      - { introv H4 H5 H6 H7.

          destruct s1' as ((sc'la & sc'ls) & ((sc'ba & sc'bs) & t')). simpl in *.

          destruct pc; simplFlattenIn H7; inverts H7; simpl in *.          
          - destruct H3 as ((H31 & H32) & H33). inverts H33. subst. inverts H31 as.

            + introv H311 H312.
              
              eexists. eexists. splits.

              * eapply ProgramStepsRefl.

              * { unfold DRel in *. simpl in *. destruct H1 as (H11 & H12 & H13). splits.
                  - assumption.
                  - assumption.
                  - intros. destruct_override.
                    + exfalso.
                      eapply lockSizeBucketSizeModEq in e; try eassumption. congruence.
                    + auto.
                }

              * { econstructor; simpl in *.
                  - assumption.
                  - eapply ProgramStepsStepImplSteps. eassumption. constructor. simpl_override. assumption.
                  - intros. destruct_override.
                    + exfalso. congruence.
                    + auto.
                }

                
              * { introv H7 H8. simpl in *.
                  split.

                  - unfold DRel in H1. simpl in H1. destruct H1 as (_ & H1 & _).
                    applys PDRelPres2 H8; eassumption.
                  - eapply PPRelPres2; eassumption.
                }

            + introv H7. exfalso. simpl in *.
              eapply noOverflow. eassumption.
                          
          - destruct H3 as ((_ & H32) & H33). inverts H32. inverts H33.
            
            eexists. eexists. splits.

            + eapply ProgramStepsRefl.

            + unfold DRel in *. simpl in *. destruct H1 as (H11 & H12 & H13). splits; assumption.

            + { econstructor; simpl in *.
                - assumption.
                - eapply ProgramStepsStepImplSteps. eassumption. constructor.
                - assumption.
              }
              
            + { introv H7 H8. simpl in *. destruct t'. destruct t.
                split. assumption. 
                - unfold DRel in H1. simpl in H1. destruct H1 as (_ & H1 & _).
                  eapply PPRelPres2'; eassumption.
              }
        }

      (* The unlock. *)
      - { introv H4 H5.
          eexists. eexists. splits.

          - unfold DRel in H1. simpl in *. destruct H1 as (H1 & _). subst.
            
            eapply ProgramStepImplProgramSteps.
            constructor.
            unfold arrayState in *.

                          
            simpl.
            unfold SHashMap.impl.
            unfold EvalProgram.
            progFrobSimpl.
            progFrobSimplAt 2.
            progFrobSimplAt 3.
            eapply ProgramStepStepsImplSteps.
            econstructor.
            eapply ProgramStepStepsImplSteps.
            econstructor.
            econstructor.
            eapply ProgramStepStepsImplSteps.
            econstructor.
            simpl.

            progFrobSimplAt 1.
            eapply ProgramStepStepsImplSteps.
            econstructor.
            simpl.

            eapply EvalProgramBindFlattenRet.

            apply EvalProgramArrayLift. eapply noOverflow'.

            unfold EvalProgram.
            eassumption.

          - unfold DRel in *. simpl in *. destruct H1 as (H11 & (H12 & H13)). subst.
            destruct s1' as ((sc'la & sc'ls) & ((sc'ba & sc'bs) & t')). simpl in *.
            destruct H3 as (H31 & H32). inverts H32. inverts H31 as; simpl.
            + introv H6 H7. inverts H7 as H7. symmetry in H7.
              splits.
              * reflexivity.
              * assumption.
              * { intro.
                  destruct_override; intros.
                  - destruct_override.
                    + f_equal.
                      symmetry.
                      assumption.
                    + auto.
                  - destruct_override.
                    + auto.
                    + auto.
                }
            + introv H6.
              exfalso.
              eapply noOverflow; eassumption.

          - econstructor. 

          - { simpl in *.
              destruct s1' as ((sc'la & sc'ls) & ((sc'ba & sc'bs) & t')). simpl in *.
              destruct H3 as (H31 & H32). inverts H32. inverts H31 as; simpl.
              - unfold DRel in H1. simpl in *. destruct H1 as (H11 & (H12 & _)). subst sas.
                + introv H6 H7 H8 H9.
                  inverts H7.
                  splits.
                  * eapply PDRelPres3; eassumption. 
                  * eapply PPRelPres3; eassumption.
              - introv H6. exfalso. simpl in *.
                eapply noOverflow. eassumption.
            }

        }

    Qed.

    Lemma AnnCHashMapConcRefSeqMap: ConcSimulates cImpl aImpl.
    Proof.
      intros.
      eapply SimulatesCoInd.simulatesCoInd with (DRel := DRel); simpl.
      - eapply PPRelSym.
      - exact initRel.
      - eapply noopRel.
      - eapply invRel.
      - eapply retRel.
      - eapply stepRel.
    Qed.

    Definition cImpl' :=
      object
        rBase
        (flattenAnnImplProgRight
          (* (objProgram (dumNoopImpl (arrayImpl bucketMapImpl))) *)
          (dumNoopImplProg (objProgram (arrayImpl (bucketMapImpl Val) nBuckets)))
          (distPairAnnImplProg (@annImpl))).

    Lemma CHashMapConcRefAnnCHashMap : ConcSimulates (cHashMapImpl Val nLocks nBuckets) cImpl.
    Proof.
      intros.
      eapply (concRefTrans _ cImpl').
      2: eapply concRefRefl.
      unfold cImpl'. unfold object. unfold rBase. unfold rLocks.

      unfold flattenAnnImplProgRight.
      unfold flattenAnnProgramRight.
      unfold flattenProgramRight.

      unfold flattenImplProgRight.
      unfold flattenProgramRight.

      eapply concRefTrans.
      * applys flattenDumNoopConcRefPair.
      * { unfold object.
          eapply concRefTrans.
          - forwards H: @flattenImplProgConcRefFlattenAnnImplProg
                          (@annImpl) (@impl Val). 2: eassumption.
            unfold impl. unfold annImpl. unfold eraseImpl.
            eapply functional_extensionality_dep. intro R. eapply functional_extensionality. intro m.
            progress repeat (progFrobSimpl; f_equal; eapply functional_extensionality; intros).
          - unfold object.
            rewrite <- annImplProgSerialImplProg.
            apply flattenImplProgConRefFlattenImplProgDist.
        }
    Qed.

    End CHashMapLinFun1.

    Module SHashMapSeqRef := SHashMap.SHashMapSeqRefFun SHashMapSeqRefArgs.
    Import SHashMapSeqRef.
    Import SHashMapSeqRefArgs.EMap.

    Lemma CHashMapSHashMap Val nL nS
      : ConcSimulates (cHashMapImpl Val nL (posMult nL nS)) (@aImpl Val nL nS).
    Proof.
      eapply concRefTrans.
      - exact (CHashMapConcRefAnnCHashMap _ _ _).
      - eapply AnnCHashMapConcRefSeqMap.
    Qed.

    Lemma CHashMapLin Val nL nS:
      Linearizable (atomicEMap Val) (cHashMapImpl Val nL (posMult nL nS)).
    Proof.
      intros.
      unfold Linearizable.
      eapply concRefTrans.
      - apply CHashMapSHashMap.
      - eapply SimulatesSerImplCoInd.SeqRefImplConRefIO.
        eapply sHashMapSeqRefEMap.
    Qed.

    (* Print Assumptions flattenImplProgConcRefFlattenAnnImplProg. *)
(*
Axioms:

  Functional extensionality and three (equivalent) versions of an equivalence
  between heterogeneous and homogeneous equality.

      functional_extensionality_dep : forall (A : Type) (B : A -> Type) (f g : forall x : A, B x),
                  (forall x : A, f x = g x) -> f = g
      Eqdep.Eq_rect_eq.eq_rect_eq : forall (U : Type) (p : U)
                                      (Q : U -> Type) (x : Q p) (h : p = p),
                   x = eq_rect p Q x p h
      JMeq.JMeq_eq : forall (A : Type) (x y : A), JMeq.JMeq x y -> x = y

  Module functor parameters (i.e., arguments of this construction and theorem, not actually axioms)
  - A type of keys [E.t] with decidable equality
  - An object of individual buckets and its simulation proof wrt.
    an [atomicEMap] specification

      E.t : Type
      E.hash : E.t -> nat
      E.eq_dec : forall x y : E.t, {x = y} + {x <> y}
      E.bucketMapImpl : forall Val : Type, Object (EMapMethod E.t Val)
      E.initBucketMapObjState : forall Val : Type,
                                 State (objBase (E.bucketMapImpl Val))
      SHashMapSeqRefArgs.bucketMapImplSeqRefAbsMap : forall Val : Type,
        Simulates.Simulates (coarsen (SHashMap.bucketMapImpl Val))
                            (atomicEMap Val)
                            (SHashMap.initBucketMapObjState Val)
                            (initEMapState Val)
*)

  End CHashMapLinFun.

End CHashMapFun.
