From Coq Require Extraction.
Require Import
        Coq.Sets.Ensembles
        Coq.Structures.Equalities.

From C4.Util Require Import
  FunOverride.

From C4.Core Require Import
  Program.

Module BMap.

  Inductive BMapMethod {Key}{Val}: Type -> Type :=
  | get: Key -> BMapMethod (option Val)
  | put: Key -> Val -> BMapMethod unit.

  Arguments BMapMethod : clear implicits.

  Definition keyArg {K V R}(m: BMapMethod K V R): K :=
    match m with
    | get k => k
    | put k v => k
    end.
  
  Inductive BMapSem' {Key eq_key_dec Val} :
    forall {A}, BMapMethod Key Val A -> (Key -> option Val) -> A -> (Key -> option Val) -> Prop :=
  | getSem map k:
      BMapSem' (get k) map (map k) map
  | putSem map k v:
      BMapSem' (put k v)
               map
               tt
               (override eq_key_dec map k (Some v)).

  Arguments BMapSem' : clear implicits.

  Module Type BMapFunType (E : MiniDecidableType).
    Definition Key := E.t.
    Definition eq_key_dec := E.eq_dec.

    Definition BMapState Val :=
      Key -> option Val.

    Definition BMapSem {Val} :
      forall {A}, BMapMethod Key Val A -> BMapState Val -> A -> BMapState Val -> Prop :=
      BMapSem' Key eq_key_dec Val.

    Arguments BMapSem : clear implicits.

    Definition atomicBMap Val : Spec (BMapMethod Key Val) :=
      Build_Spec (BMapMethod Key Val) (BMapState Val) (BMapSem Val) (fun _ => None).

    Parameter invertBMapSem : forall Val R m s r s',
      BMapSem Val R m s r s' ->

      match m in BMapMethod _ _ R' return R' -> Prop with

      | get k =>
        fun r =>
          r = s k
          /\ s = s'

      | put k v =>
        fun r =>
          r = tt
          /\ s' = (override eq_key_dec s k (Some v))

      end r.
  End BMapFunType.

  Module BMapFun (E : MiniDecidableType) <: BMapFunType E.

    Definition Key := E.t.
    Definition eq_key_dec := E.eq_dec.

    Definition BMapState Val :=
      Key -> option Val.

    Definition BMapSem {Val} :
      forall {A}, BMapMethod Key Val A -> BMapState Val -> A -> BMapState Val -> Prop :=
      BMapSem' Key eq_key_dec Val.

    Arguments BMapSem : clear implicits.

    Definition atomicBMap Val : Spec (BMapMethod Key Val) :=
      Build_Spec (BMapMethod Key Val) (BMapState Val) (BMapSem Val) (fun _ => None).

    Lemma invertBMapSem Val R m s r s':
      BMapSem Val R m s r s' ->

      match m in BMapMethod _ _ R' return R' -> Prop with

      | get k =>
        fun r =>
          r = s k
          /\ s = s'

      | put k v =>
        fun r =>
          r = tt
          /\ s' = (override eq_key_dec s k (Some v))

      end r.

    Proof.
      intros H1.
      destruct H1; auto.
    Qed.

  End BMapFun.

End BMap.


Module Map.
  Inductive MapMethod {Key}{Val}: Type -> Type :=
  | get: Key -> MapMethod (option Val)
  | put: Key -> Val -> MapMethod (option Val)
  | remove: Key -> MapMethod (option Val).

  Arguments MapMethod : clear implicits.

  Module Type MapFunType (E : MiniDecidableType).

    Definition Key := E.t.
    Definition eq_key_dec := E.eq_dec.

    Definition MapState Val :=
      Key -> option Val.

    Inductive MapSem {Val} :
      forall {A}, (MapMethod Key Val A) -> MapState Val -> A -> MapState Val -> Prop :=
    | getSem: forall map k,
        MapSem (get k) map (map k) map
    | putSem: forall map k v,
        MapSem (put k v) map (map k) (override eq_key_dec map k (Some v))
    | removeSem: forall map k,
        MapSem (remove k) map (map k) (override eq_key_dec map k None).

    Arguments MapSem : clear implicits.

    Definition atomicMap Val : Spec (MapMethod Key Val) :=
      Build_Spec (MapMethod Key Val) (MapState Val) (MapSem Val) (fun _ => None).

    Parameter invertMapSem : forall Val R m s r s',
      MapSem Val R m s r s' ->

      match m in MapMethod _ _ R' return R' -> Prop with

      | get k =>
        fun r =>
          r = s k
          /\ s = s'

      | put k v =>
        fun r =>
          r = s k
          /\ s' = (override eq_key_dec s k (Some v))

      | remove k =>
        fun r =>
          r = s k
          /\ s' = (override eq_key_dec s k None)
      end r.
  End MapFunType.

  Module MapFun (E : MiniDecidableType) <: MapFunType E.

    Definition Key := E.t.
    Definition eq_key_dec := E.eq_dec.

    Definition MapState Val :=
      Key -> option Val.

    Inductive MapSem {Val} :
      forall {A}, (MapMethod Key Val A) -> MapState Val -> A -> MapState Val -> Prop :=
    | getSem: forall map k,
        MapSem (get k) map (map k) map
    | putSem: forall map k v,
        MapSem (put k v) map (map k) (override eq_key_dec map k (Some v))
    | removeSem: forall map k,
        MapSem (remove k) map (map k) (override eq_key_dec map k None).

    Arguments MapSem : clear implicits.

    Definition atomicMap Val : Spec (MapMethod Key Val) :=
      Build_Spec (MapMethod Key Val) (MapState Val) (MapSem Val) (fun _ => None).

    Lemma invertMapSem Val R m s r s':
      MapSem Val R m s r s' ->

      match m in MapMethod _ _ R' return R' -> Prop with

      | get k =>
        fun r =>
          r = s k
          /\ s = s'

      | put k v =>
        fun r =>
          r = s k
          /\ s' = (override eq_key_dec s k (Some v))

      | remove k =>
        fun r =>
          r = s k
          /\ s' = (override eq_key_dec s k None)
      end r.

    Proof.
      intros H1.
      destruct H1; auto.
    Qed.
  End MapFun.
End Map.

Module EMap.

  Inductive EMapMethod {Key}{Val}: Type -> Type :=
  | get: Key -> EMapMethod (option Val)
  | put: Key -> Val -> EMapMethod (option Val)
  | putIfAbsent: Key -> Val -> EMapMethod (option Val)
  | remove: Key -> EMapMethod (option Val)
  | replaceIt: Key -> Val -> Val -> EMapMethod bool.
  (* | keys: EMapMethod (list (Key * Val)). *)

  Extract Inductive EMapMethod => "MapS" [ "Get" "Put" "PutIfAbsent" "Remove" "ReplaceIt" ].

  Arguments EMapMethod : clear implicits.

  Definition keyArg {K V R}(m: EMapMethod K V R): K :=
    match m with
    | get k => k
    | put k v => k
    | putIfAbsent k v => k
    | remove k => k
    | replaceIt k v v' => k
    end.

  Module Type EMapFunType (E : MiniDecidableType).

    Definition Key := E.t.
    Definition eq_key_dec := E.eq_dec.

    Definition EMapState Val :=
      Key -> option Val.

    Inductive EMapSem {Val} :
      forall {A}, EMapMethod Key Val A -> EMapState Val -> A -> EMapState Val -> Prop :=
    | getSem: forall map k,
        EMapSem (get k) map (map k) map
    | putSem: forall map k v,
        EMapSem (put k v) map (map k) (override eq_key_dec map k (Some v))
    | putIfAbsentSemSucc: forall map k v,
        map k = None ->
        EMapSem (putIfAbsent k v) map None (override eq_key_dec map k (Some v))
    | putIfAbsentSemFail: forall map k v,
        map k <> None ->
        EMapSem (putIfAbsent k v) map (map k) map
    | removeSem: forall map k,
        EMapSem (remove k) map (map k) (override eq_key_dec map k None)
    | replaceSemSucc: forall map k v v',
        map k = Some v ->
        EMapSem (replaceIt k v v') map true (override eq_key_dec map k (Some v'))
    | replaceSemFail: forall map k v v',
        map k <> Some v ->
        EMapSem (replaceIt k v v') map false map.

    Arguments EMapSem : clear implicits.

    (* | eKeysSem: forall map,
      EMapSem map eKeys map nil. (* Todo: Should change the state to finite map. *) *)

    Definition atomicEMap Val : Spec (EMapMethod Key Val) :=
      Build_Spec (EMapMethod Key Val) (EMapState Val) (EMapSem Val) (fun _ => None).

    Parameter invertEMapSem : forall Val R m s r s',
      EMapSem Val R m s r s' ->

      match m in EMapMethod _ _ R' return R' -> Prop with

      | get k =>
        fun r =>
          r = s k
          /\ s = s'

      | put k v =>
        fun r =>
          r = s k
          /\ s' = (override eq_key_dec s k (Some v))

      | putIfAbsent k v =>
        fun r =>
          (r = s k /\ (s k) = None /\ s' = (override eq_key_dec s k (Some v))) \/
          (r = s k /\ (s k) <> None /\ s' = s)

      | remove k =>
        fun r =>
          r = s k
          /\ s' = (override eq_key_dec s k None)

      | replaceIt k v v' =>
        fun r =>
          (r = true /\ (s k) = Some v /\ s' = override eq_key_dec s k (Some v'))
          \/ (r = false /\ (s k) <> Some v /\ s' = s)
      end r.
  End EMapFunType.

  Module EMapFun (E : MiniDecidableType) <: EMapFunType E.

    Definition Key := E.t.
    Definition eq_key_dec := E.eq_dec.

    Definition EMapState Val :=
      Key -> option Val.

    Inductive EMapSem {Val} :
      forall {A}, EMapMethod Key Val A -> EMapState Val -> A -> EMapState Val -> Prop :=
    | getSem: forall map k,
        EMapSem (get k) map (map k) map
    | putSem: forall map k v,
        EMapSem (put k v) map (map k) (override eq_key_dec map k (Some v))
    | putIfAbsentSemSucc: forall map k v,
        map k = None ->
        EMapSem (putIfAbsent k v) map None (override eq_key_dec map k (Some v))
    | putIfAbsentSemFail: forall map k v,
        map k <> None ->
        EMapSem (putIfAbsent k v) map (map k) map
    | removeSem: forall map k,
        EMapSem (remove k) map (map k) (override eq_key_dec map k None)
    | replaceSemSucc: forall map k v v',
        map k = Some v ->
        EMapSem (replaceIt k v v') map true (override eq_key_dec map k (Some v'))
    | replaceSemFail: forall map k v v',
        map k <> Some v ->
        EMapSem (replaceIt k v v') map false map.

    Arguments EMapSem : clear implicits.

    (* | eKeysSem: forall map,
      EMapSem map eKeys map nil. (* Todo: Should change the state to finite map. *) *)

    Definition atomicEMap Val : Spec (EMapMethod Key Val) :=
      Build_Spec (EMapMethod Key Val) (EMapState Val) (EMapSem Val) (fun _ => None).

    Lemma invertEMapSem Val R m s r s':
      EMapSem Val R m s r s' ->

      match m in EMapMethod _ _ R' return R' -> Prop with

      | get k =>
        fun r =>
          r = s k
          /\ s = s'

      | put k v =>
        fun r =>
          r = s k
          /\ s' = (override eq_key_dec s k (Some v))

      | putIfAbsent k v =>
        fun r =>
          (r = s k /\ (s k) = None /\ s' = (override eq_key_dec s k (Some v))) \/
          (r = s k /\ (s k) <> None /\ s' = s)

      | remove k =>
        fun r =>
          r = s k
          /\ s' = (override eq_key_dec s k None)

      | replaceIt k v v' =>
        fun r =>
          (r = true /\ (s k) = Some v /\ s' = override eq_key_dec s k (Some v'))
          \/ (r = false /\ (s k) <> Some v /\ s' = s)
      end r.

    Proof.
      intros H1.
      destruct H1; auto.
    Qed.
  End EMapFun.

End EMap.
