From Coq Require Import
  Equalities Arith.

From C4.Util Require Import
  MathUtil
  FunOverride
  LibTactics.

From C4.Core Require Import
  Program
  ProgramFacts
  Object
  Pair
  Simulates
  SimulatesFacts
  SeqSimulatesFacts.

From C4.App Require Import
  Array
  Maps.

Require Import C4.App.Register. (* "Register" is a keyword. *)

Import Nat.
Import EMap.

Arguments arrayCall {M} {R} _ _.
Arguments getSize {M}.

Module Type HashableType.

  Include MiniDecidableType.

  (* The hash function. *)
  Parameter hash : t -> nat.

  (* The sequential map object that is used in each bucket. *)
  Parameter bucketMapImpl : forall Val, Object (EMapMethod t Val).
End HashableType.

(* Basic object of [HashableType] with [t := nat]. *)
Module SHashMapArgs <: HashableType.
  Definition t := nat.
  Definition eq_dec := eq_nat_dec.
  Definition hash := @id nat.

  Module EMapArgs <: MiniDecidableType.
    Definition t := nat.
    Definition eq_dec := eq_nat_dec.
  End EMapArgs.

  Module CEMap := EMapFun EMapArgs.
  Import CEMap.

  Definition bucketMapImpl (Val: Type): Object (EMapMethod t Val) :=
    serialObject (atomicEMap Val).
End SHashMapArgs.

Module SHashMapFun (E : HashableType).

  Definition Key := E.t.
  Definition eq_key_dec := E.eq_dec.
  Definition hash := E.hash.
  Definition bucketMapImpl := E.bucketMapImpl.

  Section SHashMapFun1.

  Context (Val : Type).

  Definition buckets :=
    arrayImpl (bucketMapImpl Val).

  Definition BaseState len :=
    (buckets len).(objBase).(State).

  Definition BaseMethod :=
    ArrayMethod (EMapMethod Key Val).
  Definition bucketsArrayCall {R} (i: nat) (m: EMapMethod Key Val R): BaseMethod R :=
    arrayCall i m.

  Definition impl {R} (m: EMapMethod Key Val R):
    Program BaseMethod R :=

    ls <- getSize;
    let k := keyArg m in
    let i := (hash k) mod ls in
    r <- bucketsArrayCall i m;
    Return r.

  Definition hashMapImpl len :=
    flattenObject
      (buckets len) _
      (@impl).

  End SHashMapFun1.

  (* -------------------------------------------------------- *)
  Module Type SHashMapSeqRefParams.

    Module EMap := EMapFun E.
    Import EMap.

    (* Bucket map object is a sequential Simulates of an abstract map. *)
    Parameter bucketMapImplSeqRefAbsMap :
      forall Val,
        Simulates
          (coarsen (bucketMapImpl Val)) (atomicEMap Val).

  End SHashMapSeqRefParams.


  Module SHashMapSeqRefFun (SHashMapSeqRefArgs: SHashMapSeqRefParams).

    Import SHashMapSeqRefArgs.
    Import EMap.

    Import SeqRefCoInd2.

    Section SHashMapSeqRefFun1.

    Context (Val : Type).

    Definition DRel (ast : ArrayState (atomicEMap Val)) (s : EMapState Val) : Prop :=
      forall i,
        arrayMapState ast (hash i mod (proj1_sig (arraySizeState ast))) i =
        s i.

    Definition initialArray len := {| arrayMapState := fun _ => initialState (atomicEMap Val) ; arraySizeState := len |}.

    Lemma initRel: forall len, DRel (initialArray len) (fun _ => None).
    Proof.
      intros. unfold DRel. intros. simpl. reflexivity.
    Qed.

    Lemma stepRel len:
      forall R (m: EMapMethod Key Val R)
        (s1 s1': ArrayState (atomicEMap Val)) r (s2: (EMapState Val)),
        DRel s1 s2 ->
        EvalProgram (absArray (atomicEMap Val) len) _ (impl Val m) s1 r s1' ->
        exists s2',
          EMapSem Val _ m s2 r s2'
          /\ DRel s1' s2'.

    Proof.
      introv H1 H2.
      destruct m.
      - unfold impl in *. simpl in *.
        apply EvalProgramInv in H2. destruct H2 as (s'' & r' & H21 & H22). inverts H21.
        apply EvalProgramInv in H22. destruct H22 as (s''' & r' & H221 & H222). inverts H221 as.
        + introv H3 H4. inverts H4.
          apply EvalProgramInv in H222. destruct H222 as (H2221 & H2222).
          destruct s'' as (sm & (ss & sp)). simpl in *. subst.
          unfold DRel in H1. simpl in *. rewrite H1.
          eexists.
          splits.
          * econstructor.
          * unfold DRel. simpl in *. intros. destruct_override. rewrite e. auto. auto.
        + introv H3. exfalso. apply noOverflow in H3. assumption.

      - unfold impl in *. simpl in *.
        apply EvalProgramInv in H2. destruct H2 as (s'' & r' & H21 & H22). inverts H21.
        apply EvalProgramInv in H22. destruct H22 as (s''' & r' & H221 & H222). inverts H221 as.
        + introv H3 H4. inverts H4.
          apply EvalProgramInv in H222. destruct H222 as (H2221 & H2222).
          destruct s'' as (sm & (ss & sp)). simpl in *. subst.
          unfold DRel in H1. simpl in *. rewrite H1.
          eexists.
          splits.
          * econstructor.
          * unfold DRel. intros. simpl in *. destruct_override.
            destruct_override. reflexivity. rewrite e. auto.
            destruct_override. subst. exfalso. apply n. reflexivity. auto.
        + introv H3. exfalso. apply noOverflow in H3. assumption.

      - unfold impl in *. simpl in *.
        apply EvalProgramInv in H2. destruct H2 as (s'' & r' & H21 & H22). inverts H21.
        apply EvalProgramInv in H22. destruct H22 as (s''' & r' & H221 & H222). inverts H221 as.
        + introv H3 H4. inverts H4.
          * { apply EvalProgramInv in H222. destruct H222 as (H2221 & H2222).
              destruct s'' as (sm & (ss & sp)). simpl in *. subst.
              unfold DRel in H1. simpl in *. (* rewrite H1; try assumption. *)
              eexists.
              splits.
              - econstructor. rewrite <- H6. symmetry. auto.
              - unfold DRel. intros. simpl in *. destruct_override.
                destruct_override. reflexivity. rewrite e. auto.
                destruct_override. subst. exfalso. apply n. reflexivity. auto.
            }
          * { apply EvalProgramInv in H222. destruct H222 as (H2221 & H2222).
              destruct s'' as (sm & (ss & sp)). simpl in *. subst.
              unfold DRel in H1. simpl in *. (* rewrite H1; try assumption. *)
              eexists.
              splits.
              - rewrite H1. econstructor. rewrite H1 in H6. assumption.
              - unfold DRel. intros. simpl in *. destruct_override.
                rewrite e. auto.
                auto.
            }
        + introv H3. exfalso. apply noOverflow in H3. assumption.

      - unfold impl in *. simpl in *.
        apply EvalProgramInv in H2. destruct H2 as (s'' & r' & H21 & H22). inverts H21.
        apply EvalProgramInv in H22. destruct H22 as (s''' & r' & H221 & H222). inverts H221 as.
        + introv H3 H4. inverts H4.
          apply EvalProgramInv in H222. destruct H222 as (H2221 & H2222).
          destruct s'' as (sm & (ss & sp)). simpl in *. subst.
          unfold DRel in H1. simpl in *. rewrite H1.
          eexists.
          splits.
          * econstructor.
          * unfold DRel. intros. simpl in *. destruct_override.
            destruct_override. reflexivity. rewrite e. auto.
            destruct_override. subst. exfalso. apply n. reflexivity. auto.
        + introv H3. exfalso. apply noOverflow in H3. assumption.


      - unfold impl in *. simpl in *.
        apply EvalProgramInv in H2. destruct H2 as (s'' & r' & H21 & H22). inverts H21.
        apply EvalProgramInv in H22. destruct H22 as (s''' & r' & H221 & H222). inverts H221 as.
        + introv H3 H4. inverts H4.
          * { apply EvalProgramInv in H222. destruct H222 as (H2221 & H2222).
              destruct s'' as (sm & (ss & sp)). simpl in *. subst.
              unfold DRel in H1. simpl in *. (* rewrite H1; try assumption. *)
              eexists.
              splits.
              - econstructor. rewrite <- H8. symmetry. auto.
              - unfold DRel. intros. simpl in *. destruct_override.
                destruct_override. reflexivity. rewrite e. auto.
                destruct_override. subst. exfalso. apply n. reflexivity. auto.
            }
          * { apply EvalProgramInv in H222. destruct H222 as (H2221 & H2222).
              destruct s'' as (sm & (ss & sp)). simpl in *. subst.
              unfold DRel in H1. simpl in *. (* rewrite H1; try assumption. *)
              eexists.
              splits.
              - econstructor. rewrite H1 in H8. assumption.
              - unfold DRel. intros. simpl in *. destruct_override.
                rewrite e. auto.
                auto.
            }
        + introv H3. exfalso. apply noOverflow in H3. assumption.

    Qed.

    Lemma sHashMapSeqRefEMap len :
        Simulates (coarsen (hashMapImpl Val len)) (atomicEMap Val).
    Proof.
      intros.
      unfold hashMapImpl.
      eapply simulatesTrans.
      - eapply seqRefSubst. unfold buckets.
        instantiate (1 := (absArray (atomicEMap Val) len)).
        eapply @implSeqRefObjImpliesArrayImplSeqRefAbsArray.
        eapply bucketMapImplSeqRefAbsMap.
      - eapply simulatesSerImplCoInd.
        eapply initRel.
        eapply stepRel.
    Qed.

    End SHashMapSeqRefFun1.

  End SHashMapSeqRefFun.

End SHashMapFun.
