From Coq Require Import
  Eqdep Setoid SetoidClass Morphisms FunctionalExtensionality Omega Equalities.

From Paco Require Import paco.

From C4.Util Require Import
  Util
  FunOverride
  MathUtil
  LibTactics
  MTactics.

From C4 Require Import
  Program
  ProgramFacts
  ProgramOps
  ProgramOpsFacts
  Object
  Pair
  App.Register
  Lock
  Array
  Maps
  Simulates
  SimulatesFacts
  Threads
  ThreadsFacts
  Abortable
  LinCoInd
  CoarsenSimulates
  Linearizability
  Serializability
  SerializabilityAnn2
  Annotated
  AnnotatedFacts2
  Dum2
  Wrap
  WrapFacts
  RSimulates Eutt.

Import BMap.
Import SimulatesCoIndN.

Set Implicit Arguments.
Set Contextual Implicit.

Arguments arrayCall {M} {R} _ _.
Arguments getSize {M}.
Notation "a =_? b" := (eq_nat_dec a b) (at level 20).
Notation "a <>_? b" := (negb (a =_? b)) (at level 20).

Lemma modTwoZeroOrOne:
  forall n,
    n mod 2 = 0 \/
    n mod 2 = 1.
Proof.
  intro.
  forwards A: Nat.mod_bound_pos n 2;
    omega.
Qed.

Lemma modTwoNotZeroOne:
  forall n,
    n mod 2 <> 0 ->
    n mod 2 = 1.
Proof.
  introv H.
  destruct (modTwoZeroOrOne n).
  - congruence.
  - assumption.
Qed.

Lemma modTwoNotOneZero:
  forall n,
    n mod 2 <> 1 ->
    n mod 2 = 0.
Proof.
  introv H.
  destruct (modTwoZeroOrOne n).
  - assumption.
  - congruence.
Qed.

Lemma modTwoAdd:
  forall n,
    n mod 2 = 0 ->
    (n+1) mod 2 = 1.

Proof.
  introv H1.
  forwards A: Nat.add_mod n 1 2.
  omega.
  rewrite A. rewrite H1.
  reflexivity.
Qed.

Lemma foldModTwo:
  forall n,
    match snd (Nat.divmod n 1 0 1) with
    | 0 => 1
    | S _ => 0
    end = n mod 2.
Proof.
  intros.
  reflexivity.
Qed.


(**)

CoInductive opTerm {M R}: Program M R -> Prop :=
| OpTermBind R' (m: M R') f:
    (forall r, opTerm (f r)) ->
    opTerm (Bind m f)

| OpTermReturn r:
    opTerm (Return r)

| OpTermNoop p:
    opTerm p ->
    (forall r, p <> Return r) ->
    opTerm (Noop p).

Definition opTermAb {M R} (ap: Abortable (Program M) R): Prop :=
  match ap with
  | abortable p =>
    opTerm p
  end.

Definition unNoop {M R} (p q : Program M R) : Prop := Noop p = q.
Notation FiniteNoop := (Acc unNoop).

Fixpoint dropHeadNoop {M R} (p : Program M R) (w : FiniteNoop p) : Program M R :=
  match w with
  | Acc_intro _ zz =>
    match p as p0 return p0 = p -> _ with
    | Return _ | Bind _ _ => fun _ => p
    | Noop q => fun v => dropHeadNoop (zz q v)
    end Logic.eq_refl
  end.

CoFixpoint spin {M R} : Program M R := Noop spin.

Section DropNoop.

Context (classicT : forall (P : Prop), { P } + { ~ P }).

CoFixpoint dropNoop {M R} (p : Program M R) : Program M R :=
  match @classicT (FiniteNoop p) with
  | left H =>
    match dropHeadNoop H with
    | Return r => Return r
    | Bind e k => Bind e (fun x => dropNoop (k x))
    | Noop p => spin (* Unreachable *)
    end
  | right _ => spin
  end.

Inductive noNoop {M R} : Program M R -> Prop :=
| noNoop_Return r : noNoop (Return r)
| noNoop_Bind X (e : M X) k : noNoop (Bind e k)
.

Lemma noNoop_dropHeadNoop {M R} (p : Program M R) (w:  FiniteNoop p) : noNoop (dropHeadNoop w).
Proof.
  revert p w; fix SELF 2; intros p []; cbn.
  destruct p; cbn; try constructor.
  apply SELF.
Qed.

Lemma TMLOpTerm_total_ M (R : Type) (p : Program M R)
  : exists p', opTerm p' /\
      eutt ieq p' p.
Proof.
  exists (dropNoop p). split.
  - revert p; cofix SELF; intros p.
    rewrite (frobProgramId (dropNoop p)); unfold frobProgram; cbn.
    destruct classicT.
    + destruct (noNoop_dropHeadNoop a); constructor; intros ?; apply SELF.
    + cofix SELF2; cbn; constructor; [ rewrite frobProgramId; apply SELF2 | ].
      rewrite (frobProgramId spin). discriminate.
  - revert p; pcofix SELF; intros p.
    rewrite (frobProgramId (dropNoop p)); unfold frobProgram; cbn.
    destruct classicT.
    + pfold. revert p a; fix SELF2 2; intros p [].
      destruct p; cbn; constructor; auto.
    + revert p n; clear; pcofix SELF; intros [] n.
      1,2: contradiction n; constructor; discriminate.
      cbn; pfold; constructor. right; rewrite (frobProgramId spin); apply SELF.
      intros H; apply n; constructor; intros p' W.
      inversion W. apply H.
Qed.

End DropNoop.

Lemma TMLOpTerm_total M (R : Type) (p : Program M R)
  : exists p', opTerm p' /\
      eutt ieq p' p.
Proof.
  destruct classicT_inhabited as [classicT].
  apply (TMLOpTerm_total_ classicT).
Qed.

Module TMLFun (TMLArgs: MiniDecidableType).

  Import TMLArgs.

  Definition TS := nat.

  Module BMap := BMapFun TMLArgs.
  Import BMap.

  Section TTT.
  Context (Val : Type).

  Notation atomicBMap := (atomicBMap Val).
  Notation BMapState := (BMapState Val).
  Notation BMapMethod := (BMapMethod Key Val).

  Definition base :=
    pairSpec
      (atomicReg (Val := TS))
      atomicBMap.

  Definition BaseState :=
    prod
      (RegisterState (Val := TS))
      BMapState.

  Definition gtState (s: BaseState) :=
    fst s.
  Definition mapState (s: BaseState) :=
    snd s.

  Definition BaseMethod :=
    PairInstruction
      (@RegisterMethod TS)
      BMapMethod.
  Definition callGT {R} (m: @RegisterMethod nat R): BaseMethod R :=
    inl m.
  Definition callMap {R} (m: BMapMethod R): BaseMethod R :=
    inr m.

  CoFixpoint initLoop: Program BaseMethod TS :=
    lt <- callGT read;
    if ((lt mod 2) =_? 1) then
       initLoop
    else
      Return lt.

  Definition impl {R} (m: @TransMethod BMapMethod TS R): Program BaseMethod R :=
    match m with
    | init =>
      initLoop

    | abort lt =>
      Return tt

    | commit lt =>
      if ((lt mod 2) =_? 1) then
        _ <- callGT (write (lt + 1));
        Return (inl tt)
      else
        Return (inl tt)

    | liftMethod lt m =>
      match m with

      | get k =>
        v <- callMap (get k);
        gt <- callGT read;
        if (gt =_? lt) then
          Return ((Some v), lt)
        else
          Return (None, lt)

      | put k v =>
        if ((lt mod 2) =_? 0) then
          b <- callGT (cas lt (lt+1));
          if b then
            _ <- callMap (put k v);
            Return ((Some tt), lt+1)
          else
            Return (None, lt+1)
        else
          _ <- callMap (put k v);
          Return ((Some tt), lt)
      end
    end.

  Definition tml :=
    object base (@impl).

  End TTT.

  Module TMLSer.

    Section UUU.
    Context (Val : Type).

    Notation atomicBMap := (atomicBMap Val).
    Notation BMapState := (BMapState Val).
    Notation BMapMethod := (BMapMethod Key Val).

    Definition annBase :=
      Annotated TS (base (Val := Val)).

    Definition AnnBaseMethod :=
      AnnMethod
        TS
        (PairInstruction
           (@RegisterMethod TS)
           BMapMethod).
    Definition annCallGT {R} (t: TS)(m: @RegisterMethod nat R): AnnBaseMethod R :=
      annMethod t (inl m).
    Definition annCallMap {R} (t: TS) (m: BMapMethod R): AnnBaseMethod R :=
      annMethod t (inr m).


    CoFixpoint annInitLoop: Program AnnBaseMethod TS :=
      lt <- annCallGT 0 read;
      if ((lt mod 2) =_? 1) then
        annInitLoop
      else
        Return lt.

    Definition annImpl {R} (m: @TransMethod BMapMethod TS R): Program AnnBaseMethod R :=
      match m with
      | init =>
        annInitLoop

      | abort lt =>
        Return tt

      | commit lt =>
        if ((lt mod 2) =_? 1) then
          _ <- annCallGT lt (write (lt + 1));
          Return (inl tt)
        else
          Return (inl tt)

      | liftMethod lt m =>
        match m with

        | get k =>
          v <- annCallMap lt (get k);
          gt <- annCallGT lt read;
          if (gt =_? lt) then
            Return ((Some v), lt)
          else
            Return (None, lt)

        | put k v =>
          if ((lt mod 2) =_? 0) then
            b <- annCallGT lt (cas lt (lt+1));
            if b then
              _ <- annCallMap (lt+1) (put k v);
              Return ((Some tt), lt+1)
            else
              Return (None, lt+1)
          else
            _ <- annCallMap lt (put k v);
            Return ((Some tt), lt)
        end
      end.

    Definition annTML :=
      object annBase (@annImpl).

    Definition cImpl :=
      wrapImpl (@opTermAb BMapMethod) (dumInstLeftAnnFlatImpl annTML).

    (*
    Definition aImpl :=
      wrapImpl (@opTermAb BMapMethod) (coarsenImpl (dumInstLeftAnnFlatImpl annTML)).
    *)

    Definition aImpl :=
      wrapImpl (@opTermAb BMapMethod)
               (serialObject (abortableSpec (programSpec atomicBMap))).

    Definition openCImpl :=
      wrapImpl (@opTermAb BMapMethod) {|
      ObjInstruction :=
        PairInstruction
          (AnnMethod TS
                     (PairInstruction RegisterMethod BMapMethod))
          (AnnMethod TS DumMethod);
         objBase :=
           pairSpec (Annotated TS (pairSpec atomicReg atomicBMap))
                    (Annotated TS dumObj);
         objProgram :=
           flattenImplProgLeft' (@annImpl) dumInstLeftAnnImplProg |}.

    Definition openCBase :=
      pairSpec
        (Annotated TS
                   (pairSpec
                      (@atomicReg TS _)
                      (atomicBMap)))
        (Annotated TS dumObj).


    Definition openAImpl :=
      wrapImpl (@opTermAb BMapMethod) {|
      ObjInstruction :=
        Abortable (Program BMapMethod);
      objBase :=
        abortableSpec (programSpec atomicBMap);
      objProgram :=
        fun (R : Type) (m : Abortable (Program BMapMethod) R) =>
          r <- m; Return r |}.

    Definition openABase :=
      abortableSpec (programSpec atomicBMap).

    (* ----------------------------- *)

    Definition AnnDumMethod :=
      AnnMethod TS DumMethod.

    Definition annBegstate
               {R} (p: Program
                         (PairMethod
                            (@TransMethod BMapMethod TS)
                            AnnDumMethod) R) :=
      flattenProgramLeft'
        (@annImpl)
        p.

    Definition annInterstate
               {R R'}
               (p: Program AnnBaseMethod R)
               (f: R ->
                   Program
                     (PairMethod
                        (@TransMethod BMapMethod TS)
                        AnnDumMethod) R') :=
      bindFlattenProgramLeft'
        (@annImpl)
        p f.

    Lemma annbegstateFold:
      forall R (p: Program
                (PairMethod
                   (@TransMethod BMapMethod TS)
                   AnnDumMethod) R),
        flattenProgramLeft'
          (@annImpl)
          p
        =
        annBegstate p.

    Proof.
      reflexivity.
    Qed.

    Lemma anninterstateFold:
      forall R R'
        (p: Program AnnBaseMethod R)
        (f: R ->
            Program
              (PairMethod
                 (@TransMethod BMapMethod TS)
                 AnnDumMethod) R'),

        bindFlattenProgramLeft'
          (@annImpl)
          p f =
        annInterstate p f.

    Proof.
      intros. reflexivity.
    Qed.

    Definition annDumCall (ts: TS):
      (PairMethod
         (AnnMethod
            TS
            (PairInstruction
               (@RegisterMethod TS)
               BMapMethod))
         (AnnMethod TS DumMethod))
        unit :=
      inr (annMethod ts dumCall).

    Inductive TRel {R}:
      (* Program AnnBaseMethod (option R) -> *)
      Program (PairMethod
                  (AnnMethod
                     TS
                     (PairInstruction
                        (@RegisterMethod TS)
                        BMapMethod))
                  (AnnMethod TS DumMethod))
              (option R) ->
      (* prod (prod TS BMapState) unit -> *)
      prod (prod (@RegisterState TS) BMapState) DumState ->
      (* Program BaseMethod (option R) -> *)
      Program BMapMethod R ->
      BMapState ->
      Prop :=

    | TRelInit1 p sc sa:
        opTerm p ->

        TRel
          (lt <- inl (annCallGT 0 read);
           (annInterstate
              (if ((lt mod 2) =_? 1) then
                 annInitLoop
               else
                 Return lt)
              (fun lt : TS => instLeftAnn' lt (dumNoopProgram p))))
          sc
          p
          sa

    | TRelDum lt (sc: prod (prod (@RegisterState TS) BMapState) DumState) p' p sa:
        (forall r, p' <> Return r) ->
        opTerm p' ->

        lt <= gtState (fst sc) ->

        ((lt mod 2 = 0
          /\ (gtState (fst sc) = lt ->
             ProgramSteps
               atomicBMap _
               p
               sa
               (Noop p')
               (mapState (fst sc))))

         \/ ((lt mod 2 = 1)
            /\ gtState (fst sc) = lt
            /\ ProgramSteps
                atomicBMap _
                p
                sa
                (Noop p')
                (mapState (fst sc)))) ->

        TRel
          (Bind
            (annDumCall lt)
            (fun _ =>
              (annBegstate (instLeftAnn' lt (dumNoopProgram p')))))
          sc
          p
          sa

    | TRelGet1 lt k f (sc: prod (prod (@RegisterState TS) BMapState) DumState) p sa:
        (forall r, opTerm (f r)) ->

        lt <= gtState (fst sc) ->

        ((lt mod 2 = 0
          /\ (gtState (fst sc) = lt ->
             ProgramSteps
               atomicBMap _
               p
               sa
               (r <- get k; f r)
               (mapState (fst sc))))

         \/ ((lt mod 2 = 1)
            /\ gtState (fst sc) = lt
            /\ ProgramSteps
                atomicBMap _
                p
                sa
                (r <- get k; f r)
                (mapState (fst sc)))) ->

        TRel
          (v <- inl (annCallMap lt (get k));
           (annInterstate
              (gt <- annCallGT lt read;
               if (gt =_? lt) then
                 Return ((Some v), lt)
               else
                 Return (None, lt))
              (fun ret =>
                 match fst ret with
                 | Some r => instLeftAnn' (snd ret) (dumNoopProgram (f r))
                 | None => _ <- inl (abort (snd ret)); Return None
                 end)))
          sc
          p
          sa

    | TRelGet2 lt f (sc: prod (prod (@RegisterState TS) BMapState) DumState) p sa (v: option Val):
        (forall r, opTerm (f r)) ->

        lt <= gtState (fst sc) ->

        ((lt mod 2 = 0
          /\ (gtState (fst sc) = lt ->
             ProgramSteps
               atomicBMap _
               p
               sa
               (f v)
               (mapState (fst sc))))

         \/ ((lt mod 2 = 1)
            /\ gtState (fst sc) = lt
            /\ ProgramSteps
                atomicBMap _
                p
                sa
                (f v)
                (mapState (fst sc)))) ->

        TRel
          (gt <- inl (annCallGT lt read);
           (annInterstate
              (if (gt =_? lt) then
                 Return ((Some v), lt)
               else
                 Return (None, lt))
              (fun ret =>
                 match fst ret with
                 | Some r => instLeftAnn' (snd ret) (dumNoopProgram (f r))
                 | None => _ <- inl (abort (snd ret)); Return None
                 end)))
          sc
          p
          sa

    | TRelPut1 lt f (sc: prod (prod (@RegisterState TS) BMapState) DumState) p sa k v:
        (forall r, opTerm (f r)) ->

        lt <= gtState (fst sc) ->

        lt mod 2 = 0 ->

        (gtState (fst sc) = lt ->
         ProgramSteps
           atomicBMap _
           p
           sa
           (r <- put k v; (f r))
           (mapState (fst sc))) ->

        TRel
          (b <- inl (annCallGT lt (cas lt (lt+1)));
           (annInterstate
              (if b: bool then
                 _ <- annCallMap (lt+1) (put k v);
                 Return ((Some tt), lt+1)
               else
                 Return (None, lt+1))
              (fun ret =>
                 match fst ret with
                 | Some r => instLeftAnn' (snd ret) (dumNoopProgram (f r))
                 | None => _ <- inl (abort (snd ret)); Return None
                 end)))
          sc
          p
          sa

    | TRelPut3 lt f (sc: prod (prod (@RegisterState TS) BMapState) DumState) p sa k v:
        (forall r, opTerm (f r)) ->

        lt mod 2 = 1 ->
        gtState (fst sc) = lt ->
        ProgramSteps
          atomicBMap _
          p
          sa
          (r <- put k v; (f r))
          (mapState (fst sc)) ->

        TRel
          (_ <- inl (annCallMap lt (put k v));
           (annInterstate
              (Return ((Some tt), lt))
              (fun ret =>
                 match fst ret with
                 | Some r => instLeftAnn' (snd ret) (dumNoopProgram (f r))
                 | None => _ <- inl (abort (snd ret)); Return None
                 end)))
          sc
          p
          sa

    | TRelCommit lt (sc: prod (prod (@RegisterState TS) BMapState) DumState) p sa r:
        lt mod 2 = 1 ->
        gtState (fst sc) = lt ->
        ProgramSteps
          atomicBMap _
          p
          sa
          (Return r)
          (mapState (fst sc)) ->

        TRel
          (_ <- inl (annCallGT lt (write (lt + 1)));
           (annInterstate
              (Return (inl tt))
              (fun ret0 : unit + TS =>
                match ret0 with
                | inl _ => Return (Some r)
                | inr t' => _ <- inl (abort t'); Return None
                end)))
          sc
          p
          sa.


    Inductive PDRelO {R}:
      Program (PairMethod
                  (AnnMethod
                     TS
                     (PairInstruction
                        (@RegisterMethod TS)
                        BMapMethod))
                  (AnnMethod TS DumMethod))
              (option R) ->
      prod (prod (@RegisterState TS) BMapState) DumState ->
      (* Program (Abortable (Program BMapMethod)) (option R) -> *)
      (* Program (Program BaseMethod) (option R) -> *)
      Program (Abortable (Program BMapMethod)) (option R) ->
      BMapState ->
      Prop :=

    | PDRel1 r sc sa:
        PDRelO
          (Return r) sc
          (Return r) sa

    | PDRel2: forall pc sc (p: Program BMapMethod R) sa,
        TRel
          pc sc
          p sa ->
        PDRelO
          pc sc
          (r <- (abortable p); Return r)
          sa.


    Definition PDRel {R}
               (pc: Program (PairMethod
                               (AnnMethod
                                  TS
                                  (PairInstruction
                                     (@RegisterMethod TS)
                                     BMapMethod))
                               (AnnMethod TS DumMethod))
                            R)
               (sc: prod (prod (@RegisterState TS) BMapState) DumState)
               (pa: Program (Abortable (Program BMapMethod)) R)
               (* (pa: Program (Program (PairMethod AnnBaseMethod AnnDumMethod))) R) *)
               (sa: BMapState): Prop :=

      exists R',
        option R' = R
        /\ forall (e: (option R' = R)),
          match e in _ = R'' return
                Program (PairMethod AnnBaseMethod AnnDumMethod) R'' ->
                Program (Abortable (Program BMapMethod)) R'' ->
                Prop with
          | Logic.eq_refl =>
            fun pc pa =>
              PDRelO pc sc pa sa
          end pc pa.



    Definition DRel
               (sc: prod (prod (@RegisterState TS) BMapState) DumState)
               (sa: BMapState): Prop :=
      (gtState (fst sc)) mod 2 = 0 ->
      mapState (fst sc) = sa.

    Definition inCritical
               {R} (p: Program (PairMethod
                                  (AnnMethod
                                     TS
                                     (PairInstruction
                                        (@RegisterMethod TS)
                                        BMapMethod))
                                  (AnnMethod TS DumMethod)) R)
               (sc: prod (prod (@RegisterState TS) BMapState) DumState):
      Prop :=

      match p with
      | Bind (inl m) f =>
        let lt := annMethodAnn m in
        lt = gtState (fst sc)
        /\ lt mod 2 = 1

      | Bind (inr m) f =>
        let lt := annMethodAnn m in
        lt = gtState (fst sc)
        /\ lt mod 2 = 1


      | Return r =>
        False

      | Noop p' =>
        False

      end.

    Definition PPRel {R R'}
               (pc1: Program (PairMethod
                                (AnnMethod
                                   TS
                                   (PairInstruction
                                      (@RegisterMethod TS)
                                      BMapMethod))
                                (AnnMethod TS DumMethod)) R)
               (pa1: Program (Abortable (Program BMapMethod)) R)
               (pc2: Program (PairMethod
                                (AnnMethod
                                   TS
                                   (PairInstruction
                                      (@RegisterMethod TS)
                                      BMapMethod))
                                (AnnMethod TS DumMethod)) R')
               (pa2: Program (Abortable (Program BMapMethod)) R')
               (sc: prod (prod (@RegisterState TS) BMapState) DumState)
               (sa: BMapState):
      Prop :=

      (forall p, pc1 <> Noop p)
      /\ (forall p, pc2 <> Noop p)
      /\ (inCritical pc1 sc ->
         inCritical pc2 sc ->
         False).


    Arguments PDRel {R} _ _ _ _.
    Arguments NPDRel {Method} {impl1} {impl2} {PDRel} {R} _ _ _ _.
    Arguments PPRel {R R'} _ _ _ _ _ _.
    Arguments NPPRel {Method} {impl1} {impl2} {PPRel} {R R'} _ _ _ _ _ _.

    Definition init1: prod (prod (@RegisterState TS) BMapState) DumState :=
      (0, fun _ => None, tt).
    Definition init2: BMapState :=
      (fun _ => None).

    (* --------------------------------------- *)

    Ltac openPDRel :=
      unfold PDRel; eexists; split;
      [ reflexivity
      | let e := fresh "e" in
        intros e;
          let E := fresh "E" in
          assert (E: e = Logic.eq_refl);
            [ eapply UIP_refl
            | subst e;  try constructor ]].


    Ltac openPDRelIn H :=
      let H' := fresh H "_" in
      let H0 := fresh H "_" in
      let H1 := fresh H "_" in
      let H2 := fresh H "_" in
      let R := fresh "R" in
      unfold PDRel in H;
        destruct H as (R & H1 & H2);
        match type of H1 with
        | ?T1 = ?T2 => subst T2
        end;
        specialize (H2 Logic.eq_refl);
        simpl in H2;
        inverts H2 as H.

    Ltac invertRegCallIn H :=
      let H' := fresh H "_" in
      let H0 := fresh H "_" in
      let H1 := fresh H "_" in
      let H2 := fresh H "_" in
      let H1' := fresh H1 "_" in
      let H10 := fresh H1 "_" in
      let H11 := fresh H1 "_" in
      let H12 := fresh H1 "_" in

      inverts H as H _;
        inverts H as H1 H2;
        simpl in H1;
        simpl in H2;
        apply invertRegSem in H1;
        (try destruct H1 as (H11 & H12));
        try simpl in H11;
        try simpl in H12;
        simpl.
        (* unfold RegisterState in *;
        unfold DumState in *;
        try simpl in H11;
        try simpl in H12;
        simpl. *)

    Ltac invertMapCallIn H :=
      let H' := fresh H "_" in
      let H0 := fresh H "_" in
      let H1 := fresh H "_" in
      let H2 := fresh H "_" in
      let H1' := fresh H1 "_" in
      let H10 := fresh H1 "_" in
      let H11 := fresh H1 "_" in
      let H12 := fresh H1 "_" in

      inverts H as H _;
        inverts H as H1 H2;
        simpl in H1;
        simpl in H2;
        apply invertBMapSem in H1;
        destruct H1 as (H11 & H12);
        simpl in H11;
        simpl in H12;
        simpl.

    Ltac invertDumCallIn H :=
      let H' := fresh H "_" in
      let H0 := fresh H "_" in
      let H1 := fresh H "_" in

      inverts H as H1 H;
        inverts H1;
        simpl in H;
        simpl.

    Ltac invertCallIn H :=
      match type of H with
      | EvalInstruction _ (inl (annCallGT _ _)) _ _ _ =>
        invertRegCallIn H
      | EvalInstruction _ (inl (annCallMap _ _)) _ _ _ =>
        invertMapCallIn H
      | EvalInstruction _ (inr (annMethod _ _)) _ _ _ =>
        invertDumCallIn H
      | EvalInstruction _ (annDumCall _) _ _ _ =>
        invertDumCallIn H
      end.

    Ltac frob0 :=
      progFrobSimpl;
      try setoid_rewrite anninterstateFold;
      try setoid_rewrite annbegstateFold;
      try rewrite foldModTwo.

    Ltac frobIn0 H :=
      progFrobSimplIn H;
      try setoid_rewrite anninterstateFold in H;
      try setoid_rewrite annbegstateFold in H;
      try rewrite foldModTwo in H.

    Ltac frobAt0 i :=
      progFrobSimplAt i;
      try setoid_rewrite anninterstateFold;
      try setoid_rewrite annbegstateFold;
      try setoid_rewrite foldModTwo.

    Ltac frobInAt0 H i :=
      progFrobSimplAt i;
      try setoid_rewrite anninterstateFold in H;
      try setoid_rewrite annbegstateFold in H;
      try setoid_rewrite foldModTwo in H.

    Ltac frob :=
      progFrobSimpl;
      try setoid_rewrite anninterstateFold;
      try rewrite foldModTwo.

    Ltac frobIn H :=
      progFrobSimplIn H;
      try setoid_rewrite anninterstateFold in H;
      try rewrite foldModTwo in H.

    Ltac frobAt i :=
      progFrobSimplAt i;
      try setoid_rewrite anninterstateFold;
      try setoid_rewrite foldModTwo.

    Ltac frobInAt H i :=
      progFrobSimplAt i;
      try setoid_rewrite anninterstateFold in H;
      try setoid_rewrite foldModTwo in H.

    Ltac simplM :=
      try unfold gtState;
      try unfold mapState;
      try simpl;
      try unfold gtState;
      try unfold mapState;
      try simpl;
      try rewrite foldModTwo.

    Ltac simplMS :=
      try unfold gtState in *;
      try unfold mapState in *;
      try simpl in *;
      try unfold gtState in *;
      try unfold mapState in *;
      try simpl in *;
      try rewrite foldModTwo in *.

    Ltac simplMIn H :=
      try unfold gtState in H;
      try unfold mapState in H;
      try simpl in H;
      try unfold gtState in H;
      try unfold mapState in H;
      try simpl in H;
      try rewrite foldModTwo in H.

    (* ----------------------------------------------- *)
    Lemma PDRelNoop:
      forall R (p1: Program (PairMethod AnnBaseMethod AnnDumMethod) R) p2 s1 s2,
        PDRel
          (Noop p1) s1
          p2 s2 ->
        False.

    Proof.
      introv H1.
      openPDRelIn H1.
      inverts H1.
    Qed.

    Lemma PPRelSym:
      forall R (p1: Program (PairMethod AnnBaseMethod AnnDumMethod) R) p2
        R' (p1': Program (PairMethod AnnBaseMethod AnnDumMethod) R') p2' s1 s2,

        PPRel p1 p2 p1' p2' s1 s2 ->
        PPRel p1' p2' p1 p2 s1 s2.

    Proof.
      intros.
      unfold PPRel in *.
      openAnds.
      auto.
    Qed.

    Lemma PPRelNoop:
      forall R R'
        (pc1: Program (PairMethod AnnBaseMethod AnnDumMethod) R)
        (pa1: Program (Abortable (Program BMapMethod)) R)
        (pc2: Program (PairMethod AnnBaseMethod AnnDumMethod) R')
        (pa2: Program (Abortable (Program BMapMethod)) R')
        (sc: prod (prod (@RegisterState TS) BMapState) DumState)
        (sa: BMapState),

        PPRel
          (Noop pc1) pa1
          pc2 pa2
          sc sa ->
        False.

    Proof.
      introv H1.
      unfold PPRel in H1.
      openAnds.
      forwardsIn H1_1.
      apply H1_1.
      reflexivity.
    Qed.

    Lemma initRel:
      DRel init1 init2.

    Proof.
      unfold DRel.
      auto.
    Qed.

Local Opaque Nat.modulo.

    Lemma invRel':
      forall R m s1 s2,
        DRel s1 s2 ->
        @NPDRel _ cImpl aImpl (@PDRel) _
                (objProgram cImpl R m) s1
                (objProgram aImpl R m) s2
        /\ forall R' (p1: Program (PairInstruction
                               (AnnMethod
                                  TS
                                  (PairInstruction
                                     (@RegisterMethod TS)
                                     BMapMethod))
                               (AnnMethod TS DumMethod)) R') p2,
            PDRel p1 s1 p2 s2 ->
            @NPPRel _ cImpl aImpl (@PPRel) _ _
                    (objProgram cImpl R m) (objProgram aImpl R m)
                    p1 p2
                    s1 s2.

    Proof.
      introv H1.
      split.

      - unfold cImpl. unfold aImpl. simpl.
        destruct m. destruct m. simpl.
        frob.
        apply NPDRel2.
        constructor.
        openPDRel.
        frobAt 1.
        apply (@TRelInit1 R).
        unfold opTermAb in o.
        assumption.

      - introv H2.
        unfold cImpl. unfold aImpl. simpl.
        destruct m. destruct m. simpl.
        frobAt 1.
        apply NPPRel2.
        constructor.
        unfold PPRel. splits.
        + intros. frob. unfold not. intros. inverts H.
        + intros. unfold not. intros. subst. eapply PDRelNoop. eassumption.
        + introv H3 _.
          unfold inCritical in *. simpl in *.
          destruct H3 as (_ & H32).
          inverts H32.
    Qed.

    Lemma retRel:
      forall R (r: R) s1 p2 s2,
        PDRel (Return r) s1 p2 s2 ->
        DRel s1 s2 ->
        p2 = Return r.

    Proof.
      introv H1 _.
      openPDRelIn H1.
      - reflexivity.
      - inverts H1.
    Qed.


    Lemma gtIncreasing:
      forall R' (m : (PairMethod
                   (AnnMethod
                      TS
                      (PairInstruction
                         (@RegisterMethod TS)
                         BMapMethod))
                   (AnnMethod TS DumMethod)) R')
        R (f : R' -> Program (PairMethod
                               (AnnMethod
                                  TS
                                  (PairInstruction
                                     (@RegisterMethod TS)
                                     BMapMethod))
                               (AnnMethod TS DumMethod)) R)
        (s1: prod (prod (@RegisterState TS) BMapState) DumState)
        (p2: Program (Abortable (Program BMapMethod)) R)
        (s2: BMapState)
        (s1': prod (prod (@RegisterState TS) BMapState) DumState)
        (r: R'),

        PDRel (x <- m; f x) s1 p2 s2 ->
        EvalInstruction openCBase m s1 r s1' ->
        gtState (fst s1') = gtState (fst s1)
        \/ gtState (fst s1') = gtState (fst s1) + 1.

    Proof.
      introv H1 H2.
      simplM.
      openPDRelIn H1. inverts H1 as.
      - introv H1. invertCallIn H2. left. congruence.
      - introv H11 H12 H13. invertCallIn H2. left. congruence.
      (* destruct p; progFrobSimplIn H13; inverts H13. *)
      - introv H11 H12 H13. invertCallIn H2. left. congruence.
      - introv H11 H12 H13. invertCallIn H2. left. congruence.
      - introv H11 H12 H13 H14. invertCallIn H2.
        destruct H2_1 as [(H211 & H212 & H213) | (H211 & H212 & H213)].
        + right. congruence.
        + left. assumption.
      (* - introv H11 H12 H13 H14. invertCallIn H2. left. congruence. *)
      - introv H11 H12 H13 H14. invertCallIn H2. left. congruence.
      - introv H11 H12 H13. invertCallIn H2. simplMS. right. congruence.
    Qed.


    Definition annOf {R} (m: (PairMethod
                                (AnnMethod
                                   TS
                                   (PairInstruction
                                      (@RegisterMethod TS)
                                      BMapMethod))
                                (AnnMethod TS DumMethod)) R): TS :=
      match m with
      | inl m' => annMethodAnn m'
      | inr m' => annMethodAnn m'
      end.

    Lemma annLeqGt:
      forall (R : Type) (m : (PairMethod
                           (AnnMethod
                              TS
                              (PairInstruction
                                 (@RegisterMethod TS)
                                 BMapMethod))
                           (AnnMethod TS DumMethod)) R) (R' : Type)
        (f : R -> Program (PairMethod
                            (AnnMethod
                               TS
                               (PairInstruction
                                  (@RegisterMethod TS)
                                  BMapMethod))
                            (AnnMethod TS DumMethod)) (option R')) p2
        (s1: prod (prod (@RegisterState TS) BMapState) DumState)
        (s2: BMapState),

        TRel (x <- m; f x) s1 p2 s2 ->
        annOf m <= gtState (fst s1).

    Proof.
      introv H1.
      inverts H1; simpl; omega.
    Qed.


    Lemma PPRelPresHelper2:
      forall (R' : Type) (m : PairMethod AnnBaseMethod AnnDumMethod R')
        (R : Type) (f : R' -> Program (PairMethod AnnBaseMethod AnnDumMethod) R)
        (s1 : prod (prod (@RegisterState TS) BMapState) DumState)
        (p2 : Program (Abortable (Program BMapMethod)) R)
        (s2 : BMapState) (s1': prod (prod (@RegisterState TS) BMapState) DumState)
        (r : R') (R'' : Type)
        (p11 : Program (PairMethod AnnBaseMethod AnnDumMethod) R'')
        (p21 : Program (Abortable (Program BMapMethod)) R''),

        PDRel (x <- m; f x) s1 p2 s2 ->
        (* EvalInstruction (objBase cImpl) m s1 r s1' -> *)
        EvalInstruction openCBase m s1 r s1' ->
        PDRel p11 s1 p21 s2 ->
        inCritical p11 s1' ->
        inCritical p11 s1.

    Proof.
      introv H1 H2 H3 H6.

      unfold inCritical in *.
      forwards N: gtIncreasing H1 H2.
      destruct p11; try destruct p; try assumption.
      + destruct H6 as (H61 & H62).
        split; try assumption.
        destruct N as [N | N].
        * simpl in *.
          congruence.
        * openPDRelIn H3.
          forwards M: annLeqGt H3.
          exfalso.
          unfold annOf in *.
          omega.

      + destruct H6 as (H61 & H62).
        split; try assumption.
        destruct N as [N | N].
        * simpl in *.
          congruence.
        * openPDRelIn H3.
          forwards M: annLeqGt H3.
          exfalso.
          unfold annOf in *.
          omega.
    Qed.


    Lemma PPRelPres2:
      forall R' (m: (PairMethod AnnBaseMethod AnnDumMethod) R') R (f: R' -> Program (PairMethod AnnBaseMethod AnnDumMethod) R) s1 p2 s2 s1' r
        R (p11: Program (PairMethod AnnBaseMethod AnnDumMethod) R) p21
        R' (p12: Program (PairMethod AnnBaseMethod AnnDumMethod) R') p22
        s2',

        PDRel (Bind m f) s1 p2 s2 ->
        cImpl.(objBase).(EvalInstruction) m s1 r s1' ->
        PDRel p11 s1 p21 s2 ->
        PDRel p12 s1 p22 s2 ->
        PPRel p11 p21 p12 p22 s1 s2 ->
        PPRel p11 p21 p12 p22 s1' s2'.

    Proof.
      introv H1 H2 H3 H4 H5.
      unfold PPRel in *.
      openAnds.
      splits; try assumption.
      introv H6 H7.
      apply H5_3; clear H5_1 H5_2 H5_3.
      - clear H4 H7.
        eapply PPRelPresHelper2; eassumption.
      - clear H3 H6.
        eapply PPRelPresHelper2; eassumption.
    Qed.


    Lemma stepRel:
      forall R' (m: (PairMethod
                  (AnnMethod
                     TS
                     (PairInstruction
                        (@RegisterMethod TS)
                        BMapMethod))
                  (AnnMethod TS DumMethod))
                 R')
        R (f: R' -> Program (PairMethod
                              (AnnMethod
                                 TS
                                 (PairInstruction
                                    (@RegisterMethod TS)
                                    BMapMethod))
                              (AnnMethod TS DumMethod)) R)
        (s1: prod (prod (@RegisterState TS) BMapState) DumState)
        (p2: Program (Abortable (Program BMapMethod)) R)
        (s2: BMapState)
        (s1': prod (prod (@RegisterState TS) BMapState) DumState)
        (r: R'),

        DRel s1 s2 ->
        PDRel (Bind m f) s1 p2 s2 ->
        openCBase.(EvalInstruction) m s1 r s1' ->

        exists (p2': Program (Abortable (Program BMapMethod)) R)
          (s2': BMapState),

          ProgramSteps openABase _ p2 s2 p2' s2'
          /\ DRel s1' s2'
          /\ @NPDRel _ openCImpl openAImpl (@PDRel) _
                    (f r) s1' p2' s2'

          /\ (forall R (p1'': Program (PairMethod
                                   (AnnMethod
                                      TS
                                      (PairInstruction
                                         (@RegisterMethod TS)
                                         BMapMethod))
                                   (AnnMethod TS DumMethod)) R) p2'',
                PDRel p1'' s1 p2'' s2 ->
                PPRel (Bind m f) p2 p1'' p2'' s1 s2 ->
                PDRel p1'' s1' p2'' s2')

          /\ (forall R (p1'': Program (PairMethod
                                   (AnnMethod
                                      TS
                                      (PairInstruction
                                         (@RegisterMethod TS)
                                         BMapMethod))
                                   (AnnMethod TS DumMethod)) R) p2'',
                PDRel p1'' s1 p2'' s2 ->
                PPRel (Bind m f) p2 p1'' p2'' s1 s2 ->
                @NPPRel _ cImpl aImpl (@PPRel) _ _
                        (f r) p2' p1'' p2'' s1' s2')

          /\ (forall R (p11: Program (PairMethod
                                  (AnnMethod
                                     TS
                                     (PairInstruction
                                        (@RegisterMethod TS)
                                        BMapMethod))
                                  (AnnMethod TS DumMethod)) R) p21
               R' (p12: Program (PairMethod
                                   (AnnMethod
                                      TS
                                      (PairInstruction
                                         (@RegisterMethod TS)
                                         BMapMethod))
                                   (AnnMethod TS DumMethod)) R') p22,
                PDRel p11 s1 p21 s2 ->
                PDRel p12 s1 p22 s2 ->
                PPRel p11 p21 p12 p22 s1 s2 ->
                PPRel p11 p21 p12 p22 s1' s2').

    Proof.
      introv H1 H2 H3.
      assert (H2c := H2).
      openPDRelIn H2. inverts H2 as.

      - { (* Init1 *)
          introv T.
          assert (H3c := H3).
          invertCallIn H3.
          destruct s1 as (s11 & s12); destruct s12;
          destruct s1' as (s11' & s12'); destruct s12'.
          simpl in H3_1_1.
          simpl in H3_1_2.
          simpl in H3_2.

          destruct ((r mod 2) =_? 1) as [E | E].

          - { (* The loop repeats. *)
              eexists. eexists. splits.

              - eapply ProgramStepsRefl.

              - unfold DRel in *. intro H4.
                simplMS.
                forwardsIn H1.
                simplM.
                congruence.
                destruct s11; destruct s11'; congruence.

              - (* rewrite E. simpl. *)
                progFrobSimpl.
                constructor.
                openPDRel.
                constructor.
                assumption.

              - introv N1 N2.
                destruct s11; destruct s11'; simpl in *; congruence.

              - introv N1 N2.
                constructor.
                unfold PPRel. splits.
                + introv M1. simpl in M1. frobIn M1. inverts M1.
                + introv M1. subst. eapply PDRelNoop. eassumption.
                + introv N3 N4. simpl in N3; openAnds. inverts N3_2.

              - introv N1 N2 N3. eapply PPRelPres2; eassumption.
            }

          - { (* The loop is passed. *)

              apply modTwoNotOneZero in E.
              unfold DRel in H1. forwardsIn H1. simplM. congruence.
              (* destruct s2 as (s21 & s22). destruct s22. simpl in H1. subst s21. *)

              destruct p as [ R m' f' | ret | np ].

              - { (* Bind after init *)
                  eexists. eexists. splits.
                  - eapply ProgramStepsRefl.

                  - unfold DRel in *. intro H4.
                    simplMS.
                    destruct s11; destruct s11'; simpl in *; congruence.

                  - simpl.
                    frob.
                    apply NPDRel2.
                    frob.
                    apply NPDRel2.
                    destruct m'.

                    + { (* Get *)
                        frob.
                        constructor.
                        openPDRel.
                        constructor.
                        - inverts T. assumption.
                        - simpl. simplM. omega.
                        - left. splits.
                          + assumption.
                          + introv H4. simplMS. subst s2.
                            applys_eq ProgramStepsRefl 1. congruence.
                      }

                    + { (* Put *)
                        rewrite E. simpl.
                        frob.
                        constructor.
                        openPDRel.
                        constructor.
                        - inverts T. assumption.
                        - simplM. omega.
                        - assumption.
                        - introv H4. simplMS. subst s2.
                          applys_eq ProgramStepsRefl 1. congruence.
                      }

                  - introv N1 N2.
                    destruct s11; destruct s11'; simpl in *; congruence.

                  - introv N1 N2.
                    simpl. frobAt 1.
                    eapply NPPRel2.
                    frobAt 1.
                    eapply NPPRel2.
                    constructor.
                    unfold PPRel. splits.
                    + introv M1. destruct m'.
                      * frobIn M1. inverts M1.
                      * simplMIn M1. rewrite E in M1. simpl in M1. frobIn M1. inverts M1.
                    + introv M1. subst. eapply PDRelNoop. eassumption.
                    + introv N3 N4.
                      destruct m'.
                      * frobIn N3. openAnds. congruence.
                      * rewrite E in N3. simpl in N3. openAnds. simplMS. congruence.

                  - introv N1 N2 N3. eapply PPRelPres2; eassumption.
                }

              - { (* Return after init *)

                  eexists. eexists. splits.
                  - eapply ProgramStepImplProgramSteps.
                    constructor. simpl. constructor.
                    constructor.

                  - unfold DRel. introv H. simplMS. congruence.

                  - frob.
                    apply NPDRel2.
                    frob.
                    apply NPDRel2.
                    rewrite E. simpl.
                    frob.
                    apply NPDRel2.
                    frob.
                    constructor.
                    openPDRel.

                  - introv N1 N2.
                    destruct s11; destruct s11'; simpl in *; congruence.

                  - introv N1 N2.
                    frobAt 1.
                    eapply NPPRel2.
                    frobAt 1.
                    eapply NPPRel2.
                    rewrite E. simpl.
                    frobAt 1.
                    eapply NPPRel2.
                    frobAt 1.
                    constructor.
                    unfold PPRel.
                    splits.
                    + introv H. inverts H.
                    + unfold PPRel in N2. openAnds. assumption.
                    + introv N3 N4. inverts N3.

                  - introv N1 N2 N3. eapply PPRelPres2; eassumption.

                }

              - { (* Noop after init *)

                  eexists. eexists. splits.
                  - eapply ProgramStepsRefl.

                  - unfold DRel. intro H4. simpl.
                    destruct s11; destruct s11'; simpl in *; congruence.

                  - frob.
                    apply NPDRel2.
                    constructor.
                    frob.
                    openPDRel.
                    constructor.
                    + inverts T. assumption.
                    + inverts T. assumption.
                    + simplM. omega.
                    + left. splits.
                      * assumption.
                      * introv H4. simplMS.
                        subst s2. rewrite H3_2 in *.
                        eapply ProgramStepsRefl.

                  - introv N1 N2.
                    destruct s11; destruct s11'; simpl in *; congruence.

                  - introv N1 N2.
                    frobAt 1.
                    eapply NPPRel2.
                    frobAt 1.
                    constructor.
                    unfold PPRel in *.
                    openAnds. splits.
                    + introv H. inverts H.
                    + assumption.
                    + introv N3 N4.
                      unfold inCritical in N3. simpl in N3.
                      openAnds. congruence.

                  - introv N1 N2 N3. eapply PPRelPres2; eassumption.

                }
            }
        }


      - { (* Dum case *)
          introv T1 T2 N1 N2.

          invertCallIn H3.
          simplMS.
          destruct s1 as (s11 & s12); destruct s12; destruct s1' as (s11' & s12'); destruct s12'.
          simplMS.
          subst s11'.

          unfold DRel in H1. simplMIn H1.
          sort.
          destruct r.
          destruct p' as [ R m' f' | ret | np ] eqn: D.

          + { (* Bind *)
              eexists. eexists. splits.
              - eapply ProgramStepsRefl.

              - unfold DRel. simplM. auto.

              (* - destruct m' as [m' | m']. *)

              - destruct m'.

                + { (* Get *)
                    frob.
                    eapply NPDRel2.
                    frob.
                    constructor.
                    openPDRel.
                    constructor.
                    - inverts T2. assumption.
                    - simplM. omega.
                    - destruct N2 as [(N21 & N22) | (N21 & N22 & N3)].
                      + left. splits.
                        * assumption.
                        * introv H4. simplMS.
                          forwardsIn N22. assumption.
                          simplMS.
                          eapply ProgramStepsStepImplSteps.
                          eassumption.
                          constructor.

                      + right. splits; try assumption.
                        simplMS.
                        eapply ProgramStepsStepImplSteps.
                        eassumption.
                        constructor.
                  }

                + { (* Put *)
                    frob.
                    eapply NPDRel2.
                    destruct N2 as [(N21 & N22) | (N21 & N22 & N3)].
                    - rewrite N21. simpl.
                      frob.
                      constructor.
                      openPDRel.
                      constructor.
                      + inverts T2. assumption.
                      + simplM. omega.
                      + assumption.
                      + introv H4. simplMS.
                        forwardsIn N22. assumption.
                        simplMS.
                        eapply ProgramStepsStepImplSteps.
                        eassumption.
                        constructor.

                    - rewrite N21. simpl.
                      frob.
                      constructor.
                      openPDRel.
                      constructor.
                      + inverts T2. assumption.
                      + assumption.
                      + simplM. assumption.
                      + simplM.
                        eapply ProgramStepsStepImplSteps.
                        eassumption.
                        constructor.
                  }

              - introv M1 M2. assumption.

              - { introv L1 L2.
                  unfold PPRel in L2. openAnds.
                  - destruct m'.
                    + frobAt 1.
                      eapply NPPRel2.
                      frobAt 1.
                      constructor.
                      unfold PPRel. splits.
                      * introv M1. inverts M1.
                      * assumption.
                      * introv N3 N4.
                        apply L2_3.
                        -- simplMS. assumption.
                        -- assumption.

                    + { frobAt 1.
                        eapply NPPRel2.
                        destruct N2 as [(N21 & N22) | (N21 & N22 & N23)].
                        - rewrite N21. simpl.
                          frobAt 1.
                          constructor.
                          unfold PPRel. splits.
                          + introv M1. inverts M1.
                          + assumption.
                          + introv N3 N4.
                            apply L2_3.
                            * simplMS. assumption.
                            * assumption.
                        - rewrite N21. simpl.
                          frobAt 1.
                          constructor.
                          unfold PPRel. splits.
                          + introv M1. inverts M1.
                          + assumption.
                          + introv N3 N4.
                            apply L2_3.
                            * simplMS. assumption.
                            * assumption.
                      }
                }

              - introv M1 M2 M3. eapply PPRelPres2; try eassumption.
                constructor. constructor. reflexivity.

            }

          + { (* Return *)
              exfalso. eapply T1. reflexivity.
            }

          + { (* Noop *)
              eexists. eexists. splits.
              - eapply ProgramStepsRefl.

              - unfold DRel. simplM. auto.

              - frob.
                constructor.
                openPDRel.
                eapply TRelDum.
                + inverts T2. assumption.
                + inverts T2. assumption.
                + simplM. omega.
                + destruct N2 as [(N21 & N22) | (N21 & N22 & N3)].
                  * left. splits.
                    -- assumption.
                    -- introv H4. simplMS.
                       forwardsIn N22. assumption.
                       simplMS.
                       eapply ProgramStepsStepImplSteps.
                       eassumption.
                       constructor.
                  * right. splits; try assumption.
                    simplMS.
                    eapply ProgramStepsStepImplSteps.
                    eassumption.
                    constructor.

              - introv M1 M2. assumption.

              - introv L1 L2.
                unfold PPRel in L2. openAnds.
                frobAt 1.
                constructor.
                unfold PPRel. splits.
                + introv M1. inverts M1.
                + assumption.
                + introv N3 N4.
                  apply L2_3.
                  * simplMS. assumption.
                  * assumption.

              - introv M1 M2 M3. eapply PPRelPres2; try eassumption.
                constructor. constructor. reflexivity.

            }
        }


      - { (* Get1 *)

          introv T N1 N2.
          assert (H3c := H3).
          invertCallIn H3.
          destruct s1 as (s11 & s12); destruct s12;
          destruct s1' as (s11' & s12'); destruct s12'.
          simpl in H3_1_1.
          simpl in H3_1_2.
          simpl in H3_2.

          eexists. eexists. splits.
          - eapply ProgramStepsRefl.

          - unfold DRel in *. simplMS. introv H.
            forwardsIn H1. simplM. congruence.
            destruct s11; destruct s11'; simpl in *; congruence.

          - frob.
            constructor.
            openPDRel.
            constructor.
            + assumption.
            + simplMS. omega.
            + destruct N2 as [(N21 & N22) | (N21 & N22 & N3)].
              * left. splits.
                -- assumption.
                -- introv H4. simplMS.
                   forwardsIn N22. congruence.
                   simplMS.
                   eapply ProgramStepsStepImplSteps.
                   eassumption.
                   constructor.
                   rewrite <- H3_1_2 in *. subst r.
                   constructor.
              * right. splits; try assumption.
                -- simplMS. congruence.
                -- eapply ProgramStepsStepImplSteps.
                   eassumption.
                   simplMS.
                   rewrite <- H3_1_2 in *. subst r.
                   econstructor. econstructor.

          - introv M1 M2.
            destruct s11; destruct s11'; simpl in *; congruence.

          - introv M1 M2.
            frobAt 1.
            constructor.
            unfold PPRel in *.
            openAnds. splits.
            + introv H. inverts H.
            + assumption.
            + introv N3 N4.
              eapply M2_3.
              * unfold inCritical. simpl.
                unfold inCritical in N3. simpl in N3.
                openAnds. simplMS. splits; congruence.
              * destruct s11; destruct s11'. simpl in *; congruence.

          - introv M1 M2 M3. eapply PPRelPres2; eassumption.
        }


      - { (* Get2 *)
          introv T N1 N2.
          assert (H3c := H3).
          invertCallIn H3.
          destruct s1 as (s11 & s12); destruct s12;
          destruct s1' as (s11' & s12'); destruct s12'.
          simpl in H3_1_1.
          simpl in H3_1_2.
          simpl in H3_2.

          destruct (eq_nat_dec r lt) as [E | E].
          + { destruct (f0 v) as [ R m' f' | ret | np ] eqn: D.

              - { (* Bind *)
                  eexists. eexists. splits.
                  - eapply ProgramStepsRefl.

                  - unfold DRel in *. intro H4. simplMS.
                    forwardsIn H1. simplM. congruence.
                    congruence.

                  - frob.
                    apply NPDRel2.
                    rewrite D.
                    destruct m'.

                    + { (* Get *)
                        frob.
                        apply NPDRel2.
                        frob.
                        constructor.
                        openPDRel.
                        constructor.
                        - specialize (T v). rewrite D in T. inverts T. assumption.
                        - simplM. omega.
                        - destruct N2 as [(N21 & N22) | (N21 & N22 & N3)].
                          + left. splits.
                            * assumption.
                            * introv H4. simplMS.
                              forwardsIn N22. congruence.
                              simplMS.
                              rewrite <- H3_2.
                              assumption.

                          + right. splits; try assumption.
                            * simplMS. congruence.
                            * simplMS. rewrite <- H3_2. assumption.
                      }

                    + { (* Put *)
                        frob.
                        apply NPDRel2.
                        destruct N2 as [(N21 & N22) | (N21 & N22 & N3)].
                        - rewrite N21. simpl.
                          frob.
                          constructor.
                          openPDRel.
                          constructor.
                          + specialize (T v). rewrite D in T. inverts T. assumption.
                          + simplM. omega.
                          + assumption.
                          + introv H4. simplMS.
                            forwardsIn N22. congruence.
                            simplMS. rewrite <- H3_2.
                            assumption.

                        - rewrite N21. simpl.
                          frob.
                          constructor.
                          openPDRel.
                          constructor.
                          + specialize (T v). rewrite D in T. inverts T. assumption.
                          + assumption.
                          + simplMS. congruence.
                          + simplMS. rewrite <- H3_2. assumption.
                      }

                  - introv L1 L2.
                    destruct s11; destruct s11'; simpl in *; congruence.

                  - introv L1 L2.
                    frobAt 1.
                    eapply NPPRel2.
                    rewrite D.
                    destruct m'.
                    + (* Get *)
                      frobAt 1.
                      eapply NPPRel2.
                      frobAt 1.
                      constructor.
                      unfold PPRel in *. openAnds. splits.
                      * introv H. inverts H.
                      * assumption.
                      * introv M1 M2.
                        apply L2_3.
                        unfold inCritical in *. simplMS.
                        openAnds. splits; congruence.
                        destruct s11; destruct s11'; simpl in *; congruence.

                    + (* Put *)
                      frobAt 1.
                      eapply NPPRel2.
                      destruct (modTwoZeroOrOne lt) as [O|O]; rewrite O; simpl.
                      * frobAt 1.
                        constructor.
                        unfold PPRel in *. openAnds. splits.
                        -- introv H. inverts H.
                        -- assumption.
                        -- introv M1 M2.
                           apply L2_3.
                           unfold inCritical in *. simplMS.
                           openAnds. splits; congruence.
                           destruct s11; destruct s11'; simpl in *; congruence.
                      * frobAt 1.
                        constructor.
                        unfold PPRel in *. openAnds. splits.
                        -- introv H. inverts H.
                        -- assumption.
                        -- introv M1 M2.
                           apply L2_3.
                           unfold inCritical in *. simplMS.
                           openAnds. splits; congruence.
                           destruct s11; destruct s11'; simpl in *; congruence.

                  - introv L1 L2 L3. eapply PPRelPres2; eassumption.
                }

              - { (* Return *)

                  destruct (modTwoZeroOrOne lt) as [O|O].
                  + { eexists. eexists. splits.

                      - eapply ProgramStepImplProgramSteps.
                        constructor. simpl.
                        eapply abortableEvalInstS.
                        destruct N2 as [(N21 & N22) | (N21 & N22 & N3)].
                        + apply N22. simplM. congruence.
                        + apply N3.

                      - unfold DRel. introv H. simplMS. congruence.

                      - frob.
                        apply NPDRel2.
                        rewrite D.
                        frob.
                        apply NPDRel2.
                        rewrite O.
                        frob.
                        apply NPDRel2.
                        frob.
                        constructor.
                        openPDRel.

                      - introv L1 L2.
                        unfold DRel in H1. simplMS. forwardsIn H1. simplM. congruence.
                        subst s2.
                        destruct s11; destruct s11'; simpl in *; congruence.

                      - introv L1 L2.
                        frobAt 1.
                        eapply NPPRel2.
                        rewrite D.
                        frobAt 1.
                        eapply NPPRel2.
                        rewrite O. simpl.
                        frobAt 1.
                        eapply NPPRel2.
                        frobAt 1.
                        constructor.
                        unfold PPRel.
                        splits.
                        + introv H. inverts H.
                        + unfold PPRel in L2. openAnds. assumption.
                        + introv N3 N4. inverts N3.

                      - introv L1 L2 L3. eapply PPRelPres2; eassumption.

                    }

                  + { eexists. eexists. splits.

                      - eapply ProgramStepsRefl.

                      - unfold DRel. simplMS. introv H. congruence.

                      - frob.
                        apply NPDRel2.
                        rewrite D.
                        frob.
                        apply NPDRel2.
                        rewrite O.
                        frob.
                        constructor.
                        openPDRel.
                        constructor.
                        + assumption.
                        + simplM. congruence.
                        + simplM.
                          destruct N2 as [(N21 & N22) | (N21 & N22 & N3)].
                          * exfalso. congruence.
                          * simplMIn N3. congruence.

                      - introv L1 L2.
                        destruct s11; destruct s11'; simpl in *; congruence.

                      - introv L1 L2.
                        frobAt 1.
                        eapply NPPRel2.
                        rewrite D.
                        frobAt 1.
                        eapply NPPRel2.
                        rewrite O. simpl.
                        frobAt 1.
                        constructor.
                        unfold PPRel.
                        splits.
                        + introv H. inverts H.
                        + unfold PPRel in L2. openAnds. assumption.
                        + introv N3 N4.
                          apply L2.
                          unfold inCritical. unfold inCritical in N3. simplMS. openAnds. splits; congruence.
                          destruct s11; destruct s11'; simpl in *; congruence.

                      - introv L1 L2 L3. eapply PPRelPres2; eassumption.

                    }

                }

              - { (* Noop *)

                  eexists. eexists. splits.
                  - eapply ProgramStepsRefl.

                  - destruct s11; destruct s11'; simpl in *. subst. assumption.

                  - frob.
                    apply NPDRel2.
                    rewrite D.
                    frob.
                    constructor.
                    openPDRel.
                    constructor.
                    + specialize (T v). rewrite D in T. inverts T. assumption.
                    + specialize (T v). rewrite D in T. inverts T. assumption.
                    + simplMS. omega.
                    + { destruct N2 as [(N21 & N22) | (N21 & N22 & N3)].
                        + left. splits.
                          * assumption.
                          * introv H4. simplMS.
                            forwardsIn N22. congruence.
                            simplMS. rewrite <- H3_2. assumption.

                        + right. splits; try assumption.
                          * simplMS. congruence.
                          * simplMS. rewrite <- H3_2. assumption.
                      }


                  - introv L1 L2.
                    destruct s11; destruct s11'; simpl in *; congruence.

                  - introv L1 L2.
                    frobAt 1.
                    eapply NPPRel2.
                    rewrite D.
                    frobAt 1.
                    constructor.
                    unfold PPRel in *.
                    openAnds. splits.
                    + introv H. inverts H.
                    + assumption.
                    + introv N3 N4.
                      apply L2_3.
                      unfold inCritical. unfold inCritical in N3. simplMS. openAnds. splits; congruence.
                      destruct s11; destruct s11'; simpl in *; congruence.

                  - introv L1 L2 L3. eapply PPRelPres2; eassumption.

                }
            }

          + { eexists. eexists. splits.
              - eapply ProgramStepImplProgramSteps.
                unfold openABase.
                constructor.
                simpl.
                eapply abortableEvalInstA.

              - unfold DRel in *. simplMS. introv H.
                forwardsIn H1. simplM. congruence.
                destruct s11; destruct s11'; simpl in *; congruence.

              - frob.
                eapply NPDRel2.
                frob.
                eapply NPDRel2.
                frob.
                eapply NPDRel2.
                frob.
                constructor.
                openPDRel.


              - introv M1 M2.
                destruct s11; destruct s11'; simpl in *; congruence.

              - introv M1 M2.
                frobAt 1.
                eapply NPPRel2.
                frobAt 1.
                eapply NPPRel2.
                frobAt 1.
                eapply NPPRel2.
                frobAt 1.
                constructor.
                unfold PPRel in *. openAnds. splits.
                + introv H. inverts H.
                + assumption.
                + introv L1 L2.
                  inverts L1.

              - introv M1 M2 M3. eapply PPRelPres2; eassumption.
            }
        }


      - { (* Put1 *)

          intros T N1 N2 N3.
          assert (H3c := H3).
          invertCallIn H3.
          destruct s1 as (s11 & s12); destruct s12;
          destruct s1' as (s11' & s12'); destruct s12'.
          simpl in H3_1.
          simpl in H3_2.
          destruct H3_1 as [(N31 & N32 & N33) | (N31 & N32 & N33)].

          + { eexists. eexists. splits.
              - eapply ProgramStepsRefl.

              - unfold DRel in *. simplMS. introv H.
                forwardsIn H1. simplM. congruence.
                congruence.

              - subst r. simpl.
                frob.
                constructor.
                openPDRel.
                constructor.
                + assumption.
                + apply modTwoAdd. assumption.
                + simplM. assumption.
                + simplM.
                  forwardsIn N3. simplM. assumption.
                  simplMS. congruence.

              - introv M1 _.
                openPDRelIn M1.
                + openPDRel.
                + { inverts M1 as.
                    - introv L1. openPDRel. constructor; try assumption.
                    - introv L1 L2 L3 L4. openPDRel. constructor; try assumption.
                      + simplMS. omega.
                      + simplMS.
                        destruct L4 as [L4 | L4]; openAnds.
                        * left. splits. assumption. introv L. exfalso. omega.
                        * exfalso. congruence.
                    - introv L1 L2 L3. openPDRel. constructor; try assumption.
                      + simplMS. omega.
                      + simplMS.
                        destruct L3 as [L3 | L3]; openAnds.
                        * left. splits. assumption. introv L. exfalso. omega.
                        * exfalso. congruence.
                    - introv L1 L2 L3. openPDRel. constructor; try assumption.
                      + simplMS. omega.
                      + simplMS.
                        destruct L3 as [L3 | L3]; openAnds.
                        * left. splits. assumption. introv L. exfalso. omega.
                        * exfalso. congruence.
                    - introv L1 L2 L3. openPDRel. constructor; try assumption.
                      + simplMS. omega.
                      + simplMS.
                        introv L. exfalso. omega.
                    (* - introv L1 L2 L3 L4.
                      exfalso. simplMS. congruence. *)
                    - introv L1 L2 L3.
                      exfalso. simplMS. congruence.
                    - introv L1 L2.
                      exfalso. simplMS. congruence.
                  }

              - introv M1 M2.
                subst r.
                frobAt 1.
                constructor.
                unfold PPRel in *. openAnds. splits.
                + introv H. inverts H.
                + assumption.
                + introv _ L2. clear M2_1 M2_2 M2_3.
                  openPDRelIn M1.
                  * unfold inCritical in L2. assumption.
                  * { inverts M1 as.
                      - introv K1. inverts L2 as. introv L21 L22. simplMS. inverts L22.
                      - introv K1 K2 K3 K4. inverts L2 as L21 L22. simplMS.
                        destruct K4 as [K4 | K4]; openAnds.
                        + congruence.
                        + simplMS. congruence.
                      - introv K1 K2 K3. inverts L2 as L21 L22. simplMS.
                        destruct K3 as [K3 | K3]; openAnds.
                        + congruence.
                        + simplMS. congruence.
                      - introv K1 K2 K3. inverts L2 as L21 L22. simplMS.
                        destruct K3 as [K3 | K3]; openAnds.
                        + congruence.
                        + simplMS. congruence.
                      (* - introv K1 K2 K3 K4. inverts L2 as L21 L22. simplMS. congruence. *)
                      - introv K1 K2 K3 K4. inverts L2 as L21 L22. simplMS. congruence.
                      - introv K1 K2 K3. inverts L2 as L21 L22. simplMS. congruence.
                      - introv K1 K2. inverts L2 as L21 L22. simplMS. congruence.
                    }

              - introv M1 M2 M3. eapply PPRelPres2; eassumption.
            }

          + { eexists. eexists. splits.
              - eapply ProgramStepImplProgramSteps.
                unfold openABase.
                constructor.
                simpl.
                eapply abortableEvalInstA.

              - unfold DRel in *. simplMS. introv H.
                forwardsIn H1. simplM. congruence.
                congruence.

              - subst r.
                frob.
                eapply NPDRel2.
                frob.
                eapply NPDRel2.
                frob.
                eapply NPDRel2.
                frob.
                constructor.
                openPDRel.


              - introv M1 M2.
                destruct s11; destruct s11'; simpl in *; congruence.

              - introv M1 M2.
                subst r.
                frobAt 1.
                eapply NPPRel2.
                frobAt 1.
                eapply NPPRel2.
                frobAt 1.
                eapply NPPRel2.
                frobAt 1.
                constructor.
                unfold PPRel in *. openAnds. splits.
                + introv H. inverts H.
                + assumption.
                + introv L1 L2. inverts L1.

              - introv M1 M2 M3. eapply PPRelPres2; eassumption.
            }
        }

      - { (* Put2 *)
          intros T N3 N4.
          assert (H3c := H3).
          invertCallIn H3.
          destruct s1 as (s11 & s12); destruct s12;
          destruct s1' as (s11' & s12'); destruct s12'.
          simpl in H3_1_1.
          simpl in H3_1_2.
          simpl in H3_2.

          eexists. eexists. splits.
          - eapply ProgramStepsRefl.

          - unfold DRel in *. simplMS.
            introv H. congruence.

          - frob.
            eapply NPDRel2.
            destruct (f0 tt) as [ R m' f' | ret | np ] eqn: D.
            + { (* Bind *)
                destruct m'.

                - { (* Get *)
                    frob.
                    apply NPDRel2.
                    frob.
                    constructor.
                    openPDRel.
                    constructor.
                    - specialize (T tt). rewrite D in T. inverts T. assumption.
                    - simplMS. omega.
                    - right. splits; try assumption.
                      * simplMS. congruence.
                      * simplMS.
                        eapply ProgramStepsStepsImplSteps.
                        eassumption.
                        rewrite H3_1_2.
                        eapply ProgramStepStepsImplSteps.
                        econstructor. econstructor.
                        rewrite D.
                        apply ProgramStepsRefl.
                  }

                - { (* Put *)
                    frob.
                    apply NPDRel2.
                    simplMS.
                    rewrite N3. simpl.
                    frob.
                    constructor.
                    openPDRel.
                    constructor.
                    + specialize (T tt). rewrite D in T. inverts T. assumption.
                    + assumption.
                    + simplMS. congruence.
                    + simplMS.
                      eapply ProgramStepsStepsImplSteps.
                      eassumption.
                      rewrite H3_1_2.
                      eapply ProgramStepStepsImplSteps.
                      econstructor. econstructor.
                      rewrite D.
                      apply ProgramStepsRefl.
                  }
              }

            + { (* Return *)

                - frob.
                  apply NPDRel2.
                  simplMS.
                  rewrite N3.
                  frob.
                  constructor.
                  openPDRel.
                  constructor.
                  + assumption.
                  + simplMS. congruence.
                  + simplMS.
                    eapply ProgramStepsStepsImplSteps.
                    eassumption.
                    rewrite H3_1_2.
                    eapply ProgramStepStepsImplSteps.
                    econstructor. econstructor.
                    rewrite D.
                    apply ProgramStepsRefl.
              }


            + { (* Noop *)
                frob.
                constructor.
                openPDRel.
                constructor.
                + specialize (T tt). rewrite D in T. inverts T. assumption.
                + specialize (T tt). rewrite D in T. inverts T. assumption.
                + simplMS. omega.
                + right. splits; try assumption.
                  * simplMS. congruence.
                  * simplMS.
                    eapply ProgramStepsStepsImplSteps.
                    eassumption.
                    rewrite H3_1_2.
                    eapply ProgramStepStepsImplSteps.
                    econstructor. econstructor.
                    rewrite D.
                    apply ProgramStepsRefl.
              }

          - introv M1 M2.
            openPDRelIn M1.
            + openPDRel.
            + { inverts M1 as.
                - introv L1. openPDRel. constructor; try assumption.
                - introv L1 L2 L3 L4. openPDRel. constructor; try assumption.
                  + simplMS. omega.
                  + simplMS.
                    destruct L4 as [L4 | L4]; openAnds.
                    * left. splits. assumption. introv L. exfalso. congruence.
                    * exfalso. apply M2.
                      simplMS. splits; congruence.
                      simplMS. splits; congruence.
                - introv L1 L2 L3. openPDRel. constructor; try assumption.
                  + simplMS. omega.
                  + simplMS.
                    destruct L3 as [L3 | L3]; openAnds.
                    * left. splits. assumption. introv L. exfalso. congruence.
                    * exfalso. apply M2.
                      simplMS. splits; congruence.
                      simplMS. splits; congruence.
                - introv L1 L2 L3. openPDRel. constructor; try assumption.
                  + simplMS. omega.
                  + simplMS.
                    destruct L3 as [L3 | L3]; openAnds.
                    * left. splits. assumption. introv L. exfalso. congruence.
                    * exfalso. apply M2.
                      simplMS. splits; congruence.
                      simplMS. splits; congruence.
                - introv L1 L2 L3. openPDRel. constructor; try assumption.
                  + simplMS. omega.
                  + simplMS. introv L. exfalso. congruence.
                (* - introv L1 L2 L3. openPDRel. simplMS. exfalso. apply M2.
                  simplMS. splits; congruence.
                  simplMS. splits; congruence. *)
                - introv L1 L2 L3. openPDRel. simplMS. exfalso. apply M2.
                  simplMS. splits; congruence.
                  simplMS. splits; congruence.
                - introv L1 L2. openPDRel. simplMS. exfalso. apply M2.
                  simplMS. splits; congruence.
                  simplMS. splits; congruence.
              }


          - introv M1 M2.
            frobAt 1.
            eapply NPPRel2.

            destruct (f0 tt) as [ R' m' f' | ret | np ] eqn: D.
            + { (* Bind *)
                destruct m'.

                - { (* Get *)
                    frobAt 1.
                    eapply NPPRel2.
                    frobAt 1.
                    constructor.
                    unfold PPRel in *. openAnds. splits.
                    + introv H. inverts H.
                    + assumption.
                    + introv K1 K2. apply M2_3.
                      unfold inCritical in *. simplMS. openAnds. split; congruence.
                      unfold inCritical in *. destruct p1''; try destruct p0; simplMS; openAnds; try split; congruence.
                  }

                - { (* Put *)
                    frobAt 1.
                    eapply NPPRel2.
                    simplMS.
                    rewrite N3.
                    simpl.
                    frobAt 1.
                    constructor.
                    unfold PPRel in *. openAnds. splits.
                    + introv H. inverts H.
                    + assumption.
                    + introv K1 K2. apply M2_3.
                      unfold inCritical in *. simplMS. openAnds. split; congruence.
                      unfold inCritical in *. destruct p1''; try destruct p0; simplMS; openAnds; try split; congruence.
                  }
              }

            + { (* Return *)
                frobAt 1.
                eapply NPPRel2.
                simplMS.
                rewrite N3.
                simpl.
                frobAt 1.
                constructor.
                unfold PPRel in *. openAnds. splits.
                + introv H. inverts H.
                + assumption.
                + introv K1 K2. apply M2_3.
                  unfold inCritical in *. simplMS. openAnds. split; congruence.
                  unfold inCritical in *. destruct p1''; try destruct p0; simplMS; openAnds; try split; congruence.
              }

            + { (* Noop *)
                frobAt 1.
                constructor.
                unfold PPRel in *. openAnds. splits.
                + introv H. inverts H.
                + assumption.
                + introv K1 K2. apply M2_3.
                  unfold inCritical in *. simplMS. openAnds. split; congruence.
                  unfold inCritical in *. destruct p1''; try destruct p0; simplMS; openAnds; try split; congruence.
              }

          - introv M1 M2 M3. eapply PPRelPres2; eassumption.

        }


      - { (* Commit *)
          introv N1 N2.
          assert (H3c := H3).
          invertCallIn H3.
          destruct s1 as (s11 & s12); destruct s12;
          destruct s1' as (s11' & s12'); destruct s12'.
          simpl in H3_1_1.
          simpl in H3_1_2.
          simpl in H3_2.
          simplMS.

          eexists. eexists. splits.
          - eapply ProgramStepImplProgramSteps.
            constructor.
            constructor.
            simplMS.
            eassumption.

          - unfold DRel in *. simplMS. introv H. congruence.

          - frob.
            apply  NPDRel2.
            frob.
            constructor.
            openPDRel.

          - introv M1 M2.
            openPDRelIn M1.
            + openPDRel.
            + { inverts M1 as.
                - introv L1. openPDRel. constructor; try assumption.
                - introv L1 L2 L3 L4. openPDRel. constructor; try assumption.
                  + simplMS. omega.
                  + simplMS. destruct L4 as [L4 | L4]; openAnds.
                    * left. splits. assumption. introv L. omega.
                    * exfalso. apply M2.
                      simplMS. splits; congruence.
                      simplMS. splits; congruence.
                - introv L1 L2 L3. openPDRel. constructor; try assumption.
                  + simplMS. omega.
                  + simplMS. destruct L3 as [L3 | L3]; openAnds.
                    * left. splits. assumption. introv L. omega.
                    * exfalso. apply M2.
                      simplMS. splits; congruence.
                      simplMS. splits; congruence.
                - introv L1 L2 L3. openPDRel. constructor; try assumption.
                  + simplMS. omega.
                  + simplMS. destruct L3 as [L3 | L3]; openAnds.
                    * left. splits. assumption. introv L. omega.
                    * exfalso. apply M2.
                      simplMS. splits; congruence.
                      simplMS. splits; congruence.
                - introv L1 L2 L3. openPDRel. constructor; try assumption.
                  + simplMS. omega.
                  + simplMS. introv L. omega.
                - introv L1 L2 L3. exfalso. apply M2.
                  + simplMS. splits; congruence.
                  + simplMS. splits; congruence.
                - introv L1 L2. exfalso. apply M2.
                  + simplMS. splits; congruence.
                  + simplMS. splits; congruence.
              }

          - introv M1 M2.
            frobAt 1.
            apply NPPRel2.
            frobAt 1.
            constructor.
            unfold PPRel in *.
            openAnds. splits.
            + introv H. inverts H.
            + assumption.
            + introv N3 N4. inverts N3.

          - introv M1 M2 M3. eapply PPRelPres2; try eassumption.
            simpl. eassumption.

        }

        Unshelve. exact tt.
        Unshelve. exact tt.

    Qed.

    Lemma eraseInitLoop:
      programEq initLoop (erase annInitLoop).
    Proof.
      cofix CoInd.
      unfold initLoop. unfold annInitLoop.
      match goal with
      | [ |- programEq ?p ?p' ] =>
        rewriteFrobProgram p;
        rewriteFrobProgram p';
        simpl
      end.
      constructor. intros. (* rewrite foldModTwo. *)
      destruct (modTwoZeroOrOne v); simpl in H; rewrite H; simpl.
      + match goal with
        | [ |- programEq ?p ?p' ] =>
          rewriteFrobProgram p';
          simpl
        end.
        reflexivity.
      + assumption.
    Qed.

    Lemma eraseAnnTMLTML:
      forall R m,
        programEq
          (objProgram tml R m)
          (eraseImpl (objProgram annTML) R m).

    Proof.
      unfold tml. unfold annTML.
      unfold impl. unfold annImpl.
      simpl. intros.
      unfold eraseImpl; destruct m; simpl.
      - apply eraseInitLoop.
      - progFrobSimpl. reflexivity.
      - destruct (modTwoZeroOrOne t0); rewrite H; simpl.
        + progFrobSimpl. reflexivity.
        + progFrobSimpl. constructor. intros. progFrobSimpl. reflexivity.
      - destruct b.
        + progFrobSimpl. constructor. intros.
          progFrobSimpl. constructor. intros.
          destruct (eq_nat_dec v0 t0).
          * progFrobSimpl. reflexivity.
          * progFrobSimpl. reflexivity.
        + destruct (modTwoZeroOrOne t0); rewrite H; simpl.
          * progFrobSimpl. constructor. intros.
            destruct v0.
            -- progFrobSimpl. constructor. intros.
               progFrobSimpl. reflexivity.
            -- progFrobSimpl. reflexivity.
          * progFrobSimpl. constructor. intros.
            progFrobSimpl. reflexivity.
    Qed.

    Lemma dumInstLeftAnnFlatImplTMLConRefAnnTML:
      ConcSimulates
        (dumInstLeftAnnFlatImpl tml)
        (dumInstLeftAnnFlatImpl annTML).
    Proof.
      eapply dumInstLeftAnnFlatImplConRefAnnImpl'.
      apply eraseAnnTMLTML.
    Qed.

    Lemma TMLDumAnnStrictSerializable:
      ConcSimulates
        (wrapImpl (@opTermAb BMapMethod)
                  (dumInstLeftAnnFlatImpl annTML))
        (wrapImpl (@opTermAb BMapMethod)
                  (serialObject (abortableSpec (programSpec atomicBMap)))).
    Proof.
      unfold LinearizableWrap.
      unfold ConcSimulatesWrap.
      eapply (@simulatesCoInd
                _
                (wrapImpl (@opTermAb BMapMethod)
                          (dumInstLeftAnnFlatImpl annTML))
                (wrapImpl (@opTermAb BMapMethod)
                          (serialObject (abortableSpec (programSpec atomicBMap))))
                (@PDRel)
                DRel
                (@PPRel)
             ); simpl.

      - eapply PDRelNoop.
      - eapply PPRelSym.
      - eapply PPRelNoop.
      - eapply initRel.

      - eapply invRel'.
      - eapply retRel.
      - eapply stepRel.
    Qed.

    Lemma TMLStrictSerializableWrap:
      StrictSerializableWrap
        (@opTermAb _)
        atomicBMap
        tml.
    Proof.
      eapply concRefTrans.
      eapply ConcRefWrapImplConRef.
      eapply dumNoopAnnSerializable.

      eapply concRefTrans.
      eapply ConcRefWrapImplConRef.
      eapply dumInstLeftAnnFlatImplTMLConRefAnnTML.

      eapply TMLDumAnnStrictSerializable.
    Qed.

    Lemma TMLOpTermAb_total (R : Type) (p : Abortable (Program BMapMethod) R)
      : exists p', opTermAb p' /\
          eq_Abortable (eutt ieq) R p' p.
    Proof.
      destruct p as [? p].
      destruct (TMLOpTerm_total (p := p)) as [p' []].
      exists (abortable p'); cbn; split; auto.
      constructor. auto.
    Qed.

    Lemma TMLStrictSerializable : StrictSerializable atomicBMap tml instrument.
    Proof.
      eapply StrictSerializable_unwrap;
        [ apply @TMLOpTermAb_total | apply TMLStrictSerializableWrap].
    Qed.
    End UUU.

Print Assumptions TMLStrictSerializable.

  End TMLSer.

End TMLFun.
