From Coq Require String.
From Coq Require Import FunctionalExtensionality.

From C4.Util Require Import
  LibTactics.

From C4.Core Require Import
  Program
  ProgramFacts
  ProgramOps
  ProgramOpsFacts
  Object
  Pair
  Threads
  LinCoInd
  Simulates
  SimulatesFacts
  SeqSimulatesFacts
  Linearizability
  LinearizabilityFacts.

From C4.App Require Import
  Lock.

Import SimulatesCoInd.
Import SimulatesSerImplCoInd.

Module Type CoarseSyncParams.

  Parameter (M: Type -> Type).
  Parameter seqImpl: Object M.
  Parameter lockImpl: Object LockMethod.

End CoarseSyncParams.


Module CoarseSyncFun (CoarseSyncArgs: CoarseSyncParams).

  Import CoarseSyncArgs.

  Definition base :=
    pairObject lockImpl seqImpl.

  Definition BaseState :=
    prod
      lockImpl.(objBase).(State)
      seqImpl.(objBase).(State).
    (* prod (ObjState lockImpl) (ObjState seqImpl). *)
    (* Unfolded here because otherwise folded and unfolded types appear to be different during the proofs. *)
    
  (* Definition lockState (s: BaseState): ObjState lockImpl := *)
  Definition lockState (s: BaseState): lockImpl.(objBase).(State) :=
    fst s.
  (* Definition seqObjState (s: BaseState): ObjState seqImpl := *)
  Definition seqObjState (s: BaseState): seqImpl.(objBase).(State) :=
    snd s.

  Definition BaseMethod :=
    PairMethod LockMethod M.
  Definition callLock {R} (m: LockMethod R): BaseMethod R :=
    inl m.
  Definition callSeqImpl {R} (m: M R): BaseMethod R :=
    inr m.
  
  Definition impl {R} (m: M R): Program BaseMethod R :=
    _ <- callLock doLock;
    r <- callSeqImpl m;
    _ <- callLock doUnlock;
    Return r.
  
  Definition syncImpl :=
    flattenObject
      base
      _
      (@impl).

  Module Type CoarseSyncSelfLinParams.

    Parameter lockLin: Linearizable atomicLock lockImpl.

  End CoarseSyncSelfLinParams.

  Module CoarseSyncSelfLinFun (CoarseSyncSelfLinArgs: CoarseSyncSelfLinParams).

    Import CoarseSyncSelfLinArgs.

    (* 
     * Because the lock object is linearizable,
     * we can replace it with an abstract lock.
     *)
    
    (* Variable initSeqObjState: ObjState seqImpl. *)
    Variable initSeqObjState: seqImpl.(objBase).(State).
    
    (* R for reduced or replaced. *)
    Definition rBase :=
      pairSpec
        atomicLock
        seqImpl.(objBase).

    Definition RBaseState :=
      (* prod LockState (ObjState seqImpl). *)
      prod
        LockState
        seqImpl.(objBase).(State).
    Definition rLockState (s: RBaseState): LockState :=
      fst s.
    (*Definition rSeqObjState (s: RBaseState): ObjState seqImpl := *)
    Definition rSeqObjState (s: RBaseState): seqImpl.(objBase).(State) :=
      snd s.

    Definition RBaseMethod :=
      PairMethod LockMethod seqImpl.(ObjInstruction).
    Definition rCallLock {R} (m: LockMethod R): RBaseMethod R :=
      inl m.

    Definition flattenSeqImpl {R} (p: Program BaseMethod R): Program RBaseMethod R :=
      flattenProgramRight
        (objProgram seqImpl)
        p.

    Definition bindFlattenSeqImpl
               {R} (p: Program seqImpl.(ObjInstruction) R)
               {R'} (f: R -> Program BaseMethod R'):
      Program RBaseMethod R' :=
      bindFlattenProgramRight
        (objProgram seqImpl)
        (mapProgram (fun _ m => inr m) p) f.

    (* Calls on the lock remain unchanged but the calls on the interface M are expanded. *)
    Definition cImpl: Object M :=
      {| objBase := rBase;
         objProgram R m := flattenSeqImpl (impl m) |}.

    Definition aImpl: Object M :=
      serialObject (coarsen seqImpl).


    Inductive PDRel {R}:
      Program RBaseMethod R ->
      RBaseState ->
      Program M R ->
      (* ObjState seqImpl -> *)
      seqImpl.(objBase).(State) ->
      Prop :=

    | PDRel1 m sc sa:
        PDRel          
          (Noop
             (_ <- rCallLock doLock;
              (Noop
                 (flattenSeqImpl
                    (r <- callSeqImpl m;
                     _ <- callLock doUnlock;
                     Return r)))))
          sc
          (r <- m; Return r)
          sa
      
    (* At the beginning. *)
    | PDRel1' m sc sa:
        PDRel          
          (_ <- rCallLock doLock;
           (Noop
              (flattenSeqImpl
                 (r <- callSeqImpl m;
                  _ <- callLock doUnlock;
                  Return r))))
          sc
          (r <- m; Return r)
          sa

    | PDRel1'' m sc:
        rLockState sc = true ->
        PDRel          
          (Noop
             (flattenSeqImpl
                (r <- callSeqImpl m;
                 _ <- callLock doUnlock;
                 Return r)))
          sc
          (r <- m; Return r)
          (rSeqObjState sc)
          
    (* Right after the lock call. *)
    | PDRel2 m sc:
        rLockState sc = true ->
        PDRel
          (Noop
             (bindFlattenSeqImpl
                (objProgram seqImpl _ m)
                (fun r =>
                   _ <- callLock doUnlock;
                   Return r)))
          sc
          (r <- m; Return r)
          (rSeqObjState sc)

    (* At the beginning or in the middle of the object of m. *)
    (* 
     * sc: state, concrete
     * sa: state, abstract
     * pc: program, concrete
     *)
    | PDRel2' m pc sc sa:
        rLockState sc = true ->
        ProgramSteps
          (objBase seqImpl) _
          (objProgram seqImpl _ m)
          sa
          pc
          (rSeqObjState sc) ->
        PDRel
          (bindFlattenSeqImpl
             pc
             (fun r =>
                _ <- callLock doUnlock;
                Return r))
          sc
          (r <- m; Return r)
          sa

    (* At the end of the object of m, before the unlock method call. *)
    | PDRel3 m r sc sa:
        rLockState sc = true ->
        ProgramSteps
          (objBase seqImpl) _
          (objProgram seqImpl _ m)
          sa
          (Return r)
          (rSeqObjState sc) ->
        PDRel
          (Noop (
              _ <- rCallLock doUnlock;
              Noop (Return r)))
          sc
          (r <- m; Return r)
          sa

    (* Before the unlock method call. *)
    | PDRel4 m r sc sa:
        rLockState sc = true ->
        ProgramSteps
          (objBase seqImpl) _
          (objProgram seqImpl _ m)
          sa
          (Return r)
          (rSeqObjState sc) ->
        PDRel
          (_ <- rCallLock doUnlock;
           Noop (Return r))
          sc
          (r <- m; Return r)
          sa

    (* At the end. *)
    | PDRel5p r s1 s2:
        PDRel (Noop (Return r)) s1 (Return r) s2
          
    (* At the end. *)
    | PDRel5 r s1 s2:
        PDRel (Return r) s1 (Return r) s2.
    
    Hint Constructors PDRel.

      
    Definition DRel
               (s1: RBaseState)
               (* (s2: ObjState seqImpl): Prop := *)
               (s2: seqImpl.(objBase).(State)): Prop :=
      rLockState s1 = false -> rSeqObjState s1 = s2.

    
    Inductive OutCritical {R}:
      Program RBaseMethod R ->
      Prop :=
      
    | OutCritical1 m:
        OutCritical
          (Noop
             (_ <- rCallLock doLock;
              (Noop
                 (flattenProgramRight
                    (objProgram seqImpl)
                    (r <- callSeqImpl m;
                     _ <- callLock doUnlock;
                     Return r)))))

    | OutCritical1' m:          
        OutCritical
          (_ <- rCallLock doLock;
           (Noop
              (flattenProgramRight
                 (objProgram seqImpl)
                 (r <- callSeqImpl m;
                  _ <- callLock doUnlock;
                  Return r))))
          
    | OutCritical2 r:
        OutCritical
          (Noop (Return r))
          
    | OutCritical3 r:
        OutCritical
          (Return r).
    
    Hint Constructors OutCritical.

    Definition PPRel {R R'}
               (p11: Program RBaseMethod R)
               (p21: Program M R)
               (p12: Program RBaseMethod R')
               (p22: Program M R'):
      Prop :=

      OutCritical p11 \/ OutCritical p12.

    Lemma PPRelSym:
      forall R (p1: Program RBaseMethod R) p2
        R' (p1': Program RBaseMethod R') p2',

        PPRel p1 p2 p1' p2' ->
        PPRel p1' p2' p1 p2.

    Proof.
      introv H1.
      unfold PPRel in *.
      iauto.
    Qed.
    
    Definition init1: RBaseState := (initialState atomicLock, seqImpl.(objBase).(initialState)).
    Definition init2: seqImpl.(objBase).(State) := seqImpl.(objBase).(initialState).

    Lemma initRel: DRel init1 init2.
    Proof.
      unfold DRel. intro. reflexivity.
    Qed.  

    Import Coq.Strings.String.
    Ltac print_o b s :=
      match b with
      | true => idtac "!!!" s      
      | false => idtac
      end.

    Ltac print s :=
      print_o true s.

    Ltac mark :=
      print "Mark "%string.


    Ltac isOpenProg p :=
      match p with
      | _ <- callLock _; _ =>
        constr:(true)
      | mapProgram _ _ =>
        constr:(true)
      | Return _ =>
        constr:(true)
      | Noop _ =>
        constr:(true)
      | _ =>
        constr:(false)
      end.

    Ltac isOpenProg' p :=
      match p with
      (*  Without the @, it was not matching binds. *)
      | @Bind _ _ _ _ _ =>
        constr:(true)
      | mapProgram _ _ =>
        constr:(true)
      | Return _ =>
        constr:(true)
      | Noop _ =>
        constr:(true)
      | _ =>
        constr:(false)
      end.    
    
    Ltac simplFlatten :=
      unfold flattenProgramRight;
      unfold bindFlattenProgramRight;
      repeat
        (progress
           match goal with
           | [ |- context [ flattenProgram ?p _ ] ] =>
             match (isOpenProg p) with
             | true => setoid_rewrite (frobProgramId (flattenProgram p _)); simpl
             end
           | [ |- context[ bindFlattenProgram _ _ ?p ] ] =>
             match (isOpenProg' p) with             
             | true =>  setoid_rewrite (frobProgramId (bindFlattenProgram _ _ p)); simpl
             end
           end).
    
    Ltac simplFlattenIn H :=
      unfold flattenProgramRight in H;
      unfold bindFlattenProgramRight in H;
      match type of H with
      | context[ flattenProgram _ ?p ] =>
        match (isOpenProg p) with
        | true => setoid_rewrite (frobProgramId (flattenProgram _ p)) in H; simpl in H
        end
      | context[ bindFlattenProgram _ _ ?p ] =>
        match (isOpenProg' p) with             
        | true =>  setoid_rewrite (frobProgramId (bindFlattenProgram _ _ p)) in H; simpl in H
        end
      end.

    
    Lemma FlattenProgramRightRet:
      forall (R : Type) (r : R) (m : M R) (pr: Prop),
        flattenProgramRight (objProgram seqImpl) (impl m) = Return r ->
        pr.
    Proof.
      introv H1. unfold impl in H1.

      unfold flattenProgramRight in H1;
        unfold bindFlattenProgramRight in H1.
      
      match type of H1 with
      | context[ flattenProgram _ ?p ] =>
        match (isOpenProg p) with
        | true => setoid_rewrite (frobProgramId (flattenProgram _ p)) in H1; simpl in H1
        end
      | context[ bindFlattenProgram _ _ ?p ] =>
        match (isOpenProg' p) with             
        | true =>  setoid_rewrite (frobProgramId (bindFlattenProgram _ _ p)) in H; simpl in H
        end
      end.

          
      simplFlattenIn H1. inverts H1.
    Qed.

    (*
    Lemma FlattenProgramRightNoop:
      forall R (m : M R) p (pr: Prop),
        flattenProgramRight (objProgram seqImpl) (impl m) = Noop p ->
        pr.
    Proof.
      introv H1. unfold impl in H1. simplFlattenIn H1. inverts H1.
    Qed.    
    *)

    Lemma FlattenProgramRightRet':
      forall {M1 M2 I2} (impl: forall R, M2 R -> Program I2 R)
        {R R'} (p: Program (PairMethod M1 I2) R) (f: R -> Program (PairMethod M1 M2) R') r,
        
        bindFlattenProgramRight impl p f = Return r ->
        False.

    Proof.
      introv H1. destruct p; simplFlattenIn H1; inverts H1.
    Qed.    

    (*
    Lemma FlattenRightFlattenRight':
      forall R R' (m: M R) (f: R -> Program BaseMethod R'),
        programEq
          (flattenProgramRight (objProgram seqImpl)
                               (r <- callSeqImpl m;
                                f r))
          (bindFlattenProgramRight (objProgram seqImpl)
                                (seqImpl.(objProgram) _ m) f).
    
    Proof.
      intros.
      setoid_rewrite frobProgramId at 1. simpl.
      destruct (objProgram seqImpl R m).
      - setoid_rewrite frobProgramId at 4. simpl.
        constructor. intros. setoid_rewrite flattenProgramRightEq. reflexivity.
      - setoid_rewrite frobProgramId at 4. simpl. reflexivity.
      - setoid_rewrite frobProgramId at 4. simpl.
        constructor. setoid_rewrite flattenProgramRightEq. reflexivity.
    Qed.        
    *)

    Lemma flattenSeqImplNotOutCritical:
      forall R (m: M R),
        OutCritical
          (Noop
             (flattenSeqImpl
                (r <- callSeqImpl m; _ <- callLock doUnlock; Return r))) ->
        False.

    Proof.
      introv H1.
      progFrobSimplInAt H1 2.
      inverts H1.
    Qed.

    
    Lemma bindFlattenProgramRightNotOutCritical':
      forall (R : Type) p,
        OutCritical
          (Noop
             (bindFlattenProgramRight
                (objProgram seqImpl)
                (mapProgram (fun (A : Type) (m : ObjInstruction seqImpl A) => inr m) p)
                (fun r : R => _ <- callLock doUnlock; Return r))) ->
        False.

    Proof.
      introv H.      
      destruct p;
        progFrobSimplInAt H 2;
        inverts H.
    Qed.

    
    Lemma bindFlattenProgramRightNotOutCritical:
      forall R p,
        OutCritical
          (bindFlattenProgramRight
             (objProgram seqImpl)
             (mapProgram (fun (A : Type) (m : ObjInstruction seqImpl A) => inr m) p)
             (fun r : R => _ <- callLock doUnlock; Return r)) ->
        False.
    Proof.
      introv H.
      destruct p; simplFlattenIn H.
      + inverts H. 
      + progFrobSimplInAt H 2. inverts H.
      + eapply bindFlattenProgramRightNotOutCritical'; eassumption.
    Qed.
    
    Lemma noopRel:
      forall R (p1: Program RBaseMethod R) p2 s1 s2,
        PDRel (Noop p1) s1 p2 s2 ->
        (PDRel p1 s1 p2 s2
         /\ (forall R (p1t: Program RBaseMethod R) p2t,
               PDRel p1t s1 p2t s2 ->
               PPRel (Noop p1) p2 p1t p2t ->
               PPRel p1 p2 p1t p2t))
        \/ (exists p2',
              p2 = Noop p2'
              /\ PDRel p1 s1 p2' s2
              /\ (forall R (p1t: Program RBaseMethod R) p2t,
                    PDRel p1t s1 p2t s2 ->
                    PPRel (Noop p1) p2 p1t p2t ->
                    PPRel p1 p2' p1t p2t))

      \/ (exists p2',
          ProgramSteps (coarsen seqImpl) _ p2 s2 p2' s2
          /\ PDRel p1 s1 p2' s2
          /\ (forall R (p1t: Program RBaseMethod R) p2t,
                PDRel p1t s1 p2t s2 ->
                PPRel (Noop p1) p2 p1t p2t ->
                PPRel p1 p2' p1t p2t)).

    Proof.
      introv H1.
      inverts H1 as.

      (*
      - left. splits.
        + constructor.
        + introv H1 H2.
          unfold PPRel in *.
          destruct H2 as [H2 | H2].
          * left. constructor.
          * auto.
      *)

      - intros.
        left. splits.
        + econstructor.
        + introv H2 H3.
          unfold PPRel in *.
          destruct H3 as [H3 | H3].
          * left. constructor.
          * auto.

      - intros.
        left. splits.
        + progFrobSimpl. econstructor; assumption.
        + introv H2 H3.
          unfold PPRel in *.
          destruct H3 as [H3 | H3].
          * exfalso. eapply flattenSeqImplNotOutCritical. eassumption.
          * auto.

      - introv H1.
        left. splits.
        + econstructor.
          eassumption.
          constructor.
        + introv H2 H3.
          unfold PPRel in *.
          destruct H3 as [H3 | H3].
          * exfalso.
            eapply bindFlattenProgramRightNotOutCritical'; eassumption.
          * auto.

        Import Coq.Classes.Morphisms.
      - introv H1 H2 H3.
        left.
        unfold bindFlattenSeqImpl in H3;
          destruct pc;
          simplFlattenIn H3.
        + inverts H3.
        + progFrobSimplInAt H3 2.
          progFrobSimplInAt H3 3.

          lazymatch type of H3 with
          | Noop (Noop (Bind _ (fun r0: unit => ?f))) = _ =>
            replace (fun r0 => f) with (fun r0: unit => frobProgram f) in H3;
              simpl in H3
          end.
          Focus 2. 
          simpl. apply functional_extensionality. intros. progFrobSimpl. f_equal.

          lazymatch type of H3 with
          | Noop (Noop (Bind _ (fun _ => Noop ?f))) = _ =>
            replace f with (frobProgram f)
              in H3;
            simpl in H3
          end.
          Focus 2.
          simpl. progFrobSimpl. reflexivity.

          inverts H3.
          splits.
          * constructor; assumption.
          * { introv H3 H4. 
              unfold PPRel in *.
              destruct H4 as [H4 | H4].
              * exfalso. progFrobSimplIn H4. progFrobSimplInAt H4 2.
                inverts H4.
              * auto.
            }

        + inverts H3.
          splits.
          * constructor.
            assumption.
            eapply ProgramStepsStepImplSteps. eassumption. constructor.
          * { introv H3 H4. 
              unfold PPRel in *.
              destruct H4 as [H4 | H4].
              * exfalso.
                progFrobSimplIn H4.
                eapply bindFlattenProgramRightNotOutCritical'. eassumption.
              * auto.
            }

      - introv H1 H2.
        left. splits.
        + constructor; assumption.
        + introv H3 H4.
          unfold PPRel in *.
          destruct H4 as [H4 | H4].
          * inverts H4.
          * auto.

      - left. splits.
        + constructor.
        + introv H3 H4.
          unfold PPRel in *.
          destruct H4 as [H4 | H4].
          * left. constructor.
          * auto.
    Qed.

    Lemma invRel:
      forall R m sc sa,
        DRel sc sa ->
        PDRel (objProgram cImpl R m) sc
              (objProgram aImpl R m) sa
        /\ forall R' (pc: Program cImpl.(ObjInstruction) R') pa,
            PDRel pc sc pa sa ->
            PPRel (objProgram cImpl R m) (objProgram aImpl R m)
                  pc pa.

    Proof.
      introv H1.
      unfold cImpl. unfold aImpl. simpl.
      unfold impl. unfold flattenSeqImpl. simplFlatten.
      splits.
      + progFrobSimplAt 1.
        progFrobSimplAt 2.
        progFrobSimplAt 3.
        constructor.
      + introv H2. unfold PPRel. left.
        progFrobSimplAt 1.
        progFrobSimplAt 2.
        progFrobSimplAt 3.
        constructor.        
    Qed.

    
    Lemma retRel:
      forall R (r: R) sc pa sa,
        PDRel (Return r) sc pa sa ->
        DRel sc sa ->
        pa = Return r.

    Proof.
      introv H1 H2.
      inverts H1 as.
      - introv H3 H4 H5. exfalso. eapply FlattenProgramRightRet'. eassumption.
      - reflexivity.
    Qed.

    
    Lemma PDRelPres1:
      forall {R} s1 s1' s2
        (p1: Program RBaseMethod R)
        (p2: Program M R),

        rLockState s1 = false ->
        PDRel p1 s1 p2 s2 ->
        PDRel p1 s1' p2 s2.

    Proof.
      introv H1 H2.
      inverts H2.
      (* - constructor. *)
      - constructor.
      - constructor.
      - congruence.
      - congruence.
      - congruence.
      - congruence.
      - congruence.
      - constructor.
      - constructor.
      (* - constructor. *)
    Qed.    

    Lemma PPRelPres1:
      forall R s1 s2
        (m : M R) (R' : Type) (p1t : Program RBaseMethod R')
        (p2t : Program M R'),
        rLockState s1 = false ->                  
        PDRel p1t s1 p2t s2 ->
        PPRel
          (Noop (flattenSeqImpl (r <- callSeqImpl m; _ <- callLock doUnlock; Return r)))
          (r <- m; Return r) p1t p2t.

    Proof.  
      introv H1 H2.                
      unfold PPRel.
      right.
      inverts H2.
      (* + constructor. *)
      + constructor.
      + constructor.
      + congruence.
      + congruence.
      + congruence.
      + congruence.
      + congruence.
      + constructor.
      + constructor.
    Qed.

    Lemma PDRelPres2:
      forall (R' R : Type) (s1 : RBaseState) (s2 : State (objBase seqImpl))
        (s1' : LockState * State (objBase seqImpl))
        (m : M R) (i : seqImpl.(ObjInstruction) R') p
        R'' (p1t : Program RBaseMethod R'') (p2t : Program M R''),
        PPRel
          (r <- inr i;
           bindFlattenProgramRight
             (objProgram seqImpl)
             (mapProgram
                (fun (A : Type) (m : ObjInstruction seqImpl A) => inr m) (p r))
             (fun r0 : R => _ <- callLock doUnlock; Return r0))
          (r <- m; Return r)
          p1t p2t ->
        PDRel p1t s1 p2t s2 ->
        PDRel p1t s1' p2t s2.
  
    Proof.  
      introv H1 H2.
      inverts H2 as.

      (* + constructor. *)
      + constructor.
      + constructor.

      + introv H4. exfalso.
        unfold PPRel in *.
        destruct H1 as [H1 | H1].
        * inverts H1.      
        * eapply flattenSeqImplNotOutCritical; eassumption.

      + introv H4. exfalso.
        unfold PPRel in *.
        destruct H1 as [H1 | H1].
        * inverts H1.      
        * eapply bindFlattenProgramRightNotOutCritical'; eassumption.
          
      + introv H4 H5. exfalso.
        unfold PPRel in *.
        destruct H1 as [H1 | H1].
        * inverts H1.      
        * eapply bindFlattenProgramRightNotOutCritical; eassumption.

      + introv H4 H5. exfalso.
        unfold PPRel in *.
        destruct H1 as [H1 | H1]; inverts H1.      
          
      + introv H4 H5. exfalso.
        unfold PPRel in *.
        destruct H1 as [H1 | H1]; inverts H1.      

      + constructor.
      + constructor.
    Qed.
    
    Lemma PPRelPres2:
      forall (R' R : Type) (r : R') (m : M R)
        (i : seqImpl.(ObjInstruction) R') f
        (R'' : Type) (p1t : Program RBaseMethod R'') (p2t : Program M R''),
        
        PPRel (r <- inr i;
               bindFlattenProgramRight
                 (objProgram seqImpl)
                 (mapProgram
                    (fun (A : Type) (m : ObjInstruction seqImpl A) => inr m) (f r))
                 (fun r : R => _ <- callLock doUnlock; Return r))
              (r <- m; Return r)
              p1t p2t ->
        PPRel (bindFlattenProgramRight
                 (objProgram seqImpl)
                 (mapProgram
                    (fun (A : Type) (m : ObjInstruction seqImpl A) => inr m) (f r))
                 (fun r : R => _ <- callLock doUnlock; Return r))
              (r <- m; Return r)
              p1t p2t.

    Proof.
      introv H1.
      unfold PPRel in *.
      destruct H1 as [H1 | H1].
      - inverts H1.
      - right. assumption.
    Qed.

    

    Lemma PDRelPres3:
      forall (R : Type) (s1 : RBaseState) (s2 : State (objBase seqImpl))
        (s1' : LockState * State (objBase seqImpl))
        (m : M R) (r : R)
        (R'' : Type) (p1t : Program RBaseMethod R'') (p2t : Program M R''),
        PPRel (_ <- rCallLock doUnlock; Noop (Return r))
              (r <- m; Return r)
              p1t p2t ->
        PDRel p1t s1 p2t s2 ->
        PDRel p1t s1' p2t (rSeqObjState s1).

    Proof.
      introv H1 H2.
      inverts H2 as.
      
      (* + constructor. *)
      + constructor.
      + constructor.

      + introv H2. exfalso.
        unfold PPRel in *.
        destruct H1 as [H1 | H1].
        * inverts H1.
        * eapply flattenSeqImplNotOutCritical; eassumption.
        
      + introv H2. exfalso.
        unfold PPRel in *.
        destruct H1 as [H1 | H1].
        * inverts H1.
        * eapply bindFlattenProgramRightNotOutCritical'; eassumption.
        
      + introv H2 H3. exfalso.
        unfold PPRel in *.
        destruct H1 as [H1 | H1].
        * inverts H1.
        * eapply bindFlattenProgramRightNotOutCritical; eassumption.

      + introv H2 H3. exfalso.
        unfold PPRel in *.
        destruct H1 as [H1 | H1]; inverts H1.

      + introv H2 H3. exfalso.
        unfold PPRel in *.
        destruct H1 as [H1 | H1]; inverts H1.

      + constructor.
      + constructor.
    Qed.


    Lemma PPRelPres3:
      forall (R : Type)
        (m : M R) (r : R)
        (R'' : Type) (p1t : Program RBaseMethod R'') (p2t : Program M R''),
        
        PPRel (_ <- rCallLock doUnlock; Noop (Return r))
              (r <- m; Return r)
              p1t p2t ->
        PPRel (Noop (Return r))
              (r <- m; Return r)
              p1t p2t.
    
    Proof.
      introv H1.
      unfold PPRel in *.
      iauto.
    Qed.
    
    Lemma stepRel:
      forall R' m R (f: R' -> Program RBaseMethod R) s1 p2 s2 s1' r,
        DRel s1 s2 ->
        PDRel (Bind m f) s1 p2 s2 ->
        rBase.(EvalInstruction) m s1 r s1' ->
        exists p2' s2',
          ProgramSteps (coarsen seqImpl) _ p2 s2 p2' s2'
          /\ DRel s1' s2'
          /\ PDRel (f r) s1' p2' s2'
          /\ (forall R (p1t: Program RBaseMethod R) p2t,
                PDRel p1t s1 p2t s2 ->
                PPRel (Bind m f) p2 p1t p2t (* s1 s2 *) ->
                PDRel p1t s1' p2t s2'
                /\ PPRel (f r) p2' p1t p2t (* s1' s2' *)).

    Proof.
      introv H1 H2 H3.
      inverts H2 as; subst.

      (* Getting the lock. *)
      - { simpl in H3. destruct H3 as (H31 & H32). apply invertLockSem in H31.
          destruct H31 as (H311 & H312 & H313). subst.      

          eexists. eexists.
          splits.

          - eapply ProgramStepsRefl.

          - unfold DRel. unfold rLockState. intro. congruence.

            
          -  setoid_rewrite frobProgramId at 1; simpl.
             unfold DRel in *. unfold rLockState in *. unfold rSeqObjState in *. apply H1 in H312. subst. rewrite H32.
             econstructor.            
             unfold rLockState in *. assumption.             
             
          - introv H4 H5.
            split.
            + eapply PDRelPres1.
              * unfold rLockState. eassumption.
              * assumption.
                
            + eapply PPRelPres1; eassumption.
        }        

      (* A call on the seq impl. *)
      - { introv H4 H5 H6.
          destruct pc; unfold bindFlattenSeqImpl in H6; simplFlattenIn H6; inverts H6.

          simpl in *. destruct H3 as (H31 & H32).
          eexists. eexists. splits.

          - eapply ProgramStepsRefl.

          - unfold DRel. unfold rLockState in *. intro. congruence.
            
          - econstructor.
            + unfold rLockState in *. congruence.
            + eapply ProgramStepsStepImplSteps. eassumption. constructor. unfold rSeqObjState. assumption.
                
          - introv H7 H8.
            unfold bindFlattenSeqImpl in H8. simplFlattenIn H8.
            split.
            
            + applys PDRelPres2 s1 s1'; eassumption.

            + eapply PPRelPres2; eassumption.
        }

      (* The unlock. *)
      - { introv H4 H5.
          eexists. eexists. splits.
          
          - eapply ProgramStepImplProgramSteps.
            constructor. simpl. unfold EvalProgram.
            eassumption.
            
          - simpl in H3. destruct H3 as (H31 & H32).
            unfold DRel. unfold rSeqObjState.
            intro. congruence.
            
          - econstructor. 
            
          - simpl in *. destruct H3 as (H31 & H32).
            apply invertLockSem in H31. destruct H31 as (H311 & H312 & H313). subst.
            introv H7 H8.
            split.
            
            + eapply PDRelPres3; eassumption.
              
            + eapply PPRelPres3; eassumption.
        }
    Qed.

    Lemma CoarseSyncSelfLin:
      Linearizable
        (coarsen seqImpl)
        syncImpl.
    Proof.
      unfold syncImpl.
      unfold base.
      unfold Linearizable.
      eapply LinSubstLeft.
      exact lockLin.
      unfold flattenImplProgRight.

      eapply simulatesCoInd; simpl.
      - eapply PPRelSym.
      - eapply initRel.
      - eapply noopRel.
      - eapply invRel.
      - eapply retRel.
      - eapply stepRel.
    Qed.    

    (*
      A compiler bug does not accept this.
      I have worked around it below.
    *)

    (*
    Module Type CoarseSyncLinParams.
      
      Parameter AbsState : Type.
      Parameter AbsEvalInstr: forall R, M R -> AbsState -> R -> AbsState -> Prop.
      Definition absObj: Spec := Build_Spec M AbsState AbsEvalInstr.
      
      Parameter initAbsObjState: AbsState.
      
      Parameter SeqImplRefinesAbsObj:
        SeqSimulatesIO
          absObj seqImpl
          initAbsObjState initSeqObjState.

    End CoarseSyncLinParams.
    
    Module CoarseSyncLinFun (CoarseSyncLinArgs: CoarseSyncLinParams).

      Import CoarseSyncLinArgs.

      Lemma CoarseSyncSelfLin:
        Linearizable
          absObj
          syncImpl 
          initAbsObjState
          initState.

      Proof.
        unfold Linearizable.
        eapply simulatesTrans.
        apply CoarseSyncSelfLin.
          
        eapply SeqRefImplConRef.
        apply SeqImplRefinesAbsObj.
      Qed.
        
          
    End CoarseSyncLinFun.
    *)
    
  End CoarseSyncSelfLinFun.
  
End CoarseSyncFun.


Module CoarseSyncFun' (CoarseSyncArgs: CoarseSyncParams).

  Import CoarseSyncArgs.
  Module M := CoarseSyncFun CoarseSyncArgs.
  Import M.

  Module Type CoarseSyncLinParams (CoarseSyncSelfLinArgs: CoarseSyncSelfLinParams).

    Import CoarseSyncSelfLinArgs.
    Module CoarseSyncSelfLin := CoarseSyncSelfLinFun CoarseSyncSelfLinArgs.
    Import CoarseSyncSelfLin.
    
    Parameter absObj: Spec M.
    
    Parameter SeqImplRefinesAbsObj : Simulates (coarsen seqImpl) absObj.

  End CoarseSyncLinParams.

  Module CoarseSyncLinFun
         (CoarseSyncSelfLinArgs: CoarseSyncSelfLinParams)
         (CoarseSyncLinArgs: CoarseSyncLinParams CoarseSyncSelfLinArgs).

    Import CoarseSyncSelfLinArgs.
    Import CoarseSyncLinArgs.
    Import CoarseSyncLinArgs.CoarseSyncSelfLin.
    
    Lemma CoarseSyncSelfLin : Linearizable absObj syncImpl.
    Proof.
      unfold Linearizable.
      eapply simulatesTrans.
      apply CoarseSyncSelfLin.

      eapply SeqRefImplConRef.
      apply SeqImplRefinesAbsObj.
    Qed.
    
  End CoarseSyncLinFun.

End CoarseSyncFun'.
