From Coq Require Import Eqdep FunctionalExtensionality List Omega.
From C4.Util Require LibTactics.
From C4 Require Import Predefs Exec TransactionTactics.

Ltac injpair_ := LibTactics.injpair_.

Set Implicit Arguments.


(** Herein lies a general proof strategy to show that an object of transactions
  * is faithful.  It's generic in details of state and methods.
  * The basic idea is that:
  * (1) Each thread is assigned a timestemp (consecutively) at the moment that it
  *     runs some initialization code.
  * (2) Any two out-of-order (w.r.t. timestamps) actions commute.
  * Therefore, "obviously," we can fix up an interleaved schedule into a serial schedule.
  *
  * We assume that every method of a high-level spec is mapped to precisely one
  * method of the low-level object spec. *)

(** Here are a number of auxiliary inductive definitions that may be explained at
  * the points where they're used below. *)

Section Commuting.

  Variable absObj: SpecType.
  Variable absInit: State absObj.

  Variable concObj: SpecType.
  Variable concInit: State concObj.

  Variable initialize: TId -> Method concObj unit.

  Variable translateMethod: TId
                            -> forall R, Method absObj R
                                   -> Method concObj (option R).

  Variable translateReturn: TId
                            -> option nat
                            -> Method concObj unit.

  Variable timestamp : State concObj -> TId -> nat.

  Inductive step : Type -> Type :=
  | MethodStep : forall R, Method absObj R -> step (option R)
  | ReturnStep : option nat -> step unit.

  Inductive doStep (th : TId) : forall R, step R -> State concObj -> R -> State concObj -> Prop :=
  | DoMethod : forall R (m : Method absObj R) c r c',
    Exec concObj c (translateMethod th R m) r c'
    -> doStep th (MethodStep R m) c r c'
  | DoReturn : forall r c r' c',
    Exec concObj c (translateReturn th r) r' c'
    -> doStep th (ReturnStep r) c r' c'.

  Inductive doStepNoAbort (th : TId) : forall R, step R -> State concObj -> R -> State concObj -> Prop :=
  | DonaMethod : forall R (m : Method absObj R) c r c',
    Exec concObj c (translateMethod th R m) (Some r) c'
    -> doStepNoAbort th (MethodStep R m) c (Some r) c'
  | DonaReturn : forall r c r' c',
    Exec concObj c (translateReturn th (Some r)) r' c'
    -> doStepNoAbort th (ReturnStep (Some r)) c r' c'.

  Inductive doSteps (th : TId) : State concObj -> State concObj -> option (option nat) -> Prop :=
  | DsRefl : forall c,
    doSteps th c c None
  | DsReturn : forall c c' o r c'',
    doSteps th c c' None
    -> Exec concObj c' (translateReturn th o) r c''
    -> doSteps th c c'' (Some o)
  | DsMethodSome : forall R (m : Method absObj R) c c' r c'',
    doSteps th c c' None
    -> Exec concObj c' (translateMethod th R m) (Some r) c''
    -> doSteps th c c'' None
  | DsMethodNone : forall R (m : Method absObj R) c c' c'',
    doSteps th c c' None
    -> Exec concObj c' (translateMethod th R m) None c''
    -> doSteps th c c'' (Some None).

End Commuting.

Arguments ReturnStep {_} _.
Arguments MethodStep {_} _.

(** This is the signature that an object must follow. *)
Module Type S.
  (** Spec and initial state at the abstract (source) level *)
  Parameter absObj: SpecType.
  Parameter absInit: State absObj.

  (** Spec and initial state at the concrete (target) level *)
  Parameter concObj: SpecType.
  Parameter concInit: State concObj.

  (** Extract the running counter of thread timestamps from a concrete state. *)
  Parameter clock: State concObj -> nat.

  (** Run the initialization code for a given thread (assigning it a timestamp, etc.). *)
  Parameter initialize: TId -> Method concObj unit.

  (** Translate a source method to its target counterpart.
    * The idea is that the target method may return [Some], in which case
    * execution proceeds as normal; or [None], in which case we abort. *)
  Parameter translateMethod: TId
                             -> forall R, Method absObj R
                                          -> Method concObj (option R).

  (** Whenever the source program runs [Return r], in the translated code we also first
    * run the output of this function applied to [r].  For instance, for [r = None],
    * we expect this to be some rollback code. *)
  Parameter translateReturn: TId
                             -> option nat
                             -> Method concObj unit.

  (** * A set of axioms ensuring that the above translation functions
      * have disjoint ranges and are properly injective *)

  Axiom discriminate_mi : forall th1 th2 R (m1 : Method absObj R),
    existT (Method concObj) _ (translateMethod th1 R m1)
    <> existT (Method concObj) _ (initialize th2).

  Axiom discriminate_ri : forall th1 th2 r,
    translateReturn th1 r <> initialize th2.

  Axiom discriminate_mr : forall th1 th2 R (m1 : Method absObj R) r2,
    existT (Method concObj) _ (translateMethod th1 R m1)
    <> existT (Method concObj) _ (translateReturn th2 r2).

  Axiom translateMethod_inj : forall th R1 R2 (m1 : Method absObj R1) (m2 : Method absObj R2),
    existT (Method concObj) _ (translateMethod th R1 m1) = existT (Method concObj) _ (translateMethod th R2 m2)
    -> existT (Method absObj) R1 m1 = existT (Method absObj) R2 m2.

  (** This next parameter is a key part of the proof strategy.
    * We summarize any aborted transaction with this relation.
    * At the abstract level, aborted transactions should have no effects,
    * so you might think that this relation should always be equality.
    * However, an aborted transaction may leave behind some suitably
    * benign side effects in the concrete state, and this relation
    * makes "benign" formal. *)
  Parameter abortR : TId -> State concObj -> State concObj -> Prop.

  (** Here's how we look up the timestamp assigned to a thread so far. *)
  Parameter timestamp : State concObj -> TId -> nat.

  (** The initialization method had better increment the clock. *)
  Axiom clock_initialize : forall c th r c',
    Exec concObj c (initialize th) r c'
    -> clock c' = S (clock c).

  (** Steps (methods and returns) had better not change the clock. *)
  Axiom clock_step : forall c th R (s : step _ R) r c',
    doStep _ translateMethod translateReturn th s c r c'
    -> clock c' = clock c.

  (** An aborting transaction should increment the clock. *)
  Axiom clock_abortR : forall th c c',
    abortR th c c'
    -> clock c' = S (clock c).

  (** Initialization should assign the thread the next available timestamp. *)
  Axiom initialize_initializes : forall c th r c',
    Exec concObj c (initialize th) r c'
    -> timestamp c' th = clock c.

  (** Another essential element: an invariant holding of every state reachable
    * in a concrete execution, be it interleaved or serialized. *)
  Parameter invariant : State concObj -> Prop.

  (** Since it's an invariant, it had better hold in the initial state. *)
  Axiom invariant_init : invariant concInit.

  (** It should guarantee that no timestamps exceed the current clock. *)
  Axiom clock_bound : forall c th,
    invariant c
    -> timestamp c th < clock c.

  (** * And it had better be preserved by each of the key operations. *)

  Axiom invariant_initialize : forall th c r c',
    timestamp c th = 0
    -> Exec concObj c (initialize th) r c'
    -> invariant c
    -> invariant c'.

  Axiom invariant_step : forall th R (s : step _ R) c r c',
    doStep _ translateMethod translateReturn th s c r c'
    -> timestamp c th > 0
    -> invariant c
    -> invariant c'.

  Axiom invariant_abortR : forall th c c',
    abortR th c c'
    -> timestamp c th = 0
    -> invariant c
    -> invariant c'.

  (** The initial state has all timestamps zeroed out. *)
  Axiom stamps_init : forall th, timestamp concInit th = 0.

  (** Initializing a thread doesn't affect any other thread's timestamp. *)
  Axiom stamps_initialize : forall th c r c',
    Exec concObj c (initialize th) r c'
    -> forall th', th' <> th -> timestamp c' th' = timestamp c th'.

  (** Steps don't affect timestamps. *)
  Axiom stamps_step : forall th R (s : step _ R) c r c',
    doStep _ translateMethod translateReturn th s c r c'
    -> forall th', timestamp c' th' = timestamp c th'.

  (** Aborting assigns the current thread's timestamp as usual. *)
  Axiom stamps_abortR_same : forall th c c',
    abortR th c c'
    -> timestamp c' th = clock c.

  (** But it doesn't change any other timestamps. *)
  Axiom stamps_abortR_diff : forall th c c',
    abortR th c c'
    -> forall th', th' <> th -> timestamp c' th' = timestamp c th'.

  (** Here's another essential part: when a target method aborts, we can
    * achieve the same state change by aborting directly.
    * This switch is crucial for flexibility.
    * Consider the following execution tagged with timestamps, where t1 < t2.
    *   (t2: a2); (t1: a1)
    * a1 and a2 are actions of some kind.  We want to commute into the right order:
    *   (t1: a1); (t2: a2)
    * However, it's possible that a1 was a method that aborted, while now,
    * without having a2 run first and spoil the state, a1 is able to run without
    * aborting, and indeed can't abort anymore.  We will instead replace a1 in the
    * first order with a direct abort, which we then commute past a2. *)
  Axiom method_aborts : forall c th R (m : Method _ R) c',
    Exec concObj c (translateMethod th R m) None c'
    -> Exec concObj c (translateReturn th None) tt c'.

  (** * Now the most important proof obligations: a step can commute past
      * a number of other things
      * (another step, a full aborted transaction, or an initialization).
      *
      * There are some funky side conditions preventing us from needing to
      * consider some actions that abort.  For instance, we never
      * consider aborting method calls directly, since we will replace them
      * with direct aborts. *)

  (** Beyond ruling out aborting method calls, here we see that we also rule
    * out cases where the lefthand action aborts, period.  That's because
    * the next axiom handles that case, but where we must commute past a full
    * aborted transaction, rather than just its final aborting action.
    * It's a more complex condition that is nonetheless logically weaker. *)
  Axiom steps_commute : forall c th1 th2 R1 (s1 : step _ R1) R2 (s2 : step _ R2) r1 c' r2 c'',
    timestamp c th1 > timestamp c th2
    -> doStep _ translateMethod translateReturn th1 s1 c r1 c'
    -> doStep _ translateMethod translateReturn th2 s2 c' r2 c''
    -> match s1 in step _ R return R -> _ with
       | ReturnStep o => fun _ => o <> None
       | MethodStep _ _ => fun r => r <> None
       end r1
    -> match s2 in step _ R return R -> _ with
       | ReturnStep _ => fun _ => True
       | MethodStep _ _ => fun r => r <> None
       end r2
    -> invariant c
    -> exists c''', doStep _ translateMethod translateReturn th2 s2 c r2 c'''
                    /\ doStep _ translateMethod translateReturn th1 s1 c''' r1 c''.

  Axiom abort_step_commute : forall c th1 th2 R2 (s2 : step _ R2) c' r2 c'',
    abortR th1 c c'
    -> doStep _ translateMethod translateReturn th2 s2 c' r2 c''
    -> th1 <> th2
    -> match s2 in step _ R return R -> _ with
       | ReturnStep _ => fun _ => True
       | MethodStep _ _ => fun r => r <> None
       end r2
    -> invariant c
    -> exists c''', doStep _ translateMethod translateReturn th2 s2 c r2 c'''
                    /\ abortR th1 c''' c''.

  Axiom init_step_commute : forall c th1 th2 R2 (s2 : step _ R2) r1 c' r2 c'',
    Exec concObj c (initialize th1) r1 c'
    -> doStep _ translateMethod translateReturn th2 s2 c' r2 c''
    -> th1 <> th2
    -> match s2 in step _ R return R -> _ with
       | ReturnStep _ => fun _ => True
       | MethodStep _ _ => fun r => r <> None
       end r2
    -> invariant c
    -> exists c''', doStep _ translateMethod translateReturn th2 s2 c r2 c'''
                    /\ Exec _ c''' (initialize th1) r1 c''.

  (** * Last crucial ingredient: an abstraction function mapping concrete states
      * to abstract states.  We require that all of the concrete operations have
      * the expected effects on the corresponding abstract states. *)

  Parameter absOf : State concObj -> State absObj.

  Axiom absOf_init : absOf concInit = absInit.

  Axiom absOf_initialize : forall th c r c', Exec concObj c (initialize th) r c'
    -> absOf c' = absOf c.

  Axiom absOf_method : forall th R (m : Method absObj R) c r c',
    Exec concObj c (translateMethod th _ m) (Some r) c'
    -> Exec absObj (absOf c) m r (absOf c').

  Axiom absOf_return : forall th o c r c',
    Exec concObj c (translateReturn th (Some o)) r c'
    -> absOf c' = absOf c.

  Axiom absOf_abortR : forall th c c',
    abortR th c c'
    -> absOf c' = absOf c.

  (** Lastly: any aborting transaction really does satisfy [abortR]. *)
  Axiom abortR_in : forall th c r c' c'',
    invariant c
    -> timestamp c th = 0
    -> Exec concObj c (initialize th) r c''
    -> doSteps _ _ translateMethod translateReturn th c'' c' (Some None)
    -> abortR th c c'.
End S.

(** No comments yet on the object; maybe some day! *)
Module Make(Import M : S).

  Theorem clock_pos : forall c,
    invariant c
    -> clock c > 0.
  Proof.
    intros.
    assert (timestamp c 0 < clock c) by eauto using clock_bound.
    omega.
  Qed.

  Fixpoint translateTrans' (th : TId) (tr : Trans absObj) : Trans concObj :=
    match tr with
    | Return r =>
      _ <- translateReturn th r;
      Return r
    | Bind m tr' =>
      x <- translateMethod th _ m;
      match x with
      | None => Return None
      | Some v => translateTrans' th (tr' v)
      end
    end.

  Definition translateTrans (th : TId) (tr : Trans absObj) : Trans concObj :=
    _ <- initialize th;
    translateTrans' th tr.

  Module Args <: SerializabilityPars.
    Definition absObj := absObj.
    Definition absInit := absInit.
    Definition concObj := concObj.
    Definition concInit := concInit.
    Definition translateTrans := translateTrans.
  End Args.

  Module Import Ser := Serializability Args.

  Inductive orderedStep :
    CTrans concObj -> State concObj -> CTrans concObj -> State concObj -> Prop :=
  | orderedStepExec {t cp p s p' s'}:
    In t tids
    -> timestamp s t = 0
    -> progExec _ _ (translateTrans t p) s p' s'
    -> orderedStep (override cp t (translateTrans t p)) s 
                   (override cp t p') s'
  | orderedStepAbort {t cp p s s'}:
    In t tids
    -> timestamp s t = 0
    -> abortR t s s'
    -> orderedStep (override cp t (translateTrans t p)) s 
                   (override cp t (Return None)) s'.

  Inductive ordered: CTrans concObj -> State concObj -> CTrans concObj -> State concObj -> Prop :=
  | sTransExecRefl cp s: 
    ordered cp s cp s
  | orderedSteps s cp cp' s' cp'' s'': 
    ordered cp s cp' s'
    -> orderedStep cp' s' cp'' s''
    -> ordered cp s cp'' s''.

  Hint Constructors orderedStep ordered doStep.

  Hint Rewrite @override_new_val @override_old_val using (omega || congruence).

  Ltac invert H :=
    inversion H; clear H; subst; injpair_; subst.

  Definition methodOf (tr : Trans concObj) : {R : Type & Method concObj R} :=
    match tr with
    | Return _ => existT _ _ (initialize 0)
    | Bind m _ => existT _ _ m
    end.

  Definition contOf (tr : Trans concObj) : {R : Type & R -> Trans concObj} :=
    match tr with
    | Return _ => existT (fun R : Type => R -> Trans concObj) unit (fun _ => Return None)
    | Bind _ c => existT _ _ c
    end.

  Ltac invert_Bind :=
    match goal with
    | [ H : Bind _ _ = Bind _ _ |- _ ] =>
      let H' := fresh "H'" in
      generalize (f_equal methodOf H); simpl; intro H';
      apply (f_equal contOf) in H; simpl in H;
      repeat match goal with
             | [ H : existT _ ?x _ = existT _ ?y _ |- _ ] =>
               match x with
               | y => fail 1
               | _ => generalize (f_equal (@projT1 _ _) H); simpl; intro; (progress subst || clear H)
               end
             | _ => injpair_
             end; subst
    end.

  Hint Resolve invariant_initialize invariant_step initialize_initializes.

  Hint Constructors progExec progExecStep.

  Hint Extern 1 (_ <= _) => omega.

  Lemma option_not_unit : forall R,
    inhabited R
    -> option R = unit
    -> False.
  Proof.
    intros.
    assert (exists x : unit, forall y : unit, x = y) by (exists tt; destruct y; auto).
    rewrite <- H0 in H1.
    destruct H1.
    generalize (H1 None).
    destruct H.
    specialize (H1 (Some X)).
    congruence.
  Qed.

  Lemma progExec_translateTrans' : forall th p0 p s p' s',
    progExec concObj (option nat) p s p' s'
    -> p = translateTrans' th p0
    -> (exists p'', p' = translateTrans' th p'')
       \/ (exists r, p' = Return r).
  Proof.
    induction 1; solver; eauto.

    invert H0; eauto.

    destruct x; simpl in *.
    invert_Bind.
    destruct r; eauto.
    invert_Bind; eauto.

    inversion H0.
  Qed.

  Lemma progExec_translateTrans : forall th p0 p s p' s',
    progExec concObj (option nat) p s p' s'
    -> p = translateTrans th p0
    -> (s = s' /\ p' = translateTrans th p0)
       \/ (exists p'', p' = translateTrans' th p'' /\ timestamp s' th = clock s)
       \/ (exists r, p' = Return r).
  Proof.
    induction 1; solver; eauto.

    invert H0.
    erewrite initialize_initializes by eauto.
    eauto.

    invert H0.
    destruct x; simpl in *.
    invert_Bind.
    erewrite stamps_step by eauto.
    destruct r; eauto.
    invert_Bind; eauto.

    inversion H0.
  Qed.

  Lemma progExec_invariant : forall th p0 p s p' s',
    progExec concObj (option nat) p s p' s'
    -> invariant s
    -> timestamp s th = 0
    -> p = translateTrans th p0
    -> invariant s'.
  Proof.
    induction 1; solver.

    generalize H; intro He.
    eapply progExec_translateTrans in He; eauto.
    solver.
    invert H0; eauto.

    destruct x; simpl in *.
    invert H0; eauto.
    specialize (clock_pos H1).
    eauto.
    
    specialize (clock_pos H1).
    invert H0; eauto.

    invert H0.
  Qed.

  Hint Resolve progExec_invariant.

  Lemma doStep_invert : forall R (s : step _ R) th c r c',
    doStep _ translateMethod translateReturn th s c r c'
    -> match s in step _ R return R -> Prop with
       | MethodStep _ m => fun r => Exec concObj c (translateMethod th _ m) r c'
       | ReturnStep r => fun r' => Exec concObj c (translateReturn th r) r' c'
       end r.
  Proof.
    destruct 1; auto.
  Qed.


  Hint Constructors doSteps.

  Lemma progExec_doSteps' : forall t p s p' s',
    progExec concObj (option nat) p s p' s'
    -> forall p0, p = translateTrans' t p0
    -> doSteps _ _ translateMethod translateReturn t s s' (match p' with
                                                             | Return r => Some r
                                                             | _ => None
                                                           end).
  Proof.
    induction 1; intros; subst.

    destruct p0; auto.

    solver.
    specialize (IHprogExec _ eq_refl).
    destruct p'.
    2: invert H0.
    eapply progExec_translateTrans' in H; eauto; solver.
    destruct x; simpl in *.
    invert_Bind.
    invert H0.
    destruct r; try discriminate.
    eauto.
    invert_Bind.
    invert H0.
    discriminate.
    discriminate.

    specialize (IHprogExec _ eq_refl).
    destruct p'.
    2: invert H0.
    eapply progExec_translateTrans' in H; eauto; solver.
    destruct x; simpl in *.
    invert_Bind.
    invert H0.
    destruct r.
    destruct (p1 a); discriminate.
    invert H5.
    eauto.
    invert_Bind.
    invert H0.
    invert H5.
    eauto.
    discriminate.
  Qed.

  Lemma progExec_doSteps'' : forall t p s p' s',
    progExec concObj (option nat) p s p' s'
    -> forall p0, p = translateTrans t p0
    -> (p' = translateTrans t p0 /\ s' = s)
       \/ exists s0, Exec concObj s (initialize t) tt s0
                     /\ doSteps _ _ translateMethod translateReturn t s0 s' (match p' with
                                                                             | Return r => Some r
                                                                             | _ => None
                                                                             end).
  Proof.
    induction 1; intros; subst.

    eauto.

    specialize (IHprogExec _ eq_refl).
    destruct IHprogExec; intuition subst.
    invert H0.
    destruct r.
    destruct p0; simpl; eauto.

    eapply progExec_translateTrans in H; eauto.
    intuition subst.
    replace (match translateTrans t p0 with
             | Bind _ _ => None
             | Return r => Some r
             end) with (@None (option nat)) in * by (destruct p0; auto).
    invert H0.
    replace (match translateTrans' t p0 with
             | Bind _ _ => None
             | Return r => Some r
             end) with (@None (option nat)) in * by (destruct p0; auto).
    destruct r; eauto.

    destruct H1; intuition idtac.
    destruct H; subst.
    replace (match translateTrans' t x0 with
             | Bind _ _ => None
             | Return r => Some r
             end) with (@None (option nat)) in * by (destruct x0; auto).
    invert H0.
    destruct x0; simpl in *.
    destruct H.
    invert_Bind.
    destruct r.
    replace (match translateTrans' t (p a) with
             | Bind _ _ => None
             | Return r => Some r
             end) with (@None (option nat)) in * by (destruct (p a); auto).
    eauto.
    eauto.
    destruct H.
    invert_Bind.
    eauto.
    destruct H; subst.
    invert H0.
  Qed.

  Lemma progExec_abortR : forall t p s s',
    progExec concObj (option nat) (translateTrans t p) s (Return None) s'
    -> invariant s
    -> timestamp s t = 0
    -> abortR t s s'.
  Proof.
    intros.
    eapply progExec_doSteps'' in H; eauto.
    solver; eauto using abortR_in.
    discriminate.
  Qed.

  Lemma reorder_method'' : forall th1 th2 R (m : Method absObj R) p s p' s',
    invariant s
    -> timestamp s th2 = 0
    -> progExec _ (option nat) p s p' s'
    -> forall p0, p = translateTrans th2 p0
    -> forall r s'', Exec concObj s' (translateMethod th1 R m) r s''
    -> timestamp s' th2 > timestamp s' th1
    -> p' <> Return None
    -> r <> None
    -> exists s''', Exec _ s (translateMethod th1 R m) r s'''
                    /\ timestamp s''' th2 = 0
                    /\ progExec _ _ p s''' p' s''.
  Proof.
    induction 3; solver; eauto.

    assert (Hne : th1 <> th2) by (intro; subst; omega).

    generalize H1; intro He.
    eapply progExec_translateTrans in He; eauto.
    solver.

    clear H4.
    invert H2.
    eapply init_step_commute in H13; eauto.
    solver.
    destruct r; try tauto.
    invert H3.
    do 2 esplit.
    eauto.
    split.
    erewrite stamps_step by eauto; eauto.
    econstructor.
    eauto.
    unfold translateTrans.
    eapply progExecBind with (f := fun _ => translateTrans' th2 p0); eauto.

    destruct x; simpl in *.

    invert H2.
    assert (Hst : doStep _ translateMethod translateReturn th1 (MethodStep _ m) s'' r s''0) by eauto.
    destruct r0; try tauto.
    eapply steps_commute in Hst.
    3: eauto.
    3: simpl; congruence.
    3: eauto.
    2: erewrite <- stamps_step with (c' := s'') by eauto;
      erewrite <- stamps_step with (th' := th1) (c' := s'') by eauto; assumption.
    2: eauto.
    solver.
    destruct r; try tauto.
    invert H3.
    invert H9.
    eapply H4 in H18; eauto.
    3: discriminate.
    2: erewrite <- stamps_step with (c' := s'') by eauto;
      erewrite <- stamps_step with (th' := th1) (c' := s'') by eauto; assumption.
    solver.
    do 2 esplit.
    eauto.
    split; auto.
    econstructor; eauto.
    apply progExecBind with (f := fun x1 =>
                                    match x1 with
                                    | Some v => translateTrans' th2 (p v)
                                    | None => Return None
                                    end) in H20; assumption.


    invert H2.
    assert (Hst : doStep _ translateMethod translateReturn th1 (MethodStep _ m) s'' r s''0) by eauto.
    destruct o; try tauto.
    eapply steps_commute in Hst.
    3: eauto.
    3: simpl; congruence.
    3: eauto.
    2: erewrite <- stamps_step with (c' := s'') by eauto;
      erewrite <- stamps_step with (th' := th1) (c' := s'') by eauto; assumption.
    2: eauto.
    solver.
    destruct r; try tauto.
    invert H3.
    apply doStep_invert in H9.
    eapply H4 in H18; eauto.
    2: erewrite <- stamps_step with (c' := s'') by eauto;
      erewrite <- stamps_step with (th' := th1) (c' := s'') by eauto; assumption.
    solver.
    do 2 esplit.
    eauto.
    split; auto.
    econstructor; eauto.
    eapply progExecBind with (f := fun _ => Return (Some n)); eauto.
    discriminate.

    invert H2.
  Qed.

  Lemma progExec_Return : forall p s p' s',
    progExec concObj (option nat) p s p' s'
    -> forall r, p = Return r
    -> p' = Return r /\ s' = s.
  Proof.
    induction 1.

    auto.

    intros; subst.
    specialize (IHprogExec _ eq_refl); destruct IHprogExec; subst.
    invert H0.
  Qed.

  Lemma ordered_translateTrans : forall ps s ps' s',
    ordered ps s ps' s'
    -> (forall th, exists p, ps th = translateTrans th p)
    -> forall th, (exists p, ps' th = translateTrans th p)
                  \/ (exists p, ps' th = translateTrans' th p)
                  \/ (exists r, ps' th = Return r).
  Proof.
    induction 1; solver; eauto.

    invert H0.
    destruct (H1 th).
    destruct (t =_? th); solver; eauto.

    specialize (H2 th); solver.
    eapply progExec_translateTrans in H5; eauto.
    solver; eauto.
    eapply progExec_translateTrans' in H5; eauto.
    eapply progExec_Return in H5; eauto.
    solver; eauto.

    specialize (H2 th); solver; eauto.

    destruct (t =_? th); solver; eauto.
    specialize (H2 th); solver; eauto.
  Qed.

  Hint Resolve stamps_initialize stamps_step clock_initialize clock_step.

  Lemma progExec_stamp' : forall t p0 p s p' s',
    progExec concObj (option nat) p s p' s'
    -> p = translateTrans' t p0
    -> forall th, timestamp s' th = timestamp s th.
  Proof.
    induction 1; solver.

    rewrite <- H2; clear H2.
    eapply progExec_translateTrans' in H; eauto; solver.
    
    invert H0.
    destruct x; simpl in *.
    invert_Bind; eauto.
    invert_Bind; eauto.

    inversion H0.
  Qed.

  Lemma progExec_clock' : forall t p0 p s p' s',
    progExec concObj (option nat) p s p' s'
    -> p = translateTrans' t p0
    -> clock s' = clock s.
  Proof.
    induction 1; solver.

    rewrite <- H2; clear H2.
    eapply progExec_translateTrans' in H; eauto; solver.
    
    invert H0.
    destruct x; simpl in *.
    invert_Bind; eauto.
    invert_Bind; eauto.

    inversion H0.
  Qed.

  Lemma progExec_stamp'' : forall t p0 p s p' s',
    progExec concObj (option nat) p s p' s'
    -> p = translateTrans t p0
    -> (s = s' /\ p' = translateTrans t p0)
       \/ (((exists p'', p' = translateTrans' t p'')
            \/ (exists r, p' = Return r)) /\ timestamp s' t = clock s /\ clock s' = S (clock s)).
  Proof.
    induction 1; solver.

    invert H0; eauto 8.

    right; split.
    eauto using progExec_translateTrans'.
    erewrite progExec_stamp'.
    3: eauto.
    2: eapply progExecSteps; eauto.
    erewrite progExec_clock' with (s := s') (s' := s'').
    3: eauto.
    2: eapply progExecSteps; eauto.
    auto.
    
    inversion H0.
  Qed.

  Lemma progExec_stamp : forall t p0 p s p' s',
    progExec concObj (option nat) p s p' s'
    -> p = translateTrans t p0
    -> (s = s' /\ p' = translateTrans t p0) \/ timestamp s' t = clock s /\ clock s' = S (clock s).
  Proof.
    intros.
    eapply progExec_stamp'' in H; eauto; solver.
  Qed.

  Hint Extern 1 (_ >= _) => omega.

  Lemma progExec_clock'' : forall p t p0 s p' s',
    progExec concObj (option nat) p s p' s'
    -> p = translateTrans t p0
    -> clock s' >= clock s /\ (p' = translateTrans t p0
                               \/ (exists p'', p' = translateTrans' t p'')
                               \/ (exists r, p' = Return r)).
  Proof.
    induction 1; simpl; intros; subst; auto.

    destruct (IHprogExec eq_refl) as [ ? [ | [ [ ] | [ ] ] ] ]; clear IHprogExec; subst.

    invert H0.
    erewrite clock_initialize by eauto; eauto.

    invert H0; simpl in *.
    destruct x; simpl in *.
    invert_Bind.
    erewrite clock_step by eauto.
    destruct r; eauto.
    invert_Bind.
    erewrite clock_step by eauto.
    eauto.
    
    invert H0.
  Qed.

  Lemma progExec_clock : forall p t p0 s p' s',
    progExec concObj (option nat) p s p' s'
    -> p = translateTrans t p0
    -> clock s' >= clock s.
  Proof.
    intros.
    eapply progExec_clock'' in H; eauto.
  Qed.    

  Lemma progExec_stamp_bound : forall th p0 p s p' s',
    progExec _ _ p s p' s'
    -> p = translateTrans th p0
    -> p' <> translateTrans th p0
    -> clock s' > timestamp s' th.
  Proof.
    induction 1; solver.

    eapply progExec_translateTrans in H; eauto.
    solver.
    invert H0.
    erewrite clock_initialize by eauto.
    erewrite initialize_initializes by eauto.
    auto.

    erewrite progExec_clock'.
    2: eapply progExecSteps; eauto.
    2: eauto.
    erewrite progExec_stamp'.
    2: eapply progExecSteps; eauto.
    2: eauto.
    apply H3.
    intros.
    unfold translateTrans in H.
    destruct x; simpl in *.
    inversion H.
    apply discriminate_mi in H6; tauto.
    invert_Bind.
    apply discriminate_ri in H'; auto.

    inversion H0.
  Qed.

  Lemma progExec_ostamp : forall t p0 p s p' s',
    progExec concObj (option nat) p s p' s'
    -> p = translateTrans t p0
    -> forall th, th <> t -> timestamp s' th = timestamp s th.
  Proof.
    induction 1; solver.

    rewrite <- H2 by auto; clear H2.
    eapply progExec_translateTrans with (th := t) in H; eauto; solver.

    invert H0; eauto.

    invert H0.
    destruct x; simpl in *.
    invert_Bind; eauto.
    invert_Bind; eauto.

    inversion H0.
  Qed.

  Lemma ordered_stamp_bound : forall th p cp s cp' s',
    cp th = translateTrans th p
    -> ordered cp s cp' s'
    -> (forall p, cp' th <> translateTrans th p)
    -> clock s' > timestamp s' th.
  Proof.
    induction 2; solver.

    exfalso; eauto.

    destruct H1.
    destruct (t =_? th); solver.

    eauto using progExec_stamp_bound.

    erewrite progExec_ostamp; eauto.
    generalize H5.
    intro Hp.
    eapply progExec_stamp in Hp; solver.

    erewrite clock_abortR by eauto.
    destruct (t =_? th); subst.
    erewrite stamps_abortR_same by eauto.
    auto.
    erewrite stamps_abortR_diff by eauto.
    solver.
  Qed.

  Lemma ordered_invariant : forall cp s cp' s',
    ordered cp s cp' s'
    -> invariant s
    -> invariant s'.
  Proof.
    induction 1; solver.
    destruct H0; eauto using progExec_invariant, invariant_abortR.
  Qed.

  Lemma double_override_neq : forall A (f : _ -> A) k1 k2 v1 v2,
    k1 <> k2
    -> override (override f k1 v1) k2 v2 = override (override f k2 v2) k1 v1.
  Proof.
    intros; extensionality k3.
    destruct (k1 =_? k3), (k2 =_? k3); solver.
  Qed.

  Lemma translateTrans_None : forall p s p' s' th,
    progExec concObj (option nat) p s p' s'
    -> p = translateTrans th (Return None)
    -> p' = translateTrans th (Return None)
       \/ p' = (_ <- translateReturn th None; Return None)
       \/ p' = Return None.
  Proof.
    induction 1; intros.

    auto.

    intuition subst.
    invert H0; auto.

    invert H0; auto.

    invert H0.
  Qed.    

  Lemma reorder_method' : forall R (m : Method absObj R) p r cp s cp0 th s',
    invariant s
    -> (forall th, timestamp s th = 0)
    -> (forall th, exists p, cp th = translateTrans th p)
    -> ordered cp s cp0 s'
    -> forall cp' s'', cp0 = override cp' th (Bind (translateMethod th R m) p)
    -> Exec concObj s' (translateMethod th R m) (Some r) s''
    -> ordered cp s (override cp' th (p (Some r))) s''.
  Proof.
    induction 4; solver.

    destruct (H1 th); solver.
    unfold translateTrans in H2.
    inversion H2.
    apply discriminate_mi in H6; tauto.

    invert H3.
    destruct (t =_? th); solver.

    econstructor.
    eauto.
    replace (override cp'0 th (p (Some r))) with (override cp0 th (p (Some r))).
    Focus 2.
    extensionality th0.
    apply (f_equal (fun f => f th0)) in H5.
    destruct (eq_nat_dec th th0); solver.
    apply (f_equal (fun f => f th)) in H5; solver.
    eauto.

    generalize H11; intro Hs.
    eapply progExec_stamp in Hs; eauto.
    solver.

    generalize H6; intro Hm.
    assert (p' = Return None \/ p' <> Return None).
    destruct p'; intuition (try congruence).
    destruct o; intuition congruence.
    solver.

    apply progExec_abortR in H11.
    generalize H11; intro Hab.
    eapply abort_step_commute in Hab; eauto.
    2: simpl; congruence.
    2: eauto using ordered_invariant.
    2: eauto using ordered_invariant.
    2: eauto.
    solver.
    invert H12.
    econstructor.
    eapply H4.
    2: eauto.
    instantiate (1 := override cp'0 t (translateTrans t p0)).
    extensionality t'.
    apply (f_equal (fun f => f t')) in H5; destruct (t' =_? t), (t' =_? th); solver.
    rewrite double_override_neq by auto.
    replace (override cp'0 th (p (Some r))) with (override (override cp'0 th (p (Some r))) t (Return None)) at 2.
    eapply orderedStepAbort; eauto.
    erewrite stamps_step; eauto.
    extensionality t'.
    apply (f_equal (fun f => f t')) in H5; destruct (t' =_? t), (t' =_? th); solver.

    eapply reorder_method'' in Hm; try eapply H11; eauto.
    2: eauto using ordered_invariant.
    3: congruence.
    Focus 2.
    rewrite H9.
    erewrite progExec_ostamp; eauto.
    destruct (H1 th).
    eapply ordered_stamp_bound; eauto.
    solver.
    apply (f_equal (fun f => f th)) in H5; solver.
    rewrite H5 in H13.
    unfold translateTrans in H13.
    inversion H13.
    apply discriminate_mi in H16; tauto.

    solver.
    eapply H4 in H13.
    econstructor.
    apply H13.
    instantiate (1 := override cp'0 t (translateTrans t p0)).

    rewrite double_override_neq by auto.
    replace (override cp'0 th (p (Some r))) with (override (override cp'0 th (p (Some r))) t p') at 2.
    econstructor; auto.

    extensionality t'.
    apply (f_equal (fun f => f t')) in H5; destruct (t' =_? t), (t' =_? th); solver.

    extensionality t'.
    apply (f_equal (fun f => f t')) in H5; destruct (t' =_? t), (t' =_? th); solver.


    assert (t <> th).
    intro; subst.
    apply (f_equal (fun f => f th)) in H5; solver; congruence.
    generalize H11; intro Hab.
    eapply abort_step_commute in Hab; eauto.
    2: simpl; congruence.
    2: eauto using ordered_invariant.
    solver.
    invert H10.
    econstructor.
    eapply H4.
    2: eauto.
    instantiate (1 := override cp'0 t (translateTrans t p0)).
    extensionality t'.
    apply (f_equal (fun f => f t')) in H5; destruct (t' =_? t), (t' =_? th); solver.
    rewrite double_override_neq by auto.
    replace (override cp'0 th (p (Some r))) with (override (override cp'0 th (p (Some r))) t (Return None)) at 2.
    eapply orderedStepAbort; eauto.
    erewrite stamps_step; eauto.
    extensionality t'.
    apply (f_equal (fun f => f t')) in H5; destruct (t' =_? t), (t' =_? th); solver.
  Qed.

  Lemma reorder_method : forall R (m : Method absObj R) p r cp s cp' th s' s'',
    ordered cp s
            (override cp' th
                      (Bind (translateMethod th R m) p)) s'
    -> Exec concObj s' (translateMethod th R m) (Some r) s'' 
    -> invariant s
    -> (forall th, timestamp s th = 0)
    -> (forall th, exists p, cp th = translateTrans th p)
    -> ordered cp s (override cp' th (p (Some r))) s''.
  Proof.
    eauto using reorder_method'.
  Qed.

  Lemma reorder_return'' : forall th1 th2 o p s p' s',
    invariant s
    -> timestamp s th2 = 0
    -> progExec _ (option nat) p s p' s'
    -> forall p0, p = translateTrans th2 p0
    -> forall r s'', Exec concObj s' (translateReturn th1 o) r s''
    -> timestamp s' th2 > timestamp s' th1
    -> p' <> Return None
    -> exists s''', Exec _ s (translateReturn th1 o) r s'''
                    /\ timestamp s''' th2 = 0
                    /\ progExec _ _ p s''' p' s''.
  Proof.
    induction 3; solver; eauto.

    assert (Hne : th1 <> th2) by (intro; subst; omega).

    generalize H1; intro He.
    eapply progExec_translateTrans in He; eauto.
    solver.

    clear H4.
    invert H2.
    eapply init_step_commute in H12; eauto.
    2: constructor.
    solver.
    apply doStep_invert in H3.
    do 2 esplit.
    eauto.
    split.
    erewrite stamps_step by eauto; eauto.
    econstructor.
    eauto.
    unfold translateTrans.
    eapply progExecBind with (f := fun _ => translateTrans' th2 p0); eauto.

    destruct x; simpl in *.

    invert H2.
    assert (Hst : doStep _ translateMethod translateReturn th1 (ReturnStep o) s'' r s''0) by eauto.
    destruct r0; try tauto.
    eapply steps_commute in Hst.
    3: eauto.
    3: simpl; congruence.
    3: eauto.
    2: erewrite <- stamps_step with (c' := s'') by eauto;
      erewrite <- stamps_step with (th' := th1) (c' := s'') by eauto; assumption.
    2: eauto.
    solver.
    apply doStep_invert in H3.
    invert H8.
    eapply H4 in H3; eauto.
    3: discriminate.
    2: erewrite <- stamps_step with (c' := s'') by eauto;
      erewrite <- stamps_step with (th' := th1) (c' := s'') by eauto; assumption.
    solver.
    do 2 esplit.
    eauto.
    split; auto.
    econstructor; eauto.
    apply progExecBind with (f := fun x1 =>
                                    match x1 with
                                    | Some v => translateTrans' th2 (p v)
                                    | None => Return None
                                    end) in H17; assumption.

    invert H2.
    assert (Hst : doStep _ translateMethod translateReturn th1 (ReturnStep o) s'' r s''0) by eauto.
    destruct o0; try tauto.
    eapply steps_commute in Hst.
    3: eauto.
    3: simpl; congruence.
    3: eauto.
    2: erewrite <- stamps_step with (c' := s'') by eauto;
      erewrite <- stamps_step with (th' := th1) (c' := s'') by eauto; assumption.
    2: eauto.
    solver.
    apply doStep_invert in H3.
    apply doStep_invert in H8.
    eapply H4 in H3; eauto.
    2: erewrite <- stamps_step with (c' := s'') by eauto;
      erewrite <- stamps_step with (th' := th1) (c' := s'') by eauto; assumption.
    solver.
    do 2 esplit.
    eauto.
    split; auto.
    econstructor; eauto.
    eapply progExecBind with (f := fun _ => Return (Some n)); eauto.
    discriminate.

    invert H2.
  Qed.

  Lemma reorder_return' : forall o p r cp s cp0 th s',
    invariant s
    -> (forall th, timestamp s th = 0)
    -> (forall th, exists p, cp th = translateTrans th p)
    -> ordered cp s cp0 s'
    -> forall cp' s'', cp0 = override cp' th (Bind (translateReturn th o) p)
    -> Exec concObj s' (translateReturn th o) r s''
    -> ordered cp s (override cp' th (p r)) s''.
  Proof.
    induction 4; solver.

    destruct (H1 th); solver.
    unfold translateTrans in H2.
    invert_Bind.
    apply discriminate_ri in H'; tauto.

    invert H3.
    destruct (t =_? th); solver.

    econstructor.
    eauto.
    replace (override cp'0 th (p r)) with (override cp0 th (p r)).
    Focus 2.
    extensionality th0.
    apply (f_equal (fun f => f th0)) in H5.
    destruct (eq_nat_dec th th0); solver.
    apply (f_equal (fun f => f th)) in H5; solver.
    eauto.

    generalize H11; intro Hs.
    eapply progExec_stamp in Hs; eauto.
    solver.

    generalize H6; intro Hm.
    assert (p' = Return None \/ p' <> Return None).
    destruct p'; intuition (try congruence).
    destruct o0; intuition congruence.
    solver.
    
    apply progExec_abortR in H11.
    generalize H11; intro Hab.
    eapply abort_step_commute in Hab; eauto.
    2: constructor.
    2: eauto using ordered_invariant.
    2: eauto using ordered_invariant.
    2: assumption.
    solver.
    apply doStep_invert in H12.
    econstructor.
    eapply H4.
    2: eauto.
    instantiate (1 := override cp'0 t (translateTrans t p0)).
    extensionality t'.
    apply (f_equal (fun f => f t')) in H5; destruct (t' =_? t), (t' =_? th); solver.
    rewrite double_override_neq by auto.
    replace (override cp'0 th (p r)) with (override (override cp'0 th (p r)) t (Return None)) at 2.
    eapply orderedStepAbort; eauto.
    erewrite stamps_step; eauto.
    extensionality t'.
    apply (f_equal (fun f => f t')) in H5; destruct (t' =_? t), (t' =_? th); solver.

    eapply reorder_return'' in Hm; try eapply H11; eauto.
    2: eauto using ordered_invariant.
    Focus 2.
    rewrite H9.
    erewrite progExec_ostamp; eauto.
    destruct (H1 th).
    eapply ordered_stamp_bound; eauto.
    solver.
    apply (f_equal (fun f => f th)) in H5; solver.
    rewrite H5 in H13.
    unfold translateTrans in H13.
    invert_Bind.
    apply discriminate_ri in H'; tauto.

    solver.
    eapply H4 in H13.
    econstructor.
    apply H13.
    instantiate (1 := override cp'0 t (translateTrans t p0)).

    rewrite double_override_neq by auto.
    replace (override cp'0 th (p r)) with (override (override cp'0 th (p r)) t p') at 2.
    econstructor; auto.

    extensionality t'.
    apply (f_equal (fun f => f t')) in H5; destruct (t' =_? t), (t' =_? th); solver.

    extensionality t'.
    apply (f_equal (fun f => f t')) in H5; destruct (t' =_? t), (t' =_? th); solver.


    assert (t <> th).
    intro; subst.
    apply (f_equal (fun f => f th)) in H5; solver; congruence.
    generalize H11; intro Hab.
    eapply abort_step_commute in Hab; eauto.
    2: constructor.
    2: eauto using ordered_invariant.
    solver.
    apply doStep_invert in H10.
    econstructor.
    eapply H4.
    2: eauto.
    instantiate (1 := override cp'0 t (translateTrans t p0)).
    extensionality t'.
    apply (f_equal (fun f => f t')) in H5; destruct (t' =_? t), (t' =_? th); solver.
    rewrite double_override_neq by auto.
    replace (override cp'0 th (p r)) with (override (override cp'0 th (p r)) t (Return None)) at 2.
    eapply orderedStepAbort; eauto.
    erewrite stamps_step; eauto.
    extensionality t'.
    apply (f_equal (fun f => f t')) in H5; destruct (t' =_? t), (t' =_? th); solver.
  Qed.

  Lemma reorder_return : forall o p r cp s cp' th s' s'',
    ordered cp s
            (override cp' th
                      (Bind (translateReturn th o) p)) s'
    -> Exec concObj s' (translateReturn th o) r s'' 
    -> invariant s
    -> (forall th, timestamp s th = 0)
    -> (forall th, exists p, cp th = translateTrans th p)
    -> ordered cp s (override cp' th (p r)) s''.
  Proof.
    eauto using reorder_return'.
  Qed.

  Lemma double_override_eq : forall A m k (v1 v2 : A),
    override (override m k v1) k v2 = override m k v2.
  Proof.
    intros; extensionality k'; destruct (key_eq_dec k' k); solver.
  Qed.

  Hint Rewrite double_override_eq.

  Lemma reorder_method_None' : forall cp s cp0 th s',
    invariant s
    -> (forall th, timestamp s th = 0)
    -> (forall th, exists p, cp th = translateTrans th p)
    -> ordered cp s cp0 s'
    -> forall s'', (forall p, cp0 th <> translateTrans th p)
    -> (forall r, cp0 th <> Return r)
    -> Exec concObj s' (translateReturn th None) tt s''
    -> ordered cp s (override cp0 th (Return None)) s''.
  Proof.
    induction 4; solver.

    destruct (H1 th); solver.
    unfold translateTrans in H2.
    exfalso; eauto.

    invert H3.
    destruct (t =_? th); solver.

    econstructor.
    eauto.
    eapply progExec_doSteps'' in H10; eauto.
    solver.
    exfalso; eauto.
    destruct p'; try solve [ exfalso; eauto ].
    eauto using orderedStepAbort, abortR_in, ordered_invariant.

    generalize H7; intro Hm.
    assert (p' = Return None \/ p' <> Return None).
    destruct p'; intuition (try congruence).
    destruct o; intuition congruence.
    solver.
    
    apply progExec_abortR in H10.
    generalize H10; intro Hab.
    eapply abort_step_commute in Hab; eauto.
    2: constructor.
    2: eauto using ordered_invariant.
    2: eauto using ordered_invariant.
    2: assumption.
    solver.
    apply doStep_invert in H11.
    econstructor.
    eapply H4.
    2: eauto.
    eauto.
    eauto.
    rewrite double_override_neq by auto.
    replace (override (override cp0 t (Return None)) th (Return None))
            with (override (override cp0 th (Return None)) t (Return None)) by (apply double_override_neq; auto).
    eapply orderedStepAbort; eauto.
    erewrite stamps_step; eauto.

    generalize H10; intro Hpe.
    eapply progExec_translateTrans in Hpe; eauto.
    solver.
    
    eapply reorder_return'' in Hm; try eapply H10; eauto.
    2: eauto using ordered_invariant.
    Focus 2.
    replace (timestamp s'' th) with (timestamp s' th) by (symmetry; eapply progExec_ostamp; try apply H10; eauto).
    eapply progExec_stamp in H10; eauto.
    solver.
    unfold translateTrans in H12.
    destruct x; simpl in *.
    inversion H12.
    apply discriminate_mi in H14; tauto.
    invert_Bind.
    apply discriminate_ri in H'; tauto.
    eauto using ordered_invariant, clock_bound.
    solver.
    eapply H4 in H12; eauto.
    econstructor.
    apply H12.
    rewrite double_override_neq by auto.
    replace (override (override cp0 t (translateTrans' t x)) th (Return None))
            with (override (override cp0 th (Return None)) t (translateTrans' t x)) by (apply double_override_neq; auto).
    eauto.

    destruct x; try tauto.
    eapply reorder_return'' in Hm; try eapply H10; eauto.
    2: eauto using ordered_invariant.
    Focus 2.
    replace (timestamp s'' th) with (timestamp s' th) by (symmetry; eapply progExec_ostamp; try apply H10; eauto).
    eapply progExec_stamp in H10; eauto.
    solver.
    discriminate.
    eauto using ordered_invariant, clock_bound.
    solver.
    eapply H4 in H12; eauto.
    econstructor.
    apply H12.
    rewrite double_override_neq by auto.
    replace (override (override cp0 t (Return (Some n0))) th (Return None))
            with (override (override cp0 th (Return None)) t (Return (Some n0))) by (apply double_override_neq; auto).
    eauto.


    assert (t <> th) by (solver; exfalso; eauto).
    generalize H10; intro Hab.
    eapply abort_step_commute in Hab; eauto.
    2: constructor.
    2: eauto using ordered_invariant.
    solver.
    apply doStep_invert in H12.
    econstructor.
    eapply H4.
    eauto.
    eauto.
    eauto.
    rewrite double_override_neq by auto.
    replace (override (override cp0 t (Return None)) th (Return None))
            with (override (override cp0 th (Return None)) t (Return None)) by (apply double_override_neq; auto).
    eapply orderedStepAbort; eauto.
    erewrite stamps_step; eauto.
  Qed.

  Lemma reorder_method_None : forall R (m : Method _ R) p cp s cp' th s' s'',
    ordered cp s
            (override cp' th
                      (Bind (translateMethod th R m) p)) s'
    -> Exec concObj s' (translateMethod th R m) None s'' 
    -> invariant s
    -> (forall th, timestamp s th = 0)
    -> (forall th, exists p, cp th = translateTrans th p)
    -> ordered cp s (override cp' th (Return None)) s''.
  Proof.
    intros.
    eapply reorder_method_None' in H; eauto using method_aborts; solver.
    inversion H4.
    apply discriminate_mi in H7; tauto.
    discriminate.
  Qed.

  Lemma reorder' :
    forall ccp cs ccp' cs', 
      cCTransExec concObj ccp cs ccp' cs'
      -> invariant cs
      -> clock cs > 0
      -> forall acp, (forall th, ccp th = translateTrans th (acp th)
                                 /\ timestamp cs th = 0)
      -> invariant cs'
         /\ clock cs' > 0
         /\ (exists acp', (forall th, (exists r, ccp' th = Return r /\ acp' th = Return r)
                                      \/ (ccp' th = translateTrans th (acp' th)
                                       /\ timestamp cs' th = 0)
                                      \/ (ccp' th = translateTrans' th (acp' th)
                                          /\ 0 < timestamp cs' th < clock cs')))
         /\ ordered ccp cs ccp' cs'.
  Proof.
    induction 1; intros.

    solver; eauto.

    specialize (IHcCTransExec H1 H2 _ H3).
    destruct IHcCTransExec as [ ? [ ? [ [ ] ] ] ].
    destruct H0.
    generalize (H6 t); autorewrite with core; destruct 1 as [ [ ? [ ] ] | [ [ ] | [ ? [ ] ] ] ]; try congruence.

    unfold translateTrans in H9.
    invert_Bind.
    solver; eauto.
    erewrite clock_initialize by eauto; eauto.
    exists x; solver.
    destruct (t =_? th); solver; eauto.
    erewrite initialize_initializes by eauto.
    rewrite (clock_initialize _ _ _ _ H8); eauto 6.
    specialize (H6 th); solver; try congruence; eauto.
    erewrite stamps_initialize by eauto; auto.
    rewrite (clock_initialize _ _ _ _ H8); eauto.
    erewrite stamps_initialize by eauto; eauto 8.
    
    repeat (econstructor; eauto).
    eapply progExecBind with (f := fun _ => translateTrans' t (x t)); eauto.

    case_eq (x t); intros.

    rewrite H12 in H9.
    simpl in H9.
    invert_Bind.
    solver; eauto.
    erewrite clock_step by eauto; auto.
    exists (override x t (p a)); solver.
    erewrite stamps_step by eauto.
    erewrite clock_step by eauto.
    destruct (t =_? th); solver; eauto 6.
    specialize (H6 th); solver; eauto 6.
    exists (override x t (Return None)); solver.
    erewrite stamps_step by eauto.
    erewrite clock_step by eauto.
    destruct (t =_? th); solver; eauto.
    specialize (H6 th); solver; eauto 6.

    eapply reorder_method in H7; eauto; firstorder.
    eapply reorder_method_None in H7; eauto; firstorder.

    rewrite H12 in H9.
    simpl in H9.
    invert_Bind.
    solver; eauto.
    erewrite clock_step by eauto; auto.
    exists (override x t (Return o)); solver.
    erewrite stamps_step by eauto.
    erewrite clock_step by eauto.
    destruct (t =_? th); solver; eauto.
    specialize (H6 th); solver; eauto 6.

    eapply reorder_return in H7; eauto; firstorder.
  Qed.

  Hint Resolve invariant_init clock_pos stamps_init.

  Theorem reorder :
    forall acp cs ccp', 
      cCTransExec concObj
                  (translateCTrans acp) concInit
                  ccp' cs
      -> ordered (translateCTrans acp) concInit ccp' cs.
  Proof.
    intros.
    eapply reorder' in H; solver.
  Qed.

  Inductive orderedStep' :
    CTrans concObj -> State concObj -> CTrans concObj -> State concObj -> Prop :=
  | orderedStepExec' {t cp p s r s'}:
    In t tids
    -> timestamp s t = 0
    -> progExec _ _ (translateTrans t p) s (Return (Some r)) s'
    -> orderedStep' (override cp t (translateTrans t p)) s 
                   (override cp t (Return (Some r))) s'
  | orderedStepAbort' {t cp p s s'}:
    In t tids
    -> timestamp s t = 0
    -> abortR t s s'
    -> orderedStep' (override cp t (translateTrans t p)) s 
                   (override cp t (Return None)) s'.

  Inductive ordered': CTrans concObj -> State concObj -> CTrans concObj -> State concObj -> Prop :=
  | sTransExecRefl' cp s: 
    ordered' cp s cp s
  | orderedSteps' s cp cp' s' cp'' s'': 
    ordered' cp s cp' s'
    -> orderedStep' cp' s' cp'' s''
    -> ordered' cp s cp'' s''.

  Hint Constructors orderedStep' ordered'.

  Lemma ordered_ordered' : forall ps s ps' s',
    ordered ps s ps' s'
    -> (forall th, (exists p, ps th = translateTrans th p)
                   \/ (exists r, ps th = Return r))
    -> (forall th, In th tids
                   -> (exists p, ps' th = translateTrans th p)
                      \/ (exists r, ps' th = Return r))
    -> invariant s
    -> ordered' ps s ps' s'.
  Proof.
    induction 1; solver; eauto.

    destruct H0.
    specialize (progExec_translateTrans H6 eq_refl); solver.

    specialize (H2 t); solver.
    unfold translateTrans in H2.
    destruct x; simpl in *.
    invert H2.
    apply discriminate_mi in H10; tauto.
    invert_Bind.
    apply discriminate_ri in H'; tauto.
    destruct x; simpl in *.
    inversion H2.
    inversion H2.

    econstructor.
    apply H4; solver.
    specialize (H2 th); destruct (t =_? th); solver; eauto.
    destruct x.
    eauto.

    eauto using progExec_abortR, ordered_invariant.

    econstructor.
    apply H4.
    solver.
    destruct (t =_? th); solver; eauto.
    specialize (H2 th); solver; eauto.
    eauto.
    eauto.
  Qed.

  Hint Constructors sCTransExec sCTransExecStep sTransExecStep.

  Lemma progExec_ca : forall th cp c cp' c',
    progExec concObj _ cp c cp' c'
    -> forall ap, cp = translateTrans th ap
    -> cp' <> Return None
    -> exists ap', progExec absObj _ ap (absOf c) ap' (absOf c')
                   /\ (cp' = translateTrans th ap'
                       \/ cp' = translateTrans' th ap'
                       \/ exists r, ap' = Return r /\ cp' = Return r).
  Proof.
    induction 1; solver; eauto 7.

    destruct IHprogExec with (ap0 := ap); auto; clear IHprogExec; solver.

    invert H0.

    unfold translateTrans in H0.
    invert H0.
    erewrite absOf_initialize with (c' := s'') by eauto; eauto.

    destruct x; simpl in *.

    invert H0.
    destruct r.
    apply absOf_method in H9.
    eauto 10.

    tauto.

    invert H0.
    destruct o; try tauto.
    erewrite absOf_return with (c' := s'') by eauto; eauto 10.

    inversion H0.
  Qed.

  Lemma translateTrans'_inj : forall th p1 p2,
    translateTrans' th p1 = translateTrans' th p2
    -> p1 = p2.
  Proof.
    induction p1; destruct p2; solver.

    inversion H0; clear H0.
    apply translateMethod_inj in H3.
    generalize (f_equal (@projT1 _ _) H3); simpl; intro; subst.
    injpair_; subst.
    f_equal.
    extensionality x.
    apply (f_equal (fun f => f (Some x))) in H4; eauto.

    inversion H0; clear H0.
    apply discriminate_mr in H3; tauto.

    apply (f_equal methodOf) in H; simpl in H.
    symmetry in H.
    apply discriminate_mr in H; tauto.

    invert_Bind.
    apply (f_equal (fun f => f tt)) in H; congruence.
  Qed.

  Lemma translateTrans_inj : forall th p1 p2,
    translateTrans th p1 = translateTrans th p2
    -> p1 = p2.
  Proof.
    unfold translateTrans; intros.
    invert_Bind.
    eapply translateTrans'_inj.
    apply (f_equal (fun f => f tt) H).
  Qed.

  Theorem ordered'_sCTransExec : forall cp c cp' c',
    ordered' cp c cp' c'
    -> forall acp, cp = (fun th => translateTrans th (acp th))
    -> exists ap', sCTransExec _ acp (absOf c) ap' (absOf c')
                  /\ (forall th, (cp' th = cp th /\ acp th = ap' th)
                                 \/ exists r, ap' th = Return r /\ cp' th = Return r).
  Proof.
    induction 1; solver; eauto.

    edestruct IHordered'; eauto; clear IHordered'; solver.
    destruct H0.
    eapply progExec_ca in H4; eauto.
    solver; try discriminate.
    destruct x0; discriminate.
    invert H7.
    do 2 esplit.
    econstructor.
    eauto.
    replace x with (override x t p) at 1.
    econstructor; auto.
    econstructor; eauto.
    specialize (H3 t); solver.
    apply translateTrans_inj in H3; subst.
    rewrite H6.
    extensionality k.
    destruct (k =_? t); solver.
    inversion H6.

    intro th.
    destruct (th =_? t); solver; eauto.
    specialize (H3 th); solver; eauto.
    discriminate.

    eexists.
    split.
    econstructor.
    eauto.
    replace x with (override x t p) at 1.
    econstructor; auto.
    erewrite <- absOf_abortR by eauto; eauto.
    specialize (H3 t); solver.
    apply translateTrans_inj in H3; subst.
    rewrite H6.
    extensionality k.
    destruct (k =_? t); solver.
    inversion H6.

    intro th.
    destruct (th =_? t); solver; eauto.
    specialize (H3 th); solver; eauto.
  Qed.

  Theorem serialize : serializable.
  Proof.
    hnf; intros.
    apply reorder in H.
    apply ordered_ordered' in H; eauto.
    2: unfold translateCTrans; eauto.
    eapply ordered'_sCTransExec in H; eauto.
    2: unfold translateCTrans; eauto.
    rewrite absOf_init in *.
    solver; eauto.
    do 3 esplit.
    eauto.
    hnf; intros.
    apply H0 in H; solver.
    specialize (H2 t); solver; eauto.
    rewrite H in H2; discriminate.
    rewrite H in H4; injection H4; solver; eauto.
  Qed.

End Make.
