From Coq Require Import Arith.
From C4 Require Import LibTactics.

Import PeanoNat.Nat.

Notation " [ x ; .. ; y ] " := (cons x .. (cons y nil) ..).


Section Override.

  Context {Key: Type}.
  Context {Val: Type}.
  Context (eq_key_dec: forall k k': Key, {k = k'} + {k <> k'}).
  
  Definition override
             (m: Key -> Val) (k: Key) (val: Val): Key -> Val :=
    fun (k': Key) =>
      if (eq_key_dec k' k) then val else (m k').  

End Override.

Definition noverride {Val} :=
  @override nat Val eq_nat_dec.

Section Override'.

  Lemma eq_key_dec_eq:
    forall A K (n: K) (s1 s2: A) (dec: forall k k': K, {k = k'} + {k <> k'}),
      (if dec n n then s1 else s2) =
      s1.

  Proof.
    intros.
    destruct (dec n n).
    reflexivity.

    exfalso. apply n0. reflexivity.
  Qed.

  Lemma eq_key_dec_neq:
    forall (A: Type) K n n' (s1 s2: A)
      (dec: forall k k': K, {k = k'} + {k <> k'}),
      (not (n = n'))
      -> (if dec n n' then s1 else s2) =
        s2.

  Proof.
    intros.

    destruct (dec n n').
    exfalso. apply H in e. assumption.

    reflexivity.    
  Qed.


  Lemma override_new_val:
    forall {K V} dec m (k: K) (v: V),
      (override dec m k v) k = v.

    Proof.
      intros.
      unfold override.
      rewrite eq_key_dec_eq.
      reflexivity.
    Qed.

  Lemma override_new_val':
    forall {K V} dec m (k1 k2: K) (v: V),
      k1 = k2 ->
      (override dec m k1 v) k2 = v.

    Proof.
      intros.
      subst.
      unfold override.
      rewrite eq_key_dec_eq.
      reflexivity.
    Qed.

  Lemma override_new_val'':
    forall {K V} dec m (k1 k2: K) (v: V),
      k2 = k1 ->
      (override dec m k1 v) k2 = v.

    Proof.
      intros.
      subst.
      unfold override.
      rewrite eq_key_dec_eq.
      reflexivity.
    Qed.
  
  Lemma override_old_val:
    forall {K V} dec m (k k': K) (v: V),
      (not (k = k'))
      -> (override dec m k v) k' = m k'.

    Proof.
      intros.
      unfold override.
      rewrite eq_key_dec_neq.
      reflexivity.
      unfold not. intro. symmetry in H0. apply H in H0. assumption.
    Qed.

  Lemma override_old_val':
    forall {K V} dec m (k k': K) (v: V),
      (not (k' = k))
      -> (override dec m k v) k' = m k'.

  Proof.
    intros.
    unfold override.
    rewrite eq_key_dec_neq.
    reflexivity.
    assumption.
  Qed.

  Lemma override_eq_val:
    forall {K V} dec m (k k': K) (v: V),
      m k' = v ->
      override dec m k v k' = v.

  Proof.
    intros.
    destruct (dec k k'); subst. 
    apply override_new_val.
    apply override_old_val. assumption.
  Qed.

  
End Override'.


Ltac simpl_override :=
  unfold noverride;
  repeat (
      rewrite override_new_val ||
              (rewrite override_new_val'; [ | eassumption]) ||
              (rewrite override_new_val''; [ | eassumption]) ||
              (rewrite override_old_val; [ | eassumption]) ||
              (rewrite override_old_val'; [ | eassumption]));
  simpl.


Ltac simpl_override_in H :=
  unfold noverride;  
  repeat (
      rewrite override_new_val in H ||
                                    (rewrite override_new_val' in H; [ | eassumption]) ||
                                    (rewrite override_new_val'' in H; [ | eassumption]) ||
                                    (rewrite override_old_val in H; [ | eassumption]) ||
                                    (rewrite override_old_val' in H; [ | eassumption]));
  simpl in H.


Ltac simpl_override_in_star :=  
  unfold noverride in *;
  repeat
    match goal with
    | [ |- context[override ?dec ?m ?k ?v ?k] ] =>
      rewrite (override_new_val dec)
    | [ H': ?k = ?k' |- context[override ?dec ?m ?k ?v ?k'] ] =>
      rewrite (override_new_val' dec m k k' v H')
    | [ H': ?k' = ?k |- context[override ?dec ?m ?k ?v ?k'] ] =>
      rewrite (override_new_val'' dec m k' k v H')
              
    | [ H': not (?k = ?k') |- context[override ?dec ?m ?k ?v ?k'] ] =>
      rewrite (override_old_val dec m k k' v H')
    | [ H': ?k = ?k' -> False |- context[override ?dec ?m ?k ?v ?k'] ] =>
      rewrite (override_old_val dec m k k' v H')
    | [ H': not (?k' = ?k) |- context[override ?dec ?m ?k ?v ?k'] ] =>
      rewrite (override_old_val' dec m k k' v H')
    | [ H': ?k' = ?k -> False |- context[override ?dec ?m ?k ?v ?k'] ] =>
      rewrite (override_old_val' dec m k k' v H')
    end;

  repeat
    match goal with
    | [ H: context[override ?dec ?m ?k ?v ?k] |- _ ] =>
      rewrite (override_new_val dec) in H
    | [ H: context[override ?dec ?m ?k ?v ?k'], H': ?k = ?k' |- _ ] =>
      rewrite (override_new_val' dec m k k' v H') in H
    | [ H: context[override ?dec ?m ?k ?v ?k'], H': ?k' = ?k |- _ ] =>
      rewrite (override_new_val'' dec m k' k v H') in H
    | [ H: context[override ?dec ?m ?k ?v ?k'], H': not (?k = ?k') |- _ ] =>
      rewrite (override_old_val dec m k k' v H') in H
    | [ H: context[override ?dec ?m ?k ?v ?k'], H': ?k = ?k' -> False |- _ ] =>
      rewrite (override_old_val dec m k k' v H') in H
    | [ H: context[override ?dec ?m ?k ?v ?k'], H': not (?k' = ?k) |- _ ] =>
      rewrite (override_old_val' dec m k k' v H') in H
    | [ H: context[override ?dec ?m ?k ?v ?k'], H': ?k' = ?k -> False |- _ ] =>
      rewrite (override_old_val' dec m k k' v H') in H
    end.


Ltac destruct_override_in H :=
  unfold noverride in H;
  match type of H with
  | context [override ?dec _ ?k _ ?k'] =>
    destruct (dec k k');
      simpl_override_in H
  end.

Ltac destruct_override :=
  unfold noverride;
  match goal with
  | [ |- context [override ?dec _ ?k _ ?k']] =>
    destruct (dec k k');
      simpl_override
  end.








Lemma simpl_override_test:
  forall Val m k1 k2 k3 (v: Val),
    ((not (k2 = k1))
     /\ (not (k1 = k3)))
    -> ((noverride m k1 v) k1 = v
       /\ (noverride m k1 v) k2 = m k2
       /\ (noverride m k1 v) k3 = m k3).

Proof.
  intros.
  destruct H as [H1 H2].
  split.
  
  simpl_override.
  reflexivity.

  split.
  simpl_override.
  reflexivity.

  simpl_override.
  reflexivity.

Qed.


Lemma simpl_override_test2:
  forall m1 m2 k1 k2 k3 (v1: nat),
    (k1 = k3 -> False)
    -> noverride m1 k1 v1 (noverride m2 k1 k2 k3) = noverride m1 k1 v1 (m2 k3).

Proof.
  intros.
  simpl_override_in_star.
  reflexivity.
Qed.

Lemma simpl_override_test3:
  forall m k (v: nat),
    (k = 0 -> False) ->
    noverride m 0 v k = m k.

Proof.
  intros.    
  simpl_override_in_star.
  reflexivity.
Qed.


