Set Asymmetric Patterns.
From Coq Require Import Arith Omega Eqdep Equalities.

From C4.Util Require Import
  LibTactics.
From C4.Core Require Import
  Program
  ProgramFacts        
  ProgramOps
  ProgramOpsFacts
  Object
  Pair
  Threads
  LinCoInd
  Abortable
  Linearizability
  Serializability.
From C4.App Require Import
  Lock
  Maps.
From C4.AppExtra Require Import
  Counter.
Import EqdepTheory.
Import BMap.
Import SimulatesCoInd.

(* Injectivity of [option] *)
Axiom OptionInj:
  forall R R',
    option R = option R' ->
    R = R'.

(* Extensionality axiom for [Abortable] *)
Axiom invertAbortable:
  forall M R (a: Abortable M (option R)),
  exists m, a = @abortable M R m.

Module GlobalLockTM (E : MiniDecidableType).

  Module BAMap := BMapFun E.
  Import BAMap.

Section GlobalLockTM.
    
  Context (Val : Type).

  Definition initBMapState := fun (k: Key) => @None Val.
  
  Definition TId := unit.
  
  Definition tmBase :=
    pairSpec atomicLock (atomicBMap Val).

  Definition TMBaseMethod :=
    PairInstruction LockMethod (BMapMethod Key Val).
  
  Definition TMState :=
    tmBase.(State).
  Definition lockState (s: TMState): bool :=
    fst s.
  Definition mapState (s : TMState) : BMapState Val :=
    snd s.


  Definition initTMState: State tmBase :=
    pair false
         (fun k => None).

            
  Definition callLock {R} (m: LockMethod R): TMBaseMethod R :=
    inl m.

  Definition callMap {R} (m: (BMapMethod Key Val) R): TMBaseMethod R :=
    inr m.


  Definition tmImpl R (m: @TransMethod (BMapMethod Key Val) TId R): Program TMBaseMethod R :=
    match m with
    | init =>
      _ <- callLock doLock;
      Return tt

    | abort t =>
      (* Abort is never called in this algorithm. *)
      Return tt

    | commit t =>
      _ <- callLock doUnlock;
      Return (inl tt)

    | liftMethod R t m =>
      r <- callMap m;
      Return ((Some r), t)

    end.

  Definition globalLockTM: Object (@TransMethod (BMapMethod Key Val) TId) :=
    {|
      objBase := tmBase;
      objProgram := tmImpl
    |}.


  Inductive TRel {R}:
    Program TMBaseMethod (option R) ->
    TMState ->
    Program (BMapMethod Key Val) R ->
    BMapState Val
    -> Prop :=

  | TRel0 t s1 p2 s2:
      TRel
        (Noop (
             _ <- callLock doLock;
             Noop (instFlatProg' tmImpl instrument' _ p2 t)))
        s1
        p2
        s2
    
  | TRel1 t s1 p2 s2:
      TRel
        (_ <- callLock doLock;
         Noop (instFlatProg' tmImpl instrument' _ p2 t))
        s1
        p2
        s2
        
  | TRel3 t s1 p2 s2 p2':
      ProgramSteps (atomicBMap Val) _ p2 s2 p2' (mapState s1) ->
      lockState s1 = true ->
      TRel
        (instFlatProg' tmImpl instrument' _ p2' t)
        s1
        p2
        s2

           
  | TRel3p t s1 p2 s2 R' (m: BMapMethod Key Val R') (f: R' -> Program (BMapMethod Key Val) R):
      ProgramSteps (atomicBMap Val) _ p2 s2 (r <- m; f r) (mapState s1) ->
      lockState s1 = true ->
      TRel
        (r <- callMap m; Noop (instFlatProg' tmImpl instrument' _ (f r) t))
        s1
        p2
        s2

  | TRel3p' s1 p2 s2 r:
      ProgramSteps (atomicBMap Val) _ p2 s2 (Return r) (mapState s1) ->
      lockState s1 = true ->       
      TRel
        (_ <- callLock doUnlock; Noop (Return (Some r)))
        s1
        p2
        s2
        
  | TRel3' t s1 p2 s2 p2':
      ProgramSteps (atomicBMap Val) _ p2 s2 p2' (mapState s1) ->
      lockState s1 = true ->
      TRel
        (Noop (instFlatProg' tmImpl instrument' _ p2' t))
        s1
        p2
        s2.

  Hint Constructors TRel.
  
  Inductive PDRel {R'}:
            Program TMBaseMethod R' ->
            TMState ->
            Program (Abortable (Program (BMapMethod Key Val))) R' ->
            BMapState Val ->
            Prop :=

  | PDRel1 r s1 s2 R'':
      R' = option R'' ->
      PDRel (Return r) s1 (Return r) s2

  | PDRel1' r s1 s2 R'':
      R' = option R'' ->      
      PDRel (Noop (Return r)) s1 (Return r) s2
            
  | PDRel2 p1 s1 am s2:
      match am in Abortable _ R' return Program TMBaseMethod R' -> Prop with
      | abortable R m =>
        fun p1 => TRel p1 s1 m s2
      end p1 ->
      PDRel p1 s1 (x <- am; Return x) s2.
  
  Hint Constructors PDRel.

  Ltac invertPDRelTac H :=
    inverts H as;
    match goal with
    | [ |- match ?am in Abortable _ _ return _ with | _ => _ end _ -> _ ] =>
      first [ introv H; destruct am
            | introv H;
              let A := fresh "A" in
              let m := fresh "m" in
              forwards A: invertAbortable am;
              destruct A as (m' & A); subst
            ]
    | _ => idtac
    end.

  
  Definition DRel
             (s1: TMState)
             (s2: BMapState Val): Prop :=
    lockState s1 = false -> mapState s1 = s2.

  Inductive OutCritical {R}:
            Program TMBaseMethod (option R) ->
            Prop :=
  | OutCritical0 t p:
      OutCritical
        (Noop (_ <- callLock doLock;
               Noop (instFlatProg' tmImpl instrument' _ p t)))
        
  | OutCritical1 t p:
      OutCritical
        (_ <- callLock doLock;
         Noop (instFlatProg' tmImpl instrument' _ p t))
        
  | OutCritical2 r:
      OutCritical
        (Noop (Return r))
        
  | OutCritical3 r:
      OutCritical
        (Return r).
  
  Hint Constructors OutCritical.

  Definition PPRel {ROpt ROpt'}
    (p11: Program TMBaseMethod ROpt)
    (p21: Program (Abortable (Program (BMapMethod Key Val))) ROpt)
    (p12: Program TMBaseMethod ROpt')
    (p22: Program (Abortable (Program (BMapMethod Key Val))) ROpt')
    (* (s1: TMState)
    (s2: BMapState) *): 
    Prop :=

    exists R R',
      option R = ROpt
      /\ option R' = ROpt'
      /\ forall (e: (option R = ROpt))
          (e': (option R' = ROpt')),          
          match e in _ = R2 return Program TMBaseMethod R2 -> Prop with            
          | eq_refl =>
            fun p1 =>
              match e' in _ = R2' return Program TMBaseMethod R2' -> Prop with
              | eq_refl =>
                fun p2 =>
                  OutCritical p1 \/ OutCritical p2
              end p12
          end p11.
    
  Ltac openPPRel :=
    econstructor; eexists; splits;
    [ reflexivity
    | reflexivity
    | let E1 := fresh "E" in intros E1;
      let E2 := fresh "E" in intros E2;
      let A1 := fresh "A" in 
      forwards A1: UIP_refl Type E1;
      let A2 := fresh "A" in 
      forwards A2: UIP_refl Type E2;
      subst ].
    
  Ltac openPPRelIn H :=
    unfold PPRel in H;
    inverts H as H;
    let Hn := fresh H "_" in
    (* to skip H and H0 *)
    assert (Hn: True) by apply I;    
    let H0 := fresh H "_" in
    assert (H0: True) by apply I;
    let H1 := fresh H "_" in
    assert (H1: True) by apply I;
    let H2 := fresh H "_" in
    assert (H2: True) by apply I;      
    let H3 := fresh H "_" in
    clear Hn H0 H1 H2;
    let R := fresh "R" in
    destruct H as (R & H1 & H2 & H3);
    match goal with
    | [ H1 : option _ = option _ |- _ ] =>
        apply OptionInj in H1
    | _ => idtac
    end;
    match goal with
    | [ H2 : option _ = option _ |- _ ] =>
        apply OptionInj in H2
    | _ => idtac
    end;
    subst;
    specialize (H3 eq_refl eq_refl); simpl in H3.
    
  Lemma PPRelSym:
    forall R (p1: Program TMBaseMethod R) p2
      R' (p1': Program TMBaseMethod R') p2',

      PPRel p1 p2 p1' p2' ->
      PPRel p1' p2' p1 p2.

  Proof.
    introv H1.
    openPPRelIn H1.
    openPPRel.
    iauto.
  Qed.

  
  Inductive IRel {R}:
    TId ->
    Program TMBaseMethod (option R) ->
    Program (BMapMethod Key Val) R ->
    Prop :=

  | BindIRel t A (m: (BMapMethod Key Val) A) f:
      IRel t
           (Noop (r <- callMap m;
                  Noop (instFlatProg' tmImpl instrument' _ (f r) t)))
           (Bind m f)
          
  | ReturnIRel r t:
      IRel t
           (Noop (_ <- callLock doUnlock;
                  Noop (Return (Some r))))
           (Return r)
          
  | NoopIRel t p1 p2:
      (* IRel t p1 p2 -> *)
      p1 = instFlatProg' tmImpl instrument' _ p2 t ->
      IRel t (Noop p1) (Noop p2).
  
  Hint Constructors IRel.
  
  Lemma IRelInst:
    forall R t 
      (p1: Program TMBaseMethod (option R))
      (p2: Program (BMapMethod Key Val) R),
      p1 = instFlatProg' tmImpl instrument' _ p2 t ->
      IRel t p1 p2.
  
  Proof.
    introv H1.
    subst.
    unfold instFlatProg'.
    destruct p2.
    - setoid_rewrite frobProgramId at 1. simpl.
      setoid_rewrite frobProgramId at 2; simpl.
      setoid_rewrite frobProgramId at 3; simpl.
      constructor.
    - setoid_rewrite frobProgramId at 1; simpl.
      setoid_rewrite frobProgramId at 2; simpl.
      setoid_rewrite frobProgramId at 3; simpl.
      setoid_rewrite frobProgramId at 4; simpl.
      constructor.
    - match goal with
      | [ |- IRel _ ?p _] =>
        rewriteFrobProgram p
      end. simpl.
      econstructor.
      reflexivity.
  Qed.

  
  Definition init1: TMState := initTMState.
  Definition init2 : BMapState Val := initBMapState.

  Lemma initRel: DRel init1 init2.

  Proof.
    unfold DRel.
    intro.
    reflexivity.
  Qed.  

  Lemma noopRel:
    forall R (p1: Program TMBaseMethod R) p2 s1 s2,
      PDRel (Noop p1) s1 p2 s2 ->
      (PDRel p1 s1 p2 s2
       /\ (forall R (p1t: Program TMBaseMethod R) p2t,
             PDRel p1t s1 p2t s2 ->
             PPRel (Noop p1) p2 p1t p2t ->
             PPRel p1 p2 p1t p2t))
      \/ (exists p2',
            p2 = Noop p2'
            /\ PDRel p1 s1 p2' s2
            /\ (forall R (p1t: Program TMBaseMethod R) p2t,
                  PDRel p1t s1 p2t s2 ->
                  PPRel (Noop p1) p2 p1t p2t ->
                  PPRel p1 p2' p1t p2t))

      \/ (exists p2',
          ProgramSteps (abortableSpec (programSpec (atomicBMap Val))) _ p2 s2 p2' s2
          /\ PDRel p1 s1 p2' s2
          /\ (forall R (p1t: Program TMBaseMethod R) p2t,
                PDRel p1t s1 p2t s2 ->
                PPRel (Noop p1) p2 p1t p2t ->
                PPRel p1 p2' p1t p2t)).
  
  Proof.
    introv H1.    
    (* inverts H1 as. *)
    invertPDRelTac H1.

    - left. split.
      + eauto.
      + introv H1 H2.
        openPPRelIn H2.
        openPPRel.
        iauto.
          
    - left.
      inverts H1 as.

      + splits.
        * econstructor. econstructor.
        (* eauto. *)
        * { introv H2 H3.
            openPPRelIn H3.
            openPPRel.
            destruct H3_3.
            - left. inverts H; constructor.
            - right. assumption.
          }          
        

      + introv H1 H2 H3.
        split.
        * symmetry in H3. eapply IRelInst in H3. inverts H3.
          constructor. constructor. assumption. assumption.
          constructor. constructor. assumption. assumption.
          constructor. constructor. eapply ProgramStepsStepImplSteps. eassumption. constructor.
          eassumption.
        * { introv H4 H5.
            openPPRelIn H5.
            openPPRel.
            destruct H5_3.
            - left. rewrite H3 in H. inverts H; constructor.
            - right. assumption.
          }          
          
      + split.
        * constructor.
          econstructor.
          assumption.
          assumption.
        * { intros.
            openPPRelIn H1.
            openPPRel.
            destruct H1_3.
            - left.
              inverts H1; constructor.
            - right. assumption.
          }          
  Qed.

  
  Ltac invertAbortableTac m :=
    let A1 := fresh "A" in
    forwards A1: invAbort m;
    let R := fresh "R" in
    let e := fresh "e" in
    let A2 := fresh "B" in  
    destruct A1 as (R & e & A2);
    destruct e;
    destruct A2;
    subst.

       
  Lemma invRel:
    forall R m s1 s2,
      DRel s1 s2 ->
      PDRel (instFlatImplProg tmImpl instrument R m) s1
            (x <- m; Return x) s2
      /\ forall R' (p1: Program TMBaseMethod R') p2,
          PDRel p1 s1 p2 s2 ->
          PPRel (flattenProgram tmImpl (instImplProg instrument R m))
                (x <- m; Return x)
                p1 p2.

  Proof.
    simpl.
    introv H1.
    split.

    - constructor.
      destruct m.
      setoid_rewrite frobProgramId at 1; simpl.
      setoid_rewrite frobProgramId at 2; simpl.
      setoid_rewrite frobProgramId at 3. simpl.
      econstructor.

    - introv H2.
      invertAbortableTac m.
      inverts H2.
      
      + openPPRel. right. constructor.
        
      + openPPRel. right. constructor.

      + destruct am.
        setoid_rewrite frobProgramId at 2; simpl.
        setoid_rewrite frobProgramId at 1; simpl.
        setoid_rewrite frobProgramId at 2; simpl.
        setoid_rewrite frobProgramId at 3; simpl.
        openPPRel. left. constructor.
  Qed.
        
        
  Lemma retRel:
    forall R (r: R) s1 p2 s2,
      PDRel (Return r) s1 p2 s2 ->
      DRel s1 s2 ->
      p2 = Return r.

  Proof.
    introv H1 H2.
    invertPDRelTac H1.
    - reflexivity.
    - exfalso. inverts H1 as H3 H4 H5. symmetry in H5. eapply IRelInst in H5. inverts H5.
  Qed.

  
  Lemma TRelPres1:
    forall (R: Type)
      (p1: Program TMBaseMethod (option R))
      (p2: Program (BMapMethod Key Val) R)
      (s1 s1': TMState) (s2: BMapState Val),

      lockState s1 = false ->
      TRel p1 s1 p2 s2 ->
      TRel p1 s1' p2 s2.

  Proof.
    introv H1 H2.
    inverts H2;
    econstructor; congruence.
  Qed.


  Lemma PDRelPres1:
    forall (s1 : TMState) (s2 : BMapState Val) (s1' : TMState)
      (R : Type) (p1 : Program TMBaseMethod R)
      (p2 : Program (Abortable (Program (BMapMethod Key Val))) R),

      lockState s1 = false ->
      PDRel p1 s1 p2 s2 ->
      PDRel p1 s1' p2 s2.

  Proof.
    introv H1 H2.
    invertPDRelTac H2.
    - econstructor. reflexivity.
    - econstructor. reflexivity.
    - constructor.
      eapply TRelPres1; eassumption.
  Qed.


  

  Lemma PPRelPres1:
    forall (R : Type) (s1 : TMState) (s2 : BMapState Val)
      (* (s1' : State (objBase impl1)) *) (m : Program (BMapMethod Key Val) R) (t : TId)
      (R' : Type) (p1t : Program TMBaseMethod R')
      (p2t : Program (Abortable (Program (BMapMethod Key Val))) R'),
      
      lockState s1 = false ->
      PDRel p1t s1 p2t s2 ->
      PPRel
        (_ <- callLock doLock; Noop (instFlatProg' tmImpl instrument' _ m t))
        (x <- abortable m; Return x) p1t p2t (* s1 s2 *) ->
      PPRel (Noop (instFlatProg' tmImpl instrument' _ m t))
            (x <- abortable m; Return x) p1t p2t (* s1' s2 *).

  Proof.
    introv H1 H2 H3.
    openPPRelIn H3.
    destruct H3_3 as [H33 | H33].
    
    - openPPRel.
      right.
      invertPDRelTac H2.
      + constructor.
      + constructor.
      + inverts H2; try constructor; 
        unfold lockState in *; simpl in *; congruence.
      
    - openPPRel. right. assumption.
      
  Qed.
  

  Lemma InstFlatProg'NotOutCritial:
    forall R t (p : Program (BMapMethod Key Val) R),
      OutCritical (instFlatProg' tmImpl instrument' _ p t) -> False.

  Proof.
    introv H.
    destruct p.

    - setoid_rewrite frobProgramId in H at 1; simpl in H;
      setoid_rewrite frobProgramId in H at 2; simpl in H;
      inverts H.

    - setoid_rewrite frobProgramId in H at 1; simpl in H;
      setoid_rewrite frobProgramId in H at 2; simpl in H;
      inverts H.

    - setoid_rewrite frobProgramId in H at 1; simpl in H;
      destruct p; setoid_rewrite frobProgramId in H at 2; simpl in H;
      inverts H. 
  Qed.

  Lemma InstFlatProg'NotOutCritial':
    forall R t (p : Program (BMapMethod Key Val) R),
      OutCritical (Noop (instFlatProg' tmImpl instrument' _ p t)) -> False.

  Proof.
    introv H.
    destruct p.

    - setoid_rewrite frobProgramId in H at 2; simpl in H.
      inverts H.

    - setoid_rewrite frobProgramId in H at 2; simpl in H.
      inverts H.

    - setoid_rewrite frobProgramId in H at 2; simpl in H.
      inverts H. 
  Qed.
  
  Lemma TRelPres2:
  forall (s1 s1' : TMState) (s2 : BMapState Val) (R : Type)
    (p1 : Program TMBaseMethod (option R)) (R' R'' : Type) 
    (t : TId) (m : BMapMethod Key Val R'') (m' : Program (BMapMethod Key Val) R')
    (f : R'' -> Program (BMapMethod Key Val) R') (m'' : Program (BMapMethod Key Val) R),
    
    PPRel (r <- callMap m; Noop (instFlatProg' tmImpl instrument' _ (f r) t))
          (x <- abortable m'; Return x)
          p1
          (x <- abortable m''; Return x)
          (* s1 s2 *) ->
    TRel p1 s1 m'' s2 ->
    TRel p1 s1' m'' s2.
  
  Proof.  
    introv H1 H2.
    inverts H2 as.

    + constructor.
    + constructor.

    + introv H4 H5. exfalso.
      openPPRelIn H1.
      destruct H1_3 as [H13 | H13].
      * inverts H13.      
      * eapply InstFlatProg'NotOutCritial; eassumption.
      
    + introv H4 H5. exfalso.
      openPPRelIn H1.
      destruct H1_3 as [H13 | H13]; inverts H13.

    + introv H3. exfalso.
      openPPRelIn H1.
      destruct H1_3 as [H13 | H13]; inverts H13.
        
    + introv H3 H4. exfalso.
      openPPRelIn H1.
      destruct H1_3 as [H13 | H13].
      * inverts H13.
      * eapply InstFlatProg'NotOutCritial'; eassumption.
  Qed.

  Lemma PDRelPres2:
    forall (R : Type) (s1 s1': TMState) (s2 : BMapState Val)
      (p1 : Program TMBaseMethod R)
      (p2 : Program (Abortable (Program (BMapMethod Key Val))) R)
      R' R'' t (m: BMapMethod Key Val R'') (m': Program (BMapMethod Key Val) R') (f: R'' -> Program (BMapMethod Key Val) R'),

      PPRel (r <- callMap m; Noop (instFlatProg' tmImpl instrument' _ (f r) t))
            (x <- abortable m'; Return x) p1 p2 (* s1 s2 *) ->
      PDRel p1 s1 p2 s2 ->
      PDRel p1 s1' p2 s2.

  Proof.
    introv H1 H2.
    invertPDRelTac H2.
    - econstructor. reflexivity.
    - econstructor. reflexivity.
    - constructor.
      eapply TRelPres2; eassumption.
  Qed.

  Lemma PPRelPres2:
    forall (R' R : Type)
      (* (s1 : TMState) (s2 : BMapState) *)
      (* (s1' : LockState * BMapState) *)
      (r : R') (m : Program (BMapMethod Key Val) R) (t : TId) (m2 : BMapMethod Key Val R')
      (f1 : R' -> Program (BMapMethod Key Val) R) (R'' : Type)
      (p1t : Program (PairInstruction LockMethod (BMapMethod Key Val)) R'')
      (p2t : Program (Abortable (Program (BMapMethod Key Val))) R''),
      
      PPRel (r0 <- callMap m2; Noop (instFlatProg' tmImpl instrument' _ (f1 r0) t))
            (x <- abortable m; Return x) p1t p2t (* s1 s2 *) ->
      PPRel (Noop (instFlatProg' tmImpl instrument' _ (f1 r) t))
            (x <- abortable m; Return x) p1t p2t (* s1' s2 *).

  Proof.
    introv H1.
    openPPRelIn H1.
    destruct H1_3 as [H13 | H13].
    - inverts H13.
    - openPPRel. right. assumption.
  Qed.

    
  Lemma TRelPres3:
  forall (s1 s1' : TMState) (s2 s2' : BMapState Val) (R : Type)
    (p1 : Program TMBaseMethod (option R)) (R': Type) 
    (m : Program (BMapMethod Key Val) R') (m' : Program (BMapMethod Key Val) R) r,
    
    PPRel (_ <- callLock doUnlock; Noop (Return (Some r)))
          (x <- abortable m; Return x)
          p1
          (x <- abortable m'; Return x)
          (* s1 s2 *) ->
    TRel p1 s1 m' s2 ->
    TRel p1 s1' m' s2'.

  Proof.  
    introv H1 H2.
    inverts H2 as.
    
    + constructor.
    + constructor.
            
    + introv H2 H3. exfalso.
      openPPRelIn H1.
      destruct H1_3 as [H13 | H13].
      * inverts H13.
      * eapply InstFlatProg'NotOutCritial; eassumption.

    + introv H2 H3. exfalso.
      openPPRelIn H1.
      destruct H1_3 as [H13 | H13]; inverts H13.

    + introv H2 H3. exfalso.
      openPPRelIn H1.
      destruct H1_3 as [H13 | H13]; inverts H13.
      
    + introv H2 H3. exfalso.
      openPPRelIn H1.
      destruct H1_3 as [H13 | H13].
      * inverts H13.       
      * eapply InstFlatProg'NotOutCritial'; eassumption.
      
  Qed.

  
  Lemma PDRelPres3:
    forall (R : Type) (s1 s1' : TMState) (s2 s2' : BMapState Val)
      (p1 : Program TMBaseMethod R)
      (p2 : Program (Abortable (Program (BMapMethod Key Val))) R)
      R' (m': Program (BMapMethod Key Val) R') r,

      PPRel (_ <- callLock doUnlock; Noop (Return (Some r)))
            (x <- abortable m'; Return x) p1 p2 (* s1 s2 *) ->
      PDRel p1 s1 p2 s2 ->
      PDRel p1 s1' p2 s2'.

  Proof.
    introv H1 H2.
    invertPDRelTac H2.
    - econstructor. reflexivity.
    - econstructor. reflexivity.
    - constructor.
      eapply TRelPres3; eassumption.
  Qed.


  Lemma PPRelPres3:
    forall (R : Type)
      (* (s1 : TMState) (s2 : BMapState)
      (s1': LockState * BMapState) *)
      (m : Program (BMapMethod Key Val) R) (r : R) (R' : Type)
      (p1t : Program (PairInstruction LockMethod (BMapMethod Key Val)) R')
      (p2t : Program (Abortable (Program (BMapMethod Key Val))) R'),
      
      PPRel (_ <- callLock doUnlock; Noop (Return (Some r)))
            (x <- abortable m; Return x) p1t p2t (* s1 s2 *) ->
      PPRel (Noop (Return (Some r))) (Return (Some r)) p1t p2t
            (* s1' (mapState s1) *).

  Proof.
    introv H1.
    openPPRelIn H1.
    openPPRel. iauto.
  Qed.
  
  Ltac invertBindTypeTac H :=
    apply invertBindType in H;
    assert (Hn: True) by apply I;    
    let H0 := fresh H "_" in
    assert (H0: True) by apply I;
    let H1 := fresh H "_" in
    assert (H1: True) by apply I;
    let H2 := fresh H "_" in
    clear Hn H0 H1;
    destruct H as (H1 & H2); subst; specialize (H2 eq_refl); simpl in H2; destruct H2 as (H1 & H2); subst.

        
  Lemma stepRel:
    forall R' (m: TMBaseMethod R')
      R (f: R' -> Program TMBaseMethod R) s1
      p2 s2 s1' r,

      DRel s1 s2 ->
      PDRel (Bind m f) s1 p2 s2 ->
      tmBase.(EvalInstruction) m s1 r s1' ->
      exists p2' s2',
        ProgramSteps (abortableSpec (programSpec (atomicBMap Val))) _ p2 s2 p2' s2'
        /\ DRel s1' s2'                                      
        /\ PDRel (f r) s1' p2' s2'
        /\ forall R (p1t: Program TMBaseMethod R) p2t,
            PDRel p1t s1 p2t s2 ->
            PPRel (Bind m f) p2 p1t p2t (* s1 s2 *) ->
            PDRel p1t s1' p2t s2'
            /\ PPRel (f r) p2' p1t p2t (* s1' s2' *).

  Proof.
    introv H1 H2 H3.
    invertPDRelTac H2.
    remember (x <- m; f x).
    inversion H2; subst; try discriminate.

    (* Getting the lock. *)
    - invertBindTypeTac H.
      change f with (fun x => f x).
      rewrite <- H_1.
      
      simpl in H3. destruct H3 as (H31 & H32). apply invertLockSem in H31. destruct H31 as (H311 & H312 & H313). subst.      

      eexists. eexists.
      splits.

      * eapply ProgramStepsRefl.

      * unfold DRel. unfold lockState. intro. simpl in *. congruence.
        
      * { constructor. constructor.
          + unfold DRel in *. unfold lockState in *. unfold mapState in *. simpl in *.
            apply H1 in H312. subst. rewrite H32. constructor.
          + unfold lockState in *. assumption.          
        }
        
      * { introv H4 H5.
          split.
          - eapply PDRelPres1.
            + unfold lockState. eassumption.
            + assumption.
              
          - eapply PPRelPres1; eassumption.
        }

    - symmetry in H4. apply IRelInst in H4. inversion H4.
      
    (* A method call while the lock is hold. *)      
    - { (* A call on the map. *)
        invertBindTypeTac H4.
        change f with (fun x => f x).
        rewrite <- H4_1 in *.
        simpl in *. destruct H3 as (H31 & H32).
        eexists. eexists. splits.

        - eapply ProgramStepsRefl.

        - unfold DRel. unfold lockState in *. intro. simpl in *. congruence.
          
        - econstructor.
          + econstructor.
            * eapply ProgramStepsStepImplSteps. eassumption. constructor. unfold mapState. simpl. assumption.
            * unfold lockState in *. simpl in *. congruence.
              
        - introv H7 H8.
          setoid_rewrite frobProgramId in H8 at 1. simpl in H8.
          setoid_rewrite frobProgramId in H8 at 2. simpl in H8.
          split.
          
          + applys PDRelPres2 s1 s1'; eassumption.

          + eapply PPRelPres2; eassumption.
      }

    - { (* An unlock. *)
        invertBindTypeTac H4.
        change f with (fun x => f x).
        rewrite <- H4_1.
        eexists. eexists. splits.
        
        - eapply ProgramStepImplProgramSteps.
          constructor. simpl. constructor. unfold EvalProgram.
          eassumption.
          
        - simpl in H3. destruct H3 as (H31 & H32).
          unfold DRel. unfold lockState in *. unfold mapState.
          intro. simpl in *. congruence.
          
        - econstructor. reflexivity.
          
        - simpl in *. destruct H3 as (H31 & H32).
          apply invertLockSem in H31. destruct H31 as (H311 & H312 & H313).
          introv H7 H8.
          setoid_rewrite frobProgramId in H8 at 1; simpl in H8.
          setoid_rewrite frobProgramId in H8 at 2; simpl in H8.
          setoid_rewrite frobProgramId in H8 at 3. simpl in H8.
          split.
          
          + eapply PDRelPres3; eassumption.
            
          + eapply PPRelPres3; eassumption.              
      }
      
  Qed.

End GlobalLockTM.

Lemma globalLockTMStrictSerializable Val :
  StrictSerializable
    (atomicBMap Val) (globalLockTM Val) instrument.

  Proof.
    intros.
    unfold StrictSerializable.
    intros.
    unfold Linearizable.
    unfold instFlatImpl.
    unfold globalLockTM.
    unfold tmBase.
    unfold flattenObject.
    unfold abortableSpec.
    unfold serialObject.
    simpl in *.

    eapply simulatesCoInd; simpl.
    - eapply PPRelSym.
    - eapply initRel.
    - eapply noopRel.
    - eapply invRel.
    - eapply retRel.
    - eapply stepRel.
  Qed.

(* Print Assumptions globalLockTMStrictSerializable. *)

End GlobalLockTM.
