Require Import Arith Bool List String FunctionalExtensionality Omega.
From C4 Require Import Predefs Exec Map TransactionTactics Commuting.

Set Implicit Arguments.

(*
  Timestamp-based Optimistic Algorithm
  
  init_t ()
    ts(t) = clock
    clock = clock + 1

  read_t(k)
    if (wt(k) <= ts(t))
      // Only the locations that are not written since
      // the transaction has started are read.
      // The read should not be too late.
      if (c(k) = true)
        // The location should be unlocked so that
        // the latest value of the location is read.
        if (rt(k) < ts(t))
          rt(k) = ts(t)
        return m(k)
      else
        return delay
    else
      return false

  write_t (k, v)
    if (ts(t) >= rt(k))
      // Only the locations that are not read since
      // the transaction has started are written.
      // The write should not be too late.
      if (ts(t) >= wt(k))
        // If there is no transaction that is started after the current
        // transaction and has written to the location.
        cb(k) = false
        if (ts(t) != wt(k))
          ul(t).add(k, m(k), wt(k))
        m(k) = v
        wt(k) = ts(t)
      else
        // If a transaction T' that is started after the current transaction T 
        // is started has written to the location, the current write waits for
        // T' to release the lock of the location. If after the lock is released
        // the write time of the location is still greater than the start
        // time of T, we know that a new value is committed to the location.
        // The value of the current write is effectively overwritten by T'.
        // Thus, the current write does not write to the location.
        if (cb(k) = true)
          return true (* The following proof does not work with this branch right now. *)
        else
          return delay
    else
      return false   

  commit_t ()
    foreach (k, _, _) in ul(t)
      cb(k) = true

  abort_t ()
    foreach (k, t', v') in ul(t)
      wt(k) = t'
      m(k) = v'
      cb(k) = true
*)


Module TimestampAlg (MapArgs: MapPars).

  Import MapArgs.
  Module Map := Map MapArgs.
  Import Map.

  Definition TStamp := nat.
  Definition eq_ts_dec := eq_nat_dec.

  Inductive AlgMethod: Type -> Type :=
  | algInit: TId -> AlgMethod unit
  | algGet: TId -> Key -> AlgMethod (option (option Val))
  | algPut: TId -> Key -> option Val -> AlgMethod (option unit)
  | algAbort: TId -> AlgMethod unit
  | algCommit: TId -> AlgMethod unit.

  Record Undo := undo {
    undoKey: Key;
    undoTS: TStamp;
    undoVal: option Val
  }.

  Definition eq_undo_dec (u u': Undo): {u = u'} + {u <> u'}.
  Proof.
    repeat decide equality; auto using eq_key_dec, eq_val_dec.
  Qed.

  (* State that all threads can read and write *)
  Record AlgState := {
    clock: nat;                                       (* Global clock *)
    rt: Key -> TStamp;                                (* Read time *)
    wt: Key -> TStamp;                                (* Write time *)
    cb: Key -> bool;                                  (* Commit bit *)
    imap: State MapType;
    tst: TId -> TStamp;                               (* Timestamp *)
    ul: TId -> list Undo                              (* Undo log *)    
  }.

  Definition algInitState := {| clock := 1;
                                rt := fun k => 0;
                                wt := fun k => 0;
                                cb := fun k => true;
                                imap := mapInitState;
                                tst := fun _ => 0;
                                ul := fun _ => nil |}.

  Definition initTrans (ast: AlgState)(t: TId) : AlgState :=
    {| clock := S (clock ast);
       rt := rt ast;
       wt := wt ast;
       cb := cb ast;
       imap := imap ast;
       tst := override (tst ast) t (clock ast);
       ul := override (ul ast) t nil |}.

  Definition readCond (ast: AlgState)(t: TId)(k: Key): Prop :=
    wt ast k < tst ast t  /\  cb ast k = true.

  Definition readTrans (ast: AlgState)(t: TId)(k: Key) : AlgState :=
    {| clock := clock ast;
       rt := koverride (rt ast) k (max (rt ast k) (tst ast t));
       wt := wt ast;
       cb := cb ast;
       imap := imap ast;
       tst := tst ast;
       ul := ul ast |}.

  Definition readAbortCond (ast: AlgState)(t: TId)(k: Key): Prop :=
    tst ast t < wt ast k.

  Definition writeCond1 (ast: AlgState)(t: TId)(k: Key): Prop :=
    rt ast k <= tst ast t  /\  wt ast k < tst ast t  /\  cb ast k = true.

  Definition writeCond2 (ast: AlgState)(t: TId)(k: Key): Prop :=
    rt ast k <= tst ast t  /\  wt ast k > tst ast t  /\  cb ast k = true.

  Definition writeTrans (ast: AlgState)(t: TId)(k: Key)(val: option Val) : AlgState :=
    {| clock := clock ast;
       cb := koverride (cb ast) k false;
       wt := koverride (wt ast) k (tst ast t);
       rt := rt ast;
       imap := koverride (imap ast) k val;
       ul := override (ul ast) t (undo k (wt ast k) (imap ast k) :: ul ast t);
       tst := tst ast |}.

  Definition writeAbortCond (ast: AlgState)(t: TId)(k: Key): Prop :=
    tst ast t < rt ast k.

  Fixpoint reinstallImap (us : list Undo) (im : State MapType) : State MapType :=
    match us with
      | nil => im
      | u :: us' => reinstallImap us' (koverride im (undoKey u) (undoVal u))
    end.

  Fixpoint reinstallWt (us : list Undo) (wt : Key -> TStamp) : Key -> TStamp :=
    match us with
      | nil => wt
      | u :: us' => reinstallWt us' (koverride wt (undoKey u) (undoTS u))
    end.

  Fixpoint reinstallCb (us : list Undo) (cb : Key -> bool) : Key -> bool :=
    match us with
      | nil => cb
      | u :: us' => reinstallCb us' (koverride cb (undoKey u) true)
    end.

  Definition abortTrans (ast: AlgState)(t: TId) : AlgState :=
    {| clock := clock ast;
       imap := reinstallImap (ul ast t) (imap ast);
       wt := reinstallWt (ul ast t) (wt ast);
       cb := reinstallCb (ul ast t) (cb ast);
       rt := rt ast;
       tst := tst ast;
       ul := override (ul ast) t nil |}.

  Definition commitTrans (ast: AlgState)(t: TId) : AlgState :=
    {| clock := clock ast;
       cb := reinstallCb (ul ast t) (cb ast);
       rt := rt ast;
       wt := wt ast;
       imap := imap ast;
       tst := tst ast;
       ul := override (ul ast) t nil |}.
  
  Inductive AlgExec: forall (A: Type), AlgState -> AlgMethod A -> A -> AlgState -> Prop :=
    | EAlgInit (ast : AlgState) t: 
        AlgExec ast
                (algInit t)
                tt (initTrans ast t)

    | EAlgGet (ast : AlgState) t key:
        readCond ast t key ->
        AlgExec ast (algGet t key)
                (Some (imap ast key)) (readTrans ast t key)

    | EAlgGetAbort (ast : AlgState) t key:
        readAbortCond ast t key ->
        AlgExec ast (algGet t key)
                None (abortTrans ast t)

    | EAlgPut1 (ast : AlgState) t key val:
        writeCond1 ast t key ->
        AlgExec ast (algPut t key val)
                (Some tt) (writeTrans ast t key val)

    (*| EAlgPut2 (ast : AlgState) t key val:
        writeCond2 (localize ast t) key ->
        AlgExec ast (algPut t key val)
                true ast*)
    (* This case seems fundamentally incompatible with my proof strategy,
     * so I'm omitting it for now. *)

    | EAlgPutAbort (ast : AlgState) t key val:
        writeAbortCond ast t key ->
        AlgExec ast (algPut t key val)
                None (abortTrans ast t)

    | EAlgAbort (ast : AlgState) t:
        AlgExec ast
                (algAbort t)
                tt (abortTrans ast t)

    | EAlgCommit (ast : AlgState) t:
        AlgExec ast
                (algCommit t)
                tt (commitTrans ast t).

  Definition AlgType := specType AlgState AlgMethod AlgExec.

  Lemma effects : forall R (m : AlgMethod R) c r c',
    Exec AlgType c m r c'
    -> match m in AlgMethod R return R -> _ with
       | algInit th => fun _ => c' = initTrans c th
       | algGet th k => fun r =>
                          (readAbortCond c th k /\ r = None /\ c' = abortTrans c th)
                          \/ readCond c th k /\ r = Some (imap c k) /\ c' = readTrans c th k
       | algPut th k v => fun r =>
                            (writeAbortCond c th k /\ r = None /\ c' = abortTrans c th)
                            \/ (writeCond1 c th k /\ r = Some tt /\ c' = writeTrans c th k v)
       | algCommit th => fun _ => c' = commitTrans c th
       | algAbort th => fun _ => c' = abortTrans c th
       end r.
  Proof.
    destruct 1; solver; eauto.
  Qed.

  Hint Rewrite @koverride_new_val @koverride_old_val using congruence.

  Lemma reinstallCb_false : forall k uls m,
    reinstallCb uls m k = false
    -> m k = false
       /\ (forall u, In u uls -> undoKey u <> k).
  Proof.
    induction uls; simpl; intros.

    eauto.

    apply IHuls in H.
    destruct (eq_key_dec (undoKey a) k); solver; eauto.
  Qed.

  Lemma reinstallCb_true : forall k uls m,
    m k = true
    -> reinstallCb uls m k = true.
  Proof.
    induction uls; simpl; intros.

    eauto.

    apply IHuls.
    destruct (eq_key_dec (undoKey a) k); solver.
  Qed.

  Hint Resolve reinstallCb_true.

  Lemma reinstallCb_irrel : forall k uls m,
    (forall u, In u uls -> undoKey u <> k)
    -> reinstallCb uls m k = m k.
  Proof.
    induction uls; solver.
    rewrite IHuls; eauto.
    destruct (eq_key_dec (undoKey a) k); solver.
    exfalso; eauto.
  Qed.

  Lemma reinstallWt_irrel : forall k uls m,
    (forall u, In u uls -> undoKey u <> k)
    -> reinstallWt uls m k = m k.
  Proof.
    induction uls; solver.
    rewrite IHuls; eauto.
    destruct (eq_key_dec (undoKey a) k); solver.
    exfalso; eauto.
  Qed.

  Lemma reinstallImap_irrel : forall k uls m,
    (forall u, In u uls -> undoKey u <> k)
    -> reinstallImap uls m k = m k.
  Proof.
    induction uls; solver.
    rewrite IHuls; eauto.
    destruct (eq_key_dec (undoKey a) k); solver.
    exfalso; eauto.
  Qed.

  Hint Rewrite reinstallCb_irrel reinstallWt_irrel reinstallImap_irrel using tauto.

  Module M <: S.
    Definition absObj := MapType.
    Definition absInit := mapInitState.

    Definition concObj := AlgType.
    Definition concInit := algInitState.

    Definition clock := clock.
    Definition initialize := algInit.

    Definition translateMethod (th : TId) R (m : Method absObj R) : Method concObj (option R) :=
      match m with
      | mapGet k => algGet th k
      | mapPut k v => algPut th k v
      end.

    Theorem discriminate_mi : forall th1 th2 R (m1 : Method absObj R),
      existT (Method concObj) _ (translateMethod th1 m1)
      <> existT (Method concObj) _ (initialize th2).
    Proof.
      destruct m1; discriminate.
    Qed.

    Definition translateReturn (th : TId) (r : option nat) : Method concObj unit :=
      match r with
      | None => algAbort th
      | Some _ => algCommit th
      end.

    Theorem discriminate_ri : forall th1 th2 r,
      translateReturn th1 r <> initialize th2.
    Proof.
      destruct r; discriminate.
    Qed.

    Theorem discriminate_mr : forall th1 th2 R (m1 : Method absObj R) r2,
      existT (Method concObj) _ (translateMethod th1 m1)
      <> existT (Method concObj) _ (translateReturn th2 r2).
    Proof.
      destruct m1, r2; discriminate.
    Qed.

    Theorem translateMethod_inj : forall th R1 R2 (m1 : Method absObj R1) (m2 : Method absObj R2),
      existT (Method concObj) _ (translateMethod th m1) = existT (Method concObj) _ (translateMethod th m2)
      -> existT (Method absObj) R1 m1 = existT (Method absObj) R2 m2.
    Proof.
      destruct m1, m2; inversion 1; auto.
    Qed.

    Record abortR' (th : TId) (c c' : State concObj) : Prop := {
      ArFresh : tst c th = 0;
      ArClock : clock c' = S (clock c);
      ArRt : forall k, rt c' k = rt c k \/ rt c' k = max (rt c k) (clock c);
      ArWt : wt c' = wt c;
      ArCb : cb c' = cb c;
      ArImap : imap c' = imap c;
      ArTstSelf : tst c' th = clock c;
      ArTstOther : forall th', th' <> th -> tst c' th' = tst c th';
      ArUlBefore : ul c th = nil;
      ArUlAfter : forall th', ul c' th' = ul c th'
    }.

    Definition abortR := abortR'.

    Definition timestamp := tst.

    Hint Rewrite @override_new_val @override_old_val using (congruence || omega).

    Theorem clock_initialize : forall c th r c',
      Exec concObj c (initialize th) r c'
      -> clock c' = S (clock c).
    Proof.
      intros; apply effects in H; solver.
    Qed.

    Theorem clock_step : forall c th R (s : step _ R) r c',
      doStep _ translateMethod translateReturn th s c r c'
      -> clock c' = clock c.
    Proof.
      inversion 1; subst.
      apply effects in H7; destruct m; solver.
      apply effects in H7; destruct r0; solver.
    Qed.

    Theorem clock_abortR : forall th c c',
      abortR th c c'
      -> clock c' = S (clock c).
    Proof.
      destruct 1; auto.
    Qed.

    Theorem initialize_initializes : forall c th r c',
      Exec concObj c (initialize th) r c'
      -> timestamp c' th = clock c.
    Proof.
      intros; apply effects in H; solver.
    Qed.

    Record invariant' (c : State concObj) : Prop := {
      InvClock : clock c >= 1;
      InvTstBound : forall th, tst c th < clock c;
      InvTstUnique : forall th th', tst c th > 0 -> tst c th' > 0 -> tst c th = tst c th' -> th = th';

      InvWtUl : forall k, cb c k = false
                          -> exists th u, tst c th = wt c k
                                          /\ wt c k > 0
                                          /\ In u (ul c th)
                                          /\ undoKey u = k;
      InvUlWt : forall th u, In u (ul c th)
                             -> wt c (undoKey u) = tst c th
                                /\ tst c th > 0
                                /\ cb c (undoKey u) = false;

      InvWt : forall k, wt c k < clock c;
      InvUlTs : forall u th, In u (ul c th)
                             -> undoTS u < tst c th
    }.

    Definition invariant := invariant'.

    Theorem invariant_init : invariant concInit.
    Proof.
      split; solver.
    Qed.

    Theorem clock_bound : forall c th,
      invariant c
      -> timestamp c th < clock c.
    Proof.
      destruct 1; auto.
    Qed.

    Ltac u := unfold timestamp, readCond, readAbortCond, writeCond1, writeAbortCond in *.

    Ltac t' :=
      match goal with
      | [ |- context[override _ ?th1 _ ?th2] ] =>
        repeat match goal with
               | [ H : forall x, _ < _ -> _ |- _ ] => specialize (H th2)
               end;
        destruct (th1 =_? th2)
      | [ _ : context[override _ ?th1 _ ?th2] |- _ ] =>
        repeat match goal with
               | [ H : forall x, _ < _ -> _ |- _ ] => specialize (H th2)
               end;
        destruct (th1 =_? th2)
      | [ |- context[koverride _ ?k1 _ ?k2] ] => destruct (eq_key_dec k1 k2)
      | [ _ : context[koverride _ ?k1 _ ?k2] |- _ ] => destruct (eq_key_dec k1 k2)
      | [ H1 : forall k, cb _ k = _ -> _, H2 : cb _ _ = _ |- _ ] => specialize (H1 _ H2)
      | [ H1 : forall th u, In u _ -> _, H2 : In ?u _ |- _ ] =>
        match goal with
        | [ _ : context[wt _ (undoKey u)] |- _ ] => fail 1
        | _ => apply H1 in H2
        end
      | [ H : reinstallCb _ _ _ = false |- _ ] => apply reinstallCb_false in H
      end.

    Ltac cleaner :=
      repeat match goal with
             | [ H : ?X = ?Y, H' : ?Y = ?X |- _ ] => clear H'
             end.

    Ltac t := u; solver; repeat (t'; cleaner; solver).

    Ltac tryAllExists :=
      match goal with
      | [ |- @ex ?T _ ] =>
        match goal with
        | [ x : T |- _ ] => exists x; tryAllExists
        | _ => eexists; tryAllExists
        end
      | _ => solve [t; eauto]
      end.

    Ltac useBound :=
      match goal with
      | [ H : forall th, tst _ th < clock _, th : TId |- _ ] =>
        specialize (H th); unfold clock in *; omega
      end.

    Ltac useUnique :=
      match goal with
      | [ H : forall th1 th2 : TId, tst ?c th1 > 0 -> _, th1 : TId, th2 : TId |- _ ] =>
        let H' := fresh in assert (H' : tst c th1 = tst c th2) by congruence;
          apply H in H'; clear H; omega
      end.

    Theorem invariant_initialize : forall th c r c',
      timestamp c th = 0
      -> Exec concObj c (initialize th) r c'
      -> invariant c
      -> invariant c'.
    Proof.
      intros; apply effects in H0; destruct H1; solver; split; t; tryAllExists || useBound.
    Qed.

    Ltac useUniqueContra := intros;
      match goal with
      | [ H : forall x y, In _ _ -> _, H' : In _ _ |- _ ] => apply H in H'; intuition idtac;
                                                             congruence || useUnique
      end.

    Hint Rewrite reinstallWt_irrel reinstallCb_irrel using useUniqueContra.

    Lemma reinstallWt_bound : forall k c uls m,
      m k < clock c
      -> (forall u, In u uls -> undoTS u < clock c)
      -> reinstallWt uls m k < TimestampAlg.clock c.
    Proof.
      induction uls; t.
      apply IHuls; t.
    Qed.

    Lemma reinstallWt_bound' : forall c th k uls,
      tst c th < clock c
      -> wt c k < clock c
      -> (forall u, In u uls -> undoTS u < tst c th)
      -> reinstallWt uls (wt c) k < TimestampAlg.clock c.
    Proof.
      intros.
      apply reinstallWt_bound; t.
      apply H1 in H2.
      omega.
    Qed.

    Hint Resolve reinstallWt_bound'.

    Theorem invariant_step : forall th R (s : step _ R) c r c',
      doStep _ translateMethod translateReturn th s c r c'
      -> timestamp c th > 0
      -> invariant c
      -> invariant c'.
    Proof.
      destruct 1; apply effects in H.

      destruct m, 2; split; t; useUniqueContra || tryAllExists.

      destruct r, 2; split; t; tryAllExists.
    Qed.

    Hint Resolve lt_S.

    Theorem invariant_abortR : forall th c c',
      abortR th c c'
      -> timestamp c th = 0
      -> invariant c
      -> invariant c'.
    Proof.
      destruct 1, 2; split.

      t.

      intros.
      match goal with
      | [ th1 : TId, th2 : TId |- _ ] =>
        (assert (th1 <> th2) by congruence; fail 1)
          || (destruct (th1 =_? th2); t)
      end.
      repeat match goal with
             | [ H : forall th', (th' = _ -> False) -> _ |- context[tst _ ?th] ] =>
               rewrite H by congruence
             | [ H : forall th', (th' = _ -> False) -> _, H' : context[tst _ ?th] |- _ ] =>
               rewrite H in H' by congruence
             end.
      t.
      
      intros.
      repeat match goal with
             | [ th1 : TId, th2 : TId |- _ ] =>
               (assert (th1 <> th2) by congruence; fail 1)
                 || (destruct (th1 =_? th2); t)
             end;
      repeat match goal with
             | [ H : forall th', (th' = _ -> False) -> _ |- context[tst _ ?th] ] =>
               rewrite (H th) by congruence
             | [ H : forall th', (th' = _ -> False) -> _, H' : context[tst _ ?th] |- _ ] =>
               rewrite (H th) in H' by congruence
             end; eauto;
      match goal with
      | [ H : forall th : TId, _ < _, th : TId |- _ ] => specialize (H th); omega
      end.
      
      t.
      rewrite ArCb0 in *.
      t.
      destruct (x =_? th); t.
      exists x.
      rewrite ArTstOther0 by congruence.
      rewrite ArUlAfter0 by auto.
      eauto.

      intros.
      rewrite ArUlAfter0 in H0.
      destruct (th0 =_? th).
      t.
      rewrite ArWt0.
      rewrite ArTstOther0 by congruence.
      rewrite ArCb0.
      eauto.

      t.

      t.
      rewrite ArUlAfter0 in H0.
      destruct (th0 =_? th); subst.
      t.
      rewrite ArTstOther0 by congruence.
      eauto.
    Qed.

    Theorem stamps_init : forall th, timestamp concInit th = 0.
    Proof.
      auto.
    Qed.

    Theorem stamps_initialize : forall th c r c',
      Exec concObj c (initialize th) r c'
      -> forall th', th' <> th -> timestamp c' th' = timestamp c th'.
    Proof.
      intros; apply effects in H; t.
    Qed.

    Theorem stamps_step : forall th R (s : step _ R) c r c',
      doStep _ translateMethod translateReturn th s c r c'
      -> forall th', timestamp c' th' = timestamp c th'.
    Proof.
      destruct 1; apply effects in H.

      destruct m; t.
      destruct r; t.
    Qed.

    Theorem stamps_abortR_same : forall th c c',
      abortR th c c'
      -> timestamp c' th = clock c.
    Proof.
      destruct 1; auto.
    Qed.

    Theorem stamps_abortR_diff : forall th c c',
      abortR th c c'
      -> forall th', th' <> th -> timestamp c' th' = timestamp c th'.
    Proof.
      destruct 1; auto.
    Qed.

    Hint Constructors doStep AlgExec.
    Hint Extern 1 (Exec _ _ _ _ _) => progress simpl.

    Lemma reinstallWt_committed : forall c k th m,
      invariant c
      -> cb c k = true
      -> reinstallWt (ul c th) m k = m k.
    Proof.
      intros; apply reinstallWt_irrel.
      destruct H; t; congruence.
    Qed.

    Lemma reinstallImap_committed : forall c k th m,
      invariant c
      -> cb c k = true
      -> reinstallImap (ul c th) m k = m k.
    Proof.
      intros; apply reinstallImap_irrel.
      destruct H; t; congruence.
    Qed.

    Hint Rewrite reinstallWt_committed reinstallImap_committed using assumption.

    Lemma AlgExec_cong : forall c R (m : AlgMethod R) r c' r' c'',
      AlgExec c m r c'
      -> r = r'
      -> c' = c''
      -> AlgExec c m r' c''.
    Proof.
      t.
    Qed.

    Lemma algState_split : forall c1 c2,
      clock c1 = clock c2
      -> wt c1 = wt c2
      -> rt c1 = rt c2
      -> cb c1 = cb c2
      -> imap c1 = imap c2
      -> tst c1 = tst c2
      -> ul c1 = ul c2
      -> c1 = c2.
    Proof.
      destruct c1, c2; t.
    Qed.

    Lemma koverride_double_eq : forall A m k (v1 v2 : A),
      koverride (koverride m k v1) k v2
      = koverride m k v2.
    Proof.
      intros; extensionality k'; t.
    Qed.

    Hint Rewrite koverride_double_eq.

    Ltac max1 :=
      match goal with
      | [ |- context[max ?A ?B] ] => destruct (Max.max_spec A B); generalize dependent (max A B); intuition subst
      | [ _ : context[max ?A ?B] |- _ ] => destruct (Max.max_spec A B); generalize dependent (max A B); intuition subst
      end.

    Ltac max := intros; repeat max1; omega.

    Lemma max_comm : forall a b c,
      max (max a b) c = max (max a c) b.
    Proof.
      max.
    Qed.

    Hint Resolve max_comm.

    Ltac algState_eq := t; apply algState_split; t;
                        try match goal with
                            | [ |- @eq (_ -> _) _ _ ] =>
                              let x := fresh "x" in extensionality x; t
                            end.

    Hint Immediate Max.max_lub_l Max.max_lub_r.

    Lemma reinstallWt_agree : forall k uls m1 m2,
      m1 k = m2 k
      -> reinstallWt uls m1 k = reinstallWt uls m2 k.
    Proof.
      induction uls; t.
      apply IHuls; t.
    Qed.

    Lemma reinstallWt_unused : forall k k' v uls m,
      k' <> k
      -> reinstallWt uls (koverride m k' v) k = reinstallWt uls m k.
    Proof.
      intros; apply reinstallWt_agree; t.
    Qed.

    Lemma reinstallCb_agree : forall k uls m1 m2,
      m1 k = m2 k
      -> reinstallCb uls m1 k = reinstallCb uls m2 k.
    Proof.
      induction uls; t.
      apply IHuls; t.
    Qed.

    Lemma reinstallCb_unused : forall k k' v uls m,
      k' <> k
      -> reinstallCb uls (koverride m k' v) k = reinstallCb uls m k.
    Proof.
      intros; apply reinstallCb_agree; t.
    Qed.

    Lemma reinstallImap_agree : forall k uls m1 m2,
      m1 k = m2 k
      -> reinstallImap uls m1 k = reinstallImap uls m2 k.
    Proof.
      induction uls; t.
      apply IHuls; t.
    Qed.

    Hint Rewrite reinstallWt_unused reinstallCb_unused using congruence.

    Lemma reinstallCb_true' : forall c k th,
      invariant c
      -> cb c k = true
      -> reinstallCb (ul c th) (koverride (cb c) k false) k = false.
    Proof.
      intros.
      rewrite reinstallCb_irrel.
      t.
      destruct H; t; congruence.
    Qed.

    Hint Rewrite reinstallCb_true' using assumption.

    Lemma reinstallImap_koverride : forall c th k o m,
      invariant c
      -> cb c k = true
      -> koverride (reinstallImap (ul c th) m) k o = reinstallImap (ul c th) (koverride m k o).
    Proof.
      intros; extensionality k'; t.
      apply reinstallImap_agree; t.
    Qed.

    Hint Resolve reinstallImap_koverride.

    Lemma koverride_double_neq : forall k1 k2 A m (v1 v2 : A),
      k1 <> k2
      -> koverride (koverride m k1 v1) k2 v2 = koverride (koverride m k2 v2) k1 v1.
    Proof.
      intros; extensionality k; t.
    Qed.

    Hint Extern 1 (koverride _ _ _ = koverride _ _ _) => apply koverride_double_neq; congruence.

    Lemma reinstallCb_comm : forall uls1 uls2 m,
      reinstallCb uls1 (reinstallCb uls2 m) = reinstallCb uls2 (reinstallCb uls1 m).
    Proof.
      induction uls1; t.

      rewrite <- IHuls1; f_equal.
      extensionality k; t.
      symmetry; apply reinstallCb_true; t.
    Qed.

    Hint Resolve reinstallCb_comm.

    Lemma not_reinstalled : forall c th k,
      invariant c
      -> reinstallCb (ul c th) (cb c) k = true
      -> wt c k < tst c th
      -> cb c k = true.
    Proof.
      intros.
      rewrite reinstallCb_irrel in H0; auto.
      destruct H; t.
      apply InvUlWt0 in H; omega.
    Qed.

    Hint Resolve not_reinstalled.
    Hint Extern 2 (_ < _) => omega.

    Lemma not_reinstalled' : forall c th k m,
      invariant c
      -> wt c k < tst c th
      -> reinstallCb (ul c th) (koverride m k false) k = false.
    Proof.
      intros.
      rewrite reinstallCb_irrel; t.
      destruct H; t.
      apply InvUlWt0 in H2; omega.
    Qed.

    Hint Resolve not_reinstalled'.

    Theorem method_aborts : forall c th R (m : Method _ R) c',
      Exec concObj c (translateMethod th m) None c'
      -> Exec concObj c (translateReturn th None) tt c'.
    Proof.
      intros; apply effects in H; destruct m; t; discriminate.
    Qed.

    Theorem steps_commute : forall c th1 th2 R1 (s1 : step _ R1) R2 (s2 : step _ R2) r1 c' r2 c'',
      timestamp c th1 > timestamp c th2
      -> doStep _ translateMethod translateReturn th1 s1 c r1 c'
      -> doStep _ translateMethod translateReturn th2 s2 c' r2 c''
      -> match s1 in step _ R return R -> _ with
         | ReturnStep o => fun _ => o <> None
         | MethodStep _ _ => fun r => r <> None
         end r1
      -> match s2 in step _ R return R -> _ with
         | ReturnStep _ => fun _ => True
         | MethodStep _ _ => fun r => r <> None
         end r2
      -> invariant c
      -> exists c''', doStep _ translateMethod translateReturn th2 s2 c r2 c'''
                      /\ doStep _ translateMethod translateReturn th1 s1 c''' r1 c''.
    Proof.
      Ltac commute := t;
        repeat match goal with
                   [ x : unit |- _ ] => destruct x
               end;
        try max;
        (do 2 esplit; [
           solve [ constructor | constructor; simpl; constructor; t; eauto ]
         | constructor; simpl; eapply AlgExec_cong; [ constructor; t | t | algState_eq ] ]).
      
      destruct 2, 1; apply effects in H0; apply effects in H1.

      destruct m, m0; commute.
      destruct m, r0; commute.
      destruct r, m; commute.
      destruct r, r0; commute.
    Qed.

    Lemma abort_wt : forall th1 th2 c c' k,
      abortR th1 c c'
      -> invariant c
      -> wt c' k < tst c' th2
      -> th1 <> th2
      -> wt c k < tst c th2.
    Proof.
      destruct 1, 1; t.
      rewrite ArWt0 in *.
      rewrite ArTstOther0 in * by congruence.
      auto.
    Qed.

    Hint Immediate abort_wt.

    Lemma abort_cb_true : forall th c c' k,
      abortR th c c'
      -> cb c' k = true
      -> cb c k = true.
    Proof.
      intros; erewrite <- ArCb by eassumption; auto.
    Qed.

    Hint Immediate abort_cb_true.

    Lemma abortR_rt : forall th1 th2 c c' k,
      abortR th1 c c'
      -> rt c' k <= tst c' th2
      -> th1 <> th2
      -> rt c k <= tst c th2.
    Proof.
      intros.
      erewrite <- ArTstOther by eauto.
      destruct (ArRt H k); t; max.
    Qed.

    Hint Immediate abortR_rt.

    Theorem abort_step_commute : forall c th1 th2 R2 (s2 : step _ R2) c' r2 c'',
      abortR th1 c c'
      -> doStep _ translateMethod translateReturn th2 s2 c' r2 c''
      -> th1 <> th2
      -> match s2 in step _ R return R -> _ with
         | ReturnStep _ => fun _ => True
         | MethodStep _ _ => fun r => r <> None
         end r2
      -> invariant c
      -> exists c''', doStep _ translateMethod translateReturn th2 s2 c r2 c'''
                      /\ abortR th1 c''' c''.
    Proof.
      Ltac acomm := t; repeat match goal with
                              | [ x : unit |- _ ] => destruct x
                              end;
                    (do 2 esplit; [
                       solve [ constructor | constructor; simpl; try erewrite ArImap by eassumption;
                                             constructor; t; eauto ]
                     | match goal with
                       | [ H : abortR _ _ _ |- _ ] => destruct H; split; t;
                                                      try match goal with
                                                          | [ H : forall x, ul _ _ = ul _ _ |- _ ] =>
                                                            rewrite H; solve [ auto ]
                                                          end
                       end ]).
                    
      destruct 2; apply effects in H0.

      destruct m; acomm.
      rewrite ArTstOther0 by congruence.
      destruct (ArRt0 k0); t.

      destruct r; acomm.
    Qed.

    Lemma lt_clock : forall c k th,
      invariant c
      -> TimestampAlg.clock c < wt c k
      -> tst c th < wt c k.
    Proof.
      destruct 1; unfold clock in *; t.
      specialize (InvTstBound0 th).
      omega.
    Qed.

    Hint Immediate lt_clock.

    Lemma oob : forall c k,
      invariant c
      -> TimestampAlg.clock c < wt c k
      -> False.
    Proof.
      destruct 1; unfold clock in *; intros.
      specialize (InvWt0 k); omega.
    Qed.

    Hint Immediate oob.

    Theorem init_step_commute : forall c th1 th2 R2 (s2 : step _ R2) r1 c' r2 c'',
      Exec concObj c (initialize th1) r1 c'
      -> doStep _ translateMethod translateReturn th2 s2 c' r2 c''
      -> th1 <> th2
      -> match s2 in step _ R return R -> _ with
         | ReturnStep _ => fun _ => True
         | MethodStep _ _ => fun r => r <> None
         end r2
      -> invariant c
      -> exists c''', doStep _ translateMethod translateReturn th2 s2 c r2 c'''
                      /\ Exec _ c''' (initialize th1) r1 c''.
    Proof.
      Ltac icomm := t; repeat match goal with
                              | [ x : unit |- _ ] => destruct x
                              end;
                    (do 2 esplit; [
                       constructor; simpl;
                       solve [ apply EAlgGet; t
                               | apply EAlgPut1; t
                               | constructor; t ]
                     | eapply AlgExec_cong; [ eauto | t | algState_eq ] ]).

      destruct 2; apply effects in H; apply effects in H0.

      destruct m; icomm.

      destruct r; icomm.
    Qed.

    Definition absOf := imap.

    Theorem absOf_init : absOf concInit = absInit.
    Proof.
      auto.
    Qed.

    Theorem absOf_initialize : forall th c r c', Exec concObj c (initialize th) r c'
      -> absOf c' = absOf c.
    Proof.
      intros; apply effects in H; t.
    Qed.

    Hint Constructors MapExec.

    Theorem absOf_method : forall th R (m : Method absObj R) c r c',
      Exec concObj c (translateMethod th m) (Some r) c'
      -> Exec absObj (absOf c) m r (absOf c').
    Proof.
      intros; apply effects in H; destruct m; t; try discriminate;
      match goal with
      | [ H : Some _ = Some _ |- _ ] => inversion H
      end; eauto.
    Qed.

    Theorem absOf_return : forall th o c r c',
      Exec concObj c (translateReturn th (Some o)) r c'
      -> absOf c' = absOf c.
    Proof.
      intros; apply effects in H; t.
    Qed.

    Theorem absOf_abortR : forall th c c',
      abortR th c c'
      -> absOf c' = absOf c.
    Proof.
      destruct 1; auto.
    Qed.

    Lemma koverride_silly : forall A (m : Key -> A) k v,
      m k = v
      -> koverride m k v = m.
    Proof.
      intros.
      extensionality k'; t.
    Qed.

    Hint Rewrite koverride_silly using congruence.

    Lemma abortR_in' : forall th c c' r,
      ul c th = nil
      -> (forall th', tst c th' < clock c)
      -> doSteps _ _ translateMethod translateReturn th c c' r
      -> clock c' = clock c
         /\ (forall k, rt c' k = rt c k \/ rt c' k = max (rt c k) (tst c th))
         /\ (forall th', tst c' th' = tst c th')
         /\ (forall th', th' <> th -> ul c' th' = ul c th')
         /\ match r with
            | Some o =>
              match o with
                | None => wt c' = reinstallWt (ul c th) (wt c)
                          /\ imap c' = reinstallImap (ul c' th) (imap c)
                          /\ cb c' = reinstallCb (ul c th) (cb c)
                          /\ ul c' th = nil
                | _ => True
              end
            | _ =>
              wt c = reinstallWt (ul c' th) (wt c')
              /\ cb c = reinstallCb (ul c' th) (cb c')
              /\ imap c = reinstallImap (ul c' th) (imap c')
              (*reinstallWt (ul c th) (wt c) = reinstallWt (ul c' th) (wt c')
              /\ reinstallCb (ul c th) (cb c) = reinstallCb (ul c' th) (cb c')
              /\ reinstallImap (ul c th) (imap c) = reinstallImap (ul c' th) (imap c')*)
            end.
    Proof.
      induction 3.

      t.

      apply effects in H2.
      destruct o; t.

      apply effects in H2.
      destruct m; t; try discriminate.
      rewrite H8.
      destruct (H7 k0); t.

      apply effects in H2.
      destruct m; t; discriminate.
    Qed.

    Theorem abortR_in : forall th c r c' c'',
      invariant c
      -> timestamp c th = 0
      -> Exec concObj c (initialize th) r c''
      -> doSteps _ _ translateMethod translateReturn th c'' c' (Some None)
      -> abortR th c c'.
    Proof.
      intros.
      apply effects in H1; t.
      apply abortR_in' in H2.
      2: t.
      2: destruct H; t.
      t; split; auto.
      rewrite H9 in H6; auto.
      rewrite H3; t.
      intros; rewrite H3; t.
      destruct H.
      generalize (InvUlWt0 th).
      destruct (ul c th); t.
      specialize (H _ (or_introl _ eq_refl)); t.
      intros.
      destruct (th' =_? th); subst.
      rewrite H9.
      destruct H.
      generalize (InvUlWt0 th).
      destruct (ul c th); t.
      specialize (H _ (or_introl _ eq_refl)); t.
      rewrite H4; t.
    Qed.
  End M.

  Module Impl := Make(M).
End TimestampAlg.
