Require Import
        Coq.Lists.List
        Coq.Program.Equality
        Coq.Relations.Relations.

From C4.Core Require Import
  Program
  Simulates
  SimulatesFacts
  Threads.

Section FlattenBehavior.

  Context {BaseMethod : Type -> Type}
          (base : Spec (Interaction BaseMethod))
          {Method : Type -> Type}
          (impl : forall Ret, Method Ret -> Program BaseMethod Ret).

  Inductive FlattenThread :=
  | ftProgram Ret : Program BaseMethod Ret -> FlattenThread
  | ftWait A Ret : (A -> Program BaseMethod Ret) -> FlattenThread.

  Inductive FlattenAdvance' :
    forall Ret,
      option (Interaction BaseMethod Ret) ->
      list FlattenThread ->
      Ret ->
      list FlattenThread ->
      Prop :=
  | faSpawn A R m f fts :
      FlattenAdvance' unit (Some (Spawn m)) (ftProgram R (Bind m f) :: fts) tt (ftWait A R f :: fts)
  | faJoin A R a f fts :
      FlattenAdvance' A (Some Join) (ftWait A R f :: fts) a (ftProgram R (f a) :: fts)
  | faNoop R p fts :
      FlattenAdvance' unit None (ftProgram R (Noop p) :: fts) tt (ftProgram R p :: fts)
  | faSkip Ret int R p fts ret fts' :
      FlattenAdvance' Ret int fts ret fts' ->
      FlattenAdvance' Ret int (ftProgram R p :: fts) ret (ftProgram R p :: fts')
  | faUp Ret int A R f fts ret fts' :
      FlattenAdvance' Ret int fts ret fts' ->
      FlattenAdvance' Ret (option_map Up int) (ftWait A R f :: fts) ret (ftWait A R f :: fts').

  Inductive FlattenAdvance :
    list FlattenThread * base.(State) ->
    list FlattenThread * base.(State) ->
    Prop :=
  | faIntro Ret int fts st ret fts' st' :
      FlattenAdvance' Ret int fts ret fts' ->
      match int with
      | Some int' => base.(EvalInstruction) int' st ret st'
      | None => st = st'
      end ->
      FlattenAdvance (fts, st) (fts', st').

  Definition FlattenAdvances :=
    clos_refl_trans_1n (list FlattenThread * base.(State)) FlattenAdvance.

  Inductive FlattenInteraction :
    forall Ret,
      Interaction Method Ret ->
      list FlattenThread ->
      Ret ->
      list FlattenThread ->
      Prop :=
  | fiSpawn Ret m fts :
      FlattenInteraction unit (Spawn m) fts tt (ftProgram Ret (impl Ret m) :: fts)
  | fiJoin Ret ret fts :
      FlattenInteraction Ret Join (ftProgram Ret (Return ret) :: fts) ret fts
  | fiCons Ret int ft fts ret fts' :
      FlattenInteraction Ret int fts ret fts' ->
      FlattenInteraction Ret (Up int) (ft :: fts) ret (ft :: fts').

  Inductive EvalFlattenBehavior Ret (int : Interaction Method Ret) :
    list FlattenThread * base.(State) ->
    Ret ->
    list FlattenThread * base.(State) ->
    Prop :=
  | efbIntro ts fts' ret fts'' st' :
      FlattenAdvances ts (fts', st') ->
      FlattenInteraction Ret int fts' ret fts'' ->
      EvalFlattenBehavior Ret int ts ret (fts'', st').

  Definition FlattenBehavior : Spec (Interaction Method) :=
    {|
      State := list FlattenThread * base.(State);
      EvalInstruction := EvalFlattenBehavior;
      initialState := (nil, base.(initialState))
    |}.

  Lemma simulateFlattenAdvance ts ts' :
    FlattenAdvance ts ts' ->
    Simulates_ FlattenBehavior FlattenBehavior ts' ts.
  Proof.
    intro fa.
    inversion fa; subst.
    split.
    intros A int' a ts'' E.
    destruct E.
    eexists.
    - esplit.
      + econstructor 2; eassumption.
      + eassumption.
    - apply simulatesRefl.
  Qed.

End FlattenBehavior.

Arguments FlattenThread : clear implicits.


Section FlattenBehaviorSimulates.

  Context {BaseMethod : Type -> Type}
          (base1 base2 : Spec (Interaction BaseMethod))
          (Method : Type -> Type)
          (impl : forall Ret, Method Ret -> Program BaseMethod Ret).

  Lemma flattenAdvanceSimulates fts st1 fts' st1' st2 :
    FlattenAdvance base1 (fts, st1) (fts', st1') ->
    Simulates_ base1 base2 st1 st2 ->
    exists2 st2',
      FlattenAdvance base2 (fts, st2) (fts', st2') &
      Simulates_ base1 base2 st1' st2'.
  Proof.
    intros fa S.
    dependent destruction fa.
    destruct int.
    - destruct S as [S].
      edestruct S.
      + eassumption.
      + eexists.
        * esplit.
          -- eassumption.
          -- simpl.
             eassumption.
        * assumption.
    - subst.
      eexists.
      * econstructor.
        -- eassumption.
        -- reflexivity.
      * assumption.
  Qed.

  Lemma flattenAdvancesSimulates fts st1 fts' st1' st2 :
    FlattenAdvances base1 (fts, st1) (fts', st1') ->
    Simulates_ base1 base2 st1 st2 ->
    exists2 st2',
      FlattenAdvances base2 (fts, st2) (fts', st2') &
      Simulates_ base1 base2 st1' st2'.
  Proof.
    intro fas.
    remember (fts, st1) as fs1.
    remember (fts', st1') as fs1'.
    revert fts st1 st2 Heqfs1 Heqfs1'.
    induction fas; intros; dependent destruction Heqfs1; dependent destruction Heqfs1'.
    - eexists.
      constructor 1.
      assumption.
    - destruct y.
      edestruct flattenAdvanceSimulates.
      + eassumption.
      + eassumption.
      + edestruct IHfas.
        * reflexivity.
        * reflexivity.
        * eassumption.
        * eexists.
          -- econstructor 2; eassumption.
          -- assumption.
  Qed.

  Theorem flattenBehaviorSimulates'_ fts st1 st2 :
    Simulates_ base1 base2 st1 st2 ->
    Simulates_ (FlattenBehavior base1 impl) (FlattenBehavior base2 impl) (fts, st1) (fts, st2).
  Proof.
    intro.
    apply relationSimulates with
    (fun fs1 fs2 => fst fs1 = fst fs2 /\ Simulates_ base1 base2 (snd fs1) (snd fs2)).
    - clear.
      intros A int [fts1 st1] fs1'' a [fts2 st2] [H S] E.
      simpl in *.
      inversion E.
      subst.
      edestruct flattenAdvancesSimulates.
      + eassumption.
      + eassumption.
      + eexists; esplit; eauto.
    - auto.
  Qed.

  Theorem flattenBehaviorSimulates' :
    Simulates base1 base2 ->
    Simulates (FlattenBehavior base1 impl) (FlattenBehavior base2 impl).
  Proof.
    apply flattenBehaviorSimulates'_.
  Qed.

End FlattenBehaviorSimulates.
