From Coq Require Import
     List
     Program
     Relations
     RelationClasses.

From Paco Require Import paco.

From C4.Util Require Import
  Util
  LibTactics.

From C4.Core Require Import
  Program
  ProgramOps
  ProgramFacts
  ProgramOpsFacts
  Simulates
  Object
  Linearizability
  Threads
  ThreadsFacts
  Superposition
  RSimulates
  RSimulatesFacts
  Eutt.

Local Notation fI := flattenObject.

(* A strong notion of simulation: the underlying spec of [i1] simulates
   the underlying spec of [i2]; the method object of [i1]
   is eutt to the method object of [i2]. *)
Inductive HImplSimulates {I} (i1 i2 : Object I) : Prop :=
| mk_HImplSimulates (RR : IRel (ObjInstruction i1) (ObjInstruction i2))
  : HRSimulates RR (objBase i1) (objBase i2) ->
    (ieq ==> eutt RR)%isig (objProgram i1) (objProgram i2) ->
    HImplSimulates i1 i2.

Lemma Reflexive_HImplSimulates {M} (i : Object M)
  : HImplSimulates i i.
Proof.
  apply (mk_HImplSimulates i i ieq).
  - apply Simulates_RSimulates. apply superpositionSingleSimulates.
    constructor.
  - intros ? ? ? []. apply Reflexive_eutt_ieq.
Qed.

Lemma Proper_serialObject {M} (o1 o2 : Spec M)
  : HSimulates o1 o2 ->
    HImplSimulates (serialObject o1) (serialObject o2).
Proof.
  intros.
  apply (mk_HImplSimulates (serialObject o1) (serialObject o2) ieq).
  - apply Simulates_RSimulates. assumption.
  - intros ? ? ? [].
    apply Reflexive_eutt_ieq.
Qed.

Lemma mapSerialEq {M N} (f : forall R, M R -> N R) (o : Spec N)
  : HImplSimulates
      (mapObject f (serialObject o))
      (serialObject (mapSpec o f)).
Proof.
  unfold mapObject. unfold serialObject. cbn.
  match goal with
  | [ |- _ ?i1 ?i2 ] =>
    apply (mk_HImplSimulates i1 i2 (irel_fun f)); cbn
  end.
  - apply mapSpecHRSimulates.
  - intros ? ? ? [].
    unfold mapImplProg.
    pfold. constructor.
    + constructor.
    + left. pfold. constructor.
Qed.

(* A strong notion of simulation: the underlying spec of [i1] simulates
   the underlying spec of [i2]; the method object of [i1]
   is eutt to the method object of [i2]. *)
Inductive ImplSimulates {I} (i1 i2 : Object I) : Prop :=
| mk_ImplSimulates (RR : IRel (ObjInstruction i1) (ObjInstruction i2))
  : RSimulates RR (objBase i1) (objBase i2) ->
    (ieq ==> eutt RR)%isig (objProgram i1) (objProgram i2) ->
    ImplSimulates i1 i2.

Lemma EvalInstruction_superposition_nonempty {M} (o : Spec M)
      R (m : M R) (r : R) s1 ss2
  : EvalInstruction (superposition o) m s1 r ss2 ->
    exists s2, ss2 s2.
Proof.
  intros []; eauto.
Qed.

Lemma EvalProgram_Noop {M} (o : Spec M)
      {R} (p : Program M R) (r : R) (st st' : State o)
  : EvalProgram o R p st r st' ->
    EvalProgram o R (Noop p) st r st'.
Proof.
  intros. red.
  eapply ProgramStepStepsImplSteps; [ constructor | auto ].
Qed.

Lemma Proper_flattenObject {M N}
      (i1 i2 : Object N)
      (f1 f2 : forall R, M R -> Program N R)
  : HImplSimulates i1 i2 ->
    (ieq ==> eutt ieq)%isig f1 f2 ->
    HImplSimulates (fI i1 _ f1) (fI i2 _ f2).
Proof.
  intros [] ?.
  apply (mk_HImplSimulates (fI i1 _ f1) (fI i2 _ f2) RR).
  - auto.
  - intros ? ? ? ?.
    apply (Proper_flattenProgram ieq RR); auto.
Qed.

Definition StepSuperposition {S}
  : (S -> S -> Prop) ->
    (S -> Prop) -> (S -> Prop) -> Prop :=
  fun R ss ss' =>
    (exists y, ss' y)
    /\ (forall s',
           ss' s' ->
           exists s,
           ss s /\ R s s').

Lemma Reflexive_StepSuperposition {S} (RS : S -> S -> Prop)
  : (forall s, RS s s) ->
    forall P, (exists s, P s) -> StepSuperposition RS P P.
Proof.
  intros.
  constructor; auto.
  econstructor; eauto.
Qed.

Lemma Transitive_StepSuperposition {S} (RS : S -> S -> Prop)
  : (forall s1 s2 s3, RS s1 s2 -> RS s2 s3 -> RS s1 s3) ->
    forall P Q R, StepSuperposition RS P Q -> StepSuperposition RS Q R -> StepSuperposition RS P R.
Proof.
  intros H P Q R HPQ HQR.
  split.
  - apply HQR.
  - intros s Hs.
    apply HQR in Hs.
    destruct Hs as [s' [Hs' I']].
    apply HPQ in Hs'.
    destruct Hs' as [s'' [Hs'' I'']].
    eauto.
Qed.

Definition StepSuperpositionFun {S}
  : (S -> S -> Prop) ->
    (S -> Prop) -> S -> Prop :=
  fun R ss s' =>
    exists s,
      ss s /\ R s s'.

Lemma GreatestStepSuperposition {S} (R : S -> S -> Prop) (ss : S -> Prop)
  : (exists s s', ss s /\ R s s') ->
    StepSuperposition R ss (StepSuperpositionFun R ss).
Proof.
  firstorder.
Qed.

Lemma StepSuperpositionInhabited {S} (R : S -> S -> Prop) (ss ss' : S -> Prop)
  : StepSuperposition R ss ss' ->
    exists s s', ss s /\ R s s' /\ ss' s'.
Proof.
  firstorder.
Qed.

Definition all_eutt_fst {M S} (st1 st2 : list (sigT (Program M)) * S)
  : Prop := all_eutt ieq (fst st1) (fst st2) /\ snd st1 = snd st2.

Inductive TBbisim {M} (i1 i2 : Object M) RR t1 s1 t2 (ss2 : _ -> Prop) : Prop :=
| TBbisim_base
  : all_eutt RR t1 t2 ->
    (exists s2, ss2 s2) ->
    RSimulates_ RR (objBase i1) (superposition (objBase i2)) s1 ss2 ->
    TBbisim i1 i2 RR t1 s1 t2 ss2.

Lemma TBbisim_inhabited {M} (i1 i2 : Object M) RR t1 s1 t2 ss2
  : TBbisim i1 i2 RR t1 s1 t2 ss2 ->
    exists st2, ss2 st2.
Proof.
  intros []; assumption.
Qed.

(* Restrict a nondeterministic ThreadsState to the tail. *)
Definition threads_tail {M} {i1 i2 : Object M} RR {R}
           (p1 : Program (ObjInstruction i1) R)
           (ss2 : ThreadsState i2 -> Prop)
  : ThreadsState i2 -> Prop :=
  fun st2 =>
    exists p2,
      eutt RR p1 p2
      /\ ss2 (existT _ R p2 :: fst st2, snd st2).

Lemma EvalInstruction_weaken
      (M2 : Type -> Type) (o2 : Spec M2)
      (ss2 ss2' : State o2 -> Prop) (R : Type)
      (i2 : M2 R) (r : R) (x : State (superposition o2))
  : (forall s2, ss2 s2 -> ss2' s2) ->
    EvalInstruction (superposition o2) i2 ss2 r x ->
    EvalInstruction (superposition o2) i2 ss2' r x.
Proof.
  cbn.
  intros Hss2_ (Hinhabited & Hss2).
  split; [ assumption | ].
  intros st' Hst'.
  specialize (Hss2 st' Hst'). destruct Hss2 as (? & ? & ?); eauto.
Qed.

Lemma RSimulates_weaken (M1 M2 : Type -> Type)
      (RR : IRel M1 M2)
      (o1 : Spec M1) (o2 : Spec M2)
  : forall
      (s1 : State o1)
      (ss2 ss2' : State o2 -> Prop),
    (forall s2, ss2 s2 -> ss2' s2) ->
    RSimulates_ RR o1 (superposition o2) s1 ss2 ->
    RSimulates_ RR o1 (superposition o2) s1 ss2'.
Proof.
  cofix self. intros.
  destruct H0.
  constructor. intros.
  specialize H0 with (1 := H1) (2 := H2).
  destruct H0.
  exists x.
  - revert H H0. apply EvalInstruction_weaken.
  - assumption.
Qed.

(*
Definition eq_ThreadState {M} {i : Object M} (s s' : ThreadsState i)
  : Prop :=
  all_eutt ieq (fst s) (fst s') /\
  snd s = snd s'.
*)

Lemma untaus1_ProgramSteps
      (M : Type -> Type) (i2 : Object M)
      (R0 R : Type) (a : R0) (m2 : ObjInstruction i2 R0)
      (p2 : Program (ObjInstruction i2) R)
      (k2 : R0 -> Program (ObjInstruction i2) R)
  : untaus1 (fun p : Program (ObjInstruction i2) R => (x <- m2; k2 x) = p) p2 ->
    forall s x : State (objBase i2),
      EvalInstruction (objBase i2) m2 x a s ->
      ProgramSteps (objBase i2) R p2 x (k2 a) s.
Proof.
  induction 1; cbn; intros.
  - eapply ProgramStepStepsImplSteps; [ constructor | auto ].
  - apply ProgramStepImplProgramSteps. subst; constructor. assumption.
Qed.

Definition thread_nstate {M} {S}
           (t : list (sigT (Program M))) (ss : S -> Prop)
  : _ * S -> Prop :=
  fun st => fst st = t /\ ss (snd st).

Lemma threads_behavior_bisim'_Bind {M} {i1 i2 : Object M}
      RR {R0 R} (instr : _ R0) (f : R0 -> Program _ R) ps
      a s1 s1' t2 ss2
  : EvalInstruction (objBase i1) instr s1 a s1' ->
    TBbisim i1 i2 RR
           (existT (Program (ObjInstruction i1)) R (x <- instr; f x) :: ps)
           s1 t2 ss2 ->
    exists t2' ss2',
    StepSuperposition
      (ThreadsAdvance i2) (thread_nstate t2 ss2) (thread_nstate t2' ss2') /\
    TBbisim
      i1 i2 RR
      (existT (Program (ObjInstruction i1)) R (f a) :: ps) s1'
      t2' ss2'.
Proof.
  intros ? []. inv H0. injpair_; subst.
  apply eutt_Bind_inv in H7.
  destruct H7 as (m2 & k2 & Rm2 & EUTTk2 & UNTAUS).
  exists (existT _ _ (k2 a) :: t0).
  destruct H2.
  specialize H0 with (1 := Rm2) (2 := H).
  destruct H0 as [ss2' [] ?].
  exists ss2'.
  split.
  - split; cbn.
    + destruct H0. eexists (_,_); split; cbn. reflexivity. eassumption.
    + intros [] []. cbn in *; subst.
      apply H2 in H5.
      destruct H5 as (? & ? & ?).
      eexists (_, _); split; cbn. constructor; cbn. reflexivity. eassumption.
      apply ProgramStepsImplThreadsAdvance.
      revert H5; apply untaus1_ProgramSteps.
      assumption.
  - constructor; auto.
    constructor; auto.
Qed.

Lemma threads_behavior_bisim_Noop {M} {i1 i2 : Object M} RR
      R p ps s1 t2 ss2
  : TBbisim i1 i2 RR
           (existT (Program (ObjInstruction i1)) R (Noop p) :: ps) s1 t2 ss2 ->
    TBbisim i1 i2 RR
           (existT (Program (ObjInstruction i1)) R p :: ps) s1 t2 ss2.
Proof.
  intros [].
  constructor; auto.
  inv H.
  injpair_; subst.
  constructor.
  - apply eutt_Noop_L_inv; assumption.
  - assumption.
Qed.

Lemma TBbisim_cons_inv {M} (i1 i2 : Object M) RR
      p1 t1 s1 t2 ss2
  : TBbisim i1 i2 RR (p1 :: t1) s1 t2 ss2 ->
    exists p2 t2_,
    t2 = existT _ _ p2 :: t2_ /\
    eutt RR (projT2 p1) p2 /\
    TBbisim i1 i2 RR t1 s1 t2_ ss2.
Proof.
  intros [].
  inv H.
  eexists _, _.
  split; [ reflexivity | ].
  split; [ assumption | ].
  constructor; assumption.
Qed.

Lemma step_advance_cons (M : Type -> Type) (i2 : Object M)
      (p : {x : Type & Program (ObjInstruction i2) x})
      (ss2 : State (objBase i2) -> Prop)
      (t2_ t2' : list {x : Type & Program (ObjInstruction i2) x})
      (ss2' : State (objBase i2) -> Prop)
  : StepSuperposition (ThreadsAdvance i2)
                      (thread_nstate t2_ ss2)
                      (thread_nstate t2' ss2') ->
    StepSuperposition (ThreadsAdvance i2)
                      (thread_nstate (p :: t2_) ss2)
                      (thread_nstate (p :: t2') ss2').
Proof.
  intros [].
  split.
  - destruct H as [[] []].
    exists (p :: t2', s).
    constructor. reflexivity. auto.
  - intros [] []. cbn in *. subst.
    specialize (H0 (t2', s)).
    destruct H0 as [[] [[]]]; [ red; auto | ].
    cbn in *; subst.
    exists (p :: t2_, s0).
    split; [ red; auto | ].
    apply advanceCons. assumption.
Qed.

Lemma bisim_cons
      (M : Type -> Type) (i1 i2 : Object M)
      (RR : IRel (ObjInstruction i1) (ObjInstruction i2))
      (R : Type)
      (p : Program (ObjInstruction i1) R)
      (ps' : list {x : Type & Program (ObjInstruction i1) x})
      (s1' : State (objBase i1))
      (p2 : Program (ObjInstruction i2) R)
      (t2' : list {x : Type & Program (ObjInstruction i2) x})
      (ss2' : State (objBase i2) -> Prop)
  : eutt RR p p2 ->
    TBbisim i1 i2 RR ps' s1' t2' ss2' ->
    TBbisim i1 i2 RR (existT _ R p :: ps') s1'
            (existT _ R p2 :: t2') ss2'.
Proof.
  intros ? [].
  constructor; auto.
  constructor; auto.
Qed.

Lemma Reflexive_advance_nstate
      (M : Type -> Type) (i1 i2 : Object M)
      (RR : IRel (ObjInstruction i1) (ObjInstruction i2))
      x y
      (t2 : list {x : Type & Program (ObjInstruction i2) x})
      (ss2 : State (objBase i2) -> Prop)
  : TBbisim i1 i2 RR x y t2 ss2 ->
    StepSuperposition (ThreadsAdvance i2) (thread_nstate t2 ss2)
                      (thread_nstate t2 ss2).
Proof.
  intros Hss2.
  apply Reflexive_StepSuperposition.
  * apply Reflexive_ThreadsAdvance.
  * apply TBbisim_inhabited in Hss2.
    destruct Hss2; eexists (_, _); red; cbn; eauto.
Qed.

Lemma ThreadsBehavior_ThreadAdvance_bisim {M} (i1 i2 : Object M) RR t1 s1 t1' s1'
  : ThreadAdvance i1 (t1, s1) (t1', s1') ->
    forall t2 ss2,
    TBbisim i1 i2 RR t1 s1 t2 ss2 ->
    exists t2' ss2',
    StepSuperposition (ThreadsAdvance i2) (thread_nstate t2 ss2) (thread_nstate t2' ss2')
    /\ TBbisim i1 i2 RR t1' s1' t2' ss2'.
Proof.
  intros TA. dependent induction TA.
  - specialize IHTA with (1 := JMeq_refl) (2 := JMeq_refl).
    intros t2 ss2 BISIM.
    apply TBbisim_cons_inv in BISIM.
    destruct BISIM as (p2 & t2_ & ? & ? & BISIM). subst.
    apply IHTA in BISIM.
    destruct BISIM as (t2' & ss2' & ? & BISIM).
    exists (existT _ _ p2 :: t2') ss2'.
    split.
    + revert H; apply step_advance_cons.
    + revert BISIM. destruct p.
      apply bisim_cons; auto.
  - eauto using threads_behavior_bisim'_Bind.
  - intros t2 ss2 Hss2. exists t2 ss2.
    apply threads_behavior_bisim_Noop in Hss2.
    split.
    + revert Hss2; apply Reflexive_advance_nstate.
    + assumption.
Qed.

Lemma ThreadsBehavior_ThreadsAdvance_bisim' {M} (i1 i2 : Object M) RR st1 st1'
  : ThreadsAdvance i1 st1 st1' ->
    let t1 := fst st1 in
    let s1 := snd st1 in
    let t1' := fst st1' in
    let s1' := snd st1' in
    forall t2 ss2,
    TBbisim i1 i2 RR t1 s1 t2 ss2 ->
    exists t2' ss2',
    StepSuperposition (ThreadsAdvance i2) (thread_nstate t2 ss2) (thread_nstate t2' ss2')
    /\ TBbisim i1 i2 RR t1' s1' t2' ss2'.
Proof.
  induction 1; cbn.
  - intros t2 ss2 Hss2. exists t2 ss2; split; eauto.
    revert Hss2; apply Reflexive_advance_nstate.
  - destruct x as [t11 s11], y as [t12 s12]; cbn in *.
    intros t2 ss2 Hss2.
    eapply ThreadsBehavior_ThreadAdvance_bisim with (1 := H) in Hss2.
    destruct Hss2 as (t2' & ss2' & Hstep' & Hss2').
    apply IHclos_refl_trans_1n in Hss2'.
    destruct Hss2' as (t2'' & ss2'' & Hstep'' & Hss2'').
    exists t2'' ss2''. split; eauto.
    eapply Transitive_StepSuperposition; eauto.
    apply ThreadsAdvanceTrans.
Qed.

Lemma ThreadsBehavior_ThreadsAdvance_bisim {M} (i1 i2 : Object M) RR t1 s1 t1' s1'
  : ThreadsAdvance i1 (t1, s1) (t1', s1') ->
    forall t2 ss2,
    TBbisim i1 i2 RR t1 s1 t2 ss2 ->
    exists t2' ss2',
    StepSuperposition (ThreadsAdvance i2) (thread_nstate t2 ss2) (thread_nstate t2' ss2')
    /\ TBbisim i1 i2 RR t1' s1' t2' ss2'.
Proof.
  intros I t2 ss2.
  apply ThreadsBehavior_ThreadsAdvance_bisim' with (1 := I).
Qed.

Fixpoint step_preinteraction {M} {R}
         (i : Object M) (ii : Interaction M R) (r : R)
         (ts : list {X : Type & Program (ObjInstruction i) X})
  : list {X : Type & Program (ObjInstruction i) X} -> Prop :=
  match ii in Interaction _ R return R -> _ with
  | Spawn m => fun _ => eq ts
  | Join => fun r ts' =>
    exists p ts_,
    ts = existT _ _ p :: ts_ /\
    ts' = existT _ _ (Return r) :: ts_ /\
    eutt ieq p (Return r)
  | Up ii => fun r ts' =>
    exists t ts_ ts'_,
      ts = t :: ts_ /\
      ts' = t :: ts'_ /\
      step_preinteraction i ii r ts_ ts'_
  end r.

Fixpoint step_interaction {M} {R}
         (i : Object M) (ii : Interaction M R) (r : R)
         (ts : list {X : Type & Program (ObjInstruction i) X})
  : list {X : Type & Program (ObjInstruction i) X} -> Prop :=
  match ii in Interaction _ R return R -> _ with
  | Spawn m => fun _ => eq (existT _ _ (objProgram i _ m) :: ts)
  | Join => fun r ts' =>
    exists p,
    ts = existT _ _ p :: ts' /\
    eutt ieq p (Return r)
  | Up ii => fun r ts' =>
    exists t ts_ ts'_,
      ts = t :: ts_ /\
      ts' = t :: ts'_ /\
      step_interaction i ii r ts_ ts'_
  end r.

Definition on_left {A B} (R : A -> A -> Prop) (xs ys : A * B) : Prop :=
  R (fst xs) (fst ys) /\ snd xs = snd ys.

Definition sstep_preinteraction {M} {R}
           (i : Object M) (ii : Interaction M R) (r : R)
  : _ -> list {X : Type & Program (ObjInstruction i) X} * State (objBase i) -> Prop :=
  StepSuperpositionFun (on_left (step_preinteraction i ii r)).

Definition sstep_interaction {M} {R}
           (i : Object M) (ii : Interaction M R) (r : R)
  : _ -> list {X : Type & Program (ObjInstruction i) X} * State (objBase i) -> Prop :=
  StepSuperpositionFun (on_left (step_interaction i ii r)).

Section INTERACTION.

Context {M : Type -> Type} (i1 i2 : Object M)
        (RR : IRel (ObjInstruction i1) (ObjInstruction i2)).

Lemma step_preinteraction_advance
      (R : Type) (ii : Interaction M R) (r : R)
  : forall (s' : ThreadsState i2)
           (l : list {X : Type & Program (ObjInstruction i2) X}),
    step_preinteraction i2 ii r l (fst s') -> ThreadsAdvance i2 (l, snd s') s'.
Proof.
  induction ii; intros []; cbn; intros until 0.
  - intros; subst; apply Reflexive_ThreadsAdvance.
  - intros (? & ? & ? & ? & ?); subst.
    apply ProgramStepsImplThreadsAdvance.
    eapply eutt_ProgramSteps; eassumption.
  - intros (? & ? & ? & ? & ? & ?); subst.
    apply advanceCons.
    specialize (IHii r (x1, s) x0); auto.
Qed.

Lemma step_advance_preinteraction {R} {ii : Interaction M R}
      {t1 s1 t1' s1'} {r : R}
  : ThreadsInteraction i1 R ii (t1, s1) r (t1', s1') ->
    forall t2 ss2,
    TBbisim i1 i2 RR t1 s1 t2 ss2 ->
    exists t2',
    step_preinteraction i2 ii r t2 t2' /\
    StepSuperposition (ThreadsAdvance i2)
                      (thread_nstate t2 ss2)
                      (thread_nstate t2' ss2).
Proof.
  intros I; inv I. induction H2; cbn; intros t2 ss2 BISIM.
  - eexists; split; [ reflexivity | revert BISIM; apply Reflexive_advance_nstate ].
  - destruct BISIM.
    inv H. injpair_; subst.
    eexists. split.
    + eexists p2, t0. split; [ reflexivity | ]; split; [ reflexivity | ].
      apply Symmetric_eutt_ieq. revert H6. apply eutt_Ret_L_RR.
    + split.
      * destruct H0.
        eexists (_, _); red; cbn; eauto.
      * intros [] []. cbn in *; subst.
        eexists (_, _); split; [ red; cbn; eauto | ].
        apply ProgramStepsImplThreadsAdvance.
        eapply eutt_ProgramSteps.
        apply Symmetric_eutt_ieq. revert H6. apply eutt_Ret_L_RR.
  - destruct BISIM.
    inv H.
    edestruct IHThreadsInteraction' as (? & ? & ?).
    { constructor; eassumption. }
    eexists. split.
    + eexists _, _, _. split; [ | split ]; reflexivity + eassumption.
    + apply step_advance_cons. assumption.
Qed.

Lemma step_interaction_cons R
      (int : Interaction M R) (r : R)
      (p : sigT (Program (ObjInstruction i2)))
      (ss2 : State (objBase i2) -> Prop)
      (t t' : list {X : Type & Program (ObjInstruction i2) X})
  : StepSuperposition
      (fun ss20 : ThreadsState i2 => ThreadsInteraction i2 R int ss20 r)
      (thread_nstate t ss2) (thread_nstate t' ss2) ->
    StepSuperposition
      (fun ss20 : ThreadsState i2 => ThreadsInteraction i2 R (Up int) ss20 r)
      (thread_nstate (p :: t) ss2)
      (thread_nstate (p :: t') ss2).
Proof.
  intros []. split.
  - destruct H. destruct H. eexists (_, _); red; cbn in *; eauto.
  - intros [] []. cbn in *; subst.
    specialize (H0 (t', s)).
    destruct H0 as (? & ? & ?).
    + red; cbn; auto.
    + eexists (_, _); split.
      * red; cbn; eauto.
      * constructor. constructor.
        inv H1. destruct H0. cbn in *; subst; auto.
Qed.

Lemma step_interaction_interaction {R} {ii : Interaction M R}
      {t1 s1 t1' s1'} {r : R}
  : ThreadsInteraction i1 R ii (t1, s1) r (t1', s1') ->
    forall t2 ss2,
    TBbisim i1 i2 RR t1 s1 t2 ss2 ->
    forall t2',
    step_preinteraction i2 ii r t2 t2' ->
    exists t2'',
    step_interaction i2 ii r t2 t2'' /\
    StepSuperposition (fun ss20 => ThreadsInteraction i2 R ii ss20 r)
                      (thread_nstate t2' ss2)
                      (thread_nstate t2'' ss2).
Proof.
  intros I; inv I. induction H2; cbn; intros t2 ss2 BISIM t2'.
  - intros [].
    eexists; split; [ reflexivity | ].
    destruct BISIM.
    split.
    + destruct H0.
      eexists (_, x). split; cbn; eauto.
    + intros [] []. cbn in *; subst.
      exists (t2, s).
      split; [red; auto| ].
      constructor. constructor.
  - intros (? & ? & ? & ? & ?). subst.
    apply TBbisim_cons_inv in BISIM.
    destruct BISIM as (? & ? & ? & ? & ?).
    inv H. injpair_; subst.
    exists x2. split.
    + exists x1. auto.
    + split. destruct H2. destruct H2.
      eexists (_, _); red; cbn; eauto.
      intros [] []. cbn in *; subst.
      eexists (_, _); split.
      * red; cbn; eauto.
      * constructor. constructor.
  - intros (? & ? & ? & ? & ? & ?). subst.
    apply TBbisim_cons_inv in BISIM.
    destruct BISIM as (? & ? & ? & ? & ?).
    inv H.
    specialize IHThreadsInteraction' with (1 := H3) (2 := H1).
    destruct IHThreadsInteraction' as (? & ? & ?).
    eexists. split.
    + eexists _, _, _; split; eauto.
    + apply step_interaction_cons.
      assumption.
Qed.

Context (IMPL_EUTT : (ieq ==> eutt RR)%isig (objProgram i1) (objProgram i2)).

Lemma ThreadsBehavior_ThreadsInteraction_bisim_ R int t1 s1 a t1' s1'
  : ThreadsInteraction i1 R int (t1, s1) a (t1', s1') ->
    forall t2 ss2,
    TBbisim i1 i2 RR t1 s1 t2 ss2 ->
    forall t2',
    step_interaction i2 int a t2 t2' ->
    TBbisim i1 i2 RR t1' s1' t2' ss2.
Proof.
  intros I. inv I. induction H2; cbn; intros t2 ss2 BISIM t2'.
  - intros [].
    apply bisim_cons; auto.
  - intros (? & ? & ?). subst.
    apply TBbisim_cons_inv in BISIM.
    destruct BISIM as (p2 & t2_ & ? & ? & ?).
    inv H.
    injpair_; subst. assumption.
  - intros (? & ? & ? & ? & ? & ?). subst.
    apply TBbisim_cons_inv in BISIM.
    destruct BISIM as (p2 & t2_ & ? & ? & ?).
    inv H.
    destruct p. apply bisim_cons; [ assumption | ].
    revert H1. apply IHThreadsInteraction'.
    assumption.
Qed.

Lemma EvalThreadsSuperposition R int (r : R) ss2 ss2' ss2''
  : StepSuperposition (ThreadsAdvance i2) ss2 ss2' ->
    StepSuperposition (fun ss2' => ThreadsInteraction i2 R int ss2' r) ss2' ss2'' ->
    EvalInstruction (superposition (ThreadsBehavior i2)) int ss2 r ss2''.
Proof.
  intros I J; split.
  - apply J.
  - intros s'' Hs''.
    apply J in Hs''. destruct Hs'' as [ s' [ Hs' HJ ]].
    apply I in Hs'. destruct Hs' as [ s [ Hs HI ]].
    exists s. split; auto. exists s'; auto.
Qed.

Lemma ThreadsBehavior_ThreadsInteraction_bisim R int t1 s1 a t1' s1'
  : ThreadsInteraction i1 R int (t1, s1) a (t1', s1') ->
    forall t2_ ss2_,
    forall t2 ss2,
    StepSuperposition (ThreadsAdvance i2) (thread_nstate t2_ ss2_) (thread_nstate t2 ss2) ->
    TBbisim i1 i2 RR t1 s1 t2 ss2 ->
    exists t2',
    EvalInstruction (superposition (ThreadsBehavior i2))
                    int (thread_nstate t2_ ss2_) a (thread_nstate t2' ss2) /\
    TBbisim i1 i2 RR t1' s1' t2' ss2.
Proof.
  intros TI t2_ ss2_ t2 ss2 STEP BISIM.
  assert (PRE := step_advance_preinteraction TI t2 ss2 BISIM).
  destruct PRE as (t20 & PRE & STEP_PRE).
  assert (POST := step_interaction_interaction TI t2 ss2 BISIM t20 PRE).
  destruct POST as (t2' & POST & STEP_POST).
  eexists t2'.
  split.
  - eapply EvalThreadsSuperposition.
    + eapply Transitive_StepSuperposition.
      * apply ThreadsAdvanceTrans.
      * eassumption.
      * eassumption.
    + assumption.
  - revert POST. eapply ThreadsBehavior_ThreadsInteraction_bisim_; eassumption.
Qed.

Lemma ThreadsBehavior_bisim_
  : forall t1 s1 t2 ss2,
    TBbisim i1 i2 RR t1 s1 t2 ss2 ->
    Simulates.Simulates_ (ThreadsBehavior i1) (superposition (ThreadsBehavior i2))
              (t1, s1) (thread_nstate t2 ss2).
Proof.
  cofix self. intros t1 s1 t2 ss2 BISIM.
  constructor; intros R ii r [t1_ s1_] EVAL.
  destruct EVAL as [[t1' s1'] ADVANCE INTERACTION].
  eapply ThreadsBehavior_ThreadsAdvance_bisim in ADVANCE; [ | eassumption ].
  destruct ADVANCE as (t2' & ss2' & STEP & BISIM').
  eapply ThreadsBehavior_ThreadsInteraction_bisim in INTERACTION; [ | eassumption | eassumption ].
  destruct INTERACTION as (t2'' & STEP'' & BISIM'').
  exists (thread_nstate t2'' ss2').
  - assumption.
  - apply self. auto.
Qed.

End INTERACTION.

(* If two objects are equivalent then they concurrently-simulate
   each other.

   - Equivalence of objects [HImplSimulates]: they are eutt
     and the underlying specs simulate each other.

   - The proof constructs explicitly a bisimulation [TBbisim] between
     thread states and then proves the simulation diagram.
 *)
Lemma Proper_ThreadsBehavior {M} (i1 i2 : Object M)
  : HImplSimulates i1 i2 ->
    HSimulates (ThreadsBehavior i1) (ThreadsBehavior i2).
Proof.
  intros []. red.
  eapply SimulatesFacts.simulatesTrans_.
  - apply ThreadsBehavior_bisim_ with (RR0 := RR); [ assumption | ].
    eapply TBbisim_base; [ constructor | | eassumption ].
    eexists; reflexivity.
  - apply superpositionSubsetSimulates.
    intros [] []; cbn in *; subst.
    constructor.
Qed.

(**)

(** Without "H" *)

Module Forward.

Inductive TBbisim {M} (i1 i2 : Object M) RR t1 s1 t2 s2 : Prop :=
| TBbisim_base
  : all_eutt RR t1 t2 ->
    RSimulates_ RR (objBase i1) (objBase i2) s1 s2 ->
    TBbisim i1 i2 RR t1 s1 t2 s2.

Lemma bisim_cons
      (M : Type -> Type) (i1 i2 : Object M)
      (RR : IRel (ObjInstruction i1) (ObjInstruction i2))
      (R : Type)
      (p : Program (ObjInstruction i1) R)
      (ps' : list {x : Type & Program (ObjInstruction i1) x})
      (s1' : State (objBase i1))
      (p2 : Program (ObjInstruction i2) R)
      (t2' : list {x : Type & Program (ObjInstruction i2) x})
      (ss2' : State (objBase i2))
  : eutt RR p p2 ->
    TBbisim i1 i2 RR ps' s1' t2' ss2' ->
    TBbisim i1 i2 RR (existT _ R p :: ps') s1'
            (existT _ R p2 :: t2') ss2'.
Proof.
  intros ? [].
  constructor; auto.
  constructor; auto.
Qed.

Lemma TBbisim_cons_inv {M} (i1 i2 : Object M) RR
      p1 t1 s1 t2 ss2
  : TBbisim i1 i2 RR (p1 :: t1) s1 t2 ss2 ->
    exists p2 t2_,
    t2 = existT _ _ p2 :: t2_ /\
    eutt RR (projT2 p1) p2 /\
    TBbisim i1 i2 RR t1 s1 t2_ ss2.
Proof.
  intros [].
  inv H.
  eexists _, _.
  split; [ reflexivity | ].
  split; [ assumption | ].
  constructor; assumption.
Qed.

Section INTERACTION.

Context {M : Type -> Type} (i1 i2 : Object M)
        (RR : IRel (ObjInstruction i1) (ObjInstruction i2)).

(**)

Lemma threads_behavior_bisim'_Bind {R0 R} (instr : _ R0) (f : R0 -> Program _ R) ps
      a s1 s1' t2 ss2
  : EvalInstruction (objBase i1) instr s1 a s1' ->
    TBbisim i1 i2 RR
           (existT (Program (ObjInstruction i1)) R (x <- instr; f x) :: ps)
           s1 t2 ss2 ->
    exists t2' ss2',
    ThreadsAdvance i2 (t2, ss2) (t2', ss2') /\
    TBbisim
      i1 i2 RR
      (existT (Program (ObjInstruction i1)) R (f a) :: ps) s1'
      t2' ss2'.
Proof.
  intros ? []. inv H0. injpair_. subst.
  apply eutt_Bind_inv in H6.
  destruct H6 as (m2 & k2 & Rm2 & EUTTk2 & UNTAUS).
  exists (existT _ _ (k2 a) :: t0).
  destruct H1.
  specialize H0 with (1 := Rm2) (2 := H).
  destruct H0 as [ss2' ? ?].
  exists ss2'.
  split.
  - apply ProgramStepsImplThreadsAdvance.
    revert H0; apply untaus1_ProgramSteps.
    assumption.
  - constructor; auto.
    constructor; auto.
Qed.

Lemma threads_behavior_bisim_Noop R p ps s1 t2 ss2
  : TBbisim i1 i2 RR
           (existT (Program (ObjInstruction i1)) R (Noop p) :: ps) s1 t2 ss2 ->
    TBbisim i1 i2 RR
           (existT (Program (ObjInstruction i1)) R p :: ps) s1 t2 ss2.
Proof.
  intros [].
  constructor; auto.
  inv H.
  injpair_. subst.
  constructor.
  - apply eutt_Noop_L_inv; assumption.
  - assumption.
Qed.


Lemma ThreadsBehavior_ThreadAdvance_bisim ts1 ts1'
  : ThreadAdvance i1 ts1 ts1' ->
    forall t2 ss2,
    TBbisim i1 i2 RR (fst ts1) (snd ts1) t2 ss2 ->
    exists t2' ss2',
    ThreadsAdvance i2 (t2, ss2) (t2', ss2')
    /\ TBbisim i1 i2 RR (fst ts1') (snd ts1') t2' ss2'.
Proof.
  intros TA. induction TA.
  - intros t2 ss2 BISIM.
    apply TBbisim_cons_inv in BISIM.
    destruct BISIM as (p2 & t2_ & ? & ? & BISIM). subst.
    apply IHTA in BISIM.
    destruct BISIM as (t2' & ss2' & ? & BISIM).
    exists (existT _ _ p2 :: t2') ss2'.
    split.
    + apply advanceCons. assumption.
    + revert BISIM. destruct p.
      apply bisim_cons; auto.
  - intros *; apply threads_behavior_bisim'_Bind; assumption.
  - intros t2 ss2 Hss2. exists t2 ss2.
    apply threads_behavior_bisim_Noop in Hss2.
    split.
    + apply Reflexive_ThreadsAdvance.
    + assumption.
Qed.

Lemma ThreadsBehavior_ThreadsAdvance_bisim' st1 st1'
  : ThreadsAdvance i1 st1 st1' ->
    let t1 := fst st1 in
    let s1 := snd st1 in
    let t1' := fst st1' in
    let s1' := snd st1' in
    forall t2 ss2,
    TBbisim i1 i2 RR t1 s1 t2 ss2 ->
    exists t2' ss2',
    ThreadsAdvance i2 (t2, ss2) (t2', ss2')
    /\ TBbisim i1 i2 RR t1' s1' t2' ss2'.
Proof.
  induction 1; cbn.
  - intros t2 ss2 Hss2. exists t2 ss2; split; eauto.
  - destruct x as [t11 s11], y as [t12 s12]; cbn in *.
    intros t2 ss2 Hss2.
    eapply ThreadsBehavior_ThreadAdvance_bisim with (1 := H) in Hss2.
    destruct Hss2 as (t2' & ss2' & Hstep' & Hss2').
    apply IHclos_refl_trans_1n in Hss2'.
    destruct Hss2' as (t2'' & ss2'' & Hstep'' & Hss2'').
    exists t2'' ss2''. split; eauto.
    eapply ThreadsAdvanceTrans; eauto.
Qed.

Lemma step_advance_preinteraction {R} {ii : Interaction M R}
      {t1 s1 t1' s1'} {r : R}
  : ThreadsInteraction i1 R ii (t1, s1) r (t1', s1') ->
    forall t2 ss2,
    TBbisim i1 i2 RR t1 s1 t2 ss2 ->
    exists t2',
    step_preinteraction i2 ii r t2 t2' /\
    ThreadsAdvance i2 (t2, ss2) (t2', ss2).
Proof.
  intros I; inv I. induction H2; cbn; intros t2 ss2 BISIM.
  - eexists; split; [ reflexivity | apply Reflexive_ThreadsAdvance ].
  - destruct BISIM.
    inv H. injpair_. subst.
    eexists. split.
    + eexists p2, t0. split; [ reflexivity | ]; split; [ reflexivity | ].
      apply Symmetric_eutt_ieq. revert H5. apply eutt_Ret_L_RR.
    + apply ProgramStepsImplThreadsAdvance.
      eapply eutt_ProgramSteps with (RR0 := fun X m1 m2 => RR X m2 m1).
      eapply Symmetric_eutt_; [ | eassumption ]; auto.
  - destruct BISIM.
    inv H.
    edestruct IHThreadsInteraction' as (? & ? & ?).
    { constructor; eassumption. }
    eexists. split.
    + eexists _, _, _. split; [ | split ]; reflexivity + eassumption.
    + apply advanceCons. assumption.
Qed.

Lemma step_interaction_interaction {R} {ii : Interaction M R}
      {t1 s1 t1' s1'} {r : R}
  : ThreadsInteraction i1 R ii (t1, s1) r (t1', s1') ->
    forall t2 ss2,
    TBbisim i1 i2 RR t1 s1 t2 ss2 ->
    forall t2',
    step_preinteraction i2 ii r t2 t2' ->
    exists t2'',
    step_interaction i2 ii r t2 t2'' /\
    ThreadsInteraction i2 R ii (t2', ss2) r (t2'', ss2).
Proof.
  intros I; inv I. induction H2; cbn; intros t2 ss2 BISIM t2'.
  - intros [].
    eexists; split; [ reflexivity | ].
    destruct BISIM.
    constructor; constructor.
  - intros (? & ? & ? & ? & ?). subst.
    apply TBbisim_cons_inv in BISIM.
    destruct BISIM as (? & ? & ? & ? & ?).
    inv H. injpair_; subst.
    exists x2. split.
    + exists x1. auto.
    + constructor; constructor.
  - intros (? & ? & ? & ? & ? & ?). subst.
    apply TBbisim_cons_inv in BISIM.
    destruct BISIM as (? & ? & ? & ? & ?).
    inv H.
    specialize IHThreadsInteraction' with (1 := H3) (2 := H1).
    destruct IHThreadsInteraction' as (? & ? & ?).
    eexists. split.
    + eexists _, _, _; split; eauto.
    + inv H4. constructor; constructor; assumption.
Qed.

Context (IMPL_EUTT : (ieq ==> eutt RR)%isig (objProgram i1) (objProgram i2)).

Lemma ThreadsBehavior_ThreadsInteraction_bisim_ R int t1 s1 a t1' s1'
  : ThreadsInteraction i1 R int (t1, s1) a (t1', s1') ->
    forall t2 ss2,
    TBbisim i1 i2 RR t1 s1 t2 ss2 ->
    forall t2',
    step_interaction i2 int a t2 t2' ->
    TBbisim i1 i2 RR t1' s1' t2' ss2.
Proof.
  intros I. inv I. induction H2; cbn; intros t2 ss2 BISIM t2'.
  - intros [].
    apply bisim_cons; auto.
  - intros (? & ? & ?). subst.
    apply TBbisim_cons_inv in BISIM.
    destruct BISIM as (p2 & t2_ & ? & ? & ?).
    inv H.
    injpair_. subst. assumption.
  - intros (? & ? & ? & ? & ? & ?). subst.
    apply TBbisim_cons_inv in BISIM.
    destruct BISIM as (p2 & t2_ & ? & ? & ?).
    inv H.
    destruct p. apply bisim_cons; [ assumption | ].
    revert H1. apply IHThreadsInteraction'.
    assumption.
Qed.

(**)

Lemma ThreadsBehavior_ThreadsAdvance_bisim t1 s1 t1' s1'
  : ThreadsAdvance i1 (t1, s1) (t1', s1') ->
    forall t2 ss2,
    TBbisim i1 i2 RR t1 s1 t2 ss2 ->
    exists t2' ss2',
    ThreadsAdvance i2 (t2, ss2) (t2', ss2')
    /\ TBbisim i1 i2 RR t1' s1' t2' ss2'.
Proof.
  intros I t2 ss2.
  apply ThreadsBehavior_ThreadsAdvance_bisim' with (1 := I).
Qed.

Lemma ThreadsBehavior_ThreadsInteraction_bisim R int t1 s1 a t1' s1'
  : ThreadsInteraction i1 R int (t1, s1) a (t1', s1') ->
    forall t2_ ss2_,
    forall t2 ss2,
    ThreadsAdvance i2 (t2_, ss2_) (t2, ss2) ->
    TBbisim i1 i2 RR t1 s1 t2 ss2 ->
    exists t2',
    EvalInstruction (ThreadsBehavior i2)
                    int (t2_, ss2_) a (t2', ss2) /\
    TBbisim i1 i2 RR t1' s1' t2' ss2.
Proof.
  intros TI t2_ ss2_ t2 ss2 STEP BISIM.
  assert (PRE := step_advance_preinteraction TI t2 ss2 BISIM).
  destruct PRE as (t20 & PRE & STEP_PRE).
  assert (POST := step_interaction_interaction TI t2 ss2 BISIM t20 PRE).
  destruct POST as (t2' & POST & STEP_POST).
  eexists t2'.
  split.
  - econstructor.
    + eapply ThreadsAdvanceTrans; eassumption.
    + assumption.
  - revert POST. eapply ThreadsBehavior_ThreadsInteraction_bisim_; eassumption.
Qed.

Lemma ThreadsBehavior_bisim
  : forall t1 s1 t2 s2,
    TBbisim i1 i2 RR t1 s1 t2 s2 ->
    Simulates.Simulates_ (ThreadsBehavior i1) (ThreadsBehavior i2)
              (t1, s1) (t2, s2).
Proof.
  cofix self. intros t1 s1 t2 ss2 BISIM.
  constructor; intros R ii r [t1_ s1_] EVAL.
  destruct EVAL as [[t1' s1'] ADVANCE INTERACTION].
  eapply ThreadsBehavior_ThreadsAdvance_bisim in ADVANCE; [ | eassumption ].
  destruct ADVANCE as (t2' & ss2' & STEP & BISIM').
  eapply ThreadsBehavior_ThreadsInteraction_bisim in INTERACTION; [ | eassumption | eassumption ].
  destruct INTERACTION as (t2'' & STEP'' & BISIM'').
  exists (t2'', ss2').
  - assumption.
  - apply self. auto.
Qed.

End INTERACTION.

(** Without "H" *)
Lemma Proper_ThreadsBehavior {M} (i1 i2 : Object M)
  : ImplSimulates i1 i2 ->
    Simulates (ThreadsBehavior i1) (ThreadsBehavior i2).
Proof.
  intros [].
  apply ThreadsBehavior_bisim with (RR0 := RR); [ assumption | ].
  apply TBbisim_base.
  + constructor.
  + assumption.
Qed.

End Forward.
