Require Import
        Coq.Logic.Eqdep
        Coq.Logic.FunctionalExtensionality
        Coq.Program.Equality.

From C4.Core Require Import
  Prod
  Program
  Simulates
  Threads.

Section ProdBehavior.

  Context
    {T : Type}
    {Method : T -> Type -> Type}
    (B : forall t, Spec (Interaction (Method t))).

  Inductive ProdInteraction :
    forall {Ret} t,
      list T ->
      list T ->
      Interaction (Method t) Ret ->
      Interaction (ProdMethod Method) Ret ->
      Prop :=
  | prodSpawn Ret t ts (m : _ Ret) :
      ProdInteraction t ts (t :: ts) (Spawn m) (Spawn (existT _ t m))
  | prodJoin t Ret ts :
      ProdInteraction t (t :: ts) ts (@Join _ Ret) Join
  | prodUp t A ts ts' int pint :
      ProdInteraction t (Ret:=A) ts ts' int pint ->
      ProdInteraction t (t :: ts) (t :: ts') (Up int) (Up pint)
  | prodSkip t t' A ts ts' int pint :
      t <> t' ->
      ProdInteraction t (Ret:=A) ts ts' int pint ->
      ProdInteraction t (t' :: ts) (t' :: ts') int (Up pint).

  Record ProdState :=
    {
      prodTs : list T;
      prodState t : (B t).(State)
    }.

  Definition EvalProdBehavior Ret (pint : Interaction (ProdMethod Method) Ret) (pst : ProdState) (ret : Ret) (pst' : ProdState) :=
    exists t int,
      ProdInteraction t (prodTs pst) (prodTs pst') int pint /\
      (B t).(EvalInstruction) int (prodState pst t) ret (prodState pst' t) /\
      (forall t', t <> t' -> prodState pst t' = prodState pst' t').

  Definition ProdBehavior : Spec (Interaction (ProdMethod Method)) :=
    {|
      State := ProdState;
      EvalInstruction := EvalProdBehavior;
      initialState := {| prodTs := nil ; prodState t := (B t).(initialState) |}
    |}.

End ProdBehavior.

Arguments ProdState : clear implicits.


Definition overrideDep {T A t}
           (dec : forall t' : T, {t = t'} + {t <> t'})
           a f t' :=
  match dec t' with
  | left E => eq_rect t A a t' E
  | right _ => f t'
  end.

Theorem overrideDepEq T A t dec a f :
  a = @overrideDep T A t dec a f t.
Proof.
  unfold overrideDep.
  destruct dec.
  - apply eq_rect_eq.
  - congruence.
Qed.

Theorem overrideDepNEq T A t dec a f t' :
  t <> t' ->
  f t' = @overrideDep T A t dec a f t'.
Proof.
  unfold overrideDep.
  destruct dec.
  - tauto.
  - reflexivity.
Qed.

Theorem overrideDepNop T A t dec f :
  f = @overrideDep T A t dec (f t) f.
Proof.
  extensionality t'.
  unfold overrideDep.
  destruct dec.
  - destruct e.
    reflexivity.
  - reflexivity.
Qed.


Section ProdBehaviorSimulates.

  Context
    {T : Type}
    (Tdec : forall t t' : T, {t = t'} + {t <> t'})
    {Method : T -> Type -> Type}
    (B1 B2 : forall t, Spec (Interaction (Method t))).

  CoFixpoint prodBehaviorSimulates'_ (ts : list T) st1 st2 :
    (forall t, Simulates_ (B1 t) (B2 t) (st1 t) (st2 t)) ->
    Simulates_
      (ProdBehavior B1)
      (ProdBehavior B2)
      {| prodTs := ts; prodState := st1 |}
      {| prodTs := ts; prodState := st2 |}.
  Proof.
    intro S.
    split.
    intros A pint a [ts' st1'] (t & int & I & E & N).
    edestruct S as [S'].
    edestruct S' as [st2'].
    - eassumption.
    - exists
        {|
          prodTs := ts';
          prodState := overrideDep (Tdec t) st2' st2
        |}.
      + exists t, int.
        repeat split.
        * assumption.
        * simpl.
          rewrite <- overrideDepEq.
          assumption.
        * intros.
          simpl.
          apply (overrideDepNEq T (fun t => State (B2 t))).
          assumption.
      + apply prodBehaviorSimulates'_.
        intros t'.
        destruct (Tdec t t').
        * rewrite <- e, <- overrideDepEq.
          assumption.
        * rewrite <- N, <- overrideDepNEq; trivial.
  Qed.

  Theorem prodBehaviorSimulates :
    (forall t, Simulates (B1 t) (B2 t)) ->
    Simulates (ProdBehavior B1) (ProdBehavior B2).
  Proof.
    apply prodBehaviorSimulates'_.
  Qed.

End ProdBehaviorSimulates.
