From Coq Require Import
  Relations.

(** Definitions of Program, Spec, and Object *)

CoInductive Program {Instruction} {Ret : Type} : Type :=
| Bind {A} : Instruction A -> (A -> Program) -> Program
| Return : Ret -> Program
| Noop : Program -> Program.

Arguments Program : clear implicits.
Notation "x <- f ; m" := (Bind f (fun x => m)) (at level 80, right associativity).

Record Spec {Instruction : Type -> Type} :=
  { State : Type;
    EvalInstruction {A} : Instruction A -> State -> A -> State -> Prop;
    initialState : State
  }.

Arguments Spec : clear implicits.

Record Object {Method : Type -> Type} :=
  {
    ObjInstruction : Type -> Type;
    objBase : Spec ObjInstruction;
    objProgram : forall Ret, Method Ret -> Program ObjInstruction Ret
  }.

Arguments Object : clear implicits.

Definition object {M} {I} (o: Spec I) impl :=
  @Build_Object M I o impl.

Definition ObjState {M} (impl : Object M) :=
  impl.(objBase).(State).

(**)

Definition UnitInstruction (Ret : Type) := Empty_set.

Definition unitSpec : Spec UnitInstruction :=
  {|
    State := unit;
    EvalInstruction Ret i := match i : Empty_set with end;
    initialState := tt
  |}.

Definition mapSpec {Instruction} spec {Method} (impl : forall Ret, Method Ret -> Instruction Ret) :=
  {|
    State := spec.(State);
    EvalInstruction A m := spec.(EvalInstruction) (impl A m);
    initialState := spec.(initialState)
  |}.

Section EvalProgram.

  Context {Instruction} (spec : Spec Instruction).

  Inductive ProgramStep :
    forall R, Program Instruction R -> spec.(State) ->
         Program Instruction R -> spec.(State) -> Prop :=
  | EBind {R i st} {r : R} {st' B f}:
      spec.(EvalInstruction) i st r st' ->
      ProgramStep B (Bind i f) st (f r) st'
  | ENoop {R} p st:
      ProgramStep R (Noop p) st p st.

  Definition ProgramStep' R
             (s1: Program Instruction R * spec.(State))
             (s2: Program Instruction R * spec.(State)): Prop :=
    ProgramStep R (fst s1) (snd s1) (fst s2) (snd s2).
  
  
  Definition ProgramSteps' R := clos_refl_trans_1n
                                  (Program Instruction R * spec.(State))
                                  (ProgramStep' R).

  Definition ProgramSteps R
             (p1: Program Instruction R) (s1: spec.(State))
             (p2: Program Instruction R) (s2: spec.(State)): Prop :=
    ProgramSteps' R (p1, s1) (p2, s2).

  Definition EvalProgram R p s r s' :=
    ProgramSteps R p s (Return r) s'.

    
  Definition programSpec : Spec (Program Instruction) :=
    {|
      State := spec.(State);
      EvalInstruction := EvalProgram;
      initialState := spec.(initialState)
    |}.

End EvalProgram.

Definition functionSpec {Method} (f : forall Ret, Method Ret -> Ret) :
  Spec Method :=
  Build_Spec Method unit (fun Ret m _ ret _ => f Ret m = ret) tt.
