Set Asymmetric Patterns.
Require Import
        Coq.Sets.Ensembles
        Coq.Lists.List.
From C4.Util Require Import
  LibTactics.
From C4.Core Require Import
  Pair
  Program
  ProgramOps
  Object
  Threads
  Simulates
  SimulatesFacts
  Abortable
  Linearizability.

Section Serializability.

  Inductive Stateful (S : Type) (M : Type -> Type) : Type -> Type :=
  | stateful : forall R, M R -> S -> Stateful S M (R * S)%type.

  Inductive TransMethod {M TLocal} : Type -> Type :=
  | init : TransMethod TLocal
  | abort : TLocal -> TransMethod unit
  | commit : TLocal -> TransMethod (unit + TLocal)
  | liftMethod {R}: TLocal -> M R -> TransMethod (option R * TLocal).

  Arguments TransMethod : clear implicits.

  Definition instrumentBody {M TLocal} R (p : Program M R) :
    TLocal ->
    Program
      (Stateful TLocal (Abortable M))
      (option R * TLocal) :=
    foldProgram
      (stateful TLocal (Abortable M))
      (optionProgram (@abortable M) p).

  Definition instrumentEnd {M TLocal} R
             (rt : option R * TLocal) :
    Program (TransMethod M TLocal) (option R) :=
    match rt with
    | (Some r, t) =>
      ret <- commit t;
        match ret with
        | inl _ =>
          Return (Some r)
        | inr t' =>
          _ <- abort t';
            Return None
        end
    | (None, t) =>
      _ <- abort t;
        Return None
    end.

  Definition instrumentMethod {M TLocal} R
             (m : Stateful TLocal (Abortable M) R) :
    TransMethod M TLocal R :=
    let '@stateful _ _ _ m' t := m in
    let '@abortable _ A' m'' := m' in
    liftMethod t m''.

  Definition instrumentWrap' {M TLocal} R
             (p : Program (Stateful TLocal (Abortable M)) (option R * TLocal)) :
    Program (TransMethod M TLocal) (option R) :=
    bindProgram (mapProgram instrumentMethod p) (instrumentEnd R).

  Definition instrumentWrap {M TLocal} R
             (p : TLocal -> Program (Stateful TLocal (Abortable M)) (option R * TLocal)) :
    Program (TransMethod M TLocal) (option R) :=
    t <- init; instrumentWrap' R (p t).

  Definition instrument' {M TLocal} R (p : Program M R) (t : TLocal) :
    Program (TransMethod M TLocal) (option R) :=
    instrumentWrap' R (instrumentBody R p t).

  Definition instrument {M TLocal} R (p : Program M R) :
    Program (TransMethod M TLocal) (option R) :=
    instrumentWrap R (instrumentBody R p).

  (* To instrument the specProgram of an Object. *)
  Definition instImplProg {M TLocal}
             (instrument :
                forall R,
                  Program M R ->
                  Program (TransMethod M TLocal) (option R)) :
    forall R, (Abortable (Program M)) R -> Program (TransMethod M TLocal) R :=
    
    fun R m' =>
       match m' with
       | abortable R m'' =>
         (instrument R m'')
       end.
  
  (*                 Program M R
     -(instrument)-> Program (TransMethod M TLocal) (option R)
     -(conc)->       Program conc.(ObjInstruction) (option R)
   *)
  Definition instFlatImpl {M TLocal}
             (conc: Object (TransMethod M TLocal))
             (instrument :
                forall R,
                  Program M R ->
                  Program (TransMethod M TLocal) (option R)) :
    Object (Abortable (Program M)) :=
    flattenObject conc _
                          (instImplProg instrument).

  (* [Linearizable] has its arguments in the wrong order, it's confusing. *)
  (* The instrumented specect is linearizable wrt an abortable spec. *)
  Definition StrictSerializable
             {TLocal}
             {Instruction}
             (o: Spec Instruction)
             (conc: Object (TransMethod Instruction TLocal))
             (instrument :
                forall R,
                  Program Instruction R ->
                  Program (TransMethod Instruction TLocal) (option R)) : Prop :=
    Linearizable
      (abortableSpec (programSpec o))
      (instFlatImpl conc instrument).

  (* To inst and flatten a program. *)  
  Definition instFlatProg {M M' TLocal}
             (conc : forall R, TransMethod M TLocal R -> Program M' R)
             (instrument :
                forall R,
                  Program M R ->
                  Program (TransMethod M TLocal) (option R))
             R (p : Program M R):
    Program M' (option R) :=
    flattenProgram conc
                   (instrument R p).

  (* To instrument' and flatten a program. *)
  Definition instFlatProg' {M M' TLocal}
             (conc : forall R, TransMethod M TLocal R -> Program M' R)
             (instrument' :
                forall R,
                  Program M R ->
                  TLocal ->
                  Program (TransMethod M TLocal) (option R))
             R (p : Program M R) (t : TLocal) :
    Program M' (option R) :=
    flattenProgram conc
                   (instrument' R p t).

    
  (* To instrument and flatten the specProgram of an Object. *)
  Definition instFlatImplProg
             {TLocal}{M}{M''}
             (conc : forall R, TransMethod M TLocal R -> Program M'' R)
             (instrument :
                forall R,
                  Program M R ->
                  Program (TransMethod M TLocal) (option R)) :
    forall R, Abortable (Program M) R -> Program M'' R :=

    flattenImplProg conc (instImplProg instrument).

  Definition SelfStrictSerializable
             {TLocal M}
             (conc : Object (TransMethod M TLocal))
             (instrument :
                forall R,
                  Program M R ->
                  Program (TransMethod M TLocal) (option R)) : Prop :=
    Linearizable
      (coarsen (instFlatImpl conc instrument))
      (instFlatImpl conc instrument).
  
End Serializability.

Arguments TransMethod : clear implicits.

(* *)

Lemma invAbort:
  forall M R' (a: Abortable M R'),
  exists R, exists (e: option R = R'),
      match e in _ = R'p return Abortable M R'p -> Prop with
      | eq_refl =>
        fun a => exists m, a = @abortable M R m
      end a.

Proof.
  intros.
  destruct a.
  exists R.
  exists (eq_refl (option R)).
  eexists.
  reflexivity.
Qed.
