From Coq Require Import
  List
  Arith
  Setoid
  Morphisms
  FunctionalExtensionality.

From C4.Util Require LibTactics.

Ltac injpair := LibTactics.injpair.

From C4.Util Require Import Util.
From C4.Core Require Import
  Program
  ProgramOps
  Object.

Set Implicit Arguments.
Set Contextual Implicit.
Unset Strict Implicit.

Implicit Types M N : Type -> Type.
Implicit Types A B R : Type.

Notation "F ~> G" := (forall R, F R -> G R) (at level 100).

(**)

Variant Event (M : Type -> Type) : Type :=
| CallEvent   (R : Type) (m : M R)
| ReturnEvent (R : Type) (r : R)
.

Definition ThreadEvent M : Type := nat * Event M.

Inductive Thread M : Type :=
| MkThread R (p : Program M R)
| CallingThread R X (k : X -> Program M R)
| NoThread
.

Definition Threads M : Type := nat -> Thread M.

Definition noThreads {N} : Threads N := fun _ => NoThread.

Inductive PointStep {X Y A : Type} (step : Y -> A -> A -> Prop)
  (n : X * Y) (ts ts' : X -> A) : Prop :=
| MkListStep :
    step (snd n) (ts (fst n)) (ts' (fst n)) ->
    (forall m, m <> fst n -> ts m = ts' m) ->
    PointStep step n ts ts'.

Variant BiEvent M N : Type :=
| HighCallEvent {R : Type} (m : M R)
| HighReturnEvent {R : Type} (r : R)
| LowCallEvent {X : Type} (n : N X)
| LowReturnEvent {X : Type} (x : X)
| LowSilentEvent
.

Definition ThreadStep {M N : Type -> Type} (prog : M ~> Program N) (e : BiEvent M N)
    (th th' : Thread N) : Prop :=
  match e with
  | HighCallEvent m => NoThread = th /\ MkThread (prog _ m) = th'
  | HighReturnEvent r => MkThread (Return r) = th /\ NoThread = th'
  | LowCallEvent n => exists R (k : _ -> _ R), MkThread (Bind n k) = th /\ CallingThread k = th'
  | LowReturnEvent x => exists R (k : _ -> _ R), CallingThread k = th /\ MkThread (k x) = th'
  | LowSilentEvent => exists R (p : _ R), MkThread (Noop p) = th /\ MkThread p = th'
  end.

Definition BiThreadEvent M N : Type := nat * BiEvent M N.

Definition ThreadsStep M N (prog : forall R, M R -> Program N R)
  : BiThreadEvent M N -> Threads N -> Threads N -> Prop :=
  PointStep (ThreadStep prog).

Definition BiTrace M N : Type := list (BiThreadEvent M N).

Inductive Steps_ {E A} (step : E -> A -> A -> Prop) (a0 : A) : list E -> A -> Prop :=
| StepsNone : Steps_ step a0 nil a0
| StepsMore e es a1 a2 : step e a0 a1 -> Steps_ step a1 es a2 -> Steps_ step a0 (e :: es) a2
.

Definition Steps {E A} (step : E -> A -> A -> Prop) es a0 a1 := Steps_ step a0 es a1.

Definition ThreadsSteps {M N} (prog : forall R, M R -> Program N R)
  : BiTrace M N -> Threads N -> Threads N -> Prop :=
  Steps (ThreadsStep prog).

(* Lists of calls and returns *)
Definition ConcTrace (M : Type -> Type) : Type := list (ThreadEvent M).

Definition projHighEvent {M N} (t : BiThreadEvent M N) : option (ThreadEvent M) :=
  match t with
  | (i, HighCallEvent   m) => Some (i, CallEvent   m)
  | (i, HighReturnEvent r) => Some (i, ReturnEvent r)
  | (_, LowCallEvent    _) => None
  | (_, LowReturnEvent  _) => None
  | (_, LowSilentEvent   ) => None
  end.

Definition projLowEvent {M N} (ne : BiThreadEvent M N) : option (ThreadEvent N) :=
  match ne with
  | (_, HighCallEvent _) => None
  | (_, HighReturnEvent _) => None
  | (i, LowCallEvent n) => Some (i, CallEvent n)
  | (i, LowReturnEvent r) => Some (i, ReturnEvent r)
  | (_, LowSilentEvent) => None
  end.

Notation projHigh := (mapfilter projHighEvent).
Notation projLow := (mapfilter projLowEvent).

Variant IsConcTraceOf_ {M N : Type -> Type}
    (t : ConcTrace M) (t' : ConcTrace N) (o : forall R, M R -> Program N R)
  : Prop :=
| MkIsConcTraceOf_ t0 finalThreads
  : t = projHigh t0 ->
    t' = projLow t0 ->
    ThreadsSteps o t0 noThreads finalThreads ->
    IsConcTraceOf_ t t' o
.

Inductive PendingEvent M : Type :=
| MkPendingEvent (e : Event M)
| NoPendingEvent
.

Definition Pending M : Type := nat -> PendingEvent M.

Definition noPending {M} : Pending M := fun _ => NoPendingEvent.

Inductive AtomicEvent M : Type :=
| MkAtomicEvent (e : Event M)
| AtomicStepEvent R (m : M R) (r : R)
  (* This needs to be explicit because the step in the Pending record is
   * defined separately from the step in State *)
.

Coercion MkAtomicEvent : Event >-> AtomicEvent.

Definition PendingEStep {M} (e : AtomicEvent M) (pre post : PendingEvent M) : Prop :=
  match e with
  | CallEvent m => NoPendingEvent = pre /\ MkPendingEvent (CallEvent m) = post
  | AtomicStepEvent m r => MkPendingEvent (CallEvent m) = pre /\ MkPendingEvent (ReturnEvent r) = post
  | ReturnEvent r => MkPendingEvent (ReturnEvent r) = pre /\ NoPendingEvent = post
  end.

Definition PendingStep {M}
  : nat * AtomicEvent M -> Pending M -> Pending M -> Prop :=
  PointStep PendingEStep.

Definition PendingEvalStep {M} (s : Spec M) (e : AtomicEvent M) : State s -> State s -> Prop :=
  match e with
  | MkAtomicEvent e => eq
  | AtomicStepEvent m r => fun z z' => EvalInstruction s m z r z'
  end.

Definition AtomicTrace M := list (nat * AtomicEvent M).

Definition projAtomicEvent {M} (ne : nat * AtomicEvent M) : option (ThreadEvent M) :=
  match ne with
  | (i, MkAtomicEvent e) => Some (i, e)
  | (_, AtomicStepEvent _ _) => None
  end.

Notation projAtomic := (mapfilter projAtomicEvent).

Notation PendingEvalStep_ s := (fun ne => PendingEvalStep (s := s) (snd ne)).

Inductive IsConcTraceOfSpec_ {M : Type -> Type}
    (t0 : list (nat * AtomicEvent M)) (s : Spec M) (ps ps' : _) (z z' : s.(State))
  : Prop :=
| MkIsConcTraceOfSpec_
  : Steps PendingStep t0 ps ps' ->
    Steps (PendingEvalStep_ s) t0 z z' ->
    IsConcTraceOfSpec_ t0 ps ps' z z'
.
Arguments IsConcTraceOfSpec_ : clear implicits.
Arguments IsConcTraceOfSpec_ {M}.

Inductive IsConcTraceOfSpec {M : Type -> Type}
    (t : ConcTrace M) (s : Spec M)
  : Prop :=
| MkIsConcTraceOfSpec t0 z' ps'
  : Steps PendingStep t0 noPending ps' ->
    Steps (PendingEvalStep_ s) t0 (initialState s) z' ->
    t = projAtomic t0 ->
    IsConcTraceOfSpec t s
.

Definition IsConcTraceOf
  : forall {M : Type -> Type} (t : ConcTrace M) (o : Object M), Prop :=
  fun M t o =>
    exists t',
      IsConcTraceOf_ t t' o.(objProgram) /\
      IsConcTraceOfSpec t' o.(objBase).

Definition ConcTraceIncl
  : forall {M : Type -> Type} (o1 o2 : Object M), Prop :=
  fun M o1 o2 =>
    forall (t : ConcTrace M),
      IsConcTraceOf t o1 -> IsConcTraceOf t o2.

