
From Coq Require Import
  List
  Arith
  Setoid
  Morphisms
  FunctionalExtensionality.

From C4.Util Require LibTactics.

Ltac injpair := LibTactics.injpair.

From C4.Util Require Import Util.
From C4.Core Require Import
  Program
  ProgramOps
  Object
  Pair
  Threads
  Superposition
  Simulates
  BaseSimulates
  HConcSimulates
  Trace.

Set Implicit Arguments.
Set Contextual Implicit.
Unset Strict Implicit.

Theorem Transitive_ConcTraceIncl {M : Type -> Type} (o1 o2 o3 : Object M)
  : ConcTraceIncl o1 o2 -> ConcTraceIncl o2 o3 -> ConcTraceIncl o1 o3.
Proof. unfold ConcTraceIncl; eauto. Qed.

Infix "+'" := PairInstruction (at level 40) : type_scope.

Definition on_snd {A B B'} (f : B -> B') (xy : A * B) : A * B' := (fst xy, f (snd xy)).

Definition lowEvent M N (e : Event N) : BiEvent M N :=
  match e with
  | CallEvent m => LowCallEvent m
  | ReturnEvent r => LowReturnEvent r
  end.

Lemma Steps_nil {E A} (step : E -> A -> A -> Prop) x y
  : Steps step nil x y <-> x = y.
Proof.
  split.
  - inversion 1; auto.
  - intros []; constructor.
Qed.

Lemma Steps_app {E A} (step : E -> A -> A -> Prop) (xs ys : list E) s s' s''
  : Steps step xs s s' -> Steps step ys s' s'' -> Steps step (xs ++ ys) s s''.
Proof.
  unfold Steps.
  induction 1; cbn; auto.
  econstructor; eauto.
Qed.

Lemma IsConcTraceOfSpec_nil M (s : Spec M) ps ps' z z'
  : IsConcTraceOfSpec_ nil s ps ps' z z' ->
    ps = ps' /\ z = z'.
Proof.
  intros [H1 H2].
  split.
  - apply (proj1 Steps_nil H1).
  - apply (proj1 Steps_nil H2).
Qed.

Lemma Steps_cons_inv {E A} (step : E -> A -> A -> Prop) e t x y
  : Steps step (e :: t) x y -> exists z, step e x z /\ Steps step t z y.
Proof.
  inversion 1; subst; eauto.
Qed.

Lemma Steps_app_inv {E A} (step : E -> A -> A -> Prop) (xs ys : list E) s s''
  : Steps step (xs ++ ys) s s'' -> exists s', Steps step xs s s' /\ Steps step ys s' s''.
Proof.
  revert s; induction xs as [ | ? ? IH ]; cbn; intros s H.
  - exists s; split; [ constructor | auto ].
  - apply Steps_cons_inv in H; destruct H as (? & ? & ?).
    edestruct IH as (? & ? & ?); eauto.
    eexists; split; [ econstructor | ]; eauto.
Qed.

Definition highEvent {M N} (e : Event M) : BiEvent M N :=
  match e with
  | CallEvent m => HighCallEvent m
  | ReturnEvent r => HighReturnEvent r
  end.

Lemma projHigh_inv_cons {M N} e tM (btM : BiTrace M N)
  : e :: tM = projHigh btM ->
    exists btM0 btM',
      nil = projHigh btM0 /\
      tM = projHigh btM' /\
      btM = btM0 ++ on_snd highEvent e :: btM'.
Proof.
  induction btM as [ | [? []] ? IH ]; cbn.
  - discriminate.
  - clear IH. injection 1; intros -> ->.
    exists nil, btM. auto.
  - clear IH; injection 1; intros -> ->.
    exists nil, btM. auto.
  - intros H; specialize (IH H).
    destruct IH as (btM0 & btM' & H1 & H2 & H3).
    eexists (_ :: btM0), btM'.
    split; [ | split; [ | rewrite H3; reflexivity ]]; eauto.
  - intros H; specialize (IH H).
    destruct IH as (btM0 & btM' & H1 & H2 & H3).
    eexists (_ :: btM0), btM'.
    split; [ | split; [ | rewrite H3; reflexivity ]]; eauto.
  - intros H; specialize (IH H).
    destruct IH as (btM0 & btM' & H1 & H2 & H3).
    eexists (_ :: btM0), btM'.
    split; [ | split; [ | rewrite H3; reflexivity ]]; eauto.
Qed.

Lemma IsConcTraceOfSpec_cons_inv M (s : Spec M) ne t ps ps' z z'
  : IsConcTraceOfSpec_ (ne :: t) s ps ps' z z' ->
    exists ps_ z_,
      PendingStep ne ps ps_ /\
      PendingEvalStep (s := s) (snd ne) z z_ /\
      IsConcTraceOfSpec_ t s ps_ ps' z_ z'.
Proof.
  intros [H1 H2].
  apply Steps_cons_inv in H1. destruct H1 as (ps_ & ? & ?).
  apply Steps_cons_inv in H2. destruct H2 as (z_ & ? & ?).
  exists ps_, z_.
  split; auto.
  split; auto.
  constructor; auto.
Qed.

Lemma IsConcTraceOfSpec_app_inv M (s : Spec M) t t' ps ps' z z'
  : IsConcTraceOfSpec_ (t ++ t') s ps ps' z z' ->
    exists ps_ z_,
      IsConcTraceOfSpec_ t s ps ps_ z z_ /\
      IsConcTraceOfSpec_ t' s ps_ ps' z_ z'.
Proof.
  intros [H1 H2].
  apply Steps_app_inv in H1. destruct H1 as (ps_ & ? & ?).
  apply Steps_app_inv in H2. destruct H2 as (z_ & ? & ?).
  exists ps_, z_.
  split; auto.
  split; auto.
  constructor; auto.
Qed.

Lemma toObjectTrace_projLow' N ne (t0 : ConcTrace N) (t1 : AtomicTrace N)
  : ne :: t0 = projAtomic t1 ->
    exists t1a ne1 t1b,
      t1 = t1a ++ ne1 :: t1b /\
      Some ne = projAtomicEvent ne1 /\
      nil = projAtomic t1a /\
      t0 = projAtomic t1b.
Proof.
  induction t1 as [ | [n []] t1 IH]; cbn.
  - discriminate.
  - intros H; inversion H; subst; clear H IH.
    exists nil, (n, MkAtomicEvent e), t1. eauto.
  - intros H; specialize (IH H).
    destruct IH as (t1a & ne1 & t1b & H1 & H2 & H3 & H4).
    exists ((n, AtomicStepEvent m r) :: t1a), ne1, t1b.
    cbn; split_conj; eauto. f_equal; auto.
Qed.

