
From Coq Require Import
  List
  Arith
  Setoid
  Morphisms
  FunctionalExtensionality.

From C4.Util Require LibTactics.

Ltac injpair := LibTactics.injpair.

From C4.Util Require Import Util.
From C4.Core Require Import
  Program
  ProgramOps
  Object
  Pair
  Threads
  Superposition
  Simulates
  BaseSimulates
  HConcSimulates
  Trace
  Trace_Basics.

Set Implicit Arguments.
Set Contextual Implicit.
Unset Strict Implicit.

Inductive ObjectEvent M N : Type :=
| OTE_ThreadEvent (_ : BiEvent M N)
| OTE_SpecEvent R (m : N R) (r : R)
.

Definition toBiEvent {M N} (e : ObjectEvent M N) : option (BiEvent M N) :=
  match e with
  | OTE_ThreadEvent e => Some e
  | OTE_SpecEvent _ _ => None
  end.

Definition BiToSpecEvent {M N} (e : BiEvent M N) : option (AtomicEvent N) :=
  match e with
  | HighCallEvent _ => None
  | HighReturnEvent _ => None
  | LowCallEvent m => Some (MkAtomicEvent (CallEvent m))
  | LowReturnEvent m => Some (MkAtomicEvent (ReturnEvent m))
  | LowSilentEvent => None
  end.

Definition toSpecEvent {M N} (e : ObjectEvent M N) : option (AtomicEvent N) :=
  match e with
  | OTE_ThreadEvent e' => BiToSpecEvent e'
  | OTE_SpecEvent m r => Some (AtomicStepEvent m r)
  end.

Definition ObjectTrace M N : Type := list (nat * ObjectEvent M N).

Record ObjectState M (o : Object M) : Type := MkObjectState
  { OS_Threads : Threads (ObjInstruction o)
  ; OS_Pending : Pending (ObjInstruction o)
  ; OS_SpecState : State (objBase o)
  }.

Definition initialObjectState M (o : Object M) : ObjectState o := {|
  OS_Threads := fun _ => NoThread;
  OS_Pending := fun _ => NoPendingEvent;
  OS_SpecState := o.(objBase).(initialState);
|}.

Definition step_option {E A} (step : E -> A -> A -> Prop) (oe : option E) : A -> A -> Prop :=
  match oe with
  | None => eq
  | Some e => step e
  end.

Inductive ObjectThreadsStep {M} {o : Object M}
  : nat * ObjectEvent M (ObjInstruction o) -> ObjectState o -> ObjectState o -> Prop :=
| OTS_ThreadStep n (e : BiEvent M _) os1 os2
  : ThreadsStep (objProgram o) (n, e) (OS_Threads os1) (OS_Threads os2) ->
    step_option (fun e => PendingStep (n, e)) (BiToSpecEvent e) (OS_Pending os1) (OS_Pending os2) ->
    OS_SpecState os1 = OS_SpecState os2 ->
    ObjectThreadsStep (n, OTE_ThreadEvent e) os1 os2
| OTS_SpecStep n R (m : _ R) (r : R) os1 os2
  : OS_Threads os1 = OS_Threads os2 ->
    PendingStep (n, AtomicStepEvent m r) (OS_Pending os1) (OS_Pending os2) ->
    PendingEvalStep (AtomicStepEvent m r) (OS_SpecState os1) (OS_SpecState os2) ->
    ObjectThreadsStep (n, OTE_SpecEvent m r) os1 os2
.

Definition ThreadsState_ (M : Type -> Type) (o : Object M) : Type :=
  list { x : Type & Program (ObjInstruction o) x }.

Inductive eqInteraction {M} (n : nat)
  : Event M -> list nat -> list nat -> forall {R}, Interaction M R -> R -> Prop :=
| eqInteraction_Up e tid tids tids' R (i : Interaction M R) (r : R)
  : n <> tid ->
    eqInteraction n e tids tids' i r ->
    eqInteraction n e (tid :: tids) (tid :: tids') (Up i) r
| eqInteraction_Spawn tids R (m : M R)
  : eqInteraction n (CallEvent m) tids (n :: tids) (Spawn m) tt
| eqInteraction_Return tids R (r : R)
  : eqInteraction n (ReturnEvent r) (n :: tids) tids Join r
.

Fixpoint eqInteraction' {M} (n : nat) (e : Event M) (tids tids' : list nat) R (int : Interaction M R) {struct int} : R -> Prop :=
  match int, tids, tids' with
  | Up int, i :: tids, i' :: tids'  => fun r => n <> i /\ i = i' /\ eqInteraction' n e tids tids' int r
  | Spawn m, tids, i' :: tids' => fun _ => i' = n /\ tids = tids' /\ e = CallEvent m
  | Join, i :: tids, tids' => fun r => i = n /\ tids = tids' /\ e = ReturnEvent r
  | _, _, _ => fun _ => False
  end.

Lemma eqInteraction_alt {M} (n : nat) (e : Event M) tids tids' R (int : _ R) r
  : eqInteraction n e tids tids' int r -> eqInteraction' n e tids tids' int r.
Proof.
  induction 1; cbn; eauto.
Qed.

Definition SThreadsStep M (o : Object M) R (int : Interaction M R) (r : R) ths ths' : Prop :=
  exists2 ths_ : ThreadsState o,
    ThreadsAdvance o ths ths_ &
    ThreadsInteraction o R int ths_ r ths'.

Definition Thread' N : Type := {x : Type & Program N x}.

Definition Threads' N : Type := list {x : Type & Program N x}.

Inductive threadRel M : Thread M -> PendingEvent M -> Thread' M -> Prop :=
| threadRel_MkThread R (p : Program M R)
  : threadRel (MkThread p) NoPendingEvent (existT _ R p)
| threadRel_CallingThread R X (e : M X) (k : X -> Program M R)
  : threadRel (CallingThread k) (MkPendingEvent (CallEvent e)) (existT _ R (Bind e k))
| threadRel_CalledThread R X (x : X) (k : X -> Program M R)
  : threadRel (CallingThread k) (MkPendingEvent (ReturnEvent x)) (existT _ R (k x))
.

Inductive stateRel M (o : Object M)
    (tids : list nat) (ths : ThreadsState o) (zz : ObjectState o) : Prop :=
| Mk_stateRel
  : Forall2 (fun tid th => threadRel (OS_Threads zz tid) (OS_Pending zz tid) th) tids (fst ths) ->
    NoDup tids ->
    (forall i, ~ In i tids -> NoThread = OS_Threads zz i /\ NoPendingEvent = OS_Pending zz i) ->
    snd ths = OS_SpecState zz ->
    stateRel tids ths zz
.

Definition threadRel2 M (t : Thread M) (u : Thread' M) : Prop :=
  t = MkThread (projT2 u).

Inductive stateRel2 M (o : Object M)
    (tids : list nat) (ths : ThreadsState o) (zz : ObjectState o) : Prop :=
| Mk_stateRel2
  : Forall2 (fun tid th => threadRel2 (OS_Threads zz tid) th) tids (fst ths) ->
    NoDup tids ->
    (forall i, ~ In i tids -> NoThread = OS_Threads zz i) ->
    OS_Pending zz = noPending ->
    snd ths = OS_SpecState zz ->
    stateRel2 tids ths zz
.

Lemma stateRel_init M (o : Object M)
  : stateRel nil (nil, o.(objBase).(initialState)) initialObjectState.
Proof.
  constructor; auto.
  - constructor.
  - constructor.
Qed.

Lemma stateRel_init_unpack M (o : Object M) (zz : ObjectState o)
  : stateRel2 nil (nil, o.(objBase).(initialState)) zz -> zz = initialObjectState.
Proof.
  intros []. destruct zz; cbn in *.
  unfold initialObjectState. f_equal; auto.
  apply functional_extensionality; intros i. symmetry; apply H1; auto.
Qed.

Inductive IsConcTraceOf__ M (o : Object M)
    (t : ConcTrace M) (zz : ObjectState o)
  : Prop :=
| MkIsConcTraceOf__ t0 t1 zz'
  : ThreadsSteps (objProgram o) t0 (OS_Threads zz) (OS_Threads zz') ->
    IsConcTraceOfSpec_ t1 (objBase o)
      (OS_Pending zz) (OS_Pending zz') (OS_SpecState zz) (OS_SpecState zz') ->
    projHigh t0 = t ->
    projAtomic t1 = projLow t0 ->
    IsConcTraceOf__ t zz
.

Lemma IsConcTraceOf_pack M (o : Object M) (t : ConcTrace M)
  : IsConcTraceOf t o ->
    IsConcTraceOf__ (o := o) t initialObjectState.
Proof.
  intros [? [[] []]].
  subst.
  econstructor 1 with
    (zz' := {| OS_Threads := finalThreads ; OS_Pending := ps' ; OS_SpecState := z' |}); cbn.
  all: try eauto.
  constructor; eauto.
Qed.

Lemma IsConcTraceOf_unpack M (o : Object M) (t : ConcTrace M)
  : IsConcTraceOf__ (o := o) t initialObjectState ->
    IsConcTraceOf t o.
Proof.
  intros [? ? ? ? []].
  exists (projLow t0).
  constructor.
  - constructor 1 with (t0 := t0) (finalThreads := OS_Threads zz'); auto.
  - constructor 1 with (t0 := t1) (z' := OS_SpecState zz') (ps' := OS_Pending zz'); auto.
Qed.

Inductive Linearizes_invariant_0 M (o : Object M) (tids : list nat) (ths : ThreadsState o)
  : ConcTrace M -> Prop :=
| Linearizes_invariant_0_nil : Linearizes_invariant_0 tids ths nil
| Linearizes_invariant_0_cons n e t' ths' tids' Ret int (ret : Ret)
  : eqInteraction n e tids tids' int ret ->
    ThreadsInteraction o Ret int ths ret ths' ->
    NoDup tids' ->
    Linearizes_invariant_0 tids' ths' t' ->
    Linearizes_invariant_0 tids ths ((n, e) :: t')
| Linearizes_invariant_0_cons_silent t ths'
  : ThreadsAdvance o ths ths' ->
    Linearizes_invariant_0 tids ths' t ->
    Linearizes_invariant_0 tids ths t
.

Inductive Linearizes_invariant_1 M (o : Object M) (tids : list nat) (ths : ThreadsState o)
  : ConcTrace M -> Prop :=
| Linearizes_invariant_1_nil : Linearizes_invariant_1 tids ths nil
| Linearizes_invariant_1_cons n e t' ths' tids' Ret int (ret : Ret)
  : eqInteraction n e tids tids' int ret ->
    SThreadsStep int ret ths ths' ->
    NoDup tids' ->
    Linearizes_invariant_1 tids' ths' t' ->
    Linearizes_invariant_1 tids ths ((n, e) :: t')
.

Lemma Linearizes_invariant_0_to_1 M (o : Object M) (tids : list nat) (ths : ThreadsState o)
    (t : ConcTrace M)
  : Linearizes_invariant_0 tids ths t -> Linearizes_invariant_1 tids ths t.
Proof.
  induction 1.
  - constructor.
  - econstructor; eauto.
    econstructor; eauto.
  - destruct IHLinearizes_invariant_0.
    + constructor.
    + econstructor; eauto.
      destruct H2.
      econstructor; [ | eassumption ].
      eapply ThreadsFacts.ThreadsAdvanceTrans; eauto.
Qed.

Ltac SR_hyp :=
  match goal with
  | [ SR : stateRel _ _ _ |- stateRel _ _ _ ] => destruct SR; constructor; cbn; auto
  end.

Lemma stateRel_Spawn M (o : Object M) (tids : list nat) (ths : ThreadsState o) (zz : ObjectState o)
    n R p thzz' (zz' := MkObjectState thzz' (OS_Pending zz) (OS_SpecState zz))
  : NoThread = OS_Threads zz n ->
    MkThread p = thzz' n ->
    (forall m : nat, m <> n -> OS_Threads zz m = thzz' m) ->
    stateRel tids ths zz -> stateRel (n :: tids) (existT _ R p :: fst ths, snd ths) zz'.
Proof.
  intros E E' CONSTANT [].
  assert (~ In n tids).
  { intros IN. clear H0 H1. induction H; cbn in IN; auto.
    destruct IN.
    { subst. destruct H; discriminate. }
    apply IHForall2; auto.
  }
  constructor; cbn; auto.
  { constructor; auto.
    { rewrite <- E'. apply H1 in H3. rewrite <- (proj2 H3). constructor. }
    { revert H. apply Forall2_impl_.
      intros. rewrite <- CONSTANT; [ auto | congruence ]. }
  }
  { constructor; auto. }
  { intros; rewrite <- CONSTANT; [ apply H1; auto | auto ]. }
Qed.

Fixpoint ups (n : nat) {M R} (i : Interaction M R) : Interaction M R :=
  match n with
  | O => i
  | S n => Up (ups n i)
  end.

Lemma threadRel_MkThread_inv {M} R (p : Program M R) ps t
  : threadRel (MkThread p) ps t -> t = existT _ _ p.
Proof.
  inversion 1; auto.
Qed.

Lemma threadRel_CallingThread_inv {M} R S (k : R -> Program M S) (r : R) t
  : threadRel (CallingThread k) (MkPendingEvent (ReturnEvent r)) t -> t = existT _ _ (k r).
Proof.
  inversion 1; subst. injpair. auto.
Qed.

Lemma stateRel_Join M (o : Object M) (tids : list nat) (ths : ThreadsState o) (zz : ObjectState o)
    n R (r : R) thzz' (zz' := MkObjectState thzz' (OS_Pending zz) (OS_SpecState zz))
  : MkThread (Return r) = OS_Threads zz n ->
    NoThread = thzz' n ->
    (forall m : nat, m <> n -> OS_Threads zz m = thzz' m) ->
    stateRel tids ths zz ->
    exists int tids' ths',
      stateRel tids' ths' zz' /\
      eqInteraction n (ReturnEvent r) tids tids' int r /\
      ThreadsInteraction o R int ths r ths'.
Proof.
  intros E E' CONSTANT [].
  assert (W : exists tids1 tids2, tids = tids1 ++ n :: tids2).
  { destruct (in_dec eq_dec n tids).
    - apply in_split; eauto.
    - exfalso; apply H1 in n0; destruct n0; congruence.
  }
  destruct W as (tids1 & tids2 & Etids).
  destruct ths as [th s].
  exists (ups (length tids1) Join), (tids1 ++ tids2), (del_nth (length tids1) th, s).
  split_conj.
  - cbn in H2; subst; constructor.
    + revert H; cbn. eapply Forall2_del.
      intros i ? ?; rewrite CONSTANT; auto.
      intros ->. eauto using Distinct_not_In.
    + eauto using Distinct_del.
    + intros i NINi.
      destruct (eq_dec n i).
      * subst i. split; auto.
        cbn in H; apply Forall2_app_inv_l in H.
        destruct H as (? & ? & H & H' & HE).
        inversion H'; subst.
        subst zz'; cbn. destruct H4; discriminate + auto.
      * rewrite <- CONSTANT; [ eapply H1 | auto ].
        intros II; apply NINi. apply in_app_or in II. apply in_or_app.
        destruct II as [ | [ | ]]; eauto; contradiction.
    + cbn. auto.
  - subst tids; revert H0; clear; induction tids1 as [ | ? ? IH ]; cbn; constructor; auto.
    + intros ->. inversion H0. apply H2. apply in_or_app; right; cbn; auto.
    + apply IH. inversion H0; auto.
  - cbn in H, H2; subst tids s. clear H1 H0. revert th H. induction tids1 as [ | tid tids1 IH]; cbn; intros.
    + inversion H; cbn; constructor.
      rewrite <- E in H2. apply threadRel_MkThread_inv in H2. subst. constructor.
    + inversion H; cbn.
      specialize (IH l'); forward IH; eauto. inversion IH; subst.
      constructor. constructor. auto.
Qed.

Lemma stateRel_Noop M (o : Object M) (tids : list nat) (ths : ThreadsState o) (zz : ObjectState o)
    n R (p : Program (ObjInstruction o) R)
    thzz' (zz' := MkObjectState thzz' (OS_Pending zz) (OS_SpecState zz))
  : MkThread (Noop p) = OS_Threads zz n ->
    MkThread p = thzz' n ->
    (forall m : nat, m <> n -> OS_Threads zz m = thzz' m) ->
    stateRel tids ths zz ->
    exists ths',
      stateRel tids ths' zz' /\
      ThreadsAdvance o ths ths'.
Proof.
  intros E E' CONSTANT [].
  assert (W : exists tids1 tids2, tids = tids1 ++ n :: tids2).
  { destruct (in_dec eq_dec n tids).
    - apply in_split; eauto.
    - exfalso; apply H1 in n0; destruct n0; congruence.
  }
  destruct W as (tids1 & tids2 & Etids).
  destruct ths as [th s]; cbn in H.
  assert (W : exists th1 th2, th = th1 ++ existT _ R (Noop p) :: th2 /\ length th1 = length tids1).
  { subst tids. revert E H. clear.
    revert th; induction tids1 as [ | tid tids1 IH ]; cbn; intros; inversion H; subst.
    - exists nil, l'; cbn. split; auto. f_equal.
      destruct H2; try discriminate.
      inversion E; auto.
    - specialize (IH l' E H4).
      destruct IH as (th1 & th2 & IH & Hlength).
      exists (y :: th1), th2; split; (cbn; f_equal; auto).
  }
  destruct W as (th1 & th2 & -> & LENth).
  exists (th1 ++ existT _ R p :: th2, s).
  split_conj.
  - cbn in H2; subst; constructor.
    + revert H; cbn. eapply Forall2_update; try eassumption; eauto.
      { rewrite <- E, <- E'. inversion 1; constructor. }
      { intros i ? ?; rewrite CONSTANT; auto. }
    + assumption.
    + intros i NINi. specialize (H1 i NINi).
      rewrite <- CONSTANT; auto.
      intros ->; apply NINi. apply in_or_app; right; cbn; auto.
    + cbn. auto.
  - clear. apply ThreadsFacts.ThreadAdvanceImplThreadsAdvance.
    induction th1; cbn; constructor; auto.
Qed.

Lemma threadRel_MkPendingEvent_inv {M} (th : Thread M) e th'
  : threadRel th (MkPendingEvent e) th' ->
    match e with
    | CallEvent e => exists Y (k : _ -> _ Y), th = CallingThread k /\ th' = existT _ Y (Bind e k)
    | ReturnEvent r => exists Y (k : _ -> _ Y), th' = existT _ Y (k r)
    end.
Proof.
  refine (fun H =>
    match H in threadRel th p th' return
      match p return Prop with
      | NoPendingEvent => True
      | MkPendingEvent (CallEvent e) => _
      | MkPendingEvent (ReturnEvent r) => _
      end
    with
    | threadRel_CallingThread => _
    | threadRel_CalledThread => _
    | _ => I
    end); eauto.
Qed.

Lemma Linearizes_FromConcTrace_SilentSteps_ (M : Type -> Type) (o : Object M) (t0 : ConcTrace M)
    th (zz' : ObjectState o) (tids : list nat)
    (ps_ : Pending (ObjInstruction o))
    (z_ : State (objBase o))
    (zz_ := MkObjectState th ps_ z_)
    (t1a : list (nat * AtomicEvent (ObjInstruction o)))
  : forall (ths : ThreadsState o) ps z
      (SR : stateRel tids ths (MkObjectState th ps z))
      (STEPS_ : IsConcTraceOfSpec_ t1a (objBase o) ps ps_ z z_)
      (Ea : nil = projAtomic t1a)
  , exists ths_, stateRel tids ths_ zz_ /\ ThreadsAdvance o ths ths_.
Proof.
  induction t1a as [ | [n e] t1a IH]; intros ths ps z SR STEPS_ Ea.
  { exists ths. apply IsConcTraceOfSpec_nil in STEPS_. destruct STEPS_ as [<- <-].
    split; auto. }
  apply IsConcTraceOfSpec_cons_inv in STEPS_.
  destruct STEPS_ as (ps0 & z0 & STEPps & STEPz & STEPS_).
  destruct e; cbn in STEPz, Ea.
  { discriminate. }
  destruct STEPps as [[Eps Eps_] CONSTANT].
  destruct SR; cbn in *.
  assert (W : exists tids1 tids2, tids = tids1 ++ n :: tids2).
  { destruct (in_dec eq_dec n tids).
    - apply in_split; eauto.
    - exfalso; apply H1 in n0; destruct n0. congruence.
  }
  destruct W as (tids1 & tids2 & Etids).
  destruct ths as [th0 s]; cbn in H, H2; subst s.
  assert (W : exists X (k : _ -> _ X) th1 th2,
    th0 = th1 ++ existT _ X (Bind m k) :: th2 /\ length th1 = length tids1 /\ th n = CallingThread k).
  { subst tids. revert Eps H. clear.
    revert th0; induction tids1 as [ | tid tids1 IH ]; cbn; intros; inversion H; subst.
    - rewrite <- Eps in H2. apply threadRel_MkPendingEvent_inv in H2.
      destruct H2 as (Y & k & ? & ->).
      exists Y, k, nil, l'; cbn; auto.
    - specialize (IH l' Eps H4).
      destruct IH as (X & k & th1 & th2 & IH & Hlength & Eth).
      exists X, k, (y :: th1), th2; cbn; split_conj; f_equal; auto.
  }
  destruct W as (X & k & th1 & th2 & -> & LENth & Eth).
  specialize (IH (th1 ++ existT _ _ (k r) :: th2, z0) ps0 z0).
  forward IH.
  { subst tids; constructor; cbn in *; auto.
    - revert H. apply Forall2_update; eauto.
      + intros TR. rewrite <- Eps_, Eth.
        constructor.
      + intros; rewrite <- CONSTANT; auto.
    - intros; rewrite <- CONSTANT; [ apply H1 | ]; auto.
      intros ->; apply H2. apply in_or_app; right; cbn; auto.
  }
  destruct IH as (ths_ & SR_ & ADV_); auto.
  exists ths_; split; [ apply SR_ | ].
  econstructor; [ | eauto ].
  { revert STEPz. clear; induction th1; cbn; intros; constructor; auto. }
Qed.

Lemma Linearizes_FromConcTrace_SilentSteps (M : Type -> Type) (o : Object M) (t0 : ConcTrace M)
    th (zz' : ObjectState o) (tids : list nat) (ths : ThreadsState o)
    (ps_ : Pending (ObjInstruction o))
    (z_ : State (objBase o))
    (zz_ := MkObjectState th ps_ z_)
    (IH0 : forall (ths_ : ThreadsState o),
          stateRel tids ths_ zz_ ->
          Linearizes_invariant_0 (o:=o) tids ths_ t0)
    (t1a : list (nat * AtomicEvent (ObjInstruction o)))
    zz
    (Eth : th = OS_Threads zz)
    (SR : stateRel tids ths zz)
    (STEPS_ : IsConcTraceOfSpec_ t1a (objBase o) (OS_Pending zz) ps_ (OS_SpecState zz) z_)
    (Ea : nil = projAtomic t1a)
  : Linearizes_invariant_0 (o:=o) tids ths t0.
Proof.
  destruct zz as [th1 ps z]; cbn in *.
  eapply Linearizes_FromConcTrace_SilentSteps_ in STEPS_; try eassumption.
  destruct STEPS_ as (ths_ & SR_ & ADV_).
  eapply Linearizes_invariant_0_cons_silent; eauto.
  eapply IH0. subst zz_ th; auto.
Qed.

Lemma Linearizes_FromConcTrace_LowEvent (M : Type -> Type) (o : Object M) (t0 : ConcTrace M)
    (zz zz' : ObjectState o) (tids : list nat) (ths : ThreadsState o)
    (SR : stateRel tids ths zz)
    (e : Event _)
    (n : nat)
    (thzz_ : Threads (ObjInstruction o))
    ps1 z1
    (zz_ := MkObjectState thzz_ ps1 z1)
    (IH : forall (ths_ : ThreadsState o),
          stateRel tids ths_ zz_ ->
          Linearizes_invariant_0 tids ths_ t0)
    (STEP : ThreadStep (M:=M) (objProgram o) (lowEvent e) (OS_Threads zz n) (thzz_ n))
    (CONSTANT : forall i, i <> n -> OS_Threads zz i = thzz_ i)
    (STEP_ps : PendingStep (n, MkAtomicEvent e) (OS_Pending zz) ps1)
    (STEP_z : PendingEvalStep (s:=objBase o) (MkAtomicEvent e) (OS_SpecState zz) z1)
  : Linearizes_invariant_0 tids ths t0.
Proof.
  cbn in STEP_z; subst z1.
  apply IH.
  destruct SR as [SR0 SR1 SR2 SR3].
  destruct STEP_ps as [STEP_ps C_ps].
  constructor; cbn; eauto.
  { revert SR0. apply Forall2_impl_.
    intros.
    destruct (eq_dec x n) as [ -> | NE ].
    - destruct e; destruct STEP_ps as [Eps Eps_];
        destruct STEP as (? & ? & Ethn & Ethn_); cbn in *;
        rewrite <- Ethn_, <- Eps_;
        rewrite <- Ethn, <- Eps in H0.
      + apply threadRel_MkThread_inv in H0; subst y; constructor.
      + apply threadRel_CallingThread_inv in H0; subst y; constructor.
    - rewrite <- CONSTANT, <- C_ps; auto.
  }
  { intros. assert (i <> n).
    { intros ->. apply SR2 in H; destruct H.
      destruct e; cbn in *.
      - my_decompose STEP; congruence.
      - destruct STEP_ps; congruence.
    }
    - rewrite <- CONSTANT, <- C_ps; auto.
  }
Qed.

Lemma Linearizes_FromConcTrace_ M (o : Object M)
    (t : ConcTrace M) (zz : ObjectState o) tids (ths : ThreadsState o)
  : stateRel tids ths zz ->
    IsConcTraceOf__ t zz ->
    Linearizes_invariant_0 tids ths t.
Proof.
  intros SR [t0 tLow zz' STEPS STEPS_Spec PROJHIGH PROJLOW].
  apply (@proj2 (NoDup tids)).
  revert zz zz' STEPS tids ths SR t tLow STEPS_Spec PROJHIGH PROJLOW.
  induction t0 as [ | [n e] t0 IH ]; intros; (split; [ apply SR | ]).
  { subst t; constructor. }
  apply Steps_cons_inv in STEPS; destruct STEPS as (thzz_ & STEP & STEPS).
  destruct e; destruct STEP as [STEP CONSTANT]; cbn in STEP, CONSTANT, PROJHIGH, PROJLOW.
  - (* HighCall *) destruct STEP as [Ethn Ethn_].
    specialize (IH (MkObjectState thzz_ (OS_Pending zz) (OS_SpecState zz)) zz' STEPS); cbn in IH.
    specialize (IH (n :: tids) (existT _ R (objProgram o _ m) :: fst ths, snd ths)).
    forward IH. { apply stateRel_Spawn; auto. }
    specialize (IH (projHigh t0) tLow STEPS_Spec eq_refl PROJLOW); destruct IH as [IH1 IH2].
    subst t; eapply Linearizes_invariant_0_cons; try eassumption.
    { econstructor. }
    { destruct ths; cbn; constructor; constructor. }
  - (* HighReturn *) destruct STEP as [Ethn Ethn_].
    specialize (IH (MkObjectState thzz_ (OS_Pending zz) (OS_SpecState zz)) zz' STEPS); cbn in IH.
    destruct (stateRel_Join Ethn Ethn_ CONSTANT SR) as (int & tids' & ths' & SR' & Eint & STEP0).
    specialize (IH tids' ths' SR' (projHigh t0) tLow STEPS_Spec eq_refl PROJLOW).
    destruct IH as [IH1 IH2].
    subst t; eapply Linearizes_invariant_0_cons; eauto.
  - (* LowCall *) destruct STEP as (R & k & Ethn & Ethn_).
    symmetry in PROJLOW; apply toObjectTrace_projLow' in PROJLOW.
    destruct PROJLOW as (t1a & [n' []] & t1b & EtLow & Ee & Ea & Eb); inversion Ee; subst.
    apply IsConcTraceOfSpec_app_inv in STEPS_Spec.
    destruct STEPS_Spec as (ps_ & z_ & STEPS_ & STEPS_Spec).
    apply IsConcTraceOfSpec_cons_inv in STEPS_Spec.
    destruct STEPS_Spec as (ps1 & z1 & STEPS_ps & STEPS_z & STEPS_Spec).
    eapply Linearizes_FromConcTrace_SilentSteps; try eassumption + reflexivity.
    intros ths_ SR_.
    eapply Linearizes_FromConcTrace_LowEvent; try eassumption; cbn; eauto.
    cbn; intros ths1 SR1.
    apply (IH (MkObjectState _ _ _) zz' STEPS tids ths1 SR1 (projHigh t0) t1b STEPS_Spec); eauto.
  - (* LowReturn *) destruct STEP as (R & k & Ethn & Ethn_).
    symmetry in PROJLOW; apply toObjectTrace_projLow' in PROJLOW.
    destruct PROJLOW as (t1a & [n' []] & t1b & EtLow & Ee & Ea & Eb); inversion Ee; subst.
    apply IsConcTraceOfSpec_app_inv in STEPS_Spec.
    destruct STEPS_Spec as (ps_ & z_ & STEPS_ & STEPS_Spec).
    apply IsConcTraceOfSpec_cons_inv in STEPS_Spec.
    destruct STEPS_Spec as (ps1 & z1 & STEPS_ps & STEPS_z & STEPS_Spec).
    eapply Linearizes_FromConcTrace_SilentSteps; try eassumption + reflexivity.
    intros ths_ SR_.
    eapply Linearizes_FromConcTrace_LowEvent; try eassumption; cbn; eauto.
    cbn; intros ths1 SR1.
    eapply IH; try eassumption; eauto.
  - (* LowSilent *) destruct STEP as (R & p & Ethn & Ethn_).
    specialize (IH (MkObjectState thzz_ (OS_Pending zz) (OS_SpecState zz)) zz' STEPS); cbn in IH.
    destruct (stateRel_Noop Ethn Ethn_ CONSTANT SR) as (ths' & SR' & STEP0).
    specialize (IH tids ths' SR' (projHigh t0) tLow STEPS_Spec eq_refl PROJLOW).
    destruct IH as [IH1 IH2].
    subst t; eapply Linearizes_invariant_0_cons_silent; eauto.
Qed.

Lemma Linearizes_FromConcTrace M (o : Object M) (t : ConcTrace M)
  : IsConcTraceOf t o ->
    Linearizes_invariant_0 nil (nil, o.(objBase).(initialState)) t.
Proof.
  intros TRACE; eapply Linearizes_FromConcTrace_.
  - apply stateRel_init.
  - apply IsConcTraceOf_pack, TRACE.
Qed.

Definition superP {A} (step : A -> A -> Prop) : (A -> Prop) -> (A -> Prop) -> Prop :=
  fun sp sp' =>
    (exists y, sp' y) /\
    (forall s', sp' s' -> exists s, sp s /\ step s s').

Inductive Linearizes_invariant_2 M (o : Object M) (tids : list nat)
    (thsP : ThreadsState o -> Prop)
  : ConcTrace M -> Prop :=
| Linearizes_invariant_2_nil : Linearizes_invariant_2 tids thsP nil
| Linearizes_invariant_2_cons n e t' thsP' tids' Ret int (ret : Ret)
  : eqInteraction n e tids tids' int ret ->
    superP (SThreadsStep int ret) thsP thsP' ->
    NoDup tids' ->
    Linearizes_invariant_2 tids' thsP' t' ->
    Linearizes_invariant_2 tids thsP ((n, e) :: t')
.

Lemma Linearizes_Step M (o1 o2 : Object M) ths1 ths2
  : HSimulates_ (ThreadsBehavior o1) (ThreadsBehavior o2) ths1 ths2 ->
    forall R (int : Interaction M R) (r : R) ths1',
    SThreadsStep int r ths1 ths1' ->
    exists ths2',
      superP (SThreadsStep int r) ths2 ths2' /\
      HSimulates_ (ThreadsBehavior o1) (ThreadsBehavior o2) ths1' ths2'.
Proof.
  intros [SIM] R i r ths1' STEP1.
  apply SIM in STEP1.
  destruct STEP1 as [ths2' STEP2 SIM'].
  exists ths2'; split_conj; auto.
Qed.

Lemma Linearizes_invariant_1_to_2 M (o1 o2 : Object M) (t : ConcTrace M) tids ths1 ths2
  : HSimulates_ (ThreadsBehavior o1) (ThreadsBehavior o2) ths1 ths2 ->
    Linearizes_invariant_1 tids ths1 t ->
    Linearizes_invariant_2 tids ths2 t.
Proof.
  intros SIM INV. revert ths2 SIM; induction INV; intros.
  - constructor; auto.
  - apply (Linearizes_Step SIM) in H0.
    destruct H0 as [ths2' [STEP2 SIM']].
    econstructor 2; eauto.
Qed.

Fixpoint Threads_of_ths M (o : Object M) (tids : list nat)
    (th : list {x : Type & Program (ObjInstruction o) x})
  : Threads (ObjInstruction o) :=
  match tids, th with
  | nil, _ | _, nil => noThreads
  | tid :: tids, existT _ _ p :: th => set tid (MkThread p) (Threads_of_ths tids th)
  end.

Definition ObjectState_of_ths M (o : Object M) (tids : list nat) (ths : ThreadsState o)
  : ObjectState o :=
  MkObjectState (Threads_of_ths tids (fst ths)) noPending (snd ths).

Lemma stateRel_of_ths M (o : Object M) (tids : list nat) (ths : ThreadsState o)
  : length tids = length (fst ths) -> NoDup tids -> stateRel2 tids ths (ObjectState_of_ths tids ths).
Proof.
  destruct ths as [th s].
  intros LL ND; constructor; auto.
  - revert tids LL ND; induction th as [ | [] th IH ]; cbn; intros; destruct tids as [ | tid tids ];
      discriminate + constructor.
    { cbn. rewrite set_eq. constructor. }
    { cbn in IH. eapply Forall2_impl_; [ | eapply IH; auto; inversion ND; auto ].
      cbn; intros. rewrite set_neq; auto.
      intros ->; inversion ND; auto. }
  - revert tids LL ND; cbn; induction th as [ | [] th IH]; cbn; intros; destruct tids as [ | tid tids]; try discriminate.
    { constructor. }
    { cbn. rewrite set_neq.
      { apply IH; auto. { inversion ND; auto. } apply Decidable.not_or in H; apply H. }
      { intros ->; apply H; left; auto. } }
Qed.

Lemma eqInteraction_In M n (e : Event M) tids tids' R (int : Interaction M R) (r : R) i
  : i <> n -> eqInteraction n e tids tids' int r -> (In i tids <-> In i tids').
Proof.
  intros NE; induction 1 as [ * ? ? IH | | ]; cbn.
  - rewrite IH. reflexivity.
  - firstorder.
  - firstorder.
Qed.

Lemma eqInteraction_NoDup M n (e : Event M) tids tids' R (int : Interaction M R) (r : R)
  : ~ In n tids -> eqInteraction n e tids tids' int r -> NoDup tids -> NoDup tids'.
Proof.
  intros IN; induction 1; cbn; intros ND; auto.
  - inversion ND; subst.
    cbn in IN; apply Decidable.not_or in IN; destruct IN as [I1 I2].
    constructor; auto.
    erewrite <- (eqInteraction_In I1 H0); auto.
  - constructor; auto.
  - inversion ND; auto.
Qed.

Lemma eqInteraction_NoDup_inv M n (e : Event M) tids tids' R (int : Interaction M R) (r : R)
  : ~ In n tids -> eqInteraction n e tids tids' int r -> NoDup tids' -> NoDup tids.
Proof.
  intros IN; induction 1; cbn; intros ND; auto.
  - inversion ND; subst.
    cbn in IN; apply Decidable.not_or in IN; destruct IN.
    constructor; auto.
    erewrite (eqInteraction_In (not_eq_sym H) H0); auto.
  - inversion ND; auto.
  - constructor; auto. cbn in IN. apply Decidable.not_or in IN. apply IN.
Qed.

Lemma eqInteraction_Return_In M n tids tids' R' (r' : R') R (int : Interaction M R) (r : R)
  : eqInteraction n (ReturnEvent r') tids tids' int r -> In n tids.
Proof.
  remember (ReturnEvent _) as w.
  induction 1; cbn; eauto; discriminate.
Qed.

Lemma eqInteraction_In_inv M n (e : Event M) tids tids' R (int : Interaction M R) (r : R) i
  : i <> n -> eqInteraction n e tids tids' int r -> In i tids' -> In i tids.
Proof.
  induction 2; cbn; eauto; firstorder.
Qed.

Lemma eqInteraction_Return_In_inv M n tids tids' R' (r' : R') R (int : Interaction M R) (r : R)
  : NoDup tids -> eqInteraction n (ReturnEvent r') tids tids' int r -> ~In n tids'.
Proof.
  remember (ReturnEvent _) as w.
  induction 2; cbn; eauto; try discriminate.
  - intros [-> | ]; firstorder. inversion H; firstorder.
  - inversion H; auto.
Qed.

Lemma ThreadsInteraction_inv M (o : Object M) Ret (int : Interaction M Ret) p1 r p2
  : ThreadsInteraction' o Ret int p1 r p2 ->
    match int in Interaction _ Ret return Ret -> _ with
    | Spawn m => fun _ => p2 = existT _ _ (objProgram o _ m) :: p1
    | @Join _ w => fun r => p1 = existT (Program (ObjInstruction o)) w (Return r) :: p2
    | Up int => fun r => eq_hd p1 p2 /\ ThreadsInteraction' o _ int (tl p1) r (tl p2)
    end r.
Proof.
  intros []; cbn; eauto.
Qed.

Lemma Linearizes_ToConcTrace_Interaction M (o : Object M) (t : ConcTrace M) tids tids2
    (ths1 ths2 : ThreadsState o) (zz2 : ObjectState o)
    (n : nat) (e : Event M) (Ret : Type) (int : Interaction M Ret) (ret : Ret)
  : NoDup tids ->
    eqInteraction n e tids tids2 int ret ->
    ThreadsInteraction o Ret int ths1 ret ths2 ->
    stateRel2 tids2 ths2 zz2 ->
    IsConcTraceOf__ t zz2 ->
    exists zz1 : ObjectState o,
      stateRel2 tids ths1 zz1 /\ IsConcTraceOf__ ((n, e) :: t) zz1.
Proof.
  intros ND EI INT SR T.
  destruct INT as [th1 ret th2 z INT].
  exists (MkObjectState
    (set n (match e with
            | CallEvent m => NoThread
            | ReturnEvent r => MkThread (Return r)
            end) (OS_Threads zz2))
    (OS_Pending zz2)
    (OS_SpecState zz2)).
  destruct SR as [SR0 SR1 SR2 SR3].
  assert (W :
    OS_Threads zz2 n
      = match e with
        | CallEvent m => MkThread (objProgram o _ m)
        | ReturnEvent r => NoThread
        end /\
    OS_Pending zz2 n = NoPendingEvent); [ | destruct W as [Ethn2 Epsn2] ].
  { split.
    - assert (WW : forall R (r : R), e = ReturnEvent r -> NoThread = OS_Threads zz2 n).
      { intros; subst. apply eqInteraction_Return_In_inv in EI; auto. }
      clear H. apply eqInteraction_alt in EI. revert WW EI th1 th2 INT SR0; cbn; clear.
      intros WW. revert tids tids2. induction int; cbn; intros.
      + destruct tids2 as [ | i2 tids2 ]; [ contradiction | ]; my_decompose EI; subst.
        inversion SR0; subst.
        apply ThreadsInteraction_inv in INT; cbn in INT.
        inversion INT; subst.
        apply H1.
      + destruct tids as [ | i1 tids ]; [ contradiction | ]; my_decompose EI; subst.
        symmetry; eapply WW; reflexivity.
      + destruct tids; [ contradiction | ]; destruct tids2; [ contradiction | ];
          my_decompose EI.
        apply ThreadsInteraction_inv in INT; cbn in INT; my_decompose INT.
        eapply IHint; try eassumption.
        inversion SR0; auto.
    - rewrite SR3; auto. }
  split.
  { constructor; cbn; auto.
    - clear H. apply eqInteraction_alt in EI. revert EI ND SR1 th1 th2 INT SR0; cbn; clear.
      revert tids tids2. induction int; cbn; intros.
      + destruct tids2 as [ | i2 tids2 ]; [ contradiction | ]; my_decompose EI; subst.
        inversion SR0; subst.
        apply ThreadsInteraction_inv in INT; cbn in INT.
        inversion INT; subst.
        revert H3; apply Forall2_impl_.
        intros. rewrite set_neq; auto.
        { intros ->; inversion SR1; subst; contradiction. }
      + destruct tids as [ | i1 tids ]; [ contradiction | ]; my_decompose EI; subst.
        apply ThreadsInteraction_inv in INT; cbn in INT; subst.
        constructor.
        { rewrite set_eq; constructor. }
        { revert SR0; apply Forall2_impl_.
          intros. rewrite set_neq; auto.
          { intros ->; inversion ND; subst; contradiction. } }
      + destruct tids; [ contradiction | ]; destruct tids2; [ contradiction | ];
          my_decompose EI.
        apply ThreadsInteraction_inv in INT; cbn in INT; destruct INT.
        destruct th1; [ contradiction | ].
        inversion SR0; subst; cbn in H; subst. constructor.
        { rewrite set_neq; auto. }
        eapply IHint; try eassumption.
        { inversion ND;  auto. }
        { inversion SR1; auto. }
    - intros. destruct (@set_or_not n i) as [[-> SON] | [NE SON]]; rewrite SON.
      + destruct e; auto.
        exfalso; eauto using eqInteraction_Return_In.
      + eauto using eqInteraction_In_inv.
  }
  { destruct T as [t0 t1 zz' *].
    econstructor 1 with (t0 := (n, highEvent e) :: t0) (t1 := t1); cbn; eauto.
    - econstructor; eauto. constructor; auto.
      + rewrite set_eq; destruct e; cbn; auto.
      + intros; rewrite set_neq; auto.
    - destruct e; cbn; f_equal; auto.
    - destruct e; cbn; auto. }
Qed.

Inductive Advance M (o : Object M)
  : {X : Type & Program (ObjInstruction o) X} -> State (objBase o) ->
    {X : Type & Program (ObjInstruction o) X} -> State (objBase o) -> Prop :=
| Advance_Noop s R p : Advance (existT _ R (Noop p)) s (existT _ R p) s
| Advance_Bind s s' S R m (r : S) (k : S -> _ R)
  : EvalInstruction (objBase o) m s r s' ->
    Advance (existT _ R (Bind m k)) s (existT _ R (k r)) s'
.

Lemma ThreadAdvance_inv M (o : Object M) th1 s1 th2 s2
  : ThreadAdvance o (th1, s1) (th2, s2) ->
    exists tha p1 p2 thb,
      th1 = tha ++ p1 :: thb /\
      th2 = tha ++ p2 :: thb /\
      Advance p1 s1 p2 s2.
Proof.
  remember (th1, s1) eqn:E1.
  remember (th2, s2) eqn:E2.
  intros ADV; revert th1 th2 E1 E2.
  induction ADV; cbn; intros; inversion E1; inversion E2; subst.
  - specialize (IHADV _ _ eq_refl eq_refl).
    my_decompose IHADV.
    eexists (_ :: _), _, _, _; cbn.
    split_conj; [ | | eassumption ]; f_equal; eauto.
  - eexists nil, _, _, _; cbn; split_conj; [ | | constructor ]; eauto.
  - eexists nil, _, _, _; cbn; split_conj; [ | | constructor ]; eauto.
Qed.

Ltac auto_step :=
  constructor; cbn; [ rewrite !set_eq; eauto | intros; rewrite !set_neq; auto ].

Local Lemma Forall2_lll {A B} (R : A -> B -> Prop) xs ys y ys'
  : Forall2 R xs (ys ++ y :: ys') -> length ys < length xs.
Proof.
  intros H; apply Forall2_length in H; rewrite H;
  rewrite app_length. apply Nat.lt_add_pos_r. apply Nat.lt_0_succ.
Qed.

Lemma Linearizes_ToConcTrace_Advance M (o : Object M) (t : ConcTrace M) tids
    (ths1 ths2 : ThreadsState o) (zz2 : ObjectState o)
  : ThreadsAdvance o ths1 ths2 ->
    stateRel2 tids ths2 zz2 ->
    IsConcTraceOf__ t zz2 ->
    exists zz1 : ObjectState o,
      stateRel2 tids ths1 zz1 /\ IsConcTraceOf__ t zz1.
Proof.
  intros ADV SR2 T2.
  induction ADV as [ ths1 | ths1 ths2 ths3 ADV12 ADV23 IH ]; [ eauto | ].
  forward IH; [ assumption | ].
  destruct IH as (zz1 & SR1 & T1).
  destruct ths1 as [th1 s1], ths2 as [th2 s2].
  apply ThreadAdvance_inv in ADV12.
  destruct ADV12 as (tha & p1 & p2 & thb & E1 & E2 & ADV12).
  pose (n := nth (length tha) tids 0).
  destruct SR1 as []; cbn in *.
  exists (MkObjectState (set n (MkThread (projT2 p1)) (OS_Threads zz1)) noPending s1).
  split.
  { constructor; cbn; auto.
    - clear ADV23 SR2 H1. subst th1 th2; revert tids n H H0; induction tha; cbn; intros.
      { inversion H; subst. constructor.
        { rewrite set_eq; constructor. }
        { revert H7; apply Forall2_impl_; intros.
          rewrite set_neq; cbn; auto.
          intros ->; inversion H0; subst; contradiction. }
      }
      { inversion H; subst. constructor.
        { rewrite set_neq; auto. inversion H0; subst; cbn.
          eauto using nth_not_In, Forall2_lll. }
        { apply IHtha; auto. inversion H0; auto. } }
    - intros; rewrite set_neq; auto.
      subst th2; eauto using nth_not_In, Forall2_lll.
  }
  assert (W : MkThread (projT2 p2) = OS_Threads zz1 n).
  { subst th2. revert H; clear. revert tids n; induction tha; intros.
    - inversion H; subst. auto.
    - inversion H; subst. cbn in n. apply IHtha. auto.
  }
  destruct ADV12.
  { destruct T1.
    constructor 1 with (t0 := (n, LowSilentEvent) :: t0) (t1 := t1) (zz' := zz'); cbn; auto.
    - econstructor; [ | eassumption ]; auto_step.
    - congruence. }
  { destruct T1.
    constructor 1 with (t0 := (n, LowCallEvent m) :: (n, LowReturnEvent r) :: t0)
      (t1 := (n, MkAtomicEvent (CallEvent m)) :: (n, AtomicStepEvent m r)
          :: (n, MkAtomicEvent (ReturnEvent r)) :: t1) (zz' := zz'); cbn; auto.
    { constructor 2 with (a1 := set n (CallingThread k) (OS_Threads zz1));
        [ | econstructor; [ | eassumption ] ]; auto_step. }
    { rewrite H2 in H6. destruct H6. constructor.
      { constructor 2 with (a1 := set n (MkPendingEvent (CallEvent m)) noPending);
          [ | constructor 2 with (a1 := set n (MkPendingEvent (ReturnEvent r)) noPending);
          [ | econstructor; [ | eassumption ]]]; auto_step. }
      { constructor 2 with (a1 := s);
          [ | constructor 2 with (a1 := s');
          [ | econstructor; [ | eassumption ]]]; cbn; try congruence. }
    }
    { f_equal; congruence. }
  }
Qed.

Fixpoint adjust {M R} (i : Interaction M R) (n : nat) : nat :=
  match i, n with
  | Up i, S n => S (adjust i n)
  | Spawn _, n => S n
  | Join, S n => n
  | _, _ => O (* Unreachable *)
  end.

Lemma SThreadsStep_length M (o : Object M) R (r : R) (int : Interaction M R) t t'
  : SThreadsStep (o:=o) int r t t' -> length (fst t') = adjust int (length (fst t)).
Proof.
  intros [t_ ADV INT].
  transitivity (adjust int (length (fst t_))).
  - clear ADV. destruct INT as [* H]. cbn; induction H; cbn; auto.
  - clear INT.  f_equal. induction ADV.
    + reflexivity.
    + rewrite IHADV. clear ADV IHADV.
      induction H; cbn; auto.
Qed.

Lemma eqInteraction_length M n (e : Event M) tids tids' R i (r : R)
  : eqInteraction n e tids tids' i r ->
    length tids' = adjust i (length tids).
Proof.
  induction 1; cbn; auto.
Qed.

Lemma Linearizes_ToConcTrace_ M (o : Object M) (t : ConcTrace M) tids thsP
  : (exists ths, thsP ths) ->
    (forall ths, thsP ths -> length tids = length (fst ths)) ->
    NoDup tids ->
    Linearizes_invariant_2 (o := o) tids thsP t ->
    exists ths zz,
      thsP ths /\
      stateRel2 tids ths zz /\
      IsConcTraceOf__ t zz.
Proof.
  intros H LL NODUP; induction 1 as [ | ].
  - destruct H as [ths H]. exists ths. eexists.
    split_conj; auto.
    { apply stateRel_of_ths; auto. }
    { econstructor 1 with (t0 := nil) (t1 := nil); eauto; constructor; constructor. }
  - destruct H1 as [H' H1].
    rename IHLinearizes_invariant_2 into IH.
    forward IH. { auto. }
    forward IH.
    { intros tw Htw. apply H1 in Htw. destruct Htw as [tw0 [Htw0 Htw]].
      rewrite (SThreadsStep_length Htw).
      rewrite (eqInteraction_length H0).
      f_equal. apply LL, Htw0. }
    forward IH. { auto. }
    destruct IH as (ths2 & zz2 & Hths2 & SR2 & T2).
    specialize (H1 ths2 Hths2).
    destruct H1 as (ths & Hths & STEPS).
    destruct STEPS as [ths1 ADV INT].
    eapply (Linearizes_ToConcTrace_Interaction (tids := tids)) in INT; try eassumption.
    destruct INT as (zz1 & SR1 & T1).
    eapply (Linearizes_ToConcTrace_Advance) in ADV; try eassumption.
    destruct ADV as (zz & SR & T).
    exists ths, zz. auto.
Qed.

Lemma Linearizes_ToConcTrace M (o : Object M) (t : ConcTrace M)
  : Linearizes_invariant_2 nil (eq (nil, o.(objBase).(initialState))) t ->
    IsConcTraceOf t o.
Proof.
  intros INV. apply Linearizes_ToConcTrace_ in INV; eauto using NoDup_nil.
  2: intros ? <-; reflexivity.
  destruct INV as (_ & zz & [] & SR & TRACE).
  apply stateRel_init_unpack in SR. rewrite SR in TRACE.
  apply IsConcTraceOf_unpack, TRACE.
Qed.

Theorem ConcSimulates_ConcTraceIncl {M : Type -> Type} (o1 o2 : Object M)
  : HConcSimulates o1 o2 -> ConcTraceIncl o1 o2.
Proof.
  intros LIN t TRACE.
  apply Linearizes_FromConcTrace in TRACE.
  apply Linearizes_invariant_0_to_1 in TRACE.
  apply Linearizes_invariant_1_to_2 with (1 := LIN) in TRACE.
  apply Linearizes_ToConcTrace in TRACE.
  apply TRACE.
Qed.

(* Print Assumptions ConcSimulates_ConcTraceIncl. *)
