
From Coq Require Import
  List
  Arith
  Setoid
  Morphisms
  FunctionalExtensionality.

From C4.Util Require LibTactics.

Ltac injpair := LibTactics.injpair.

From C4.Util Require Import Util.
From C4.Core Require Import
  Program
  ProgramOps
  Object
  Pair
  Threads
  Superposition
  Simulates
  BaseSimulates
  HConcSimulates
  Trace
  Trace_Basics.

Set Implicit Arguments.
Set Contextual Implicit.
Unset Strict Implicit.

Definition flattenThread {N P : Type -> Type} (fNP : N ~> Program P) (th : Thread P)
   (thN : Thread N) (thP : Thread P) : Prop :=
  match th with
  | NoThread => NoThread = thN /\ NoThread = thP
  | MkThread p =>
    (exists pN, programEq p (flattenProgram fNP pN) /\ MkThread pN = thN /\ NoThread = thP) \/
    (exists X (k : X -> Program N _) (pP : Program P X),
      programEq p (bindProgram pP (fun x => Noop (flattenProgram fNP (k x)))) /\
      CallingThread k = thN /\ MkThread pP = thP)
  | CallingThread k =>
    exists Y (kN : Y -> Program N _) (kP : _ -> Program P Y),
      (forall x, programEq (k x) (bindProgram (kP x) (fun y => Noop (flattenProgram fNP (kN y))))) /\
      CallingThread kN = thN /\ CallingThread kP = thP
  end.

Definition flattenThread_low {N P} (fNP : N ~> Program P) (th : Thread P)
    (thN : Thread N) (thP : Thread P) : Prop :=
  match thP with
  | NoThread => True (* Uninteresting case *)
  | MkThread pP =>
    exists X (p : Program P X) (k : _ -> Program N X),
      programEq p (bindProgram pP (fun x => Noop (flattenProgram fNP (k x)))) /\
      MkThread p = th /\ CallingThread k = thN
  | CallingThread kP =>
    exists Y (k : _ -> Program P Y) (kN : _ -> Program N Y),
      (forall x, programEq (k x) (bindProgram (kP x) (fun y => Noop (flattenProgram fNP (kN y))))) /\
      CallingThread k = th /\ CallingThread kN = thN
  end.

Definition flattenThread_high {N P} (fNP : N ~> Program P) (th : Thread P)
    (thN : Thread N) (thP : Thread P) : Prop :=
  match thN with
  | NoThread => NoThread = th /\ NoThread = thP
  | MkThread pN =>
    exists (p : Program P _),
      programEq p (flattenProgram fNP pN) /\
      MkThread p = th /\ NoThread = thP
  | CallingThread kN =>
    (exists p (pP : Program P _),
      programEq p (bindProgram pP (fun x => Noop (flattenProgram fNP (kN x)))) /\
      MkThread p = th /\ MkThread pP = thP) \/
    (exists X (k : X -> Program P _) (kP : _ -> Program P _),
      (forall x, programEq (k x) (bindProgram (kP x) (fun y => Noop (flattenProgram fNP (kN y))))) /\
      CallingThread k = th /\ CallingThread kP = thP)
  end.

Definition flattenThreads {N P : Type -> Type} (fNP : N ~> Program P)
    (th : Threads P) (thN : Threads N) (thP : Threads P) : Prop :=
  forall i, flattenThread fNP (th i) (thN i) (thP i).

Definition flattenThreads_low {N P : Type -> Type} (fNP : N ~> Program P)
    (th : Threads P) (thN : Threads N) (thP : Threads P) : Prop :=
  forall i, flattenThread_low fNP (th i) (thN i) (thP i).

Definition flattenThreads_high {N P : Type -> Type} (fNP : N ~> Program P)
    (th : Threads P) (thN : Threads N) (thP : Threads P) : Prop :=
  forall i, flattenThread_high fNP (th i) (thN i) (thP i).

Lemma flattenThreads_to_low {N P : Type -> Type} (fNP : N ~> Program P)
    {th : Threads P} {thN : Threads N} {thP : Threads P}
  : flattenThreads fNP th thN thP -> flattenThreads_low fNP th thN thP.
Proof.
  intros TH i; specialize (TH i); destruct th; cbn in TH;
    my_decompose TH;
    repeat match goal with
    | [ H : _ = _ |- _ ] => rewrite <- H
    end; cbn; eauto 6.
Qed.

Lemma flattenThreads_to_high {N P : Type -> Type} (fNP : N ~> Program P)
    {th : Threads P} {thN : Threads N} {thP : Threads P}
  : flattenThreads fNP th thN thP -> flattenThreads_high fNP th thN thP.
Proof.
  intros TH i; specialize (TH i); destruct th; cbn in TH;
    my_decompose TH;
    repeat match goal with
    | [ H : _ = _ |- _ ] => rewrite <- H
    end; cbn; eauto 6. right; eauto 6.
Qed.

Lemma flattenNoThreads {N P} (fNP : N ~> Program P) : flattenThreads fNP noThreads noThreads noThreads.
Proof.
  intros i; cbn; auto.
Qed.

Lemma flattenThreads_HighCall {M N P : Type -> Type} (fMN : M ~> Program N) (fNP : N ~> Program P)
    th th_ thN thP R (e : M R) n
  : flattenThreads fNP th thN thP ->
    NoThread = th n /\ MkThread (flattenImplProg fNP fMN R e) = th_ n ->
    (forall m : nat, m <> n -> th m = th_ m) ->
    flattenThreads fNP th_ (set n (MkThread (fMN R e)) thN) thP.
Proof.
  intros TH [Ethn Ethn_] CONSTANT i.
  specialize (TH i).
  destruct (@set_or_not n i) as [[<- SON] | [NEQ SON]]; rewrite SON.
  - rewrite <- Ethn_. rewrite <- Ethn in TH. cbn in TH |- *. destruct TH as [EthNn EthPn].
    left; eexists. split_conj; auto. reflexivity.
  - rewrite <- CONSTANT; auto.
Qed.

Definition eqeq {A B : Type} (P : Type -> Type) (e : A = B) : P A -> P B -> Prop :=
  match e with eq_refl => eq end.

Lemma programEq_bind_inv {M X1 X2 R}
    (e1 : M X1) (k1 : X1 -> Program M R) (e2 : M X2) (k2 : X2 -> Program M R)
  : programEq (Bind e1 k1) (Bind e2 k2) ->
    exists (E : X1 = X2), eqeq E e1 e2 /\ forall x1 x2, eqeq E x1 x2 -> programEq (k1 x1) (k2 x2).
Proof.
  refine (fun H =>
    match H in programEq U V return
      match U, V return Prop with
      | Bind e1 k1, Bind e2 k2 => _
      | _, _ => True
      end with
    | BindEq _ _ _ _ _ => _
    | _ => I
    end).
  exists eq_refl; split; [ | intros ? _ [] ]; cbn; auto.
Qed.

Lemma ConcTrace_flatten_inv_ {M N P : Type -> Type}
    (fNP : N ~> Program P) (fMN : M ~> Program N)
    tMP th th'
  : ThreadsSteps (M:=M) (flattenImplProg fNP fMN) tMP th th' ->
    forall thN thP,
    flattenThreads fNP th thN thP ->
    exists tMN tNP thN' thP',
      ThreadsSteps fMN tMN thN thN' /\
      ThreadsSteps fNP tNP thP thP' /\
      projHigh tMP = projHigh tMN /\ projLow tMN = projHigh tNP /\ projLow tMP = projLow tNP.
Proof.
  induction 1 as [ | th e t th_ th' STEP STEPS IH ]; cbn; intros thN thP TH.
  { (* nil trace *) exists nil, nil, thN, thP. split_conj; constructor. }
  (* cons trace *)
  destruct e as [n []].
  destruct STEP as [STEP CONSTANT]; cbn in STEP, CONSTANT.
  - (* High Call *)
    specialize (IH (set n (MkThread (fMN _ m)) thN) thP).
    forward IH. { eauto using flattenThreads_HighCall. }
    destruct IH as (tMN & tNP & thN' & thP' & IH); my_decompose IH.
    exists ((n, HighCallEvent m) :: tMN), tNP, thN', thP'.
    split_conj; cbn; auto; [ | congruence ].
    econstructor; eauto.
    constructor; cbn. split.
    { specialize (TH n); rewrite <- (proj1 STEP) in TH; inversion TH; auto. }
    { rewrite set_eq; auto. }
    { intros j NEQ; rewrite set_neq; auto. }
  - (* High Return *)
    destruct STEP as [[Ethn Ethn_] CONSTANT]; cbn in Ethn, Ethn_, CONSTANT.
    assert (THn := TH n). rewrite <- Ethn in THn; cbn in THn.
    destruct THn as [ (pN & Ep & EpN & EpP) | (X & k & p & Ep & EpN & EpP) ].
    + rewrite (frobProgramId (flattenProgram _ _)) in Ep.
      destruct pN; cbn in Ep; inversion Ep; subst.
      specialize (IH (set n NoThread thN) thP).
      forward IH.
      { intros i. specialize (TH i).
        destruct (@set_or_not n i) as [[<- SON] | [NEQ SON]]; rewrite SON.
        - rewrite <- Ethn_, <- EpP. cbn; auto.
        - rewrite <- CONSTANT; auto. }
      destruct IH as (tMN & tNP & thN' & thP' & IH); my_decompose IH.
      exists ((n, HighReturnEvent r0) :: tMN), tNP, thN', thP'.
      split_conj; cbn; auto; [ | congruence ].
      econstructor; eauto.
      constructor; cbn.
      { split; auto. rewrite set_eq; auto. }
      intros; rewrite set_neq; auto.
    + rewrite (frobProgramId (bindProgram _ _)) in Ep.
      destruct p; cbn in Ep; try solve [ inversion Ep ].
  - (* Low Call *)
    destruct STEP as [STEP CONSTANT]; cbn in STEP, CONSTANT.
    destruct STEP as (R & k & Ethn & Ethn_).
    assert (THn := TH n). rewrite <- Ethn in THn; cbn in THn.
    destruct THn as [ (pN & Ep & EpN & EpP) | (Y & kN & pP & Ep & EpN & EpP) ].
    { rewrite (frobProgramId (flattenProgram _ _)) in Ep.
      destruct pN; cbn in Ep; inversion Ep; subst. }
    rewrite (frobProgramId (bindProgram _ _)) in Ep.
    destruct pP as [ X' e' k' | | ]; cbn in Ep; try solve [ inversion Ep ].
    { apply programEq_bind_inv in Ep. destruct Ep as [<- [<- Ep]].
      specialize (IH thN (set n (CallingThread k') thP)).
      forward IH.
      { intros i; specialize (TH i).
        destruct (@set_or_not n i) as [[<- SON] | [NEQ SON]]; rewrite SON.
        - rewrite <- Ethn_. cbn. exists Y, kN, k'. split; auto.
          intros x; apply Ep; cbn; auto.
        - rewrite <- CONSTANT; auto. }
      destruct IH as (tMN & tNP & thN' & thP' & IH); my_decompose IH.
      exists tMN, ((n, LowCallEvent n0) :: tNP), thN', thP'.
      split_conj; cbn; auto; [ | f_equal; assumption ].
      econstructor; eauto.
      constructor; cbn.
      { rewrite set_eq; eauto. }
      { intros j NEj; rewrite set_neq; auto. }
    }
  - (* Low Return *)
    destruct STEP as [STEP CONSTANT]; cbn in STEP, CONSTANT.
    destruct STEP as (R & k & Ethn & Ethn_).
    assert (THn := TH n). rewrite <- Ethn in THn; cbn in THn.
    destruct THn as (Y & kN & kP & Ek & EkN & EkP).
    specialize (Ek x).
    rewrite (frobProgramId (bindProgram _ _)) in Ek.
    specialize (IH thN (set n (MkThread (kP x)) thP)).
    forward IH.
    { intros i; specialize (TH i).
      destruct (@set_or_not n i) as [[<- SON] | [NEQ SON]]; rewrite SON.
      - rewrite <- Ethn_. cbn. right. exists Y, kN, (kP x); split_conj; auto.
        rewrite (frobProgramId (bindProgram _ _)); auto.
      - rewrite <- CONSTANT; auto. }
    destruct IH as (tMN & tNP & thN' & thP' & IH); my_decompose IH.
    exists tMN, ((n, LowReturnEvent x) :: tNP), thN', thP'.
    split_conj; cbn; auto; [ | f_equal; assumption ].
    econstructor; eauto.
    constructor; cbn.
    { rewrite set_eq; eauto. }
    { intros j NEj; rewrite set_neq; auto. }
  - destruct STEP as [STEP CONSTANT]; cbn in STEP, CONSTANT.
    destruct STEP as (R & p & Ethn & Ethn_).
    assert (THn := TH n); rewrite <- Ethn in THn; cbn in THn.
    destruct THn as [ (pN & Ep & EpN & EpP) | (Y & kN & pP & Ep & EpN & EpP) ].
    { rewrite (frobProgramId (flattenProgram _ _)) in Ep.
      destruct pN as [ X eN kN | pN | ]; cbn in Ep; inversion Ep; subst.
      { rewrite ProgramOpsFacts.bindFlattenProgramBindProgram in H1.
        specialize (IH (set n (CallingThread kN) thN) (set n (MkThread (fNP X eN)) thP)).
        forward IH.
        { intros i; specialize (TH i).
          destruct (@set_or_not n i) as [[<- SON] | [NEQ SON]]; rewrite SON.
          - rewrite <- Ethn_. cbn. right. exists X, kN, (fNP X eN). rewrite set_eq; auto.
          - rewrite <- CONSTANT, set_neq; auto. }
        destruct IH as (tMN & tNP & thN' & thP' & IH); my_decompose IH.
        exists ((n, LowCallEvent eN) :: tMN), ((n, HighCallEvent eN) :: tNP), thN', thP'.
        split_conj; cbn; auto; [ .. | f_equal; assumption ].
        { econstructor; eauto.
          constructor; cbn.
          { rewrite set_eq; eauto. }
          { intros j NEj; rewrite set_neq; auto. } }
        { econstructor; eauto.
          constructor; cbn.
          { rewrite set_eq; auto. }
          { intros j NEj; rewrite set_neq; auto. } }
      }
      { specialize (IH (set n (MkThread pN) thN) thP).
        forward IH.
        { intros i; specialize (TH i).
          destruct (@set_or_not n i) as [[<- SON] | [NEQ SON]]; rewrite SON.
          - rewrite <- Ethn_. cbn. left. eauto.
          - rewrite <- CONSTANT; auto. }
        destruct IH as (tMN & tNP & thN' & thP' & IH); my_decompose IH.
        exists ((n, LowSilentEvent) :: tMN), tNP, thN', thP'.
        split_conj; auto.
        { econstructor; eauto.
          constructor; cbn.
          { rewrite set_eq; eauto. }
          { intros j NEj; rewrite set_neq; auto. } }
      }
    }
    { rewrite (frobProgramId (bindProgram _ _)) in Ep.
      destruct pP as [ | r | pP ]; cbn in Ep; inversion Ep; subst.
      { specialize (IH (set n (MkThread (kN r)) thN) (set n NoThread thP)).
        forward IH.
        { intros i; specialize (TH i).
          destruct (@set_or_not n i) as [[<- SON] | [NEQ SON]]; rewrite SON.
          - rewrite <- Ethn_. cbn. left. exists (kN r). rewrite set_eq; auto.
          - rewrite <- CONSTANT, set_neq; auto. }
        destruct IH as (tMN & tNP & thN' & thP' & IH); my_decompose IH.
        exists ((n, LowReturnEvent r) :: tMN), ((n, HighReturnEvent r) :: tNP), thN', thP'.
        split_conj; cbn; auto; [ .. | f_equal; assumption ].
        { econstructor; eauto. constructor; cbn.
          - rewrite set_eq; eauto.
          - intros j NEj; rewrite set_neq; auto. }
        { econstructor; eauto. constructor; cbn.
          - rewrite set_eq; auto.
          - intros j NEj; rewrite set_neq; auto. }
      }
      { specialize (IH thN (set n (MkThread pP) thP)).
        forward IH.
        { intros i; specialize (TH i).
          destruct (@set_or_not n i) as [[<- SON] | [NEQ SON]]; rewrite SON.
          - rewrite <- Ethn_. cbn. right. eauto 7.
          - rewrite <- CONSTANT; auto. }
        destruct IH as (tMN & tNP & thN' & thP' & IH); my_decompose IH.
        exists tMN, ((n, LowSilentEvent) :: tNP), thN', thP'.
        split_conj; cbn; auto.
        { econstructor; eauto. constructor; cbn.
          - rewrite set_eq; eauto.
          - intros j NEj; rewrite set_neq; auto. }
      }
   }
Qed.

Theorem ConcTrace_flatten_inv {M N P : Type -> Type}
    (fNP : forall R, N R -> Program P R)
    (fMN : forall R, M R -> Program N R)
    tM tP
  : IsConcTraceOf_ tM tP (flattenImplProg fNP fMN) ->
    exists tN,
      IsConcTraceOf_ tM tN fMN /\
      IsConcTraceOf_ tN tP fNP.
Proof.
  intros [].
  eapply ConcTrace_flatten_inv_ in H1.
  { my_decompose H1.
    eexists; split; econstructor; eauto; congruence.
  }
  { apply flattenNoThreads. }
Qed.

Lemma programEq_inv {M R} (p q : Program M R)
  : programEq p q ->
    match q with
    | Bind e k' => exists k, p = Bind e k /\ forall x, programEq (k x) (k' x)
    | Noop q' => exists p', p = Noop p' /\ programEq p' q'
    | Return r => p = Return r
    end.
Proof.
  intros []; eauto.
Qed.

Ltac conclude :=
  split_conj; cbn; auto;
  repeat lazymatch goal with
  | [ |- ThreadsSteps _ (_ :: _) _ _ ] =>
    econstructor; [ | eassumption ];
      try solve [ constructor; cbn; [rewrite set_eq | intros j NEj; rewrite set_neq]; eauto ]
  | [ |- Steps _ (_ :: _) _ _ ] =>
    econstructor; [ | eassumption ];
      try solve [ constructor; cbn; [rewrite set_eq | intros j NEj; rewrite set_neq]; eauto ]
  | [ |- ThreadsSteps _ (_ ++ _) _ _ ] =>
    eapply Steps_app; [ eassumption | ]
  | [ |- context [ mapfilter _ (_ ++ _) ] ] => rewrite mapfilter_app; cbn
  | [ H : nil = ?x |- context [ ?x ] ] => rewrite <- H; cbn
  | [ H : ?x = nil |- context [ ?x ] ] => rewrite H; cbn
  end; cbn; auto; try (f_equal; assumption).

Ltac TH_hyp :=
  lazymatch goal with
  | [ TH : flattenThreads _ ?th ?thN ?thP |- context [ set ?n ] ] =>
    intros i; specialize (TH i);
      destruct (@set_or_not n i) as [[<- SON] | [NEi SON]]; try rewrite !SON;
      [ repeat lazymatch goal with
        | [ E : _ = ?x |- context [ ?x ] ] => rewrite <- E
        | [ E : _ = ?x, TH : flattenThread _ ?x _ _ |- _ ] => rewrite <- E in TH
        | [ E : _ = ?x, TH : flattenThread _ _ ?x _ |- _ ] => rewrite <- E in TH
        | [ E : _ = ?x, TH : flattenThread _ _ _ ?x |- _ ] => rewrite <- E in TH
        end; cbn
      | repeat lazymatch goal with
        | [ CONSTANT : forall j, j <> _ -> _ = ?th j |- context [ ?th ] ] => rewrite <- CONSTANT; cbn; eauto
        end ]
  end.

Lemma ConcTrace_flatten_low {M N P : Type -> Type}
    (fNP : N ~> Program P) (fMN : M ~> Program N)
    tNP thP thP'
  : ThreadsSteps fNP tNP thP thP' ->
    nil = projHigh tNP ->
    forall th thN,
    flattenThreads fNP th thN thP ->
    exists tMP th',
    projHigh tMP = nil /\
    projLow tMP = projLow tNP /\
    ThreadsSteps (M:=M) (flattenImplProg fNP fMN) tMP th th' /\
    flattenThreads fNP th' thN thP'.
Proof.
  induction 1 as [ | thP [n e] tNP thP_ thP' STEP_NP STEPS_NP IH];
    intros PROJ_NP th thN TH.
  { exists nil, th. split_conj; auto; constructor. }
  destruct e; destruct STEP_NP as [STEP_NP CONSTANT]; cbn in STEP_NP, CONSTANT, PROJ_NP;
    try discriminate.
  - (* LowCall *) destruct STEP_NP as (R & k & EthP & EthP_).
    assert (THn := flattenThreads_to_low TH (i := n)); rewrite <- EthP in THn; cbn in THn.
    destruct THn as (X0 & p & kN & Ep & Eth & EthN).
    apply programEq_inv in Ep; cbn in Ep; destruct Ep as (k0 & -> & Ek).
    specialize (IH PROJ_NP (set n (CallingThread k0) th) thN).
    forward IH. { TH_hyp; eauto 6. }
    destruct IH as (tMP & th' & PROJLOW_MP & PROJHIGH_MP & STEPS & TH').
    exists ((n, LowCallEvent n0) :: tMP), th'.
    conclude.
  - (* LowReturn *) destruct STEP_NP as (R & k & EthP & EthP_).
    assert (THn := flattenThreads_to_low TH (i := n)); rewrite <- EthP in THn; cbn in THn.
    destruct THn as (Y & k0 & kN & Ek & Eth & EthN).
    specialize (IH PROJ_NP (set n (MkThread (k0 x)) th) thN).
    forward IH. { TH_hyp. right; eauto 6. }
    destruct IH as (tMP & th' & PROJLOW_MP & PROJHIGH_MP & STEPS & TH').
    exists ((n, LowReturnEvent x) :: tMP), th'.
    conclude.
  - (* LowSilent *) destruct STEP_NP as (R & p & EthP & EthP_).
    assert (THn := flattenThreads_to_low TH (i := n)); rewrite <- EthP in THn; cbn in THn.
    destruct THn as (X & p0 & kN & Ep & Eth & EthN).
    apply programEq_inv in Ep; cbn in Ep; destruct Ep as (p' & -> & Ep).
    specialize (IH PROJ_NP (set n (MkThread p') th) thN).
    forward IH. { TH_hyp. right; eauto 6. }
    destruct IH as (tMP & th' & PROJLOW_MP & PROJHIGH_MP & STEPS & TH').
    exists ((n, LowSilentEvent) :: tMP), th'.
    conclude.
Qed.

Lemma ConcTrace_flatten_ {M N P : Type -> Type}
    (fNP : N ~> Program P) (fMN : M ~> Program N)
    tMN thN thN'
  : ThreadsSteps fMN tMN thN thN' ->
    forall tNP thP thP' th,
    projLow tMN = projHigh tNP ->
    ThreadsSteps fNP tNP thP thP' ->
    flattenThreads fNP th thN thP ->
    exists tMP th',
    projHigh tMP = projHigh tMN /\
    projLow tMP = projLow tNP /\
    ThreadsSteps (M:=M) (flattenImplProg fNP fMN) tMP th th'.
Proof.
  induction 1 as [ | thN [n e] tMN thN_ thN' STEP_MN STEPS_MN IH];
    intros tNP thP thP' th PROJ_MN STEPS_NP TH.
  { (* nil *) eapply ConcTrace_flatten_low in STEPS_NP; eauto; my_decompose STEPS_NP; eauto. }
  (* cons *)
  destruct e; destruct STEP_MN as [STEP_MN CONSTANT]; cbn in STEP_MN, CONSTANT, PROJ_MN.
  - (* HighCall *) destruct STEP_MN as [EthN EthN_].
    assert (THn := flattenThreads_to_high TH (i := n)); rewrite <- EthN in THn; cbn in THn.
    destruct THn as [Eth EthP].
    specialize (IH tNP thP thP' (set n (MkThread (flattenProgram fNP (fMN R m))) th) PROJ_MN STEPS_NP).
    forward IH. { TH_hyp. left; eexists; split_conj; reflexivity. }
    destruct IH as (tMP & th' & PROJHIGH_MP & PROJLOW_MP & STEPS).
    exists ((n, HighCallEvent m) :: tMP), th'.
    conclude.
  - (* HighReturn *) destruct STEP_MN as [EthN EthN_].
    assert (THn := flattenThreads_to_high TH (i := n)); rewrite <- EthN in THn; cbn in THn.
    destruct THn as (p & Ep & Eth & EthP).
    apply programEq_inv in Ep; cbn in Ep; subst p.
    specialize (IH tNP thP thP' (set n NoThread th) PROJ_MN STEPS_NP).
    forward IH. { TH_hyp; eauto. }
    destruct IH as (tMP & th' & PROJHIGH_MP & PROJLOW_MP & STEPS).
    exists ((n, HighReturnEvent r) :: tMP), th'.
    conclude.
  - (* LowCall *) destruct STEP_MN as (R & k & EthN & EthN_).
    apply projHigh_inv_cons in PROJ_MN. destruct PROJ_MN as (tNP0 & tNP' & PROJ_ & PROJ_MN & ->).
    apply Steps_app_inv in STEPS_NP; destruct STEPS_NP as (thP_ & STEPS_ & STEPS_NP).
    apply Steps_cons_inv in STEPS_NP; destruct STEPS_NP as (thP__ & STEP2 & STEPS_NP).
    destruct STEP2 as [STEP2 CONSTANT2]; cbn in STEP2, CONSTANT2; destruct STEP2 as [EthP_ EthP__].
    eapply ConcTrace_flatten_low with (3 := TH) in STEPS_; auto.
    destruct STEPS_ as (tMP_ & th_ & PROJ_MP_ & PROJLOW_MP_ & STEPS_ & TH_).
    clear TH.
    assert (THn := flattenThreads_to_high TH_ (i := n)); rewrite <- EthN in THn; cbn in THn.
    destruct THn as (p & Ep & Eth & EthP).
    apply programEq_inv in Ep; cbn in Ep; destruct Ep as (p' & -> & Ep).
    rewrite ProgramOpsFacts.bindFlattenProgramBindProgram in Ep.
    specialize (IH tNP' thP__ thP' (set n (MkThread p') th_)).
    specialize (IH PROJ_MN STEPS_NP).
    forward IH. { TH_hyp. right; eauto 6. }
    destruct IH as (tMP & th' & PROJHIGH_MP & PROJLOW_MP & STEPS).
    exists (tMP_ ++ (n, LowSilentEvent) :: tMP), th'.
    conclude.
  - (* LowReturn *) destruct STEP_MN as (R & k & EthN & EthN_).
    apply projHigh_inv_cons in PROJ_MN. destruct PROJ_MN as (tNP0 & tNP' & PROJ_ & PROJ_MN & ->).
    apply Steps_app_inv in STEPS_NP; destruct STEPS_NP as (thP_ & STEPS_ & STEPS_NP).
    apply Steps_cons_inv in STEPS_NP; destruct STEPS_NP as (thP__ & STEP2 & STEPS_NP).
    destruct STEP2 as [STEP2 CONSTANT2]; cbn in STEP2, CONSTANT2; destruct STEP2 as [EthP_ EthP__].
    eapply ConcTrace_flatten_low with (3 := TH) in STEPS_; auto.
    destruct STEPS_ as (tMP_ & th_ & PROJ_MP_ & PROJLOW_MP_ & STEPS_ & TH_).
    clear TH.
    assert (THn := flattenThreads_to_high TH_ (i := n)); rewrite <- EthN in THn; cbn in THn.
    destruct THn as [ (p & pP & Ep & Eth_ & EthP_') | THn ]; [ | my_decompose THn; congruence ].
    assert (pP = Return x); [ | subst pP; clear EthP_' ].
    { rewrite <- EthP_' in EthP_. inversion EthP_.
      injpair; auto. }
    apply programEq_inv in Ep; cbn in Ep; destruct Ep as (p' & -> & Ep).
    specialize (IH tNP' thP__ thP' (set n (MkThread p') th_)).
    specialize (IH PROJ_MN STEPS_NP).
    forward IH. { TH_hyp. left; eauto 6. }
    destruct IH as (tMP & th' & PROJHIGH_MP & PROJLOW_MP & STEPS).
    exists (tMP_ ++ (n, LowSilentEvent) :: tMP), th'.
    conclude.
  - (* LowSilent *) destruct STEP_MN as (R & k & EthN & EthN_).
    assert (THn := flattenThreads_to_high TH (i := n)); rewrite <- EthN in THn; cbn in THn.
    destruct THn as  (p & Ep & Eth & EthP).
    apply programEq_inv in Ep; cbn in Ep; destruct Ep as (p' & -> & Ep).
    specialize (IH tNP thP thP' (set n (MkThread p') th)).
    specialize (IH PROJ_MN STEPS_NP).
    forward IH. { TH_hyp. left; eauto 6. }
    destruct IH as (tMP & th' & PROJHIGH_MP & PROJLOW_MP & STEPS).
    exists ((n, LowSilentEvent) :: tMP), th'.
    conclude.
Qed.

Theorem ConcTrace_flatten {M N P : Type -> Type}
    (fNP : forall R, N R -> Program P R)
    (fMN : forall R, M R -> Program N R)
    tM tN tP
  : IsConcTraceOf_ tM tN fMN ->
    IsConcTraceOf_ tN tP fNP ->
    IsConcTraceOf_ tM tP (flattenImplProg fNP fMN).
Proof.
  intros [] [].
  eapply ConcTrace_flatten_ with (1:= H1) in H4;
    try (eassumption + congruence + auto using flattenNoThreads).
  my_decompose H4; econstructor; eauto; congruence.
Qed.

Theorem ConcTrace_flattenObject {M N : Type -> Type}
    (oN : Object N) (fMN : forall R, M R -> Program N R) tM tN
  : IsConcTraceOf_ tM tN fMN ->
    IsConcTraceOf tN oN ->
    IsConcTraceOf tM (flattenObject oN _ fMN).
Proof.
  intros TRACE_M [tN' [TRACE_N TRACE_N']].
  exists tN'.
  split; auto.
  eapply ConcTrace_flatten; eassumption.
Qed.

Theorem ConcTrace_flattenObject_inv {M N : Type -> Type}
    (oN : Object N) (fMN : forall R, M R -> Program N R) tM
  : IsConcTraceOf tM (flattenObject oN _ fMN) ->
    exists tN,
      IsConcTraceOf_ tM tN fMN /\
      IsConcTraceOf tN oN.
Proof.
  intros [t_ [H1 H2]]; cbn in H1, H2.
  apply ConcTrace_flatten_inv in H1.
  destruct H1 as [tN [HN1 HN2]].
  exists tN; split; eauto.
  econstructor; split; eauto.
Qed.

(* Lemma 4.5 *)
Lemma ConcTraceIncl_flatten {M N} (impl : M ~> Program N) (o o' : Object N)
  : ConcTraceIncl o o' -> ConcTraceIncl (flattenObject o _ impl) (flattenObject o' _ impl).
Proof.
  intros H. red. intros t Ht.
  apply ConcTrace_flattenObject_inv in Ht.
  destruct Ht as [tN [Hf Ho]].
  eauto using ConcTrace_flattenObject.
Qed.
