From Coq Require Import
  List
  Arith
  Setoid
  Morphisms
  FunctionalExtensionality.

From C4.Util Require LibTactics.

Ltac injpair := LibTactics.injpair.

From C4.Util Require Import Util.
From C4.Core Require Import
  Program
  ProgramOps
  Object
  Pair
  Threads
  Superposition
  Simulates
  BaseSimulates
  HConcSimulates
  Trace
  Trace_Basics.

Set Implicit Arguments.
Set Contextual Implicit.
Unset Strict Implicit.

Module Import Dir.
Inductive dir := L | R.
End Dir.

Module ODir.
Variant odir : Type := L | Neither | R.
End ODir.

Definition left_TE {M N} (e : ThreadEvent M) : ThreadEvent (M +' N) :=
  match snd e with
  | CallEvent m => (fst e, CallEvent (inl m))
  | ReturnEvent r => (fst e, ReturnEvent r)
  end.

Instance EqDec_dir : EqDec dir.
Proof. refine
  (fun x y =>
    match x, y with
    | L, L | R, R => left eq_refl
    | _, _ => right _
    end); discriminate.
Qed.

Record pairing : Type :=
  { pairing_pair : nat -> option (dir * nat)
  ; pairing_unpair : dir * nat -> option nat
  ; pairing_pair_unpair : forall i j, pairing_pair i = Some j <-> pairing_unpair j = Some i
  }.

Definition pairing_empty : pairing.
Proof. refine
  {| pairing_pair := fun _ => None
  ;  pairing_unpair := fun _ => None
  |}.
  all: split; discriminate.
Defined.

Variant rel := Free | Paired.

Definition fliprel (r : rel) : rel :=
  match r with Free => Paired | Paired => Free end.

Class evalR (r : rel) (d : dir) (p : pairing) (i j : nat) : Prop :=
  evalR_def
    : match r with
      | Paired => pairing_pair p i = Some (d, j)
      | Free => pairing_pair p i = None /\ pairing_unpair p (d, j) = None
      end.

Arguments evalR : simpl never.

Definition is_unpaired_ (d : ODir.odir) (p : pairing) (i : nat) : Prop :=
  match d with
  | ODir.Neither => pairing_pair p i = None
  | ODir.L => pairing_unpair p (L, i) = None
  | ODir.R => pairing_unpair p (R, i) = None
  end.

Arguments is_unpaired_ : simpl never.

Notation is_unpaired := (is_unpaired_ ODir.Neither).
Notation is_unpaired_L := (is_unpaired_ ODir.L).
Notation is_unpaired_R := (is_unpaired_ ODir.R).

Definition switchR r d p i j : evalR r d p i j -> pairing.
Proof. refine
  match r with
  | Free => fun e =>
    {| pairing_pair := set i (Some (d, j)) (pairing_pair p)
    ;  pairing_unpair := set (d, j) (Some i) (pairing_unpair p)
    |}
  | Paired => fun e =>
    {| pairing_pair := set i None (pairing_pair p)
    ;  pairing_unpair := set (d, j) None (pairing_unpair p)
    |}
  end; unfold evalR in e; intros.
  - unfold set, eq_dec at 2, EqDec_pair; cbn [fst snd].
    destruct (eq_dec d _).
    + destruct (eq_dec j _).
      * destruct (eq_dec i _).
        ** subst; split; intros; f_equal; destruct j0; auto.
        ** rewrite pairing_pair_unpair. destruct j0; cbn in *; subst.
           split; intros W; [rewrite W in e; discriminate (proj2 e) | injection W; contradiction ].
      * destruct (eq_dec i _).
        ** rewrite <- pairing_pair_unpair. subst.
           split; intros W; [ injection W; intros X; rewrite <- X in n; contradiction | rewrite W in e; discriminate (proj1 e)].
        ** apply pairing_pair_unpair.
    + destruct (eq_dec i _).
      ** rewrite <- pairing_pair_unpair. subst.
          split; intros W; [ injection W; intros <-; contradiction | rewrite W in e; discriminate (proj1 e)].
      ** apply pairing_pair_unpair.
  - unfold set, eq_dec at 2, EqDec_pair; cbn [fst snd].
    destruct (eq_dec d _).
    + destruct (eq_dec j _).
      * destruct (eq_dec i _).
        ** split; discriminate.
        ** rewrite pairing_pair_unpair. rewrite pairing_pair_unpair in e. destruct j0; cbn in *; subst.
           rewrite e. split; intros W; [injection W; contradiction | discriminate ].
      * destruct (eq_dec i _).
        ** rewrite <- pairing_pair_unpair. subst. rewrite e.
           split; [ discriminate | injection 1; intros ].
           rewrite <- H0 in n; contradiction.
        ** apply pairing_pair_unpair.
    + destruct (eq_dec i _).
      * rewrite <- pairing_pair_unpair, <- e0, e.
        split; [ discriminate | injection 1; intros HH; rewrite <- HH in n; contradiction ].
      * apply pairing_pair_unpair.
Defined.

Arguments switchR : simpl never.

Notation is_free := (evalR Free).
Notation is_paired := (evalR Paired).
Notation add d p i j := (@switchR Free d p i j _).
Notation rm d p i j := (@switchR Paired d p i j _).

Lemma is_free_unpaired_L p i j
  : is_free L p i j <-> (is_unpaired p i /\ is_unpaired_L p j).
Proof. firstorder. Qed.

Lemma is_free_unpaired_R p i j
  : is_free R p i j <-> (is_unpaired p i /\ is_unpaired_R p j).
Proof. firstorder. Qed.

Definition pairing_incl p q : Prop :=
  forall i dj, pairing_pair p i = Some dj -> pairing_pair q i = Some dj.

Lemma incl_paired p q d i j :
  pairing_incl p q -> is_paired d p i j -> is_paired d q i j.
Proof. firstorder. Qed.

Lemma incl_unpaired d p q i :
  pairing_incl p q -> is_unpaired_ d q i -> is_unpaired_ d p i.
Proof.
  destruct d; unfold is_unpaired_; cbn; intros PI PP.
  - destruct (pairing_unpair p _) as [ a | ] eqn:Ea; auto.
    apply pairing_pair_unpair, PI, pairing_pair_unpair in Ea.
    congruence.
  - destruct (pairing_pair p _) as [ a | ] eqn:Ea; auto.
    apply PI in Ea. congruence.
  - destruct (pairing_unpair p _) as [ a | ] eqn:Ea; auto.
    apply pairing_pair_unpair, PI, pairing_pair_unpair in Ea.
    congruence.
Qed.

Inductive pairConcTrace_ {M N : Type -> Type}
    (p : pairing) (tM : ConcTrace M) (tN : ConcTrace N) (tMN : ConcTrace (M +' N)) : Prop :=
| pairConcTrace_nil : tM = nil -> tN = nil -> tMN = nil -> pairConcTrace_ p tM tN tMN
| pairConcTrace_call_L i j (FREE : is_free L p i j) R (m : M R) tM' tMN'
  : tM = (j, CallEvent m) :: tM' ->
    tMN = (i, CallEvent (inl m)) :: tMN' ->
    pairConcTrace_ (add L p i j) tM' tN tMN' ->
    pairConcTrace_ p tM tN tMN
| pairConcTrace_call_R i j (FREE : is_free R p i j) S (m : N S) tN' tMN'
  : tN = (j, CallEvent m) :: tN' ->
    tMN = (i, CallEvent (inr m)) :: tMN' ->
    pairConcTrace_ (add Dir.R p i j) tM tN' tMN' ->
    pairConcTrace_ p tM tN tMN
| pairConcTrace_return_L i j (PAIRED : is_paired L p i j) R (r : R) tM' tMN'
  : tM = (j, ReturnEvent r) :: tM' ->
    tMN = (i, ReturnEvent r) :: tMN' ->
    pairConcTrace_ (rm L p i j) tM' tN tMN' ->
    pairConcTrace_ p tM tN tMN
| pairConcTrace_return_R i j (PAIRED : is_paired R p i j) S (r : S) tN' tMN'
  : tN = (j, ReturnEvent r) :: tN' ->
    tMN = (i, ReturnEvent r) :: tMN' ->
    pairConcTrace_ (rm R p i j) tM tN' tMN' ->
    pairConcTrace_ p tM tN tMN
.

Definition pairConcTrace {M N}
    (tM : ConcTrace M) (tN : ConcTrace N) (tMN : ConcTrace (M +' N)) : Prop :=
  pairConcTrace_ pairing_empty tM tN tMN.

(* The program part of pairObject *)
Definition pairProgram {M1 M2 N1 N2 : Type -> Type}
    (fM : forall R, M1 R -> Program M2 R)
    (fN : forall R, N1 R -> Program N2 R)
  : forall R, PairInstruction M1 N1 R -> Program (PairInstruction M2 N2) R :=
  fun (R : Type) (m : PairInstruction M1 N1 R) =>
    match m with
    | inl mL =>
        ProgramOps.mapProgram (fun A : Type => inl) (fM R mL)
    | inr mR =>
        ProgramOps.mapProgram (fun A : Type => inr) (fN R mR)
    end.

Arguments pairProgram : clear implicits.
Arguments pairProgram {M1 M2 N1 N2} fM fN.

Definition addL {A B} (a : A) (xs : list A * B) : list A * B :=
  (a :: fst xs, snd xs).

Definition addR {A B} (b : B) (xs : A * list B) : A * list B :=
  (fst xs, b :: snd xs).

Fixpoint unzipTrace {M M' N N'} (athread : nat -> ODir.odir)
    (tMN : BiTrace (M +' N) (M' +' N')) : BiTrace M M' * BiTrace N N' :=
  match tMN with
  | nil => (nil, nil)
  | e :: tMN =>
    match snd e, athread (fst e) with
    | HighCallEvent (inl m), _ =>
      addL (fst e, HighCallEvent m) (unzipTrace (set (fst e) ODir.L athread) tMN)
    | HighCallEvent (inr m), _ =>
      addR (fst e, HighCallEvent m) (unzipTrace (set (fst e) ODir.R athread) tMN)
    | HighReturnEvent r, ODir.L =>
      addL (fst e, HighReturnEvent r) (unzipTrace (set (fst e) ODir.Neither athread) tMN)
    | HighReturnEvent r, ODir.R =>
      addR (fst e, HighReturnEvent r) (unzipTrace (set (fst e) ODir.Neither athread) tMN)
    | LowCallEvent (inl m), _ =>
      addL (fst e, LowCallEvent m) (unzipTrace athread tMN)
    | LowCallEvent (inr m), _ =>
      addR (fst e, LowCallEvent m) (unzipTrace athread tMN)
    | LowReturnEvent r, ODir.L =>
      addL (fst e, LowReturnEvent r) (unzipTrace athread tMN)
    | LowReturnEvent r, ODir.R =>
      addR (fst e, LowReturnEvent r) (unzipTrace athread tMN)
    | LowSilentEvent, ODir.L =>
      addL (fst e, LowSilentEvent) (unzipTrace athread tMN)
    | LowSilentEvent, ODir.R =>
      addR (fst e, LowSilentEvent) (unzipTrace athread tMN)
    | _, _ => (nil, nil)
    end
  end.

Definition projLeft_ {M M' N N'} (athread : nat -> ODir.odir)
    (tMN : BiTrace (M +' N) (M' +' N')) : BiTrace M M' :=
  fst (unzipTrace athread tMN).

Definition projLeft {M M' N N'} : BiTrace (M +' N) (M' +' N') -> BiTrace M M' :=
  projLeft_ (fun _ => ODir.Neither).

Definition projRight_ {M M' N N'} (athread : nat -> ODir.odir)
    (tMN : BiTrace (M +' N) (M' +' N')) : BiTrace N N' :=
  snd (unzipTrace athread tMN).

Definition projRight {M M' N N'} : BiTrace (M +' N) (M' +' N') -> BiTrace N N' :=
  projRight_ (fun _ => ODir.Neither).

Require Import Eqdep.

Inductive eq_Thread {M N} (eq_Program : forall X, Program M X -> Program N X -> Prop)
  : Thread M -> Thread N -> Prop :=
| eq_MkThread R (p p' : Program _ R)
  : eq_Program R p p' ->
    eq_Thread eq_Program (MkThread p) (MkThread p')
| eq_CallingThread X R (k k' : X -> Program _ R)
  : (forall x, eq_Program R (k x) (k' x)) ->
    eq_Thread eq_Program (CallingThread k) (CallingThread k')
.

Inductive matchPairThread {M N} (th : Thread (M +' N)) (thM : Thread M) (thN : Thread N)
  : ODir.odir -> Prop :=
| matchPairThread_L
  : NoThread = thN ->
    eq_Thread (fun X p p' => p = mapProgram (fun _ => inl) p') th thM ->
    matchPairThread th thM thN ODir.L
| matchPairThread_R
  : NoThread = thM ->
    eq_Thread (fun X p p' => p = mapProgram (fun _ => inr) p') th thN ->
    matchPairThread th thM thN ODir.R
| matchPairThread_Neither
  : NoThread = th -> NoThread = thM -> NoThread = thN -> matchPairThread th thM thN ODir.Neither
.

Definition matchPairThreads {M N} (th : Threads (M +' N)) (thM : Threads M) (thN : Threads N) (f : nat -> ODir.odir) : Prop :=
  forall n, matchPairThread (th n) (thM n) (thN n) (f n).

Theorem ConcTrace_pairProgram_inv_aux {M M' N N' : Type -> Type}
    (fM : M ~> Program M') (fN : N ~> Program N')
    (tMN : BiTrace (M +' N) (M' +' N'))
    th th_
  : ThreadsSteps (pairProgram fM fN) tMN th th_ ->
    forall (thM : Threads M') (thN : Threads N') (athread : nat -> ODir.odir),
    matchPairThreads th thM thN athread ->
    exists thM' thN',
      ThreadsSteps fM (projLeft_  athread tMN) thM thM' /\
      ThreadsSteps fN (projRight_ athread tMN) thN thN'.
Proof.
  induction 1 as [| th [n e] thMN th1 th_ [STEP CONSTANT] STEPS IH ].
  { (* Base case *) intros; do 2 eexists. split; constructor. }
  (* Step *)
  intros thM thN athread MATCH. cbn in STEP, CONSTANT.
  destruct e; cbn in STEP.
  - (* highcall *)
    destruct STEP as [Eth Eth1].
    assert (NOPE : NoThread = thM n /\ NoThread = thN n).
    { destruct (MATCH n); auto. all: inversion H0; congruence. }
    destruct m as [m | m].
    + (* highcall inl *) cbn in *.
      specialize (IH (set n (MkThread (fM _ m)) thM) thN (set n ODir.L athread)).
      forward IH.
      { intros tid. specialize (MATCH tid).
        destruct (@set_or_not n tid) as [[EQ SON] | [NEQ SON]].
        - subst n. rewrite !SON, <- Eth1.
          constructor.
          + apply NOPE.
          + constructor; auto.
        - rewrite !SON. rewrite <- CONSTANT; auto.
      }
      destruct IH as (thM' & thN' & [STEPS_M STEPS_N]). exists thM', thN'.
      split; [ | auto ].
      econstructor; [ | eassumption ].
      constructor; cbn; [ | apply @set_neq_flip].
      rewrite set_eq.
      rewrite <- (proj1 NOPE).
      constructor; auto.
    + (* highcall inl *) cbn in *.
      specialize (IH thM (set n (MkThread (fN _ m)) thN) (set n ODir.R athread)).
      forward IH.
      { intros tid. specialize (MATCH tid).
        destruct (@set_or_not n tid) as [[EQ SON] | [NEQ SON]].
        - subst n. rewrite !SON, <- Eth1.
          constructor.
          + apply NOPE.
          + constructor; auto.
        - rewrite !SON. rewrite <- CONSTANT; auto.
      }
      destruct IH as (thM' & thN' & [STEPS_M STEPS_N]). exists thM', thN'.
      split; [ auto | ].
      econstructor; [ | eassumption ].
      constructor; cbn; [ | apply @set_neq_flip].
      rewrite set_eq.
      rewrite <- (proj2 NOPE).
      constructor; auto.
  - (* highreturn *)
    destruct STEP as [Eth Eth1].
    specialize (IH
      (match athread n with ODir.L => set n NoThread thM | _ => thM end)
      (match athread n with ODir.R => set n NoThread thN | _ => thN end)
      (set n ODir.Neither athread)
    ).
    forward IH.
    { intros tid; destruct (MATCH n); [ | | congruence ].
      - destruct (@set_or_not n tid) as [[EQ SON] | [NEQ SON]]; rewrite !SON.
        + subst n.
          rewrite <- Eth1, <- H. constructor; auto.
        + rewrite <- CONSTANT; auto.
      - destruct (@set_or_not n tid) as [[EQ SON] | [NEQ SON]]; rewrite !SON.
        + subst n.
          rewrite <- Eth1, <- H. constructor; auto.
        + rewrite <- CONSTANT; auto.
    }
    destruct IH as (thM' & thN' & [STEPS_M STEPS_N]). exists thM', thN'.
    unfold projLeft_, projRight_; cbn.
    destruct (MATCH n); [ | | congruence ]; cbn; split; auto.
    + econstructor; eauto.
      constructor; cbn. 2:{ apply @set_neq_flip. }
      rewrite set_eq.
      destruct H0; try discriminate.
      revert p' H0.
      (* LOL *)
      lazymatch goal with
      | [ |- ?G ] =>
        lazymatch (eval pattern R1, p in G) with
        | ?f _ _ =>
          refine match Eth in _ = ttt return
            match ttt with
            | MkThread p => f _ p
            | _ => False
            end with eq_refl => _ end
        end
      end.
      intros ? HHH.
      apply eq_sym, ProgramOpsFacts.mapProgramEqRet in HHH. rewrite HHH.
      constructor; auto.
    + econstructor; eauto.
      constructor; cbn. 2:{ apply @set_neq_flip. }
      rewrite set_eq.
      destruct H0; try discriminate.
      revert p' H0.
      (* LOL *)
      lazymatch goal with
      | [ |- ?G ] =>
        lazymatch (eval pattern R1, p in G) with
        | ?f _ _ =>
          refine match Eth in _ = ttt return
            match ttt with
            | MkThread p => f _ p
            | _ => False
            end with eq_refl => _ end
        end
      end.
      intros ? HHH.
      apply eq_sym, ProgramOpsFacts.mapProgramEqRet in HHH. rewrite HHH.
      constructor; auto.
  - (* lowcall *)
    destruct STEP as (R' & k & Eth & Eth1).
    destruct n0 as [m | m].
    + assert (III : athread n = ODir.L /\ exists k_,
        k = (fun x => mapProgram (fun _ m => inl m) (k_ x)) /\ thM n = MkThread (Bind m k_)).
      { destruct (MATCH n); try congruence.
        - split; auto. destruct H0; try discriminate.
          revert p' Eth1 H0.
          lazymatch goal with
          | [ |- ?G ] =>
            lazymatch (eval pattern R0, p in G) with
            | ?f _ _ =>
              refine match Eth in _ = ttt return
                match ttt with
                | MkThread p => f _ p
                | _ => False
                end with eq_refl => _ end
            end
          end; clear R0 p Eth.
          intros p' Eth1 H0.
          apply eq_sym, ProgramOpsFacts.mapProgramEqBind in H0.
          destruct H0 as (m' & k' & ? & ? & ?).
          exists k'. split; auto.
          congruence.
        - exfalso. destruct H0; try discriminate.
          subst p. rewrite (frobProgramId (mapProgram _ _)) in Eth.
          destruct p'; cbn in Eth; try discriminate.
      }
      destruct III as [NOPE [k_ [Ek Ek_]]].
      specialize (IH (set n (CallingThread k_) thM) thN athread).
      forward IH.
      { intros tid; specialize (MATCH tid).
        destruct (@set_or_not n tid) as [[En SON]|[NEn SON]].
        - rewrite !SON.
          subst n.
          rewrite NOPE in *.
          constructor.
          + inversion MATCH; auto.
          + rewrite <- Eth1. constructor. rewrite Ek. auto.
        - rewrite !SON, <- CONSTANT; auto.
      }
      destruct IH as (thM' & thN' & [STEPS_M STEPS_N]); exists thM', thN'.
      cbn; split; auto.
      econstructor; [ | eassumption ].
      constructor; cbn; [ | apply @set_neq_flip].
      rewrite set_eq. rewrite Ek_; eauto.
    (* COPY *)
    + assert (III : athread n = ODir.R /\ exists k_,
        k = (fun x => mapProgram (fun _ => inr) (k_ x)) /\ thN n = MkThread (Bind m k_)).
      { destruct (MATCH n); try congruence.
        - exfalso. destruct H0; try discriminate.
          subst p. rewrite (frobProgramId (mapProgram _ _)) in Eth.
          destruct p'; cbn in Eth; try discriminate.
        - split; auto. destruct H0; try discriminate.
          revert p' Eth1 H0.
          lazymatch goal with
          | [ |- ?G ] =>
            lazymatch (eval pattern R0, p in G) with
            | ?f _ _ =>
              refine match Eth in _ = ttt return
                match ttt with
                | MkThread p => f _ p
                | _ => False
                end with eq_refl => _ end
            end
          end; clear R0 p Eth.
          intros p' STEP H0.
          apply eq_sym, ProgramOpsFacts.mapProgramEqBind in H0.
          destruct H0 as (m' & k' & ? & ? & ?).
          exists k'. split; auto.
          congruence.
      }
      destruct III as [NOPE [k_ [Ek Ek_]]].
      specialize (IH thM (set n (CallingThread k_) thN) athread).
      forward IH.
      { intros tid; specialize (MATCH tid).
        destruct (@set_or_not n tid) as [[En SON]|[NEn SON]].
        - rewrite !SON.
          subst n.
          rewrite NOPE in *.
          constructor.
          + inversion MATCH; auto.
          + rewrite <- Eth1. constructor. rewrite Ek. auto.
        - rewrite !SON, <- CONSTANT; auto.
      }
      destruct IH as (thM' & thN' & [STEPS_M STEPS_N]); exists thM', thN'.
      cbn; split; auto.
      econstructor; [ | eassumption ].
      constructor; cbn; [ | apply @set_neq_flip].
      rewrite set_eq. rewrite Ek_. eauto.
  - (* low return *)
    destruct STEP as (R0 & k & Eth & Eth1).
    assert (W := MATCH n); remember (athread n) as an eqn:Ean; destruct W.
    3: congruence.
    { inversion H0.
      { congruence. }
      rewrite <- Eth in H1; clear Eth.
      revert x Eth1 H2.
      lazymatch goal with
      | [ |- ?G ] =>
        lazymatch (eval pattern R0, X, k in G) with
        | ?f _ _ _ =>
          refine match H1 in _ = ttt return
            match ttt with
            | CallingThread k => f _ _ k
            | _ => False
            end with eq_refl => _ end
        end
      end; clear R0 X k H1.
      intros x STEP H2.
      specialize (IH (set n (MkThread (k' x)) thM) thN athread).
      forward IH.
      { intros tid.
        destruct (@set_or_not n tid) as [[EQ SON] | [NEQ SON]].
        - subst n; rewrite SON, <- STEP, <- H, <- Ean.
          constructor; auto. constructor. auto.
        - rewrite SON. rewrite <- CONSTANT; auto.
      }
      destruct IH as (thM' & thN' & [STEPS_M STEPS_N]); exists thM', thN'.
      unfold projLeft_, projRight_; cbn; rewrite <- Ean; cbn; split; auto.
      econstructor; [| eassumption ].
      constructor; cbn; [ | apply @set_neq_flip ].
      rewrite <- H2, set_eq. eauto.
    }
    (* COPY *)
    { inversion H0.
      { congruence. }
      rewrite <- Eth in H1; clear Eth.
      revert x Eth1.
      lazymatch goal with
      | [ |- ?G ] =>
        lazymatch (eval pattern R0, X, k in G) with
        | ?f _ _ _ =>
          refine match H1 in _ = ttt return
            match ttt with
            | CallingThread k => f _ _ k
            | _ => False
            end with eq_refl => _ end
        end
      end; clear R0 X k H1.
      intros x Eth1.
      specialize (IH thM (set n (MkThread (k' x)) thN) athread).
      forward IH.
      { intros tid.
        destruct (@set_or_not n tid) as [[EQ SON] | [NEQ SON]].
        - subst n; rewrite SON, <- H, <- Eth1, <- Ean.
          constructor; auto. constructor. auto.
        - rewrite SON. rewrite <- CONSTANT; auto.
      }
      destruct IH as (thM' & thN' & [STEPS_M STEPS_N]); exists thM', thN'.
      unfold projLeft_, projRight_; cbn; rewrite <- Ean; cbn; split; auto.
      econstructor; [| eassumption ].
      constructor; cbn; [ | apply @set_neq_flip ].
      rewrite <- H2, set_eq. eauto.
    }
  - (* low silent *)
    destruct STEP as (R0 & p & Eth & Eth1).
    assert (W := MATCH n); remember (athread n) as an eqn:Ean; destruct W.
    3: congruence.
    { inversion H0. 2:congruence.
      rewrite <- H1 in Eth; clear H1.
      revert p' H3 H2.
      lazymatch goal with
      | [ |- ?G ] =>
        lazymatch (eval pattern R1, p0 in G) with
        | ?f _ _ =>
          refine match Eth in _ = ttt return
            match ttt with
            | MkThread p => f _ p
            | _ => False
            end with eq_refl => _ end
        end
      end; clear R1 p0 Eth.
      intros p' H3 H2.
      apply eq_sym, ProgramOpsFacts.mapProgramEqNoop in H3.
      destruct H3 as [p'' [-> ]].
      specialize (IH (set n (MkThread p'') thM) thN athread).
      forward IH.
      { intros tid.
        destruct (@set_or_not n tid) as [[EQ SON] | [NEQ SON]].
        - subst n; rewrite SON, <- Ean, <- Eth1, <- H.
          constructor; auto. constructor. auto.
        - rewrite SON. rewrite <- CONSTANT; auto.
      }
      destruct IH as (thM' & thN' & [STEPS_M STEPS_N]); exists thM', thN'.
      unfold projLeft_, projRight_; cbn. rewrite <- Ean; cbn; split; auto.
      econstructor; [| eassumption ].
      constructor; cbn; [ | apply @set_neq_flip ].
      rewrite set_eq. eauto.
    }
    { inversion H0. 2:congruence.
      rewrite <- H1 in Eth; clear H1.
      revert p' H3 H2.
      lazymatch goal with
      | [ |- ?G ] =>
        lazymatch (eval pattern R1, p0 in G) with
        | ?f _ _ =>
          refine match Eth in _ = ttt return
            match ttt with
            | MkThread p => f _ p
            | _ => False
            end with eq_refl => _ end
        end
      end; clear R1 p0 Eth.
      intros p' H5 H4.
      apply eq_sym, ProgramOpsFacts.mapProgramEqNoop in H5.
      destruct H5 as [p'' [-> ]].
      specialize (IH thM (set n (MkThread p'') thN) athread).
      forward IH.
      { intros tid.
        destruct (@set_or_not n tid) as [[EQ SON] | [NEQ SON]].
        - subst n; rewrite SON, <- Ean, <- Eth1, <- H.
          constructor; auto. constructor. auto.
        - rewrite SON. rewrite <- CONSTANT; auto.
      }
      destruct IH as (thM' & thN' & [STEPS_M STEPS_N]); exists thM', thN'.
      unfold projLeft_, projRight_; cbn; rewrite <- Ean; cbn; split; auto.
      econstructor; [| eassumption ].
      constructor; cbn; [ | apply @set_neq_flip ].
      rewrite set_eq. eauto.
    }
Qed.

Lemma matchPairThreads_noThreads M N : @matchPairThreads M N noThreads noThreads noThreads (fun _ => ODir.Neither).
Proof.
  constructor; auto.
Qed.

Definition to_odir (d : dir) : ODir.odir :=
  match d with | L => ODir.L | R => ODir.R end.

From Paco Require Import paco.

Definition onesideF (d : Dir.dir) {M N R} (self : Program (M +' N) R -> Prop)
  (p : Program (M +' N) R) : Prop :=
  match p with
  | Noop p => self p
  | Bind m k =>
    (match m, d with
     | inl _, L | inr _, R => True
     | _, _ => False
     end) /\ forall x, self (k x)
  | Return _ => True
  end.

Lemma monotone_onesideF d {M N R} : monotone1 (@onesideF d M N R).
Proof.
  intros [ X m k | | ] self self'; cbn; auto.
  firstorder.
Qed.
Hint Resolve monotone_onesideF : paco.

Definition oneside d {M N R} : Program (M +' N) R -> Prop := paco1 (onesideF d) bot1.

Definition onesideT d {M N} (t : Thread (M +' N)) : Prop :=
  match t with
  | MkThread p => oneside d p
  | CallingThread k => forall x, oneside d (k x)
  | NoThread => False
  end.

Record pct_pair {M M'} (athread : nat -> ODir.odir) (th : Threads (M +' M')) : Prop :=
  { pct_pair_0 : forall i,
      match athread i with
      | ODir.Neither => NoThread = th i
      | ODir.L => onesideT Dir.L (th i)
      | ODir.R => onesideT Dir.R (th i)
      end
  }.

Definition lt_odir (d d' : ODir.odir) : Prop :=
  match d, d' with
  | ODir.L, ODir.L | ODir.R, ODir.R | ODir.Neither, _ => True
  | _, _ => False
  end.

Delimit Scope odir_scope with odir.
Infix "<=" := lt_odir : odir_scope.

Record pct_pair_low {M M'} (athreadH athread : nat -> ODir.odir) (th : Threads (M +' M')) : Prop :=
  { pct_pair_low_impl : forall i, (athread i <= athreadH i)%odir
  ; pct_pair_low_0 : forall i, ODir.Neither <> athreadH i ->
      match athread i, th i with
      | ODir.Neither, MkThread _ => True
      | (ODir.L | ODir.R), CallingThread _ => True
      | _, _ => False
      end
  }.

Definition athread_to_pairing (athread : nat -> ODir.odir) : pairing.
Proof.
  refine
    {| pairing_pair := fun i =>
         match athread i with
         | ODir.Neither => None
         | ODir.L => Some (L, i)
         | ODir.R => Some (R, i)
         end
    ;  pairing_unpair := fun dj =>
         match athread (snd dj), fst dj with
         | ODir.L, L | ODir.R, R => Some (snd dj)
         | _, _ => None
         end
    |}.
  intros i [d j]; cbn; split.
  { destruct (athread i) eqn:E; inversion 1; cbn; subst; rewrite E; auto. }
  { destruct d, (athread j) eqn:E; inversion 1; cbn; subst; rewrite E; auto. }
Defined.

Definition eq_pairing (p q : pairing) : Prop :=
  forall i, pairing_pair p i = pairing_pair q i.

Delimit Scope pairing_scope with pairing.

Infix "=" := eq_pairing : pairing_scope.

Lemma eq_pairing_unpair {p q}
  : eq_pairing p q -> forall dj, pairing_unpair p dj = pairing_unpair q dj.
Proof.
  intros E dj.
  destruct (pairing_unpair p _) eqn:Ep.
  - apply pairing_pair_unpair in Ep.
    rewrite E in Ep.
    apply pairing_pair_unpair in Ep.
    congruence.
  - destruct (pairing_unpair q _) eqn:Eq; try reflexivity.
    apply pairing_pair_unpair in Eq.
    rewrite <- E in Eq.
    apply pairing_pair_unpair in Eq.
    congruence.
Qed.

Lemma Proper_is_free (p q : pairing) d i j
  : (p = q)%pairing -> is_free d p i j -> is_free d q i j.
Proof.
  intros E; unfold is_free.
  rewrite E, (eq_pairing_unpair E). auto.
Qed.

Lemma Proper_is_paired (p q : pairing) d i j
  : (p = q)%pairing -> is_paired d p i j -> is_paired d q i j.
Proof.
  intros E; unfold is_paired.
  rewrite E. auto.
Qed.

Lemma Proper_set {X Y} `{EqDec X} (x : X) (y : Y)
  : Proper (pointwise_relation _ eq ==> pointwise_relation _ eq) (set x y).
Proof.
  intros f g Ef i. unfold set. destruct eq_dec; auto.
Qed.

Lemma Proper_add (p q : pairing) d i j (FREE_p : is_free d p i j) (FREE_q : is_free d q i j)
  : (p = q)%pairing -> (add d p i j = add d q i j)%pairing.
Proof.
  intros E i0; apply Proper_set. auto.
Qed.

Lemma Proper_rm (p q : pairing) d i j (PAIRED_p : is_paired d p i j) (PAIRED_q : is_paired d q i j)
  : (p = q)%pairing -> (rm d p i j = rm d q i j)%pairing.
Proof.
  intros E i0; apply Proper_set; auto.
Qed.

Lemma Proper_pairConcTrace {M N} (p q : pairing) (tM : ConcTrace M) (tN : ConcTrace N) tMN
  : (p = q)%pairing -> pairConcTrace_ p tM tN tMN -> pairConcTrace_ q tM tN tMN.
Proof.
  intros E P; revert q E; induction P; intros; subst.
  - apply pairConcTrace_nil; auto.
  - assert (FREE' := Proper_is_free E FREE).
    eapply pairConcTrace_call_L; eauto using @Proper_add.
  - assert (FREE' := Proper_is_free E FREE).
    eapply pairConcTrace_call_R; eauto using @Proper_add.
  - assert (PAIRED' := Proper_is_paired E PAIRED).
    eapply pairConcTrace_return_L; eauto using @Proper_rm.
  - assert (PAIRED' := Proper_is_paired E PAIRED).
    eapply pairConcTrace_return_R; eauto using @Proper_rm.
Qed.

Lemma free_pct_pair {M M'} d athread (th : Threads (M +' M')) n
  : pct_pair athread th -> NoThread = th n -> is_free d (athread_to_pairing athread) n n.
Proof.
  intros PCT Eth. red; cbn.
  assert (H := pct_pair_0 PCT (i := n)).
  destruct (athread n); solve [ auto | rewrite <- Eth in H; contradiction ].
Qed.

Lemma free_pct_pair_low {M M'} d athread athreadL (th : Threads (M +' M')) n
  : pct_pair athread th -> pct_pair_low athread athreadL th ->
    (exists R (p : _ R), MkThread p = th n) -> is_free d (athread_to_pairing athreadL) n n.
Proof.
  intros PCT PCTL (R & p & Eth); red; cbn.
  assert (H := pct_pair_0 PCT (i := n)).
  assert (H1 := pct_pair_low_impl PCTL (i := n)). red in H1.
  assert (H2 := pct_pair_low_0 PCTL (i := n)).
  rewrite <- Eth in H, H2.
  destruct (athread n); try discriminate.
  all: destruct (athreadL n); auto.
  all: contradiction H2; discriminate.
Qed.

Lemma paired_pct_pair_low {M M'} d athread athreadL (th : Threads (M +' M')) n
  : pct_pair athread th -> pct_pair_low athread athreadL th ->
    athread n = to_odir d ->
    (exists X R (k : X -> _ R), CallingThread k = th n) -> is_paired d (athread_to_pairing athreadL) n n.
Proof.
  intros PCT PCTL Ea (X & R & k & Eth). red; cbn.
  assert (H := pct_pair_0 PCT (i := n)).
  assert (H1 := pct_pair_low_impl PCTL (i := n)). red in H1.
  assert (H2 := pct_pair_low_0 PCTL (i := n)).
  rewrite <- Eth in H, H2.
  destruct (athread n); try discriminate.
  all: destruct (athreadL n); try contradiction; auto.
  all: try (destruct d; auto; discriminate).
  all: try (contradiction H2; discriminate).
Qed.

Lemma add_athread d athread i (FREE : is_free d (athread_to_pairing athread) i i)
  : (athread_to_pairing (set i (to_odir d) athread)
  = add d (athread_to_pairing athread) i i)%pairing.
Proof.
  intros j. unfold add; cbn.
  unfold set; destruct eq_dec as [ <- | NE ].
  - destruct d; auto.
  - auto.
Qed.

Lemma rm_athread d athread i (FREE : is_paired d (athread_to_pairing athread) i i)
  : (athread_to_pairing (set i ODir.Neither athread)
  = rm d (athread_to_pairing athread) i i)%pairing.
Proof.
  intros j. unfold rm; cbn.
  unfold set; destruct eq_dec as [ <- | NE ].
  - destruct d; auto.
  - auto.
Qed.

Arguments add_athread : clear implicits.
Arguments rm_athread : clear implicits.

Lemma pct_pair_set {M N} athread (th th' : Threads (M +' N)) n d
  : match d with
    | ODir.Neither => NoThread = th' n
    | ODir.L => onesideT L (th' n)
    | ODir.R => onesideT R (th' n)
    end ->
    (forall i, i <> n -> th i = th' i) ->
    pct_pair athread th ->
    pct_pair (set n d athread) th'.
Proof.
  intros Eth CONSTANT PCT.
  constructor; intros i. unfold set; destruct eq_dec as [ <- | NE ].
  - auto.
  - rewrite <- CONSTANT; auto. apply PCT.
Qed.

Lemma pct_pair_set_quiet {M N} athread (th th' : Threads (M +' N)) n
  : (exists X R (m : _ X) (k : _ -> _ R), MkThread (Bind m k) = th n /\ CallingThread k = th' n)
    \/ (exists X R (x : X) (k : X -> _ R), CallingThread k = th n /\ MkThread (k x) = th' n)
    \/ (exists R (p : _ R), MkThread (Noop p) = th n /\ MkThread p = th' n)  ->
    (forall i, i <> n -> th i = th' i) ->
    pct_pair athread th ->
    pct_pair athread th'.
Proof.
  intros Eth CONSTANT PCT.
  constructor; intros i.
  assert (H := pct_pair_0 PCT (i := i)).
  destruct (eq_dec n i) as [ <- | NE ].
  - destruct Eth as
      [ (X & R & m & k & Eth & Eth') | [ (X & R & x & k & Eth & Eth') | (R & p & Eth & Eth')]];
      rewrite <- Eth in H; rewrite <- Eth'; destruct (athread n); cbn in *; try discriminate; auto.
    + punfold H; cbn in H; destruct H; pclearbot; auto.
    + punfold H; cbn in H; destruct H; pclearbot; auto.
    + punfold H; cbn in H; pclearbot; auto.
    + punfold H; cbn in H; pclearbot; auto.
  - rewrite <- CONSTANT; auto.
Qed.

Lemma pct_pair_low_set_high {M N} athread athreadL (th th' : Threads (M +' N)) n d
  : match d with
    | ODir.Neither => exists R (p : _ R), MkThread p = th n /\ NoThread = th' n
    | ODir.L | ODir.R => exists R (p : _ R), NoThread = th n /\ MkThread p = th' n
    end ->
    (forall i, i <> n -> th i = th' i) ->
    pct_pair athread th ->
    pct_pair_low athread athreadL th ->
    pct_pair_low (set n d athread) athreadL th'.
Proof.
  intros Eth CONSTANT PCT PCTL.
  constructor.
  - intros i.
    assert (H := pct_pair_0 PCT (i := i)).
    assert (H1 := pct_pair_low_impl PCTL (i := i)).
    assert (H2 := pct_pair_low_0 PCTL (i := i)).
    unfold set; destruct (eq_dec n i) as [ <- | NE ]; [ | auto ].
    destruct d; destruct Eth as (R & p & Eth & Eth'); rewrite <- Eth in H, H2;
      destruct (athread n), (athreadL n); contradiction + auto.
    all: contradiction H2; discriminate.
  - intros i.
    assert (H := pct_pair_0 PCT (i := i)).
    assert (H1 := pct_pair_low_impl PCTL (i := i)).
    assert (H2 := pct_pair_low_0 PCTL (i := i)).
    unfold set; destruct (eq_dec n i) as [ <- | NE ]; [ | rewrite <- CONSTANT; auto ].
    destruct d; destruct Eth as (R & p & Eth & Eth'); rewrite <- Eth in H, H2;
      rewrite <- Eth';
      destruct (athread n), (athreadL n); contradiction + auto.
Qed.

Lemma pct_pair_low_set {M N} athread athreadL (th th' : Threads (M +' N)) n d
  : match d with
    | ODir.Neither => exists R (p : Program (M +' N) R), MkThread p = th' n
    | ODir.L => exists X R m (k : X -> _ R), MkThread (Bind (inl m) k) = th n /\ CallingThread k = th' n
    | ODir.R => exists X R m (k : X -> _ R), MkThread (Bind (inr m) k) = th n /\ CallingThread k = th' n
    end ->
    (forall i, i <> n -> th i = th' i) ->
    pct_pair athread th ->
    pct_pair_low athread athreadL th ->
    pct_pair_low athread (set n d athreadL) th'.
Proof.
  intros Eth CONSTANT PCT PCTL.
  constructor.
  - intros i.
    assert (H := pct_pair_0 PCT (i := i)).
    assert (H1 := pct_pair_low_impl PCTL (i := i)).
    assert (H2 := pct_pair_low_0 PCTL (i := i)).
    unfold set; destruct (eq_dec n i) as [ <- | NE ]; [ | auto ].
    destruct d; my_decompose Eth; [ rewrite <- Eth0 in H, H2 | | rewrite <- Eth0 in H, H2 ];
      destruct (athread n), (athreadL n); discriminate + contradiction + auto.
    all: punfold H; cbn in H; contradiction (proj1 H).
  - intros i.
    assert (H := pct_pair_0 PCT (i := i)).
    assert (H1 := pct_pair_low_impl PCTL (i := i)).
    assert (H2 := pct_pair_low_0 PCTL (i := i)).
    unfold set; destruct (eq_dec n i) as [ <- | NE ]; [ | rewrite <- CONSTANT; auto ].
    destruct d; my_decompose Eth; try rewrite <- Eth0 in H, H2;
      rewrite <- Eth;
      destruct (athread n), (athreadL n); contradiction + auto.
Qed.

Lemma pct_pair_low_set_quiet {M N} athread athreadL (th th' : Threads (M +' N)) n
  : (exists R (p : _ R), MkThread (Noop p) = th n /\ MkThread p = th' n) ->
    (forall i, i <> n -> th i = th' i) ->
    pct_pair_low athread athreadL th ->
    pct_pair_low athread athreadL th'.
Proof.
  intros (R & p & Eth & Eth') CONSTANT PCTL.
  constructor.
  - intros i; apply (pct_pair_low_impl PCTL).
  - intros i; assert (H := pct_pair_low_0 PCTL (i := i)).
    destruct (eq_dec n i) as [ <- | ]; [ | rewrite <- CONSTANT; auto ].
    rewrite <- Eth in H; rewrite <- Eth'; destruct athreadL; auto.
Qed.

Lemma oneside_map_L {M N} X (p : Program M X)
  : oneside (N := N) L (mapProgram (fun _ => inl) p).
Proof.
  revert p; pcofix SELF.
  intros [ Y m k | | ]; pfold; cbn; auto.
Qed.

Lemma oneside_map_R {M N} X (p : Program N X)
  : oneside (M := M) R (mapProgram (fun _ => inr) p).
Proof.
  revert p; pcofix SELF.
  intros [ Y m k | | ]; pfold; cbn; auto.
Qed.

Lemma paired_athread d athread n
  : athread n = d ->
    let p := athread_to_pairing athread in
    match d with
    | ODir.Neither => True
    | ODir.L => is_paired L p n n
    | ODir.R => is_paired R p n n
    end.
Proof.
  intros <-; cbn.
  unfold is_paired; cbn.
  destruct (athread n); auto.
Qed.

Hint Resolve @oneside_map_L @oneside_map_R : oneside.

Ltac solve_oneside :=
  match goal with
  | [ E : _ = ?t |- onesideT _ ?t ] => rewrite <- E; cbn; eauto with oneside
  end.

Lemma discriminate_athread_Neither M N (th : Threads (M +' N)) athread n
  : pct_pair athread th -> (exists R (p : _ R), MkThread p = th n) \/ (exists R S (k : R -> _ S), CallingThread k = th n) -> athread n = ODir.Neither -> False.
Proof.
  intros PCT Eth E.
  assert (H := pct_pair_0 PCT (i := n)).
  rewrite E in H.
  my_decompose Eth; congruence.
Qed.

Hint Resolve @discriminate_athread_Neither : discriminate_athread.

Ltac discriminate_athread :=
  lazymatch goal with
  | [ Eth : MkThread _ = ?th ?n, PCT : pct_pair ?athread ?th, Ea : ?athread ?n = ODir.Neither |- _ ] =>
    exfalso; eauto with discriminate_athread
  | [ Eth : CallingThread _ = ?th ?n, PCT : pct_pair ?athread ?th, Ea : ?athread ?n = ODir.Neither |- _ ] =>
    exfalso; eauto 9 with discriminate_athread
  end.

Ltac unstick :=
  unfold projLeft_, projRight_; cbn;
    match goal with
    | [ athread : nat -> ODir.odir |- _ ] =>
      match goal with
      | [ |- context [ athread ?n ] ] =>
        let E := fresh "E" in destruct (athread n) eqn:E; cbn;
        try discriminate_athread
      end
    end.

Lemma pairConcTrace_projHigh_ {M M' N N'} fM fN th th_
    (btMN : BiTrace (M +' N) (M' +' N'))
  : ThreadsSteps (pairProgram fM fN) btMN th th_ ->
    forall (tMN : ConcTrace (M +' N)),
    tMN = projHigh btMN ->
    forall athread,
    pct_pair athread th ->
    pairConcTrace_ (athread_to_pairing athread) (projHigh (projLeft_ athread btMN))
      (projHigh (projRight_ athread btMN)) tMN.
Proof.
  induction 1 as [ | th [n e] btMN th' th_ [STEP CONSTANT] STEPS IH];
    intros tMN PROJ athread PCT; subst; cbn.
  { constructor; auto. }
  specialize (IH _ eq_refl).
  destruct e as [ ? m | | | |]; cbn in *.
  - (* High Call *) destruct STEP as [Eth Eth'].
    destruct m as [m | m]; cbn.
    + assert (FREE := free_pct_pair (d := L) PCT Eth).
      eapply pairConcTrace_call_L; try reflexivity; [].
      eapply Proper_pairConcTrace; [ apply add_athread | apply IH ].
      eapply pct_pair_set; eauto; cbn.
      solve_oneside.
    + assert (FREE := free_pct_pair (d := R) PCT Eth).
      eapply pairConcTrace_call_R; try reflexivity; [].
      eapply Proper_pairConcTrace; [ apply add_athread | apply IH ].
      eapply pct_pair_set; eauto; cbn.
      solve_oneside.
  - (* High Return *) destruct STEP as [Eth Eth'].
    unstick.
    + assert (PAIRED := paired_athread E); cbn in PAIRED.
      eapply pairConcTrace_return_L; try reflexivity; [].
      eapply Proper_pairConcTrace; [ apply rm_athread | apply IH ].
      eapply pct_pair_set; eauto; cbn.
    + assert (PAIRED := paired_athread E); cbn in PAIRED.
      eapply pairConcTrace_return_R; try reflexivity; [].
      eapply Proper_pairConcTrace; [ apply rm_athread | apply IH ].
      eapply pct_pair_set; eauto; cbn.
  - (* Low Call *) destruct STEP as (R & k & Eth & Eth').
    destruct n0 as [m | m]; cbn.
    + apply IH. eapply pct_pair_set_quiet; eauto 7.
    + apply IH. eapply pct_pair_set_quiet; eauto 7.
  - (* Low Return *) destruct STEP as (R & k & Eth & Eth').
    unstick.
    + apply IH. eapply pct_pair_set_quiet; eauto 8.
    + apply IH. eapply pct_pair_set_quiet; eauto 8.
  - (* Low Silent *) destruct STEP as (R & p & Eth & Eth').
    unstick.
    + apply IH. eapply pct_pair_set_quiet; eauto 8.
    + apply IH. eapply pct_pair_set_quiet; eauto 8.
Qed.

Lemma pairConcTrace_projLow_ {M M' N N'} fM fN th th_
    (btMN : BiTrace (M +' N) (M' +' N'))
  : ThreadsSteps (pairProgram fM fN) btMN th th_ ->
    forall (tMN : ConcTrace (M' +' N')),
    tMN = projLow btMN ->
    forall athread athreadL,
    pct_pair athread th ->
    pct_pair_low athread athreadL th ->
    pairConcTrace_ (athread_to_pairing athreadL) (projLow (projLeft_ athread btMN))
      (projLow (projRight_ athread btMN)) tMN.
Proof.
  induction 1 as [ | th [n e] btMN th' th_ [STEP CONSTANT] STEPS IH];
    intros tMN PROJ athread athreadL PCT PCTL; subst; cbn.
  { constructor; auto. }
  specialize (IH _ eq_refl).
  destruct e as [ ? m | | | |]; cbn in *.
  - (* High Call *) destruct STEP as [Eth Eth'].
    destruct m as [m | m]; cbn.
    + eapply IH; [ eapply pct_pair_set; eauto; solve_oneside | ].
      eapply pct_pair_low_set_high; eauto.
    + eapply IH; [ eapply pct_pair_set; eauto; solve_oneside | ].
      eapply pct_pair_low_set_high; eauto.
  - (* High Return *) destruct STEP as [Eth Eth'].
    unstick.
    + apply IH; [ eapply pct_pair_set; eauto | ].
      eapply pct_pair_low_set_high; eauto.
    + apply IH; [ eapply pct_pair_set; eauto | ].
      eapply pct_pair_low_set_high; eauto.
  - (* Low Call *) destruct STEP as (R & k & Eth & Eth').
    destruct n0 as [m | m]; cbn.
    + assert (FREE := free_pct_pair_low (d := L) PCT PCTL (ex_intro _ _ (ex_intro _ _ Eth))).
      eapply pairConcTrace_call_L; try reflexivity; [].
      eapply Proper_pairConcTrace; [ apply add_athread | ].
      apply IH; [ eapply pct_pair_set_quiet; eauto 8 |].
      eapply pct_pair_low_set; cbn; eauto 8.
    + assert (FREE := free_pct_pair_low (d := Dir.R) PCT PCTL (ex_intro _ _ (ex_intro _ _ Eth))).
      eapply pairConcTrace_call_R; try reflexivity; [].
      eapply Proper_pairConcTrace; [ apply add_athread | ].
      apply IH; [ eapply pct_pair_set_quiet; eauto 8 |].
      eapply pct_pair_low_set; cbn; eauto 8.
  - (* Low Return *) destruct STEP as (R & k & Eth & Eth').
    unstick.
    + assert (PAIRED := paired_pct_pair_low (d := L) PCT PCTL E (ex_intro _ _ (ex_intro _ _ (ex_intro _ _ Eth)))).
      eapply pairConcTrace_return_L; try reflexivity; [].
      eapply Proper_pairConcTrace; [ apply rm_athread | ].
      apply IH; [ eapply pct_pair_set_quiet; eauto 8 | ].
      eapply pct_pair_low_set; cbn; eauto 8.
    + assert (PAIRED := paired_pct_pair_low (d := Dir.R) PCT PCTL E (ex_intro _ _ (ex_intro _ _ (ex_intro _ _ Eth)))).
      eapply pairConcTrace_return_R; try reflexivity; [].
      eapply Proper_pairConcTrace; [ apply rm_athread | ].
      apply IH; [ eapply pct_pair_set_quiet; eauto 8 | ].
      eapply pct_pair_low_set; cbn; eauto 8.
  - (* Low Silent *) destruct STEP as (R & p & Eth & Eth').
    unstick.
    + apply IH; [ eapply pct_pair_set_quiet; eauto 8 | ].
      eapply pct_pair_low_set_quiet; eauto.
    + apply IH; [ eapply pct_pair_set_quiet; eauto 8 | ].
      eapply pct_pair_low_set_quiet; eauto.
Qed.

Lemma athread_empty : (athread_to_pairing (fun _ => ODir.Neither) = pairing_empty)%pairing.
Proof.
  intros i; cbn; reflexivity.
Qed.

Lemma pct_pair_empty {M N} : @pct_pair M N (fun _ => ODir.Neither) noThreads.
Proof.
  constructor; intros; reflexivity.
Qed.

Lemma pct_pair_low_empty {M N} : @pct_pair_low M N (fun _ => ODir.Neither) (fun _ => ODir.Neither) noThreads.
Proof.
  constructor; intros; cbn; auto.
Qed.

Lemma pairConcTrace_projHigh {M M' N N'} fM fN
    (btMN : BiTrace (M +' N) (M' +' N')) (tMN : ConcTrace (M +' N)) th_
  : ThreadsSteps (pairProgram fM fN) btMN noThreads th_ ->
    tMN = projHigh btMN ->
    pairConcTrace (projHigh (projLeft_ (fun _ : nat => ODir.Neither) btMN))
      (projHigh (projRight_ (fun _ : nat => ODir.Neither) btMN)) tMN.
Proof.
  intros STEPS PROJ; eapply Proper_pairConcTrace, pairConcTrace_projHigh_;
    eauto using athread_empty, pct_pair_empty.
Qed.

Lemma pairConcTrace_projLow {M M' N N'} fM fN
    (btMN : BiTrace (M +' N) (M' +' N')) (tMN : ConcTrace (M' +' N')) th_
  : ThreadsSteps (pairProgram fM fN) btMN noThreads th_ ->
    tMN = projLow btMN ->
    pairConcTrace (projLow (projLeft_ (fun _ : nat => ODir.Neither) btMN))
      (projLow (projRight_ (fun _ : nat => ODir.Neither) btMN)) tMN.
Proof.
  intros STEPS PROJ; eapply Proper_pairConcTrace, pairConcTrace_projLow_;
    eauto using athread_empty, pct_pair_empty, pct_pair_low_empty.
Qed.

Theorem ConcTrace_pairProgram_inv {M M' N N' : Type -> Type}
  : forall (fM : M ~> Program M') (fN : N ~> Program N')
           tMN tMN',
      IsConcTraceOf_ tMN tMN' (pairProgram fM fN) ->
      exists tM tM' tN tN',
        IsConcTraceOf_ tM tM' fM /\
        IsConcTraceOf_ tN tN' fN /\
        pairConcTrace tM tN tMN /\
        pairConcTrace tM' tN' tMN'.
Proof.
  intros fM fN tMN tMN'.
  intros TRACE_MN.
  destruct TRACE_MN as [btMN th_ HtMN HtMN' STEPS].
  assert (STEPS0 := STEPS).
  eapply ConcTrace_pairProgram_inv_aux in STEPS;
    [ | apply @matchPairThreads_noThreads ].
  destruct STEPS as (thM' & thN' & STEPS_M & STEPS_N).
  repeat eexists; try eassumption.
  + eapply pairConcTrace_projHigh; eauto.
  + eapply pairConcTrace_projLow; eauto.
Qed.

Inductive pairBiTrace_ {M1 M2 N1 N2 : Type -> Type} (p : pairing)
  : BiTrace (M1 +' N1) (M2 +' N2) -> BiTrace M1 M2 -> BiTrace N1 N2 -> Prop :=
| pairBiTrace_nil : pairBiTrace_ p nil nil nil
| pairBiTrace_HighCall_L R (m : M1 R) i j (FREE : is_free L p i j) tMN tM tN
  : pairBiTrace_ (add L p i j) tMN tM tN ->
    pairBiTrace_ p ((i, HighCallEvent (inl m)) :: tMN) ((j, HighCallEvent m) :: tM) tN
| pairBiTrace_HighCall_R R (m : N1 R) i j (FREE : is_free Dir.R p i j) tMN tM tN
  : pairBiTrace_ (add Dir.R p i j) tMN tM tN ->
    pairBiTrace_ p ((i, HighCallEvent (inr m)) :: tMN) tM ((j, HighCallEvent m) :: tN)
| pairBiTrace_HighReturn_L R (r : R) i j (PAIRED : is_paired L p i j) tMN tM tN
  : pairBiTrace_ (rm L p i j) tMN tM tN ->
    pairBiTrace_ p ((i, HighReturnEvent r) :: tMN) ((j, HighReturnEvent r) :: tM) tN
| pairBiTrace_HighReturn_R R (r : R) i j (PAIRED : is_paired Dir.R p i j) tMN tM tN
  : pairBiTrace_ (rm Dir.R p i j) tMN tM tN ->
    pairBiTrace_ p ((i, HighReturnEvent r) :: tMN) tM ((j, HighReturnEvent r) :: tN)
| pairBiTrace_LowCall_L R (m : M2 R) i j tMN tM tN
  : i = unwrap 0 (pairing_unpair p (L, j)) ->
    pairBiTrace_ p tMN tM tN ->
    pairBiTrace_ p ((i, LowCallEvent (inl m)) :: tMN) ((j, LowCallEvent m) :: tM) tN
| pairBiTrace_LowCall_R R (m : N2 R) i j tMN tM tN
  : i = unwrap 0 (pairing_unpair p (Dir.R, j)) ->
    pairBiTrace_ p tMN tM tN ->
    pairBiTrace_ p ((i, LowCallEvent (inr m)) :: tMN) tM ((j, LowCallEvent m) :: tN)
| pairBiTrace_LowReturn_L R (r : R) i j tMN tM tN
  : i = unwrap 0 (pairing_unpair p (Dir.L, j)) ->
    pairBiTrace_ p tMN tM tN ->
    pairBiTrace_ p ((i, LowReturnEvent r) :: tMN) ((j, LowReturnEvent r) :: tM) tN
| pairBiTrace_LowReturn_R R (r : R) i j tMN tM tN
  : i = unwrap 0 (pairing_unpair p (Dir.R, j)) ->
    pairBiTrace_ p tMN tM tN ->
    pairBiTrace_ p ((i, LowReturnEvent r) :: tMN) tM ((j, LowReturnEvent r) :: tN)
| pairBiTrace_LowSilent_L i j tMN tM tN
  : i = unwrap 0 (pairing_unpair p (L, j)) ->
    pairBiTrace_ p tMN tM tN ->
    pairBiTrace_ p ((i, LowSilentEvent) :: tMN) ((j, LowSilentEvent) :: tM) tN
| pairBiTrace_LowSilent_R i j tMN tM tN
  : i = unwrap 0 (pairing_unpair p (Dir.R, j)) ->
    pairBiTrace_ p tMN tM tN ->
    pairBiTrace_ p ((i, LowSilentEvent) :: tMN) tM ((j, LowSilentEvent) :: tN)
.

Definition pairBiTrace {M1 M2 N1 N2 : Type -> Type}
  : BiTrace (M1 +' N1) (M2 +' N2) -> BiTrace M1 M2 -> BiTrace N1 N2 -> Prop :=
  pairBiTrace_ pairing_empty.

Inductive pairThread {M N : Type -> Type}
  : option (nat + nat) -> option (nat + nat) -> Thread (M +' N) -> Thread M -> Thread N -> Prop :=
| pairThread_no : pairThread None None NoThread NoThread NoThread
| pairThread_MkThread_L i R (p : Program M R)
  : pairThread (Some (inl i)) None (MkThread (mapProgram (fun _ => inl) p)) (MkThread p) NoThread
| pairThread_MkThread_R i R (p : Program N R)
  : pairThread (Some (inr i)) None (MkThread (mapProgram (fun _ => inr) p)) NoThread (MkThread p)
| pairThread_Calling_L i R X (k : X -> Program M R)
  : pairThread (Some (inl i)) (Some (inl i))
      (CallingThread (fun x => mapProgram (fun _ => inl) (k x))) (CallingThread k) NoThread
| pairThread_Calling_R i R X (k : X -> Program N R)
  : pairThread (Some (inr i)) (Some (inr i))
      (CallingThread (fun x => mapProgram (fun _ => inr) (k x))) NoThread (CallingThread k)
.

Section PAIR_THREADS.

Context {M N : Type -> Type}
    (p q : pairing) (thMN : Threads (M +' N)) (thM : Threads M) (thN : Threads N).

Record pairThreads_ : Prop :=
  { pairThreads_L : forall i j,
      is_paired L p i j ->
      match thM j with
      | NoThread => False
      | MkThread f => is_free L q i j
          /\ MkThread (mapProgram (fun _ => inl) f) = thMN i
      | CallingThread k => is_paired L q i j
          /\ CallingThread (fun x => mapProgram (fun _ => inl) (k x)) = thMN i
      end
  ; pairThreads_R : forall i j,
      is_paired R p i j ->
      match thN j with
      | NoThread => False
      | MkThread f => is_free R q i j
          /\ MkThread (mapProgram (fun _ => inr) f) = thMN i
      | CallingThread k => is_paired R q i j
          /\ CallingThread (fun x => mapProgram (fun _ => inr) (k x)) = thMN i
      end
  ; pairThreads_NoThread_MN : forall i, is_unpaired p i -> NoThread = thMN i
  ; pairThreads_NoThread_M : forall j, is_unpaired_L p j -> NoThread = thM j
  ; pairThreads_NoThread_N : forall j, is_unpaired_R p j -> NoThread = thN j
  ; pairThreads_subpairing : pairing_incl q p
  }.

Lemma pairThreads_free_L (TH : pairThreads_) i j
  : is_free L p i j ->
    is_free L q i j /\ NoThread = thM j /\ NoThread = thMN i.
Proof.
  intros FREE. rewrite is_free_unpaired_L in FREE.
  destruct FREE as [Ui Uj].
  split; [ | split].
  - apply is_free_unpaired_L.
    split; eapply incl_unpaired; eassumption + eapply pairThreads_subpairing; eauto.
  - apply pairThreads_NoThread_M; auto.
  - apply pairThreads_NoThread_MN; auto.
Qed.

Lemma pairThreads_free_R (TH : pairThreads_) i j
  : is_free R p i j ->
    is_free R q i j /\ NoThread = thN j /\ NoThread = thMN i.
Proof.
  intros FREE. rewrite is_free_unpaired_R in FREE.
  destruct FREE as [Ui Uj].
  split; [ | split].
  - apply is_free_unpaired_R.
    split; eapply incl_unpaired; eassumption + eapply pairThreads_subpairing; eauto.
  - apply pairThreads_NoThread_N; auto.
  - apply pairThreads_NoThread_MN; auto.
Qed.

End PAIR_THREADS.

Arguments pairThreads_subpairing : clear implicits.
Arguments pairThreads_subpairing {M N p q thMN thM thN}.

Lemma pairThreads_L' M N p q i j (P : is_paired L p i j)
    (thMN : Threads (M +' N)) thM thN
    thj
  : pairThreads_ p q thMN thM thN ->
    thj = thM j ->
    match thj with
    | NoThread => NoThread (* unreachable *)
    | MkThread (Return r) => MkThread (Return r)
    | MkThread (Noop f) => MkThread (Noop (mapProgram (fun _ => inl) f))
    | MkThread (Bind m k) => MkThread (Bind (inl m) (fun x => mapProgram (fun _ => inl) (k x)))
    | CallingThread k => CallingThread (fun x => mapProgram (fun _ => inl) (k x))
    end = thMN i.
Proof.
  intros H ->; generalize (pairThreads_L H P). destruct (thM j) as [? f | | ]; intros []; auto.
  rewrite (frobProgramId (mapProgram _ _)) in H1; destruct f; auto.
Qed.

Lemma pairThreads_R' M N p q i j (P : is_paired R p i j)
    (thMN : Threads (M +' N)) thM thN
    thj
  : pairThreads_ p q thMN thM thN ->
    thj = thN j ->
    match thj with
    | NoThread => NoThread (* unreachable *)
    | MkThread (Return r) => MkThread (Return r)
    | MkThread (Noop f) => MkThread (Noop (mapProgram (fun _ => inr) f))
    | MkThread (Bind m k) => MkThread (Bind (inr m) (fun x => mapProgram (fun _ => inr) (k x)))
    | CallingThread k => CallingThread (fun x => mapProgram (fun _ => inr) (k x))
    end = thMN i.
Proof.
  intros H ->; generalize (pairThreads_R H P). destruct (thN j) as [? f | | ]; intros []; auto.
  rewrite (frobProgramId (mapProgram _ _)) in H1; destruct f; auto.
Qed.

Definition leftBiEvent (M1 M2 N1 N2 : Type -> Type)
    (p : pairing) (e : BiThreadEvent M1 M2) : BiThreadEvent (M1 +' N1) (M2 +' N2) :=
  ( unwrap 0 (pairing_unpair p (L, fst e))
  , match snd e with
    | HighCallEvent m => HighCallEvent (inl m)
    | HighReturnEvent r => HighReturnEvent r
    | LowCallEvent m => LowCallEvent (inl m)
    | LowReturnEvent r => LowReturnEvent r
    | LowSilentEvent => LowSilentEvent
    end).

Definition leftBiTrace (M1 M2 N1 N2 : Type -> Type)
    (p : pairing) : BiTrace M1 M2 -> BiTrace (M1 +' N1) (M2 +' N2) :=
  map (leftBiEvent p).

Definition rightBiEvent (M1 M2 N1 N2 : Type -> Type)
    (p : pairing) (e : BiThreadEvent N1 N2) : BiThreadEvent (M1 +' N1) (M2 +' N2) :=
  ( unwrap 0 (pairing_unpair p (Dir.R, fst e))
  , match snd e with
    | HighCallEvent m => HighCallEvent (inr m)
    | HighReturnEvent r => HighReturnEvent r
    | LowCallEvent m => LowCallEvent (inr m)
    | LowReturnEvent r => LowReturnEvent r
    | LowSilentEvent => LowSilentEvent
    end).

Definition rightBiTrace (M1 M2 N1 N2 : Type -> Type)
    (p : pairing) : BiTrace N1 N2 -> BiTrace (M1 +' N1) (M2 +' N2) :=
  map (rightBiEvent p).

Lemma pairBiTrace_left {M1 M2 N1 N2 : Type -> Type}
    (p : pairing) tMN btM0 btM btN
  : nil = projHigh btM0 ->
    (exists btMN,
      tMN = projHigh btMN /\
      pairBiTrace_ p btMN btM btN) ->
    (exists btMN : BiTrace (M1 +' N1) (M2 +' N2),
      tMN = projHigh btMN /\
      pairBiTrace_ p btMN (btM0 ++ btM) btN).
Proof.
  induction btM0 as [ | [? []] ? IH ].
  1: solve [auto].
  all: try discriminate.
  all: intros H H'; cbn in H; specialize (IH H H'); destruct IH as [btMN [PROJ PAIR]].
  all: eexists; split; [ | cbn; constructor ]; cbn; reflexivity + eassumption.
Qed.

Lemma pairBiTrace_right {M1 M2 N1 N2 : Type -> Type}
    (p : pairing) tMN btN0 btM btN
  : nil = projHigh btN0 ->
    (exists btMN,
      tMN = projHigh btMN /\
      pairBiTrace_ p btMN btM btN) ->
    (exists btMN : BiTrace (M1 +' N1) (M2 +' N2),
      tMN = projHigh btMN /\
      pairBiTrace_ p btMN btM (btN0 ++ btN)).
Proof.
  induction btN0 as [ | [? []] ? IH ].
  1: solve [auto].
  all: try discriminate.
  all: intros H H'; cbn in H; specialize (IH H H'); destruct IH as [btMN [PROJ PAIR]].
  all: eexists; split; [ | cbn; constructor ]; cbn; reflexivity + eassumption.
Qed.

Lemma pairBiTrace_pairConcTrace_ {M1 M2 N1 N2 : Type -> Type}
    (p : pairing) tM tN tMN (btM : BiTrace M1 M2) (btN : BiTrace N1 N2)
  : tM = projHigh btM ->
    tN = projHigh btN ->
    pairConcTrace_ p tM tN tMN ->
    exists btMN,
      tMN = projHigh btMN /\
      pairBiTrace_ p btMN btM btN.
Proof.
  intros PROJ_M PROJ_N PAIR; revert btM btN PROJ_M PROJ_N; induction PAIR; cbn; intros.
  - subst.
    rewrite <- (app_nil_r btM); apply pairBiTrace_left; auto.
    rewrite <- (app_nil_r btN); apply pairBiTrace_right; auto.
    exists nil; do 2 constructor.
  - subst.
    destruct (projHigh_inv_cons PROJ_M) as (btM0 & btM' & PROJ_M0 & PROJ_M' & EM).
    subst btM.
    apply pairBiTrace_left; auto.
    specialize (IHPAIR _ _ PROJ_M' eq_refl).
    destruct IHPAIR as (btMN & PROJ_MN & PAIR_).
    exists ((i, HighCallEvent (inl m)) :: btMN).
    cbn; split; [ rewrite <- PROJ_MN; reflexivity | ].
    econstructor; eassumption.
  - subst.
    destruct (projHigh_inv_cons PROJ_N) as (btN0 & btN' & PROJ_N0 & PROJ_N' & EN).
    subst btN.
    apply pairBiTrace_right; auto.
    specialize (IHPAIR _ _ eq_refl PROJ_N').
    destruct IHPAIR as (btMN & PROJ_MN & PAIR_).
    exists ((i, HighCallEvent (inr m)) :: btMN).
    cbn; split; [ rewrite <- PROJ_MN; reflexivity | ].
    econstructor; eassumption.
  - subst.
    destruct (projHigh_inv_cons PROJ_M) as (btM0 & btM' & PROJ_M0 & PROJ_M' & EM).
    subst btM.
    apply pairBiTrace_left; auto.
    specialize (IHPAIR _ _ PROJ_M' eq_refl).
    destruct IHPAIR as (btMN & PROJ_MN & PAIR_).
    exists ((i, HighReturnEvent r) :: btMN).
    cbn; split; [ rewrite <- PROJ_MN; reflexivity | ].
    econstructor; eassumption.
  - subst.
    destruct (projHigh_inv_cons PROJ_N) as (btN0 & btN' & PROJ_N0 & PROJ_N' & EN).
    subst btN.
    apply pairBiTrace_right; auto.
    specialize (IHPAIR _ _ eq_refl PROJ_N').
    destruct IHPAIR as (btMN & PROJ_MN & PAIR_).
    exists ((i, HighReturnEvent r) :: btMN).
    cbn; split; [ rewrite <- PROJ_MN; reflexivity | ].
    econstructor. eassumption.
Qed.

Lemma pairBiTrace_pairConcTrace {M1 M2 N1 N2 : Type -> Type}
    tM tN tMN (btM : BiTrace M1 M2) (btN : BiTrace N1 N2)
  : tM = projHigh btM ->
    tN = projHigh btN ->
    pairConcTrace tM tN tMN ->
    exists btMN,
      tMN = projHigh btMN /\
      pairBiTrace btMN btM btN.
Proof.
  apply pairBiTrace_pairConcTrace_.
Qed.

Lemma no_paired_empty_L i j : ~ is_paired L pairing_empty i j.
Proof. discriminate. Qed.

Lemma no_paired_empty_R i j : ~ is_paired R pairing_empty i j.
Proof. discriminate. Qed.

Lemma is_paired_is_unpaired p d i j d' i'
  : is_paired d p i j -> is_unpaired_ d' p i' ->
    match d, d' with
    | _, ODir.Neither => i <> i'
    | L, ODir.L | R, ODir.R => j <> i'
    | _, _ => True
    end.
Proof.
  destruct d, d'; unfold is_paired, is_unpaired; try constructor; intros H1 H2.
  all: intros <-; try congruence.
  all: apply pairing_pair_unpair in H1; congruence.
Qed.

Lemma is_paired_is_free p d i j d' i' j'
  : is_paired d p i j -> is_free d' p i' j' ->
    match d, d' with
    | L, L | R, R => i <> i' /\ j <> j'
    | _, _ => i <> i'
    end.
Proof.
  destruct d, d'; unfold is_paired, is_free; intros H1 [H2 H3]; try split.
  all: intros <-; try congruence.
  all: apply pairing_pair_unpair in H1; congruence.
Qed.

Lemma is_paired_is_paired p d i j d' i' j'
  : is_paired d p i j -> is_paired d' p i' j' ->
    match d, d' with
    | L, L | R, R => (i = i' /\ j = j') \/ (i <> i' /\ j <> j')
    | _, _ => i <> i'
    end.
Proof.
  destruct d, d'; unfold is_paired, is_free; intros H1 H2.
  - destruct (eq_dec i i'); [ subst i'; left | right ].
    + rewrite H2 in H1. injection H1; split; congruence.
    + split; auto. intros <-.
      apply pairing_pair_unpair in H1.
      apply pairing_pair_unpair in H2.
      rewrite H2 in H1; injection H1; auto.
  - intros <-. rewrite H2 in H1; discriminate.
  - intros <-. rewrite H2 in H1; discriminate.
  - destruct (eq_dec i i'); [ subst i'; left | right ].
    + rewrite H2 in H1. injection H1; split; congruence.
    + split; auto. intros <-.
      apply pairing_pair_unpair in H1.
      apply pairing_pair_unpair in H2.
      rewrite H2 in H1; injection H1; auto.
Qed.

Lemma pairing_exclusive d p i j j'
  : is_paired d p i j -> is_free d p i j' -> False.
Proof.
  destruct d; exact (fun X Y => proj1 (is_paired_is_free X Y) eq_refl).
Qed.

Lemma is_paired_add d d' p i j (FREE : is_free d' p i j) i0 j0
  : is_paired d p i0 j0 -> is_paired d (add d' p i j) i0 j0.
Proof.
  unfold is_paired, add; cbn.
  destruct (@set_or_not i i0) as [[-> ->] | [NE ->]]; auto.
  rewrite (proj1 FREE); discriminate.
Qed.

Lemma is_paired_add_eq d p i j (FREE : is_free d p i j)
  : is_paired d (add d p i j) i j.
Proof.
  unfold is_paired, add; cbn. rewrite set_eq; auto.
Qed.

Hint Resolve is_paired_add_eq : core.

Lemma is_paired_L_add_R p i j (FREE : is_free R p i j) i0 j0
  : i <> i0 ->
    is_paired L p i0 j0 -> is_paired L (add R p i j) i0 j0.
Proof.
  unfold is_paired, add; cbn. intros; rewrite set_neq; auto.
Qed.

Lemma is_paired_R_add_L p i j (FREE : is_free L p i j) i0 j0
  : i <> i0 ->
    is_paired R p i0 j0 -> is_paired R (add L p i j) i0 j0.
Proof.
  unfold is_paired, add; cbn. intros; rewrite set_neq; auto.
Qed.

Lemma is_free_add d p i j (FREE : is_free d p i j) i0 j0
  : i <> i0 -> j <> j0 ->
    is_free d p i0 j0 -> is_free d (add d p i j) i0 j0.
Proof.
  unfold is_free, add; cbn. intros; rewrite 2 set_neq; auto. congruence.
Qed.

Lemma is_free_L_add_R p i j (FREE : is_free R p i j) i0 j0
  : i <> i0 ->
    is_free L p i0 j0 -> is_free L (add R p i j) i0 j0.
Proof.
  unfold is_free, add; cbn. intros; rewrite 2 set_neq; auto. congruence.
Qed.

Lemma is_free_R_add_L p i j (FREE : is_free L p i j) i0 j0
  : i <> i0 ->
    is_free R p i0 j0 -> is_free R (add L p i j) i0 j0.
Proof.
  unfold is_free, add; cbn. intros; rewrite 2 set_neq; auto. congruence.
Qed.

Lemma is_paired_rm d d' p i j (FREE : is_paired d' p i j) i0 j0
  : match d, d' with
    | L, L | R, R => i <> i0 /\ j <> j0
    | _, _ => i <> i0
    end ->
    is_paired d p i0 j0 -> is_paired d (rm d' p i j) i0 j0.
Proof.
  unfold is_free, rm; cbn. destruct d, d'; intros; rewrite set_neq; firstorder.
Qed.

Lemma is_free_rm d d' p i j (FREE : is_paired d' p i j) i0 j0
  : is_free d p i0 j0 ->
    is_free d (rm d' p i j) i0 j0.
Proof.
  unfold is_free, rm; cbn.
  intros [].
  destruct (eq_dec i i0); [subst; rewrite set_eq | rewrite set_neq; auto]; split; auto.
  all: destruct (eq_dec (d', j) (d, j0)) as [EE | ]; [rewrite EE, set_eq | rewrite set_neq]; auto.
Qed.

Lemma is_free_rm_eq d p i j (FREE : is_paired d p i j)
  : is_free d (rm d p i j) i j.
Proof.
  unfold is_free, rm; cbn; rewrite 2 set_eq; auto.
Qed.

Hint Resolve is_free_rm_eq : core.

Lemma is_paired_add_inv d d' p i j (FREE : is_free d' p i j) i0 j0
  : is_paired d (add d' p i j) i0 j0 ->
    match d, d' with
    | L, L | R, R => (i = i0 /\ j = j0) \/ (i <> i0 /\ j <> j0 /\ is_paired d p i0 j0)
    | _, _ => i <> i0 /\ is_paired d p i0 j0
    end.
Proof.
  unfold is_paired, add; cbn.
  destruct (eq_dec i i0); [ subst; rewrite set_eq | rewrite set_neq; auto ].
  - injection 1; intros <- <-.
    destruct d'; eauto.
  - intros H. assert (d = d' -> j <> j0).
    { intros <- <-. apply pairing_pair_unpair in H. rewrite (proj2 FREE) in H. discriminate. }
    destruct d, d'; eauto.
Qed.

Lemma is_unpaired_add_inv d d' p i j (FREE : is_free d' p i j) i0
  : is_unpaired_ d (add d' p i j) i0 ->
    match d, d' with
    | ODir.Neither, _ => i <> i0
    | ODir.L, L | ODir.R, R => j <> i0
    | _, _ => True
    end /\ is_unpaired_ d p i0.
Proof.
  unfold is_unpaired, add; cbn.
  destruct d.
  - destruct d'; [ | rewrite set_neq; auto; congruence ].
    intros H; assert (j <> i0). { intros <-. rewrite set_eq in H. discriminate. }
    rewrite set_neq in H; eauto; congruence.
  - intros H; assert (i <> i0). { intros <-; rewrite set_eq in H; discriminate. }
    rewrite set_neq in H; eauto.
  - destruct d'; [ rewrite set_neq; auto; congruence | ].
    intros H; assert (j <> i0). { intros <-. rewrite set_eq in H. discriminate. }
    rewrite set_neq in H; eauto; congruence.
Qed.

Lemma is_unpaired_rm_inv d d' p i j (FREE : is_paired d' p i j) i0
  : is_unpaired_ d (rm d' p i j) i0 ->
    match d, d' with
    | ODir.Neither, _ => i = i0 \/ (i <> i0 /\ is_unpaired_ d p i0)
    | ODir.L, L | ODir.R, R => j = i0 \/ (j <> i0 /\ is_unpaired_ d p i0)
    | _, _ => is_unpaired_ d p i0
    end.
Proof.
  unfold is_unpaired_, rm; cbn; destruct d.
  - destruct d'; [ | rewrite set_neq; auto; congruence ].
    destruct (eq_dec j i0); [ auto | rewrite set_neq; auto; congruence ].
  - destruct (eq_dec i i0); [ auto | rewrite set_neq; auto ].
  - destruct d'; [ rewrite set_neq; auto; congruence | ].
    destruct (eq_dec j i0); [ auto | rewrite set_neq; auto; congruence ].
Qed.

Lemma is_free_add_inv d d' p i j (FREE : is_free d' p i j) i0 j0
  : is_free d (add d' p i j) i0 j0 ->
    match d, d' with
    | L, L | R, R => (i <> i0 /\ j <> j0 /\ is_free d p i0 j0)
    | _, _ => i <> i0 /\ is_free d p i0 j0
    end.
Proof.
  unfold is_free, add; cbn.
  destruct d, d'.
  all: destruct (eq_dec i i0); [ subst; rewrite set_eq | rewrite set_neq; auto ].
  all: intros [H1 H2]; try discriminate.
  all: try match goal with
       | [ H : set (L, _) _ _ (R, _) = _ |- _ ] => rewrite set_neq in H; [ | congruence ]; auto
       | [ H : set (R, _) _ _ (L, _) = _ |- _ ] => rewrite set_neq in H; [ | congruence ]; auto
       end.
  all: destruct (eq_dec j j0); [ subst; rewrite set_eq in H2 | rewrite set_neq in H2; auto ].
  all: discriminate + congruence.
Qed.

(**)

Lemma is_paired_rm_inv d d' p i j (FREE : is_paired d' p i j) i0 j0
  : is_paired d (rm d' p i j) i0 j0 ->
    match d, d' with
    | L, L | R, R => (i <> i0 /\ j <> j0 /\ is_paired d p i0 j0)
    | _, _ => i <> i0 /\ is_paired d p i0 j0
    end.
Proof.
  unfold is_paired, rm; cbn.
  intros H.
  assert (i <> i0); [ intros <-; rewrite set_eq in H; discriminate | ].
  rewrite set_neq in H; auto.
  destruct d, d'; auto.
  all: assert (j <> j0); [ intros <-; apply pairing_pair_unpair in H; apply pairing_pair_unpair in FREE; congruence | auto ].
Qed.

Lemma pairing_add_incl d p q i j (FREE : is_free d p i j)
  : pairing_incl q p -> pairing_incl q (add d p i j).
Proof.
  intros INCL i0 [d0 j0] PP; apply INCL in PP; apply is_paired_add; auto.
Qed.

Lemma pairing_add_incl' d p q i j (FREE : is_free d q i j)
  : is_paired d p i j -> pairing_incl q p -> pairing_incl (add d q i j) p.
Proof.
  intros PP INCL i0 [d0 j0] PQ.
  unfold add in PQ; cbn in PQ.
  destruct (eq_dec i i0); [ subst; rewrite set_eq in PQ | rewrite set_neq in PQ; auto ].
  rewrite <- PQ. apply PP.
Qed.

Lemma pairing_rm_incl d p q i j (FREE : is_paired d p i j)
  : is_free d q i j -> pairing_incl q p -> pairing_incl q (rm d p i j).
Proof.
  intros FF INCL i0 [d0 j0] PQ.
  unfold rm; cbn. destruct (eq_dec i i0); [ subst; rewrite set_eq | rewrite set_neq; auto ].
  rewrite (proj1 FF) in PQ. discriminate.
Qed.

Lemma pairing_rm_incl' d p q i j (FREE : is_paired d q i j)
  : pairing_incl q p -> pairing_incl (rm d q i j) p.
Proof.
  intros INCL i0 [d0 j0] PP.
  unfold rm in PP; cbn in PP.
  destruct (eq_dec i i0); [ subst; rewrite set_eq in PP | rewrite set_neq in PP; auto ].
  discriminate.
Qed.

Hint Resolve pairing_add_incl pairing_add_incl' pairing_rm_incl pairing_rm_incl' : core.

Lemma pairThreads_empty {M N : Type -> Type}
  : @pairThreads_ M N pairing_empty pairing_empty noThreads noThreads noThreads.
Proof.
  constructor; auto; discriminate.
Qed.

Lemma pairThreads_unwrap_L M N p q i j (thMN : Threads (M +' N)) thM thN
  : pairThreads_ p q thMN thM thN ->
    i = unwrap 0 (pairing_unpair p (L, j)) ->
    NoThread <> thM j ->
    is_paired L p i j.
Proof.
  intros TH PAIRED EthM.
  destruct (pairing_unpair _ _) as [ i0 | ] eqn:E; cbn in PAIRED; subst i.
  - apply pairing_pair_unpair in E. apply E.
  - contradiction EthM. apply TH. auto.
Qed.

Lemma pairThreads_unwrap_R M N p q i j (thMN : Threads (M +' N)) thM thN
  : pairThreads_ p q thMN thM thN ->
    i = unwrap 0 (pairing_unpair p (Dir.R, j)) ->
    NoThread <> thN j ->
    is_paired R p i j.
Proof.
  intros TH PAIRED EthM.
  destruct (pairing_unpair _ _) as [ i0 | ] eqn:E; cbn in PAIRED; subst i.
  - apply pairing_pair_unpair in E. apply E.
  - contradiction EthM. apply TH. auto.
Qed.

Lemma pairThreads_MkThread_free_low_L M N p q i j (P : is_paired L p i j)
    (thMN : Threads (M +' N)) thM thN
    R (p0 : _ R)
  : pairThreads_ p q thMN thM thN ->
    MkThread p0 = thM j ->
    is_free L q i j.
Proof.
  intros TH EthM. apply TH in P. rewrite <- EthM in P. apply P.
Qed.

Lemma pairThreads_MkThread_free_low_R M N p q i j (P : is_paired R p i j)
    (thMN : Threads (M +' N)) thM thN
    R (p0 : _ R)
  : pairThreads_ p q thMN thM thN ->
    MkThread p0 = thN j ->
    is_free Dir.R q i j.
Proof.
  intros TH EthM. apply TH in P. rewrite <- EthM in P. apply P.
Qed.

Lemma and_accum {A B : Prop} : A -> (A -> B) -> A /\ B.
Proof. auto. Qed.

Lemma pairThreads_CallingThread_paired_L M N p q i j (thMN : Threads (M +' N)) thM thN R S (k : R -> _ S)
  : pairThreads_ p q thMN thM thN ->
    i = unwrap 0 (pairing_unpair p (L, j)) ->
    CallingThread k = thM j ->
    is_paired L p i j /\ is_paired L q i j.
Proof.
  intros TH P EthM.
  apply and_accum.
  - eapply pairThreads_unwrap_L; eassumption + congruence.
  - intros X. apply TH in X. rewrite <- EthM in X. apply X.
Qed.

Lemma pairThreads_CallingThread_paired_R M N p q i j (thMN : Threads (M +' N)) thM thN R S (k : R -> _ S)
  : pairThreads_ p q thMN thM thN ->
    i = unwrap 0 (pairing_unpair p (Dir.R, j)) ->
    CallingThread k = thN j ->
    is_paired Dir.R p i j /\ is_paired Dir.R q i j.
Proof.
  intros TH P EthM.
  apply and_accum.
  - eapply pairThreads_unwrap_R; eassumption + congruence.
  - intros X. apply TH in X. rewrite <- EthM in X. apply X.
Qed.

Ltac preserve :=
  lazymatch goal with
  | [ |- pairThreads_ _ _ _ _ _ -> pairThreads_ _ _ _ _ _ ] =>
    let TH := fresh "TH" in
    let i0 := fresh "i0" in
    let j0 := fresh "j0" in
    let TH' := fresh "TH" in
    intros TH;
    let hintro := intros Hp; try (specialize (TH' Hp); my_decompose TH') in
    constructor;
      [ intros i0 j0; assert (TH' := pairThreads_L TH (i := i0) (j := j0)); hintro
      | intros i0 j0; assert (TH' := pairThreads_R TH (i := i0) (j := j0)); hintro
      | intros i0; assert (TH' := pairThreads_NoThread_MN TH (i := i0)); hintro
      | intros j0; assert (TH' := pairThreads_NoThread_M TH (j := j0)); hintro
      | intros j0; assert (TH' := pairThreads_NoThread_N TH (j := j0)); hintro
      | assert (TH' := pairThreads_subpairing TH); eauto
      ]
  end.

Ltac pairing_inv :=
  try lazymatch goal with
  | [ H : is_paired ?d (add ?d' _ _ _) _ _ |- _ ] =>
    apply is_paired_add_inv in H;
    cbn in H;
    my_decompose H
  | [ H : is_paired ?d (rm ?d' _ _ _) _ _ |- _ ] =>
    apply is_paired_rm_inv in H; cbn in H;
    my_decompose H
  | [ H : is_free ?d (add ?d' _ _ _) _ _ |- _ ] =>
    apply is_free_add_inv in H; cbn in H;
    my_decompose H
  | [ H : is_unpaired_ _ (add _ _ _ _) _ |- _ ] =>
    apply is_unpaired_add_inv in H; cbn in H;
    my_decompose H
  | [ H : is_unpaired_ _ (rm _ _ _ _) _ |- _ ] =>
    apply is_unpaired_rm_inv in H; cbn in H;
    my_decompose H
  | [ H1 : is_paired _ ?p _ _, H2 : is_free _ ?p _ _ |- _ ] =>
    let H := fresh "H" in
    assert (H := is_paired_is_free H1 H2); cbn in H; my_decompose H
  | [ H1 : is_paired _ ?p _ _, H2 : is_unpaired_ _ ?p _ |- _ ] =>
    let H := fresh "H" in
    assert (H := is_paired_is_unpaired H1 H2); cbn in H; my_decompose H
  | [ H1 : is_paired _ ?p _ _, H2 : is_paired _ ?p _ _ |- _ ] =>
    let H := fresh "H" in
    assert (H := is_paired_is_paired H1 H2); cbn in H; my_decompose H
  end.

Ltac magic :=
  rewrite ?set_eq;
  repeat lazymatch goal with
  | [ H : _ = ?y |- context [ ?y ] ] =>
    rewrite <- H; cbn
  | [ H : ?i <> ?j |- context [ set ?i _ _ ?j ] ] => rewrite (set_neq (not_eq_sym H))
  | [ H : ?j <> ?i |- context [ set ?i _ _ ?j ] ] => rewrite (set_neq H)
  | [ CONSTANT : forall _, _ -> ?x _ = ?y _, H : _ <> ?i |- context [ ?y ?i ] ] =>
    rewrite <- CONSTANT by congruence
  | [ H1 : ?A, H2 : ?A -> ?B |- ?B ] => exact (H2 H1)
  | [ |- _ /\ _ ] => split; [ | solve [ auto ]]
  end.

Theorem ConcTrace_pairProgram_ {M1 M2 N1 N2 : Type -> Type}
    (fM : M1 ~> Program M2) (fN : N1 ~> Program N2)
    tM tN thM thM_ thN thN_ (p : pairing) (q : pairing) tMN thMN
  : ThreadsSteps fM tM thM thM_ ->
    ThreadsSteps fN tN thN thN_ ->
    pairBiTrace_ p tMN tM tN ->
    pairThreads_ p q thMN thM thN ->
    exists thMN_,
      ThreadsSteps (pairProgram fM fN) tMN thMN thMN_ /\
      pairConcTrace_ q (projLow tM) (projLow tN) (projLow tMN).
Proof.
  intros STEPS_L STEPS_R TRACE TH.
  revert thMN thM thN q STEPS_L STEPS_R TH.
  induction TRACE; intros.
  - exists thMN. split; constructor; auto.
  - (* highcall L *) apply Steps_cons_inv in STEPS_L.
    destruct STEPS_L as [thM' [STEP STEPS_L]].
    destruct STEP as [STEP CONSTANT]; cbn in STEP, CONSTANT.
    destruct STEP as [EthM EthM'].
    specialize (IHTRACE (set i (MkThread (mapProgram (fun _ => inl) (fM _ m))) thMN) thM' thN q).
    forward IHTRACE; auto.
    forward IHTRACE; auto.
    forward IHTRACE.
    { revert TH; preserve; pairing_inv; magic; auto.
      - split; auto; apply (pairThreads_free_L TH FREE).
    }
    destruct IHTRACE as (thMN_ & STEPS_ & TRACE_); exists thMN_.
    split; [ | auto].
    econstructor; [ | eassumption ].
    constructor; cbn; [ | apply (@set_neq_flip) ].
    rewrite set_eq. split; [ apply (pairThreads_free_L TH FREE) | auto ].
  - (* highcall R *) apply Steps_cons_inv in STEPS_R.
    destruct STEPS_R as [thM' [STEP STEPS_R]].
    destruct STEP as [STEP CONSTANT]; cbn in STEP, CONSTANT.
    destruct STEP as [EthM EthM'].
    specialize (IHTRACE (set i (MkThread (mapProgram (fun _ => inr) (fN _ m))) thMN) thM thM' q).
    forward IHTRACE; auto.
    forward IHTRACE; auto.
    forward IHTRACE.
    { revert TH; preserve; pairing_inv; magic; auto.
      - apply (pairThreads_free_R TH FREE).
    }
    destruct IHTRACE as (thMN_ & STEPS_ & TRACE_); exists thMN_.
    split; [ | auto].
    econstructor; [ | eassumption ].
    constructor; cbn; [ | apply (@set_neq_flip) ].
    rewrite set_eq. split; auto. apply TH. rewrite is_free_unpaired_R in FREE. apply FREE.
  - (* highreturn L *) apply Steps_cons_inv in STEPS_L.
    destruct STEPS_L as [thM' [STEP STEPS_L]].
    destruct STEP as [STEP CONSTANT]; cbn in STEP, CONSTANT.
    destruct STEP as [EthM EthM'].
    specialize (IHTRACE (set i NoThread thMN) thM' thN q).
    forward IHTRACE; auto.
    forward IHTRACE; auto.
    forward IHTRACE.
    { revert TH; preserve; pairing_inv; magic; auto.
      - eauto using pairThreads_MkThread_free_low_L.
    }
    destruct IHTRACE as (thMN_ & STEPS_ & TRACE_); exists thMN_.
    split; [ | auto].
    econstructor; [ | eassumption ].
    constructor; cbn; [ | apply (@set_neq_flip) ].
    rewrite set_eq. eauto using (pairThreads_L' _ TH EthM).
  - (* highreturn R *) apply Steps_cons_inv in STEPS_R.
    destruct STEPS_R as [thM' [STEP STEPS_R]].
    destruct STEP as [STEP CONSTANT]; cbn in STEP, CONSTANT.
    destruct STEP as [EthM EthM'].
    specialize (IHTRACE (set i NoThread thMN) thM thM' q).
    forward IHTRACE; auto.
    forward IHTRACE; auto.
    forward IHTRACE.
    { revert TH; preserve; pairing_inv; magic; auto.
      - eauto using pairThreads_MkThread_free_low_R.
    }
    destruct IHTRACE as (thMN_ & STEPS_ & TRACE_); exists thMN_.
    split; [ | auto].
    econstructor; [ | eassumption ].
    constructor; cbn; [ | apply (@set_neq_flip) ].
    rewrite set_eq. eauto using (pairThreads_R' _ TH EthM).
  - (* lowcall L *) apply Steps_cons_inv in STEPS_L.
    destruct STEPS_L as [thM' [STEP STEPS_L]].
    destruct STEP as [STEP CONSTANT]; cbn in STEP, CONSTANT.
    destruct STEP as (X & k & EthM & EthM').
    assert (PAIRED : is_paired L p i j).
    { eapply pairThreads_unwrap_L; eassumption + congruence. }
    assert (FREE_q : is_free L q i j). { eauto using pairThreads_MkThread_free_low_L. }
    assert (III := pairThreads_L' PAIRED TH EthM); cbn in III.
    specialize (IHTRACE (set i (CallingThread (fun x => mapProgram (fun _ => inl) (k x))) thMN) thM' thN (add L q i j)).
    forward IHTRACE; auto.
    forward IHTRACE; auto.
    forward IHTRACE.
    { revert TH; preserve; pairing_inv; magic; auto.
      - destruct (thM j0), TH0; split; auto.
        + apply is_free_add; auto.
        + apply is_paired_add; auto.
      - destruct (thN j0), TH0; split; auto.
        + apply is_free_R_add_L; auto.
        + apply is_paired_R_add_L; auto.
    }
    destruct IHTRACE as (thMN_ & STEPS_ & TRACE_); exists thMN_. cbn.
    split; [ | eapply pairConcTrace_call_L; eauto].
    econstructor; [ | eassumption ].
    constructor; cbn; [ | apply (@set_neq_flip) ].
    rewrite set_eq. eauto.
  - (* lowcall R *) apply Steps_cons_inv in STEPS_R.
    destruct STEPS_R as [thM' [STEP STEPS_R]].
    destruct STEP as [STEP CONSTANT]; cbn in STEP, CONSTANT.
    destruct STEP as (X & k & EthM & EthM').
    assert (PAIRED : is_paired R p i j).
    { eapply pairThreads_unwrap_R; eassumption + congruence. }
    assert (FREE_q : is_free R q i j). { eauto using pairThreads_MkThread_free_low_R. }
    assert (III := pairThreads_R' PAIRED TH EthM); cbn in III.
    specialize (IHTRACE (set i (CallingThread (fun x => mapProgram (fun _ => inr) (k x))) thMN) thM thM' (add R q i j)).
    forward IHTRACE; auto.
    forward IHTRACE; auto.
    forward IHTRACE.
    { revert TH; preserve; pairing_inv; magic; auto.
      - destruct (thM j0), TH0; split; auto.
        + apply is_free_L_add_R; auto.
        + apply is_paired_L_add_R; auto.
      - destruct (thN j0), TH0; split; auto.
        + apply is_free_add; auto.
        + apply is_paired_add; auto.
    }
    destruct IHTRACE as (thMN_ & STEPS_ & TRACE_); exists thMN_. cbn.
    split; [ | eapply pairConcTrace_call_R; eauto].
    econstructor; [ | eassumption ].
    constructor; cbn; [ | apply (@set_neq_flip) ].
    rewrite set_eq. eauto.
  - (* lowreturn L *) apply Steps_cons_inv in STEPS_L.
    destruct STEPS_L as [thM' [STEP STEPS_L]].
    destruct STEP as [STEP CONSTANT]; cbn in STEP, CONSTANT.
    destruct STEP as (R & k & EthM & EthM').
    assert (PAIRED : is_paired L p i j /\ is_paired L q i j). { eauto using pairThreads_CallingThread_paired_L. }
    destruct PAIRED as [PAIRED PAIRED_q].
    assert (III : CallingThread (fun x => mapProgram (fun _ => inl) (k x)) = thMN i).
    { eapply pairThreads_L' with (2 := TH) in EthM; eauto. }
    specialize (IHTRACE (set i (MkThread (mapProgram (fun _ => inl) (k r))) thMN) thM' thN (rm L q i j)).
    forward IHTRACE; auto.
    forward IHTRACE; auto.
    forward IHTRACE.
    { revert TH; preserve; pairing_inv; magic; auto.
      - destruct (thM j0); my_decompose TH0; split; auto.
        + apply is_free_rm; auto.
        + apply is_paired_rm; auto.
      - destruct (thN j0); my_decompose TH0; split; auto.
        + apply is_free_rm; auto.
        + apply is_paired_rm; auto.
    }
    destruct IHTRACE as (thMN_ & STEPS_ & TRACE_); exists thMN_. cbn.
    split; [ | eapply pairConcTrace_return_L; eauto].
    econstructor; [ | eassumption ].
    constructor; cbn; [ | apply (@set_neq_flip) ].
    rewrite set_eq. eauto.
  - (* lowcall R *) apply Steps_cons_inv in STEPS_R.
    destruct STEPS_R as [thM' [STEP STEPS_R]].
    destruct STEP as [STEP CONSTANT]; cbn in STEP, CONSTANT.
    destruct STEP as (R & k & EthM & EthM').
    assert (PAIRED : is_paired Dir.R p i j /\ is_paired Dir.R q i j). { eauto using pairThreads_CallingThread_paired_R. }
    destruct PAIRED as [PAIRED PAIRED_q].
    assert (III : CallingThread (fun x => mapProgram (fun _ => inr) (k x)) = thMN i).
    { eapply pairThreads_R' with (2 := TH) in EthM; eauto. }
    specialize (IHTRACE (set i (MkThread (mapProgram (fun _ => inr) (k r))) thMN) thM thM' (rm Dir.R q i j)).
    forward IHTRACE; auto.
    forward IHTRACE; auto.
    forward IHTRACE.
    { revert TH; preserve; pairing_inv; magic; auto.
      - destruct (thM j0); my_decompose TH0; split; auto.
        + apply is_free_rm; auto.
        + apply is_paired_rm; auto.
      - destruct (thN j0); my_decompose TH0; split; auto.
        + apply is_free_rm; auto.
        + apply is_paired_rm; auto.
    }
    destruct IHTRACE as (thMN_ & STEPS_ & TRACE_); exists thMN_. cbn.
    split; [ | eapply pairConcTrace_return_R; eauto].
    econstructor; [ | eassumption ].
    constructor; cbn; [ | apply (@set_neq_flip) ].
    rewrite set_eq. eauto.
  - (* silent L *) apply Steps_cons_inv in STEPS_L.
    destruct STEPS_L as [thM' [STEP STEPS_L]].
    destruct STEP as [STEP CONSTANT]; cbn in STEP, CONSTANT.
    destruct STEP as (R1 & p1 & EthM & EthM').
    assert (PAIRED : is_paired L p i j).
    { eapply pairThreads_unwrap_L; eauto; rewrite <- EthM; discriminate. }
    assert (III : MkThread (Noop (mapProgram (fun _ => inl) p1)) = thMN i).
    { eapply pairThreads_L' in EthM; eauto. }
    specialize (IHTRACE (set i (MkThread (mapProgram (fun _ => inl) p1)) thMN) thM' thN q).
    forward IHTRACE; auto.
    forward IHTRACE; auto.
    forward IHTRACE.
    { revert TH; preserve; pairing_inv; revert TH0; magic; auto.
      - intros []; split; auto.
    }
    destruct IHTRACE as (thMN_ & STEPS_ & TRACE_); exists thMN_. cbn.
    split; [ | eauto ].
    econstructor; [ | eassumption ].
    constructor; cbn; [ | apply (@set_neq_flip) ].
    rewrite set_eq. eauto.
  - (* silent R *) apply Steps_cons_inv in STEPS_R.
    destruct STEPS_R as [thM' [STEP STEPS_R]].
    destruct STEP as [STEP CONSTANT]; cbn in STEP, CONSTANT.
    destruct STEP as (R1 & p1 & EthM & EthM').
    assert (PAIRED : is_paired R p i j).
    { eapply pairThreads_unwrap_R; eauto; rewrite <- EthM; discriminate. }
    assert (III : MkThread (Noop (mapProgram (fun _ => inr) p1)) = thMN i).
    { eapply pairThreads_R' in EthM; eauto. }
    specialize (IHTRACE (set i (MkThread (mapProgram (fun _ => inr) p1)) thMN) thM thM' q).
    forward IHTRACE; auto.
    forward IHTRACE; auto.
    forward IHTRACE.
    { revert TH; preserve; pairing_inv; revert TH0; magic; auto.
      - intros []; split; auto.
    }
    destruct IHTRACE as (thMN_ & STEPS_ & TRACE_); exists thMN_. cbn.
    split; [ | eauto ].
    econstructor; [ | eassumption ].
    constructor; cbn; [ | apply (@set_neq_flip) ].
    rewrite set_eq. eauto.
Qed.

Theorem ConcTrace_pairProgram {M1 M2 N1 N2 : Type -> Type}
    (fM : forall R, M1 R -> Program M2 R)
    (fN : forall R, N1 R -> Program N2 R)
    tM tN tM' tN' tMN
  : IsConcTraceOf_ tM tM' fM ->
    IsConcTraceOf_ tN tN' fN ->
    pairConcTrace tM tN tMN ->
    exists tMN',
      pairConcTrace tM' tN' tMN' /\
      IsConcTraceOf_ tMN tMN' (pairProgram fM fN).
Proof.
  intros [] [] PAIR.
  eapply pairBiTrace_pairConcTrace in PAIR; eauto.
  destruct PAIR as (btMN & EtMN & PAIR).
  eapply ConcTrace_pairProgram_ in PAIR; eauto using pairThreads_empty.
  destruct PAIR as [thMN_ [STEPS PAIR]].
  exists (projLow btMN).
  split; [ subst; apply PAIR | ].
  econstructor; eassumption + reflexivity.
Qed.

Definition mapEvent {M N} (f : M ~> N) (e : Event M) : Event N :=
  match e with
  | CallEvent m => CallEvent (f _ m)
  | ReturnEvent r => ReturnEvent r
  end.

Definition mapPendingEvent {M N} (f : M ~> N) (e : PendingEvent M) : PendingEvent N :=
  match e with
  | MkPendingEvent e => MkPendingEvent (mapEvent f e)
  | NoPendingEvent => NoPendingEvent
  end.

Definition pairPending {M N} (p : pairing) (psM : Pending M) (psN : Pending N) : Pending (M +' N) :=
  fun i =>
    match pairing_pair p i with
    | None => NoPendingEvent
    | Some (Dir.L, j) => mapPendingEvent (fun _ => inl) (psM j)
    | Some (Dir.R, j) => mapPendingEvent (fun _ => inr) (psN j)
    end.

Definition mapAtomicEvent {M N} (f : M ~> N) (e : AtomicEvent M) : AtomicEvent N :=
  match e with
  | MkAtomicEvent e => MkAtomicEvent (mapEvent f e)
  | AtomicStepEvent m r => AtomicStepEvent (f _ m) r
  end.

Definition leftAtomicTrace {M N} (p : pairing) : AtomicTrace M -> AtomicTrace (M +' N) :=
  List.map (fun ne => ( unwrap 0 (pairing_unpair p (Dir.L, fst ne))
                      , mapAtomicEvent (fun _ => inl) (snd ne))).

Definition rightAtomicTrace {M N} (p : pairing) : AtomicTrace N -> AtomicTrace (M +' N) :=
  List.map (fun ne => ( unwrap 0 (pairing_unpair p (Dir.R, fst ne))
                      , mapAtomicEvent (fun _ => inr) (snd ne))).

Definition toAtomicTrace {M} : ConcTrace M -> AtomicTrace M :=
  List.map (on_snd (fun e => MkAtomicEvent e)).

Record pairedPending {M N} (p : pairing) (psM : Pending M) (psN : Pending N) : Prop :=
  { pairedPending_L : forall i j, is_paired Dir.L p i j -> NoPendingEvent <> psM j
  ; pairedPending_R : forall i j, is_paired Dir.R p i j -> NoPendingEvent <> psN j
  ; pairedPending_free_L : forall j, is_unpaired_L p j -> NoPendingEvent = psM j
  ; pairedPending_free_R : forall j, is_unpaired_R p j -> NoPendingEvent = psN j
  }.

Lemma pairedPending_silent_L {M N} (p : pairing) (psM psM_ : Pending M) (psN : Pending N)
    (f g : Event M) n
  : MkPendingEvent f = psM n /\ MkPendingEvent g = psM_ n ->
    (forall m : nat, m <> n -> psM m = psM_ m) ->
    pairedPending p psM psN ->
    pairedPending p psM_ psN.
Proof.
  intros [En En_] CONSTANT PP.
  constructor; try apply PP.
  - intros i j Pij; destruct (eq_dec j n).
    + subst j; rewrite <- En_; discriminate.
    + rewrite <- CONSTANT by auto. eapply pairedPending_L; eassumption.
  - intros j Pij; destruct (eq_dec j n).
    + exfalso. subst j. rewrite <- (pairedPending_free_L PP Pij) in En; discriminate.
    + rewrite <- CONSTANT by auto; eapply pairedPending_free_L; eassumption.
Qed.

Lemma pairedPending_silent_R {M N} (p : pairing) (psM : Pending M) (psN psN_ : Pending N)
    (f g : Event N) n
  : MkPendingEvent f = psN n /\ MkPendingEvent g = psN_ n ->
    (forall m : nat, m <> n -> psN m = psN_ m) ->
    pairedPending p psM psN ->
    pairedPending p psM psN_.
Proof.
  intros [En En_] CONSTANT PP.
  constructor; try apply PP.
  - intros i j Pij; destruct (eq_dec j n).
    + subst j; rewrite <- En_; discriminate.
    + rewrite <- CONSTANT by auto. eapply pairedPending_R; eassumption.
  - intros j Pij; destruct (eq_dec j n).
    + exfalso. subst j. rewrite <- (pairedPending_free_R PP Pij) in En; discriminate.
    + rewrite <- CONSTANT by auto; eapply pairedPending_free_R; eassumption.
Qed.

Lemma pairPending_paired_L {M N} (p : pairing) (psM : Pending M) (psN : Pending N) i j
  : is_paired L p i j -> pairPending p psM psN i = mapPendingEvent (fun _ => inl) (psM j).
Proof.
  intros Pij. unfold pairPending. rewrite Pij. auto.
Qed.

Lemma pairPending_paired_R {M N} (p : pairing) (psM : Pending M) (psN : Pending N) i j
  : is_paired R p i j -> pairPending p psM psN i = mapPendingEvent (fun _ => inr) (psN j).
Proof.
  intros Pij. unfold pairPending. rewrite Pij. auto.
Qed.

Lemma pairPending_lookup_L {M N} (p : pairing) (psM : Pending M) (psN : Pending N) j e
  : MkPendingEvent e = psM j ->
    pairedPending p psM psN ->
    MkPendingEvent (mapEvent (fun _ => inl) e)
    = pairPending p psM psN (unwrap 0 (pairing_unpair p (Dir.L, j))).
Proof.
  intros Eps Pij. unfold pairPending.
  destruct (pairing_unpair _ _) eqn:Ej; cbn.
  - apply pairing_pair_unpair in Ej. rewrite Ej, <- Eps. auto.
  - exfalso. eapply (pairedPending_free_L Pij) in Ej. congruence.
Qed.

Lemma pairPending_lookup_R {M N} (p : pairing) (psM : Pending M) (psN : Pending N) j e
  : MkPendingEvent e = psN j ->
    pairedPending p psM psN ->
    MkPendingEvent (mapEvent (fun _ => inr) e)
    = pairPending p psM psN (unwrap 0 (pairing_unpair p (Dir.R, j))).
Proof.
  intros Eps Pij. unfold pairPending.
  destruct (pairing_unpair _ _) eqn:Ej; cbn.
  - apply pairing_pair_unpair in Ej. rewrite Ej, <- Eps. auto.
  - exfalso. eapply (pairedPending_free_R Pij) in Ej. congruence.
Qed.

Lemma pairing_split p i : is_unpaired p i \/ exists j, is_paired L p i j \/ is_paired R p i j.
Proof.
  destruct (pairing_pair p i) as [[[] j] | ] eqn:E; eauto.
Qed.

Lemma pairPending_unpaired {M N} (p : pairing) (psM : Pending M) (psN : Pending N) i
  : is_unpaired p i -> pairPending p psM psN i = NoPendingEvent.
Proof.
  intros A; unfold pairPending. rewrite A. reflexivity.
Qed.

Lemma paired_unwrap_disjoint d p i j j'
  : i <> unwrap 0 (pairing_unpair p (d, j)) ->
    is_paired d p i j' ->
    j <> j'.
Proof.
  intros A B <-. apply pairing_pair_unpair in B.
  rewrite B in A. contradiction.
Qed.

Theorem ConcTraceSpec_pairSpec_silent_L {M N : Type -> Type}
    (sM : Spec M) (sN : Spec N)
    (zN : State sN) (psN : Pending N)
    p (t0M : AtomicTrace M) (PROJ : nil = projAtomic t0M)
  : forall (zM zM' : State sM)
      (psM psM' : Pending M)
      (PP : pairedPending p psM psN)
      (STEPS_psM : Steps PendingStep t0M psM psM')
      (STEPS_zM : Steps (fun ne : nat * AtomicEvent M =>
        PendingEvalStep (s:=sM) (snd ne)) t0M zM zM'),
    Steps PendingStep (leftAtomicTrace p t0M)
      (pairPending p psM psN) (pairPending p psM' psN)
    /\ Steps (fun ne : nat * AtomicEvent (M +' N) => PendingEvalStep (s:=pairSpec sM sN) (snd ne))
         (leftAtomicTrace p t0M) (zM, zN) (zM', zN)
    /\ pairedPending p psM' psN.
Proof.
  induction t0M as [ | [n e] t0M IH ]; cbn; intros.
  - apply Steps_nil in STEPS_psM; apply Steps_nil in STEPS_zM.
    subst; split_conj; solve [ auto | constructor ].
  - apply Steps_cons_inv in STEPS_psM; apply Steps_cons_inv in STEPS_zM.
    destruct STEPS_psM as [psM_ [[STEP_ps CONSTANT_ps] STEPS_psM]].
    destruct STEPS_zM as [zM_ [STEP_z STEPS_zM]].
    destruct e as [ | R m r]; [ discriminate | ].
    cbn in *.
    specialize (IH PROJ zM_ zM' psM_ psM').
    assert (PP_ := pairedPending_silent_L STEP_ps CONSTANT_ps PP).
    forward IH; [ auto | ].
    specialize (IH STEPS_psM STEPS_zM).
    destruct IH as ( ? & ? & ? ).
    split_conj; [ econstructor .. | ] ; eauto.
    + constructor; cbn.
      * destruct STEP_ps as [S1 S2].
        eapply pairPending_lookup_L in S1; eauto.
        eapply pairPending_lookup_L in S2; eauto.
      * intros i0 NEi. destruct (@pairing_split p i0) as [ | [ j [Hj | Hj] ] ].
        { rewrite 2 pairPending_unpaired by auto. reflexivity. }
        { assert (j <> n) by eauto using paired_unwrap_disjoint, not_eq_sym.
          do 2 erewrite pairPending_paired_L by eauto.
          rewrite <- CONSTANT_ps; auto. }
        { rewrite 2 (pairPending_paired_R Hj). auto. }
    + cbn; split; auto.
Qed.

Theorem ConcTraceSpec_pairSpec_silent_R {M N : Type -> Type}
    (sM : Spec M) (sN : Spec N)
    (zM : State sM) (psM : Pending M)
    p (t0N : AtomicTrace N) (PROJ : nil = projAtomic t0N)
  : forall (zN zN' : State sN)
      (psN psN' : Pending N)
      (PP : pairedPending p psM psN)
      (STEPS_psN : Steps PendingStep t0N psN psN')
      (STEPS_zN : Steps (fun ne : nat * AtomicEvent N =>
        PendingEvalStep (s:=sN) (snd ne)) t0N zN zN'),
    Steps PendingStep (rightAtomicTrace p t0N)
      (pairPending p psM psN) (pairPending p psM psN')
    /\ Steps (fun ne : nat * AtomicEvent (M +' N) => PendingEvalStep (s:=pairSpec sM sN) (snd ne))
         (rightAtomicTrace p t0N) (zM, zN) (zM, zN')
    /\ pairedPending p psM psN'.
Proof.
  induction t0N as [ | [n e] t0N IH ]; cbn; intros.
  - apply Steps_nil in STEPS_psN; apply Steps_nil in STEPS_zN.
    subst; split_conj; solve [ auto | constructor ].
  - apply Steps_cons_inv in STEPS_psN; apply Steps_cons_inv in STEPS_zN.
    destruct STEPS_psN as [psN_ [[STEP_ps CONSTANT_ps] STEPS_psN]].
    destruct STEPS_zN as [zN_ [STEP_z STEPS_zN]].
    destruct e as [ | R m r]; [ discriminate | ].
    cbn in *.
    specialize (IH PROJ zN_ zN' psN_ psN').
    assert (PP_ := pairedPending_silent_R STEP_ps CONSTANT_ps PP).
    forward IH; [ solve [ auto ] | ].
    specialize (IH STEPS_psN STEPS_zN).
    destruct IH as (? & ? & ?).
    split_conj; [ econstructor .. | ] ; eauto.
    + constructor; cbn.
      * destruct STEP_ps as [S1 S2].
        eapply pairPending_lookup_R in S1; eauto.
        eapply pairPending_lookup_R in S2; eauto.
      * intros i0 NEi. destruct (@pairing_split p i0) as [ | [ j [Hj | Hj] ] ].
        { rewrite 2 pairPending_unpaired by auto. reflexivity. }
        { rewrite 2 (pairPending_paired_L Hj). auto. }
        { assert (j <> n) by eauto using paired_unwrap_disjoint, not_eq_sym.
          do 2 erewrite pairPending_paired_R by eauto.
          rewrite <- CONSTANT_ps; auto. }
    + cbn; split; auto.
Qed.

Lemma proj_leftAtomicTrace_nil {M N} p t
  : nil = projAtomic t ->
    mapfilter projAtomicEvent (leftAtomicTrace (M := M) (N := N) p t) = nil.
Proof.
  induction t as [ | [ ? [] ] ]; cbn; auto. discriminate.
Qed.

Lemma proj_rightAtomicTrace_nil {M N} p t
  : nil = projAtomic t ->
    mapfilter projAtomicEvent (rightAtomicTrace (M := M) (N := N) p t) = nil.
Proof.
  induction t as [ | [ ? [] ] ]; cbn; auto. discriminate.
Qed.

Lemma projAtomic_cons_inv {M} (e : ThreadEvent M) t t0
  : e :: t = projAtomic t0 ->
    exists t01 e' t02,
      t0 = t01 ++ e' :: t02 /\
      nil = projAtomic t01 /\
      Some e = projAtomicEvent e' /\
      t = projAtomic t02.
Proof.
  intros W.
  induction t0 as [ | [i [e'| X m r]] t0 IH ]; cbn; discriminate + auto.
  - injection W; intros -> ->.
    exists nil, (i, MkAtomicEvent e'), t0.
    auto.
  - specialize (IH W).
    destruct IH as (t01 & e' & t02 & H & ? & ? & ?).
    rewrite H.
    exists ((i, AtomicStepEvent m r) :: t01), e', t02.
    auto.
Qed.

Lemma pairedPending_add_L {M N} p (psM psM_ : Pending M) (psN : Pending N) i j (_ : is_free Dir.L p i j) e
  : MkPendingEvent e = psM_ j ->
    (forall j', j' <> j -> psM j' = psM_ j') ->
    pairedPending p psM psN -> pairedPending (add Dir.L p i j) psM_ psN.
Proof.
  intros Eps CONSTANT PP.
  constructor.
  - intros i0 j0 Pi0. pairing_inv.
    + rewrite <- Eps; discriminate.
    + rewrite <- CONSTANT by auto. apply (pairedPending_L PP Pi0).
  - intros i0 j0 Pi0. pairing_inv.
    apply (pairedPending_R PP Pi0).
  - intros j0 Pj0. pairing_inv.
    rewrite <- CONSTANT by auto. apply (pairedPending_free_L PP Pj0).
  - intros j0 Pj0. pairing_inv.
    apply (pairedPending_free_R PP Pj0).
Qed.

Lemma pairedPending_add_R {M N} p (psM : Pending M) (psN psN_ : Pending N) i j (_ : is_free Dir.R p i j) e
  : MkPendingEvent e = psN_ j ->
    (forall j', j' <> j -> psN j' = psN_ j') ->
    pairedPending p psM psN -> pairedPending (add Dir.R p i j) psM psN_.
Proof.
  intros Eps CONSTANT PP.
  constructor.
  - intros i0 j0 Pi0. pairing_inv.
    apply (pairedPending_L PP Pi0).
  - intros i0 j0 Pi0. pairing_inv.
    + rewrite <- Eps; discriminate.
    + rewrite <- CONSTANT by auto. apply (pairedPending_R PP Pi0).
  - intros j0 Pj0. pairing_inv.
    apply (pairedPending_free_L PP Pj0).
  - intros j0 Pj0. pairing_inv.
    rewrite <- CONSTANT by auto. apply (pairedPending_free_R PP Pj0).
Qed.

Lemma pairedPending_rm_L {M N} p (psM psM_ : Pending M) (psN : Pending N) i j (_ : is_paired Dir.L p i j)
  : NoPendingEvent = psM_ j ->
    (forall j', j' <> j -> psM j' = psM_ j') ->
    pairedPending p psM psN -> pairedPending (rm Dir.L p i j) psM_ psN.
Proof.
  intros Eps CONSTANT PP.
  constructor.
  - intros i0 j0 Pi0. pairing_inv.
    rewrite <- CONSTANT by auto. apply (pairedPending_L PP Pi0).
  - intros i0 j0 Pi0. pairing_inv.
    apply (pairedPending_R PP Pi0).
  - intros j0 Pj0. pairing_inv.
    + auto.
    + rewrite <- CONSTANT by auto. apply (pairedPending_free_L PP Pj0).
  - intros j0 Pj0. pairing_inv.
    apply (pairedPending_free_R PP Pj0).
Qed.

Lemma pairedPending_rm_R {M N} p (psM : Pending M) (psN psN_ : Pending N) i j (_ : is_paired Dir.R p i j)
  : NoPendingEvent = psN_ j ->
    (forall j', j' <> j -> psN j' = psN_ j') ->
    pairedPending p psM psN -> pairedPending (rm Dir.R p i j) psM psN_.
Proof.
  intros Eps CONSTANT PP.
  constructor.
  - intros i0 j0 Pi0. pairing_inv.
    apply (pairedPending_L PP Pi0).
  - intros i0 j0 Pi0. pairing_inv.
    rewrite <- CONSTANT by auto. apply (pairedPending_R PP Pi0).
  - intros j0 Pj0. pairing_inv.
    apply (pairedPending_free_L PP Pj0).
  - intros j0 Pj0. pairing_inv.
    + auto.
    + rewrite <- CONSTANT by auto. apply (pairedPending_free_R PP Pj0).
Qed.

Lemma projAtomicEvent_Some_inv {M} e0 j (e : _ M)
  : Some (j, e) = projAtomicEvent e0 -> e0 = (j, MkAtomicEvent e).
Proof.
  destruct e0 as [n []]; cbn; discriminate + injection 1; intros -> ->; reflexivity.
Qed.

Lemma pairing_pair_switch p r d i j (rr : evalR r d p i j) i0
    (x := pairing_pair p i0)
    (y := pairing_pair (switchR rr) i0)
  : (i <> i0 /\ x = y /\ x <> Some (d, j))
  \/ (i = i0 /\
      match r with
      | Free => x = None /\ y = Some (d, j)
      | Paired => x = Some (d, j) /\ y = None
      end).
Proof.
  unfold switchR in y. destruct r; cbn in y.
  + destruct (@set_or_not i i0) as [ [ <- SON ] | ( NEi & SON ) ].
    * right; subst x y. rewrite SON. split; auto. split; auto. apply rr.
    * left; subst x y. rewrite SON. split; auto. split; auto.
      intros EE. apply pairing_pair_unpair in EE.
      destruct rr as [ rr1 rr2 ]. rewrite rr2 in EE. discriminate.
  + destruct (@set_or_not i i0) as [ [ <- SON ] | ( NEi & SON ) ].
    * right; subst x y. rewrite SON. split; auto.
    * left; subst x y. rewrite SON. split; auto. split; auto.
      intros EE. apply pairing_pair_unpair in EE.
      apply pairing_pair_unpair in rr.
      rewrite rr in EE.
      injection EE; contradiction.
Qed.

Lemma pairPending_op_L {M N} r p i j (psM psM_ : Pending M) (psN : Pending N) (rr : evalR r L p i j)
  : (forall m, m <> j -> psM m = psM_ m) ->
    (forall m : nat, m <> i -> pairPending p psM psN m = pairPending (switchR rr) psM_ psN m).
Proof.
  intros CONSTANT m NEm. unfold pairPending.
  assert (PPS := pairing_pair_switch (rr := rr) (i0 := m)); cbn in PPS.
  destruct PPS as [(NEi & E1 & E2) | [Ei ?]]; [ | congruence ].
  rewrite <- E1.
  destruct (pairing_pair p m) as [ [[] j'] | ]; auto.
  rewrite <- CONSTANT; auto.
  intros ->; contradiction.
Qed.

Lemma pairPending_op_R {M N} r p i j (psM : Pending M) (psN psN_ : Pending N) (rr : evalR r Dir.R p i j)
  : (forall m, m <> j -> psN m = psN_ m) ->
    (forall m : nat, m <> i -> pairPending p psM psN m = pairPending (switchR rr) psM psN_ m).
Proof.
  intros CONSTANT m NEm. unfold pairPending.
  assert (PPS := pairing_pair_switch (rr := rr) (i0 := m)); cbn in PPS.
  destruct PPS as [(NEi & E1 & E2) | [Ei ?]]; [ | congruence ].
  rewrite <- E1.
  destruct (pairing_pair p m) as [ [[] j'] | ]; auto.
  rewrite <- CONSTANT; auto.
  intros ->; contradiction.
Qed.

Lemma is_unpaired_rm d p i j (_ : is_paired d p i j)
  : is_unpaired (rm d p i j) i.
Proof.
  red. unfold rm. cbn. rewrite set_eq. reflexivity.
Qed.

Theorem ConcTraceSpec_pairSpec_ {M N : Type -> Type}
    (sM : Spec M) (sN : Spec N)
    p
    (tM : ConcTrace M) (tN : ConcTrace N) (tMN : ConcTrace (M +' N))
  : pairConcTrace_ p tM tN tMN ->
  forall
    (zM : State sM) (zM' : State sM)
    (zN : State sN) (zN' : State sN)
    (psM psM' : Pending M) (psN psN' : Pending N)
    (PP : pairedPending p psM psN)
    (t0M : list (nat * AtomicEvent M))
    (t0N : list (nat * AtomicEvent N))
    (PROJ_tM : tM = projAtomic t0M)
    (PROJ_tN : tN = projAtomic t0N)
    (STEPS_psM : Steps PendingStep t0M psM psM')
    (STEPS_zM : Steps (fun ne : nat * AtomicEvent M =>
      PendingEvalStep (s:=sM) (snd ne)) t0M zM zM')
    (STEPS_psN : Steps PendingStep t0N psN psN')
    (STEPS_zN : Steps (fun ne : nat * AtomicEvent N =>
      PendingEvalStep (s:=sN) (snd ne)) t0N zN zN')
  , exists t0MN psMN',
      tMN = projAtomic t0MN /\
      IsConcTraceOfSpec_ t0MN (pairSpec sM sN) (pairPending p psM psN) psMN' (zM, zN) (zM', zN').
Proof.
  induction 1; intros.
  - (* nil *) subst.
    eassert (STEPS_L := ConcTraceSpec_pairSpec_silent_L PROJ_tM PP STEPS_psM STEPS_zM).
    destruct STEPS_L as (STEPS_ps_L & STEPS_z_L & PP').
    eassert (STEPS_R := ConcTraceSpec_pairSpec_silent_R PROJ_tN PP' STEPS_psN STEPS_zN).
    destruct STEPS_R as (STEPS_ps_R & STEPS_z_R & _).
    exists (leftAtomicTrace p t0M ++ rightAtomicTrace p t0N), (pairPending p psM' psN').
    split.
    + unfold projAtomic.
      rewrite mapfilter_app, proj_leftAtomicTrace_nil, proj_rightAtomicTrace_nil by auto.
      reflexivity.
    + constructor; eapply Steps_app; eauto.
  - (* call_L *)
    rewrite H in *; clear H. rewrite H0 in *; clear H0.
    apply projAtomic_cons_inv in PROJ_tM.
    destruct PROJ_tM as (t01M & [n eM] & t02M & Et0M & Et01M & EeM & EtM').
    rewrite Et0M in STEPS_psM, STEPS_zM.
    apply Steps_app_inv in STEPS_psM. apply Steps_app_inv in STEPS_zM.
    destruct STEPS_psM as (psM_ & STEPS_psM_ & STEPS_psM).
    destruct STEPS_zM as (zM_ & STEPS_zM_ & STEPS_zM).
    apply Steps_cons_inv in STEPS_psM. apply Steps_cons_inv in STEPS_zM.
    destruct STEPS_psM as (psM__ & STEPS_psM__ & STEPS_psM).
    destruct STEPS_zM as (zM__ & STEPS_zM__ & STEPS_zM).
    assert (STEPS_L := ConcTraceSpec_pairSpec_silent_L (zN := zN) (psN := psN) Et01M PP STEPS_psM_ STEPS_zM_).
    destruct STEPS_L as (STEPS_ps_L & STEPS_z_L & PP').
    destruct STEPS_psM__ as [STEPS_psM__ CONSTANT].
    cbn in CONSTANT.
    apply projAtomicEvent_Some_inv in EeM. injection EeM; clear EeM; intros -> ->.
    cbn in STEPS_zM__, STEPS_psM__.
    destruct STEPS_psM__ as [En En_].
    specialize (IHpairConcTrace_ zM__ zM' zN zN' psM__ psM' psN psN').
    forward IHpairConcTrace_. { eauto using @pairedPending_add_L. }
    specialize (IHpairConcTrace_ _ _ EtM' PROJ_tN).
    forward IHpairConcTrace_; [ eauto | ].
    forward IHpairConcTrace_; [ eauto | ].
    forward IHpairConcTrace_; [ eauto | ].
    forward IHpairConcTrace_; [ eauto | ].
    destruct IHpairConcTrace_ as (t0MN & psMN' & EtMN' & [IHSTEPS_ps IHSTEPS_z]).
    exists (leftAtomicTrace p t01M ++ (i, MkAtomicEvent (CallEvent (inl m))) :: t0MN), psMN'.
    split; [ | constructor ].
    + rewrite mapfilter_app. rewrite proj_leftAtomicTrace_nil by auto; cbn. f_equal; auto.
    + eapply Steps_app; eauto. econstructor; eauto.
      constructor; cbn.
      { split; auto.
        { rewrite pairPending_unpaired by (eapply is_free_unpaired_L; eauto).
          reflexivity. }
        { erewrite pairPending_paired_L by eauto. rewrite <- En_; cbn. reflexivity. }
      }
      { apply pairPending_op_L. assumption. }
    + eapply Steps_app; eauto. econstructor; eauto.
      cbn. subst zM_. auto.
  - (* call_R *)
    rewrite H in *; clear H. rewrite H0 in *; clear H0.
    apply projAtomic_cons_inv in PROJ_tN.
    destruct PROJ_tN as (t01N & [n eN] & t02N & Et0N & Et01N & EeN & EtN').
    rewrite Et0N in STEPS_psN, STEPS_zN.
    apply Steps_app_inv in STEPS_psN. apply Steps_app_inv in STEPS_zN.
    destruct STEPS_psN as (psN_ & STEPS_psN_ & STEPS_psN).
    destruct STEPS_zN as (zN_ & STEPS_zN_ & STEPS_zN).
    apply Steps_cons_inv in STEPS_psN. apply Steps_cons_inv in STEPS_zN.
    destruct STEPS_psN as (psN__ & STEPS_psN__ & STEPS_psN).
    destruct STEPS_zN as (zN__ & STEPS_zN__ & STEPS_zN).
    assert (STEPS_L := ConcTraceSpec_pairSpec_silent_R (zM := zM) (psM := psM) Et01N PP STEPS_psN_ STEPS_zN_).
    destruct STEPS_L as (STEPS_ps_L & STEPS_z_L & PP').
    destruct STEPS_psN__ as [STEPS_psN__ CONSTANT].
    cbn in CONSTANT.
    apply projAtomicEvent_Some_inv in EeN. injection EeN; clear EeN; intros -> ->.
    cbn in STEPS_zN__, STEPS_psN__.
    destruct STEPS_psN__ as [En En_].
    specialize (IHpairConcTrace_ zM zM' zN__ zN' psM psM' psN__ psN').
    forward IHpairConcTrace_. { eauto using @pairedPending_add_R. }
    specialize (IHpairConcTrace_ _ _ PROJ_tM EtN').
    forward IHpairConcTrace_; [ eauto | ].
    forward IHpairConcTrace_; [ eauto | ].
    forward IHpairConcTrace_; [ eauto | ].
    forward IHpairConcTrace_; [ eauto | ].
    destruct IHpairConcTrace_ as (t0MN & psMN' & EtMN' & [IHSTEPS_ps IHSTEPS_z]).
    exists (rightAtomicTrace p t01N ++ (i, MkAtomicEvent (CallEvent (inr m))) :: t0MN), psMN'.
    split; [ | constructor ].
    + rewrite mapfilter_app. rewrite proj_rightAtomicTrace_nil by auto; cbn. f_equal; auto.
    + eapply Steps_app; eauto. econstructor; eauto.
      constructor; cbn.
      { split; auto.
        { rewrite pairPending_unpaired by (eapply is_free_unpaired_R; eauto).
          reflexivity. }
        { erewrite pairPending_paired_R by eauto. rewrite <- En_; cbn. reflexivity. }
      }
      { apply pairPending_op_R. assumption. }
    + eapply Steps_app; eauto. econstructor; eauto.
      cbn. subst zN_. auto.
  - (* return_L *)
    rewrite H in *; clear H. rewrite H0 in *; clear H0.
    apply projAtomic_cons_inv in PROJ_tM.
    destruct PROJ_tM as (t01M & [n eM] & t02M & Et0M & Et01M & EeM & EtM').
    rewrite Et0M in STEPS_psM, STEPS_zM.
    apply Steps_app_inv in STEPS_psM. apply Steps_app_inv in STEPS_zM.
    destruct STEPS_psM as (psM_ & STEPS_psM_ & STEPS_psM).
    destruct STEPS_zM as (zM_ & STEPS_zM_ & STEPS_zM).
    apply Steps_cons_inv in STEPS_psM. apply Steps_cons_inv in STEPS_zM.
    destruct STEPS_psM as (psM__ & STEPS_psM__ & STEPS_psM).
    destruct STEPS_zM as (zM__ & STEPS_zM__ & STEPS_zM).
    assert (STEPS_L := ConcTraceSpec_pairSpec_silent_L (zN := zN) (psN := psN) Et01M PP STEPS_psM_ STEPS_zM_).
    destruct STEPS_L as (STEPS_ps_L & STEPS_z_L & PP').
    destruct STEPS_psM__ as [STEPS_psM__ CONSTANT].
    cbn in CONSTANT.
    apply projAtomicEvent_Some_inv in EeM. injection EeM; clear EeM; intros -> ->.
    cbn in STEPS_zM__, STEPS_psM__.
    destruct STEPS_psM__ as [En En_].
    specialize (IHpairConcTrace_ zM__ zM' zN zN' psM__ psM' psN psN').
    forward IHpairConcTrace_. { eauto using @pairedPending_rm_L. }
    specialize (IHpairConcTrace_ _ _ EtM' PROJ_tN).
    forward IHpairConcTrace_; [ eauto | ].
    forward IHpairConcTrace_; [ eauto | ].
    forward IHpairConcTrace_; [ eauto | ].
    forward IHpairConcTrace_; [ eauto | ].
    destruct IHpairConcTrace_ as (t0MN & psMN' & EtMN' & [IHSTEPS_ps IHSTEPS_z]).
    exists (leftAtomicTrace p t01M ++ (i, MkAtomicEvent (ReturnEvent r)) :: t0MN), psMN'.
    split; [ | constructor ].
    + rewrite mapfilter_app. rewrite proj_leftAtomicTrace_nil by auto; cbn. f_equal; auto.
    + eapply Steps_app; eauto. econstructor; eauto.
      constructor; cbn.
      { split; auto.
        { erewrite pairPending_paired_L by eauto.
          rewrite <- En; reflexivity. }
        { erewrite pairPending_unpaired by apply is_unpaired_rm. reflexivity. }
      }
      { apply pairPending_op_L. assumption. }
    + eapply Steps_app; eauto. econstructor; eauto.
      cbn. subst zM_. auto.
  - (* return_L *)
    rewrite H in *; clear H. rewrite H0 in *; clear H0.
    apply projAtomic_cons_inv in PROJ_tN.
    destruct PROJ_tN as (t01N & [n eN] & t02N & Et0N & Et01N & EeN & EtN').
    rewrite Et0N in STEPS_psN, STEPS_zN.
    apply Steps_app_inv in STEPS_psN. apply Steps_app_inv in STEPS_zN.
    destruct STEPS_psN as (psN_ & STEPS_psN_ & STEPS_psN).
    destruct STEPS_zN as (zN_ & STEPS_zN_ & STEPS_zN).
    apply Steps_cons_inv in STEPS_psN. apply Steps_cons_inv in STEPS_zN.
    destruct STEPS_psN as (psN__ & STEPS_psN__ & STEPS_psN).
    destruct STEPS_zN as (zN__ & STEPS_zN__ & STEPS_zN).
    assert (STEPS_L := ConcTraceSpec_pairSpec_silent_R (zM := zM) (psM := psM) Et01N PP STEPS_psN_ STEPS_zN_).
    destruct STEPS_L as (STEPS_ps_L & STEPS_z_L & PP').
    destruct STEPS_psN__ as [STEPS_psN__ CONSTANT].
    cbn in CONSTANT.
    apply projAtomicEvent_Some_inv in EeN. injection EeN; clear EeN; intros -> ->.
    cbn in STEPS_zN__, STEPS_psN__.
    destruct STEPS_psN__ as [En En_].
    specialize (IHpairConcTrace_ zM zM' zN__ zN' psM psM' psN__ psN').
    forward IHpairConcTrace_. { eauto using @pairedPending_rm_R. }
    specialize (IHpairConcTrace_ _ _ PROJ_tM EtN').
    forward IHpairConcTrace_; [ eauto | ].
    forward IHpairConcTrace_; [ eauto | ].
    forward IHpairConcTrace_; [ eauto | ].
    forward IHpairConcTrace_; [ eauto | ].
    destruct IHpairConcTrace_ as (t0MN & psMN' & EtMN' & [IHSTEPS_ps IHSTEPS_z]).
    exists (rightAtomicTrace p t01N ++ (i, MkAtomicEvent (ReturnEvent r)) :: t0MN), psMN'.
    split; [ | constructor ].
    + rewrite mapfilter_app. rewrite proj_rightAtomicTrace_nil by auto; cbn. f_equal; auto.
    + eapply Steps_app; eauto. econstructor; eauto.
      constructor; cbn.
      { split; auto.
        { erewrite pairPending_paired_R by eauto.
          rewrite <- En; reflexivity. }
        { erewrite pairPending_unpaired by apply is_unpaired_rm. reflexivity. }
      }
      { apply pairPending_op_R. assumption. }
    + eapply Steps_app; eauto. econstructor; eauto.
      cbn. subst zN_. auto.
Qed.

Lemma pairedPending_empty {M N} : @pairedPending M N pairing_empty noPending noPending.
Proof.
  constructor; intros; try reflexivity.
  - contradiction (no_paired_empty_L H).
  - contradiction (no_paired_empty_R H).
Qed.

Theorem ConcTraceSpec_pairSpec {M N : Type -> Type}
    (sM : Spec M) (sN : Spec N)
    tM tN tMN
  : IsConcTraceOfSpec tM sM ->
    IsConcTraceOfSpec tN sN ->
    pairConcTrace tM tN tMN ->
    IsConcTraceOfSpec tMN (pairSpec sM sN).
Proof.
  intros [t0M zM' psM' STEPS_psM STEPS_zM PROJ_tM].
  intros [t0N zN' psN' STEPS_psN STEPS_zN PROJ_tN].
  intros PAIR.
  edestruct (@ConcTraceSpec_pairSpec_ M N) as (t0MN & psMN' & PROJ & []); try eassumption.
  - apply pairedPending_empty.
  - econstructor; eauto.
Qed.

Theorem ConcTrace_pairObject {M N : Type -> Type} (oM : Object M) (oN : Object N) tM tN tMN
  : pairConcTrace tM tN tMN ->
    IsConcTraceOf tM oM ->
    IsConcTraceOf tN oN ->
    IsConcTraceOf tMN (pairObject oM oN).
Proof.
  intros PAIRTRACE [tM' [TRACE_M TRACE_M']] [tN' [TRACE_N TRACE_N']].
  assert (TRACE_MN' := ConcTrace_pairProgram TRACE_M TRACE_N PAIRTRACE).
  destruct TRACE_MN' as [tMN' [PAIRTRACE_MN' TRACE_MN']].
  exists tMN'.
  split; auto.
  eapply ConcTraceSpec_pairSpec; try eassumption.
Qed.

Definition leftPendingEvent {M N : Type -> Type} (e : PendingEvent (M +' N)) : PendingEvent M :=
  match e with
  | MkPendingEvent (CallEvent (inl m)) => MkPendingEvent (CallEvent m)
  | MkPendingEvent (ReturnEvent r) => MkPendingEvent (ReturnEvent r)
  | _ => NoPendingEvent
  end.

Definition leftPending {M N : Type -> Type} (p : pairing) (ps : Pending (M +' N)) : Pending M :=
  fun j =>
    match pairing_unpair p (Dir.L, j) with
    | None => NoPendingEvent
    | Some i => leftPendingEvent (ps i)
    end.

Definition rightPendingEvent {M N : Type -> Type} (e : PendingEvent (M +' N)) : PendingEvent N :=
  match e with
  | MkPendingEvent (CallEvent (inr m)) => MkPendingEvent (CallEvent m)
  | MkPendingEvent (ReturnEvent r) => MkPendingEvent (ReturnEvent r)
  | _ => NoPendingEvent
  end.

Definition rightPending {M N : Type -> Type} (p : pairing) (ps : Pending (M +' N)) : Pending N :=
  fun j =>
    match pairing_unpair p (Dir.R, j) with
    | None => NoPendingEvent
    | Some i => rightPendingEvent (ps i)
    end.

Lemma unpair_add_neq r d p i j (rr : evalR r d p i j) dj
  : dj <> (d, j) -> pairing_unpair (switchR rr) dj = pairing_unpair p dj.
Proof.
  intros NEd. unfold switchR; cbn.
  destruct r; cbn.
  - rewrite set_neq by congruence; auto.
  - rewrite set_neq by congruence; auto.
Qed.

Lemma rightPending_left {M N : Type -> Type} r (p : pairing) i j (rr : evalR r Dir.L p i j)
    (ps ps_ : Pending (M +' N))
  : (forall i0, i0 <> i -> ps i0 = ps_ i0) ->
    rightPending (switchR rr) ps_ = rightPending p ps.
Proof.
  intros CONSTANT.
  apply functional_extensionality; intros j0.
  unfold rightPending.
  rewrite unpair_add_neq by discriminate.
  destruct pairing_unpair as [ i0 | ] eqn:E; auto.
  f_equal; rewrite <- CONSTANT; auto.
  apply pairing_pair_unpair in E.
  destruct r.
  - apply (is_paired_is_free E rr).
  - apply (is_paired_is_paired E rr).
Qed.

Lemma leftPending_right {M N : Type -> Type} r (p : pairing) i j (rr : evalR r Dir.R p i j)
    (ps ps_ : Pending (M +' N))
  : (forall i0, i0 <> i -> ps i0 = ps_ i0) ->
    leftPending (switchR rr) ps_ = leftPending p ps.
Proof.
  intros CONSTANT.
  apply functional_extensionality; intros j0.
  unfold leftPending.
  rewrite unpair_add_neq by discriminate.
  destruct pairing_unpair as [ i0 | ] eqn:E; auto.
  f_equal; rewrite <- CONSTANT; auto.
  apply pairing_pair_unpair in E.
  destruct r.
  - apply (is_paired_is_free E rr).
  - apply (is_paired_is_paired E rr).
Qed.

Lemma leftPending_constant {M N} (p : pairing) i j (rr : is_paired Dir.L p i j)
    (ps ps_ : Pending (M +' N))
  : (forall i0, i0 <> i -> ps i0 = ps_ i0) ->
    (forall j0, j0 <> j -> leftPending p ps j0 = leftPending p ps_ j0).
Proof.
  intros CONSTANT j0 NEj.
  unfold leftPending.
  destruct pairing_unpair eqn:E; auto.
  rewrite CONSTANT; auto.
  intros <-.
  apply pairing_pair_unpair in E.
  rewrite rr in E.
  congruence.
Qed.

Lemma rightPending_constant {M N} (p : pairing) i j (rr : is_paired Dir.R p i j)
    (ps ps_ : Pending (M +' N))
  : (forall i0, i0 <> i -> ps i0 = ps_ i0) ->
    (forall j0, j0 <> j -> rightPending p ps j0 = rightPending p ps_ j0).
Proof.
  intros CONSTANT j0 NEj.
  unfold rightPending.
  destruct pairing_unpair eqn:E; auto.
  rewrite CONSTANT; auto.
  intros <-.
  apply pairing_pair_unpair in E.
  rewrite rr in E.
  congruence.
Qed.

Lemma leftPending_constant_ {M N} (p : pairing) i j (rr : is_paired Dir.R p i j)
    (ps ps_ : Pending (M +' N))
  : (forall i0, i0 <> i -> ps i0 = ps_ i0) ->
    leftPending p ps = leftPending p ps_.
Proof.
  intros CONSTANT. apply functional_extensionality; intros j0.
  unfold leftPending.
  destruct pairing_unpair eqn:E; auto.
  rewrite CONSTANT; auto.
  intros <-.
  apply pairing_pair_unpair in E.
  rewrite rr in E.
  congruence.
Qed.

Lemma rightPending_constant_ {M N} (p : pairing) i j (rr : is_paired Dir.L p i j)
    (ps ps_ : Pending (M +' N))
  : (forall i0, i0 <> i -> ps i0 = ps_ i0) ->
    rightPending p ps = rightPending p ps_.
Proof.
  intros CONSTANT. apply functional_extensionality; intros j0.
  unfold rightPending.
  destruct pairing_unpair eqn:E; auto.
  rewrite CONSTANT; auto.
  intros <-.
  apply pairing_pair_unpair in E.
  rewrite rr in E.
  congruence.
Qed.

Lemma leftPending_add {M N : Type -> Type} r d (p : pairing) i j (rr : evalR r d p i j)
    (ps ps_ : Pending (M +' N))
  : (forall i0, i0 <> i -> ps i0 = ps_ i0) ->
    (forall j0, j0 <> j -> leftPending p ps j0 = leftPending (switchR rr) ps_ j0).
Proof.
  intros CONSTANT j0 NEj.
  unfold leftPending.
  rewrite unpair_add_neq by congruence.
  destruct pairing_unpair eqn:E; auto.
  rewrite CONSTANT; auto.
  intros <-.
  apply pairing_pair_unpair in E.
  destruct r.
  - rewrite (proj1 rr) in E. discriminate.
  - rewrite rr in E. congruence.
Qed.

Lemma rightPending_add {M N : Type -> Type} r d (p : pairing) i j (rr : evalR r d p i j)
    (ps ps_ : Pending (M +' N))
  : (forall i0, i0 <> i -> ps i0 = ps_ i0) ->
    (forall j0, j0 <> j -> rightPending p ps j0 = rightPending (switchR rr) ps_ j0).
Proof.
  intros CONSTANT j0 NEj.
  unfold rightPending.
  rewrite unpair_add_neq by congruence.
  destruct pairing_unpair eqn:E; auto.
  rewrite CONSTANT; auto.
  intros <-.
  apply pairing_pair_unpair in E.
  destruct r.
  - rewrite (proj1 rr) in E. discriminate.
  - rewrite rr in E. congruence.
Qed.

Lemma leftPending_free_L {M N : Type -> Type} (p : pairing) i j (rr : is_free Dir.L p i j)
    (ps : Pending (M +' N))
  : NoPendingEvent = leftPending p ps j.
Proof.
  unfold leftPending. rewrite (proj2 rr). reflexivity.
Qed.

Lemma leftPending_paired_L {M N : Type -> Type} (p : pairing) i j (rr : is_paired Dir.L p i j)
    (ps : Pending (M +' N))
  : leftPending p ps j = leftPendingEvent (ps i).
Proof.
  unfold leftPending. apply pairing_pair_unpair in rr. rewrite rr. reflexivity.
Qed.

Lemma rightPending_free_R {M N : Type -> Type} (p : pairing) i j (rr : is_free Dir.R p i j)
    (ps : Pending (M +' N))
  : NoPendingEvent = rightPending p ps j.
Proof.
  unfold rightPending. rewrite (proj2 rr). reflexivity.
Qed.

Lemma rightPending_paired_R {M N : Type -> Type} (p : pairing) i j (rr : is_paired Dir.R p i j)
    (ps : Pending (M +' N))
  : rightPending p ps j = rightPendingEvent (ps i).
Proof.
  unfold rightPending. apply pairing_pair_unpair in rr. rewrite rr. reflexivity.
Qed.

Record pairedPending' {M N} (p : pairing) (ps : Pending (M +' N)) : Prop :=
  { pairedPending'_L : forall i j, is_paired L p i j ->
      match ps i with
      | MkPendingEvent (CallEvent (inr _)) => False
      | _ => True
      end
  ; pairedPending'_R : forall i j, is_paired R p i j ->
      match ps i with
      | MkPendingEvent (CallEvent (inl _)) => False
      | _ => True
      end
  ; pairedPending'_free : forall i, is_unpaired p i -> NoPendingEvent = ps i
  }.

Lemma pairedPending'_paired_inv_L {M N} p (ps : Pending (M +' N)) i X (m : M X)
  : MkPendingEvent (CallEvent (inl m)) = ps i ->
    pairedPending' p ps ->
    exists j, is_paired L p i j.
Proof.
  intros H PP.
  destruct (pairing_pair p i) as [ [[] j] | ] eqn:E; try solve [ exists j; auto ].
  - exfalso. apply PP in E. rewrite <- H in E. auto.
  - exfalso. apply PP in E. rewrite <- H in E; discriminate.
Qed.

Lemma pairedPending'_paired_inv_R {M N} p (ps : Pending (M +' N)) i X (m : N X)
  : MkPendingEvent (CallEvent (inr m)) = ps i ->
    pairedPending' p ps ->
    exists j, is_paired R p i j.
Proof.
  intros H PP.
  destruct (pairing_pair p i) as [ [[] j] | ] eqn:E; try solve [ exists j; auto ].
  - exfalso. apply PP in E. rewrite <- H in E. auto.
  - exfalso. apply PP in E. rewrite <- H in E; discriminate.
Qed.

Lemma pp_add_L {M N} p (ps ps_ : Pending (M +' N)) r d i j (rr : evalR r d p i j)
  : match r, d with
    | Free, L => exists X (m : M X), MkPendingEvent (CallEvent (inl m)) = ps_ i
    | Free, R => exists X (m : N X), MkPendingEvent (CallEvent (inr m)) = ps_ i
    | Paired, _ => NoPendingEvent = ps_ i
    end ->
    (forall i0, i0 <> i -> ps i0 = ps_ i0) ->
    pairedPending' p ps -> pairedPending' (switchR rr) ps_.
Proof.
  intros Hr CONSTANT PP.
  constructor.
  - intros i0 j0 Pij0. destruct r.
    + pairing_inv; destruct d; my_decompose Pij0; my_decompose Hr.
      { rewrite <- Hr. constructor. }
      { rewrite <- CONSTANT by auto. apply PP in Pij0; auto. }
      { rewrite <- CONSTANT by auto. apply PP in Pij0; auto. }
    + pairing_inv; destruct d; my_decompose Pij0; my_decompose Hr.
      { rewrite <- CONSTANT by auto. apply PP in Pij0; auto. }
      { rewrite <- CONSTANT by auto. apply PP in Pij0; auto. }
  - intros i0 j0 Pij0. destruct r.
    + pairing_inv; destruct d; my_decompose Pij0; my_decompose Hr.
      { rewrite <- CONSTANT by auto. apply PP in Pij0; auto. }
      { rewrite <- Hr. constructor. }
      { rewrite <- CONSTANT by auto. apply PP in Pij0; auto. }
    + pairing_inv; destruct d; my_decompose Pij0; my_decompose Hr.
      { rewrite <- CONSTANT by auto. apply PP in Pij0; auto. }
      { rewrite <- CONSTANT by auto. apply PP in Pij0; auto. }
  - intros i0 Pij0. destruct r.
    + pairing_inv. my_decompose Pij0; my_decompose Hr.
      { rewrite <- CONSTANT by auto. apply PP in Pij0; auto. }
    + pairing_inv; my_decompose Hr; auto.
      { rewrite <- CONSTANT by auto. apply PP in Pij0; auto. }
Qed.

Lemma pp_constant {M N} p (ps ps_ : Pending (M +' N)) i
  : ( (NoPendingEvent = ps i /\ NoPendingEvent = ps_ i) \/
      (exists X (m : M X) (r : X),
        MkPendingEvent (CallEvent (inl m)) = ps i /\ MkPendingEvent (ReturnEvent r) = ps_ i) \/
      (exists X (m : N X) (r : X),
        MkPendingEvent (CallEvent (inr m)) = ps i /\ MkPendingEvent (ReturnEvent r) = ps_ i)
    ) ->
    (forall i0, i0 <> i -> ps i0 = ps_ i0) ->
    pairedPending' p ps -> pairedPending' p ps_.
Proof.
  intros EQ CONSTANT PP. constructor.
  - intros i0 j0 Pij0. destruct (eq_dec i i0) as [EQi | NEQi].
    + subst i. destruct EQ as [ [E E_] | [(X & m & r & E & E_) | (X & m & r & E & E_)]];
       rewrite <- E_; auto.
    + rewrite <- CONSTANT by auto. apply (pairedPending'_L PP) in Pij0; auto.
  - intros i0 j0 Pij0. destruct (eq_dec i i0) as [EQi | NEQi].
    + subst i. destruct EQ as [ [E E_] | [(X & m & r & E & E_) | (X & m & r & E & E_)]];
       rewrite <- E_; auto.
    + rewrite <- CONSTANT by auto. apply (pairedPending'_R PP) in Pij0; auto.
  - intros i0 Pi0. destruct (eq_dec i i0) as [EQi | NEQi].
    + subst i. destruct EQ as [ [E E_] | [(X & m & r & E & E_) | (X & m & r & E & E_)]];
       rewrite <- E_; auto.
      all: exfalso; apply (pairedPending'_free PP) in Pi0; congruence.
    + rewrite <- CONSTANT by auto. apply (pairedPending'_free PP) in Pi0; auto.
Qed.

Ltac solve_pp :=
  solve [ eapply pp_add_L + eapply pp_constant; congruence + eauto ].

Theorem ConcTraceSpec_pairSpec_inv_ {M N : Type -> Type} (sM : Spec M) (sN : Spec N)
    atMN
  : forall p ps ps' zM zN zM' zN' tMN tM tN,
    tMN = projAtomic atMN ->
    pairedPending' p ps ->
    pairConcTrace_ p tM tN tMN ->
    Steps PendingStep atMN ps ps' ->
    Steps (PendingEvalStep_ (pairSpec sM sN)) atMN (zM, zN) (zM', zN') ->
    exists atM atN psM' psN',
      tM = projAtomic atM /\
      tN = projAtomic atN /\
      Steps PendingStep atM (leftPending p ps) psM' /\
      Steps (PendingEvalStep_ sM) atM zM zM' /\
      Steps PendingStep atN (rightPending p ps) psN' /\
      Steps (PendingEvalStep_ sN) atN zN zN' .
Proof.
  induction atMN as [ | [n e] atMN IH ]; intros * EtMN PP PT STEPSp STEPSz.
  { (* nil *)
    exists nil, nil, (leftPending p ps), (rightPending p ps).
    apply Steps_nil in STEPSp; apply Steps_nil in STEPSz; rewrite STEPSp; inversion STEPSz.
    subst tMN; inversion PT; try discriminate.
    repeat constructor; auto. }
  (* cons *)
  eapply Steps_cons_inv in STEPSp. destruct STEPSp as [ps_ [[STEPp CONSTANT] STEPSp]].
  eapply Steps_cons_inv in STEPSz. destruct STEPSz as [z_ [STEPz STEPSz]].
  cbn in CONSTANT.
  destruct e as [ [ X [ m | m ] | X r ] | X [m | m ] r ]; cbn in STEPp, STEPz.
  - (* call L *)
    cbn in EtMN; subst tMN.
    inversion PT; try discriminate.
    inversion H0; subst; clear H0.
    revert PT H1.
    refine match H5 with
      | eq_refl => _
      end; clear m0 H5.
    intros PT H1.
    specialize (IH (add L p i j) ps_ ps' zM zN zM' zN' _ tM' tN eq_refl).
    destruct STEPp as [Sp Sp_].
    forward IH; [ solve_pp | ].
    forward IH; [ auto | ].
    forward IH; [ auto | ].
    forward IH; [ auto | ].
    my_decompose IH.
    exists ((j, MkAtomicEvent (CallEvent m)) :: atM), atN, psM', psN'.
    split_conj; cbn; try f_equal; auto.
    + econstructor; eauto. constructor; cbn; [ split | ].
      * eapply leftPending_free_L; eauto.
      * erewrite leftPending_paired_L by apply is_paired_add_eq.
        rewrite <- Sp_; cbn. reflexivity.
      * auto using leftPending_add.
    + econstructor; eauto. constructor.
    + erewrite rightPending_left in IH4; eassumption.
  - (* call R *)
    cbn in EtMN; subst tMN.
    inversion PT; try discriminate.
    inversion H0; subst; clear H0.
    revert PT H1.
    refine match H5 with
      | eq_refl => _
      end; clear m0 H5.
    intros PT H1.
    specialize (IH (add Dir.R p i j) ps_ ps' zM zN zM' zN' _ tM tN' eq_refl).
    destruct STEPp as [Sp Sp_].
    forward IH; [ solve_pp | ].
    forward IH; [ auto | ].
    forward IH; [ auto | ].
    forward IH; [ auto | ].
    my_decompose IH.
    exists atM, ((j, MkAtomicEvent (CallEvent m)) :: atN), psM', psN'.
    split_conj; cbn; try f_equal; auto.
    + erewrite leftPending_right in IH2; eassumption.
    + econstructor; eauto. constructor; cbn; [ split | ].
      * eapply rightPending_free_R; eauto.
      * erewrite rightPending_paired_R by apply is_paired_add_eq.
        rewrite <- Sp_; cbn. reflexivity.
      * auto using rightPending_add.
    + econstructor; eauto. constructor.
  - cbn in EtMN; subst tMN.
    inversion PT; try discriminate.
    + (* return L *)
      inversion H0; subst; clear H0.
      revert PT.
      refine match H5 with eq_refl => _ end; clear r0 H5. intros PT.
      specialize (IH (rm L p i j) ps_ ps' zM zN zM' zN' _ tM' tN eq_refl).
      destruct STEPp as [Sp Sp_].
      forward IH; [ solve_pp | ].
      forward IH; [ auto | ].
      forward IH; [ auto | ].
      forward IH; [ auto | ].
      my_decompose IH.
      exists ((j, MkAtomicEvent (ReturnEvent r)) :: atM), atN, psM', psN'.
      split_conj; cbn; try f_equal; auto.
      { econstructor; eauto. constructor; cbn; [ split | ].
        * erewrite leftPending_paired_L by eauto.
          rewrite <- Sp; cbn. reflexivity.
        * eapply leftPending_free_L, is_free_rm_eq.
        * eauto using leftPending_add.
      }
      { econstructor; eauto. constructor. }
      { erewrite rightPending_left in IH4; eassumption. }
    + (* return R *)
      inversion H0; subst; clear H0.
      revert PT.
      refine match H5 with eq_refl => _ end; clear r0 H5. intros PT.
      specialize (IH (rm R p i j) ps_ ps' zM zN zM' zN' _ tM tN' eq_refl).
      destruct STEPp as [Sp Sp_].
      forward IH; [ solve_pp | ].
      forward IH; [ auto | ].
      forward IH; [ auto | ].
      forward IH; [ auto | ].
      my_decompose IH.
      exists atM, ((j, MkAtomicEvent (ReturnEvent r)) :: atN), psM', psN'.
      split_conj; cbn; try f_equal; auto.
      { erewrite leftPending_right in IH2; eassumption. }
      { econstructor; eauto. constructor; cbn; [ split | ].
        * erewrite rightPending_paired_R by eauto.
          rewrite <- Sp; cbn. reflexivity.
        * eapply rightPending_free_R, is_free_rm_eq.
        * eauto using rightPending_add.
      }
      { econstructor; eauto. constructor. }
  - (* silent L *)
    cbn in EtMN; subst tMN.
    destruct STEPz as [STEPz ->].
    destruct z_ as [zM_ zN_].
    specialize (IH p ps_ ps' zM_ zN_ zM' zN' _ tM tN eq_refl).
    destruct STEPp as [Sp Sp_].
    forward IH.
    { eapply pp_constant; try eassumption. right; left. eauto. }
    forward IH; [ auto | ].
    forward IH; [ auto | ].
    forward IH; [ auto | ].
    my_decompose IH.
    assert (Pj : exists j, is_paired L p n j). { eapply pairedPending'_paired_inv_L; eauto. }
    destruct Pj as [j Pj].
    exists ((j, AtomicStepEvent m r) :: atM), atN, psM', psN'.
    split_conj; cbn; auto.
    { econstructor; eauto. constructor; cbn; [ | ].
      * erewrite 2 leftPending_paired_L by eauto.
        rewrite <- Sp, <- Sp_. cbn. auto.
      * eauto using leftPending_constant.
    }
    { econstructor; cbn; eauto. }
    { erewrite rightPending_constant_; eauto. }
  - (* silent L *)
    cbn in EtMN; subst tMN.
    destruct STEPz as [STEPz ->].
    destruct z_ as [zM_ zN_].
    specialize (IH p ps_ ps' zM_ zN_ zM' zN' _ tM tN eq_refl).
    destruct STEPp as [Sp Sp_].
    forward IH.
    { eapply pp_constant; try eassumption. right; right. eauto. }
    forward IH; [ auto | ].
    forward IH; [ auto | ].
    forward IH; [ auto | ].
    my_decompose IH.
    assert (Pj : exists j, is_paired R p n j). { eapply pairedPending'_paired_inv_R; eauto. }
    destruct Pj as [j Pj].
    exists atM, ((j, AtomicStepEvent m r) :: atN), psM', psN'.
    split_conj; cbn; auto.
    { erewrite leftPending_constant_; eauto. }
    { econstructor; eauto. constructor; cbn; [ | ].
      * erewrite 2 rightPending_paired_R by eauto.
        rewrite <- Sp, <- Sp_. cbn. auto.
      * eauto using rightPending_constant.
    }
    { econstructor; cbn; eauto. }
Qed.

Lemma pairedPending'_empty M N : pairedPending' (M := M) (N := N) pairing_empty noPending.
Proof.
  constructor; cbn; auto.
Qed.

Theorem ConcTraceSpec_pairSpec_inv {M N : Type -> Type} (sM : Spec M) (sN : Spec N) tMN tM tN
  : pairConcTrace tM tN tMN ->
    IsConcTraceOfSpec tMN (pairSpec sM sN) ->
    IsConcTraceOfSpec tM sM /\
    IsConcTraceOfSpec tN sN.
Proof.
  intros PT [t0M [zM' zN'] psM' STEPS_psM STEPS_zM PROJ_tM].
  eapply ConcTraceSpec_pairSpec_inv_ in PT; eauto using pairedPending'_empty.
  2: exact STEPS_zM.
  my_decompose PT.
  split; econstructor; eauto.
Qed.

Lemma ConcTrace_pairObject_inv {M N} (o1 : Object M) (o2 : Object N) (t : ConcTrace (PairMethod M N))
  : IsConcTraceOf t (pairObject o1 o2) ->
    exists t1 t2,
      pairConcTrace t1 t2 t /\
      IsConcTraceOf t1 o1 /\
      IsConcTraceOf t2 o2.
Proof.
  intros [t_ [Ht Ht_]].
  apply ConcTrace_pairProgram_inv in Ht.
  destruct Ht as (tM & tM_ & tN & tN_ & HM & HN & P & P_).
  eapply ConcTraceSpec_pairSpec_inv in Ht_; [ | apply P_ ].
  destruct Ht_ as (HM_ & HN_).
  exists tM, tN. split; [ | split ].
  - auto.
  - econstructor; eauto.
  - econstructor; eauto.
Qed.

(* Lemma 4.4 *)
Lemma ConcTraceIncl_pairObject {M N} (o1 o1' : Object M) (o2 o2' : Object N)
  : ConcTraceIncl o1 o1' ->
    ConcTraceIncl o2 o2' ->
    ConcTraceIncl (pairObject o1 o2) (pairObject o1' o2').
Proof.
  intros H1 H2.
  red.
  intros t Ht.
  apply ConcTrace_pairObject_inv in Ht.
  destruct Ht as (t1 & t2 & P & Ht1 & Ht2).
  apply H1 in Ht1. apply H2 in Ht2.
  eapply ConcTrace_pairObject; eauto.
Qed.
