
From Coq Require Import
  List
  Arith
  Setoid
  Morphisms
  FunctionalExtensionality.

From C4.Util Require LibTactics.

Ltac injpair := LibTactics.injpair.

From C4.Util Require Import Util.
From C4.Core Require Import
  Program
  ProgramOps
  Object
  Pair
  Threads
  Superposition
  Simulates
  BaseSimulates
  HConcSimulates
  Trace
  Trace_Basics.

Set Implicit Arguments.
Set Contextual Implicit.
Unset Strict Implicit.

Fixpoint atomicBiTrace {M} (t0 : list (nat * AtomicEvent M)) : BiTrace M M :=
  match t0 with
  | nil => nil
  | ne :: t0 =>
    let '(n, e) := ne in
    match e with
    | MkAtomicEvent e =>
      match e with
      | CallEvent m => (n, HighCallEvent m) :: (n, LowCallEvent m) :: atomicBiTrace t0
      | ReturnEvent r => (n, LowReturnEvent r) :: (n, HighReturnEvent r) :: atomicBiTrace t0
      end
    | AtomicStepEvent _ _ => atomicBiTrace t0
    end
  end.

Lemma projLow_atomicBiTrace {M} (t0 : list (nat * AtomicEvent M))
  : projLow (atomicBiTrace t0) = projAtomic t0.
Proof.
  induction t0 as [ | [? [ [] | ]] ? ]; cbn; congruence.
Qed.

Lemma projHigh_atomicBiTrace {M} (t0 : list (nat * AtomicEvent M))
  : projHigh (atomicBiTrace t0) = projAtomic t0.
Proof.
  induction t0 as [ | [? [ [] | ]] ? ]; cbn; congruence.
Qed.

Definition PendingEvent_to_serial_Thread {M} (e : PendingEvent M) : Thread M :=
  match e with
  | MkPendingEvent e =>
    match e with
    | @CallEvent _ X m => CallingThread (X := X) Return
    | ReturnEvent r => MkThread (Return r)
    end
  | NoPendingEvent => NoThread
  end.

Definition Pending_to_serial_Threads {M} (p : Pending M) : Threads M :=
  fun n => PendingEvent_to_serial_Thread (p n).

Arguments Pending_to_serial_Threads M p n /.

Definition PT_serial_rel_ {M} : PendingEvent M -> Thread M -> Prop :=
  fun p t =>
    match p with
    | NoPendingEvent => t = NoThread
    | MkPendingEvent (@CallEvent _ X _ | @ReturnEvent _ X _) => t = CallingThread (X := X) Return
    end.

Definition PT_serial_rel {M} (p : Pending M) (t : Threads M) : Prop :=
  forall n, PT_serial_rel_ (p n) (t n).

Lemma ThreadsSteps_serialObject {M} (t0 : list (nat * AtomicEvent M))
    (ps ps' : Pending M) ts
  : PT_serial_rel ps ts ->
    Steps PendingStep t0 ps ps' ->
    exists ts',
    ThreadsSteps (M:=M) (fun (Ret : Type) (m : M Ret) => x <- m; Return x)
       (atomicBiTrace t0) ts ts'.
Proof.
  revert ps ts; induction t0 as [ | [n e] t0 IH ]; intros ps ts HPT H; cbn.
  - apply Steps_nil in H. subst. eexists. constructor.
  - apply Steps_cons_inv in H.
    destruct H as [ps_ []].
    destruct e as [ [] | ].
    + destruct H; cbn in H.
      specialize (IH ps_ (set n (CallingThread (X := R) Return) ts)).
      forward IH.
      { intros i. destruct (@set_or_not n i) as [[<- ->] | [NEQ ->]].
        - rewrite <- (proj2 H). constructor.
        - rewrite <- H1; auto. }
      forward IH; [ assumption | ].
      destruct IH as [ts' ?].
      exists ts'.
      apply (StepsMore (a1 := set n (MkThread (Bind m Return)) ts)).
      { constructor; cbn.
        + specialize (HPT n). rewrite <- (proj1 H) in HPT; cbn in HPT.
          split; auto. rewrite set_eq. auto.
        + intros; rewrite set_neq; auto. }
      apply (StepsMore (a1 := set n (CallingThread (X := R) Return) ts)).
      { constructor; cbn.
        + eexists _, _. rewrite 2 set_eq. eauto.
        + intros; rewrite 2 set_neq; auto. }
      assumption.
    + destruct H; cbn in H.
      specialize (IH ps_ (set n NoThread ts)).
      forward IH.
      { intros i. destruct (@set_or_not n i) as [[<- ->] | [NEQ ->]].
        - rewrite <- (proj2 H). constructor.
        - rewrite <- H1; auto. }
      forward IH; [ assumption | ].
      destruct IH as [ts' ?].
      exists ts'.
      apply (StepsMore (a1 := set n (MkThread (Return r)) ts)).
      { constructor; cbn.
        + specialize (HPT n). rewrite <- (proj1 H) in HPT; cbn in HPT.
          eexists _, _. rewrite (set_eq (f := ts)). eauto.
        + intros; rewrite set_neq; auto. }
      apply (StepsMore (a1 := set n NoThread ts)).
      { constructor; cbn.
        + rewrite 2 set_eq. eauto.
        + intros; rewrite 2 set_neq; auto. }
      assumption.
    + destruct H; cbn in H.
      specialize (IH ps_ ts).
      forward IH.
      { intros i. specialize (HPT i). destruct (@set_or_not n i) as [[<- EE] | [NEQ EE]].
        - rewrite <- (proj2 H). rewrite <- (proj1 H) in HPT. apply HPT.
        - rewrite <- H1; auto. }
      forward IH; [ assumption | ].
      apply IH.
Qed.

Lemma PT_serial_rel_init {M} : PT_serial_rel (M := M) noPending noThreads.
Proof. constructor. Qed.

Theorem ConcTrace_serialObject {M} (s : Spec M) (t : ConcTrace M)
  : IsConcTraceOfSpec t s -> IsConcTraceOf t (serialObject s).
Proof.
  intros []. red; cbn.
  exists (projAtomic t0).
  split.
  - apply (ThreadsSteps_serialObject (@PT_serial_rel_init _)) in H. destruct H.
    eexists (atomicBiTrace t0) _.
    + subst; eauto using projHigh_atomicBiTrace.
    + eauto using projLow_atomicBiTrace.
    + eassumption.
  - eexists; eassumption + reflexivity.
Qed.
