From Coq Require Import List Equalities Arith.

From C4.Core Require Import
  Program.
From C4.App Require Import
  Maps.
From C4.Extraction Require Import
  Extraction
  ExtractionIO
  ExtractionMaps
  TMLExt.
Import ListNotations.

Import BMap.
Import ProgramOps.

Module TMLArgs <: MiniDecidableType.
  Definition t := nat.
  Definition eq_dec := eq_nat_dec.
End TMLArgs.

Module Import TML := TMLFun TMLArgs.

Module Import TML_BMap_Ext := ExtractionMaps.BMapFunExt TMLArgs TML.BMap.

Create HintDb relm discriminated.
Create HintDb relcon discriminated.

Module Type MonadContext.

Parameter M : Type -> Type.
Parameter E0 : Type -> Type.
Parameter RM : forall (E : Type -> Type) (A : Type), E A -> M A -> Prop.
Arguments RM {E A}.

End MonadContext.

Definition Key := nat.
Definition Val := nat.

Module TMLContext <: MonadContext.

Parameter M : Type -> Type.
Extract Constant M "a" => "TMLContext.M a".

Parameter read : M TS.
Extract Inlined Constant read => "TMLContext.read".

Parameter write : TS -> M unit.
Extract Inlined Constant write => "TMLContext.write".

Parameter cas : TS -> TS -> M bool.
Extract Inlined Constant cas => "TMLContext.cas".

Parameter get : Key -> M (option Val).
Extract Inlined Constant get => "TMLContext.get".

Parameter put : Key -> Val -> M unit.
Extract Inlined Constant put => "TMLContext.put".

Definition E0 : Type -> Type := @BaseMethod Val.

(* This should be an Inductive but they don't work with modules *)
Parameter RM : forall (E : Type -> Type) (A : Type), E A -> M A -> Prop.
Arguments RM {E A}.

Notation R0 := (@RM E0 _).

Parameter R_read : R0 (callGT _ Register.read) read.
Parameter R_write : forall ts, R0 (callGT _ (Register.write ts)) (write ts).
Parameter R_cas : forall ts ts', R0 (callGT _ (Register.cas ts ts')) (cas ts ts').

Parameter R_get : forall k, R0 (callMap _ (BMap.get k)) (get k).
Parameter R_put : forall k v, R0 (callMap _ (BMap.put k v)) (put k v).

Hint Resolve R_read : relm.
Hint Resolve R_write : relm.
Hint Resolve R_cas : relm.
Hint Resolve R_get : relm.
Hint Resolve R_put : relm.

End TMLContext.

Module MonadExt (Context : MonadContext).
Include Context.

Parameter bindM : forall A B, M A -> (A -> M B) -> M B.
Arguments bindM {A B}.
Extract Inlined Constant bindM => "(Prelude.>>=)".

Parameter retM : forall A, A -> M A.
Arguments retM {A}.
Extract Inlined Constant retM => "Prelude.return".

Parameter fixM0 : forall A, (M A -> M A) -> M A.
Arguments fixM0 {A}.
Extract Inlined Constant fixM0 => "Data.Function.fix".

Parameter fixM : forall A B, ((A -> M B) -> A -> M B) -> A -> M B.
Arguments fixM {A B}.
Extract Inlined Constant fixM => "Data.Function.fix".

Parameter PR : forall (E : Type -> Type) (A : Type), Program E A -> M A -> Prop.
Arguments PR {E A}.

Parameter PR_bind1 : forall (E : Type -> Type) (A B : Type)
  (m : E A) (u' : M A)
  (k : A -> Program E B) (k' : A -> M B),
  RM m u' ->
  (forall a, PR (k a) (k' a)) ->
  PR (Bind m k) (bindM u' k').

Parameter PR_bind : forall (E : Type -> Type) (A B : Type)
  (u : Program E A) (u' : M A)
  (k : A -> Program E B) (k' : A -> M B),
  PR u u' ->
  (forall a, PR (k a) (k' a)) ->
  PR (bindProgram u k) (bindM u' k').

Parameter PR_ret : forall (E : Type -> Type) (A : Type) (a : A),
  @PR E A (Return a) (retM a).

(* Kinda unsound *)
Parameter PR_fix0 : forall (E : Type -> Type) (A : Type)
  (f : Program E A -> Program E A)
  (g : M A -> M A)
  (u : Program E A),
  u = f u ->
  (forall u' v', PR u' v' -> PR (f u') (g v')) ->
  PR u (fixM0 g).
Arguments PR_fix0 {E A}.

Definition DM {E : Type -> Type} {A : Type} (u : Program E A) : Type :=
  { p : M A | PR u p }.

Extraction Implicit DM [ 1 3 ].

Definition bindDM1 {E : Type -> Type} {A B : Type}
  (m : E A) (k : A -> Program E B)
  (dm : M A) (R_m : RM m dm)
  (dk : forall a, DM (k a)) : DM (Bind m k).
Proof.
  econstructor; apply PR_bind1.
  - apply R_m.
  - apply (fun a => proj2_sig (dk a)).
Defined.

Extraction Implicit bindDM1 [ 4 5 ].

Definition bindDM {E : Type -> Type} {A B : Type}
  (m : Program E A) (k : A -> Program E B)
  (dm : DM m) (dk : forall a, DM (k a)) : DM (bindProgram m k).
Proof.
  econstructor; apply PR_bind.
  - apply (proj2_sig dm).
  - apply (fun a => proj2_sig (dk a)).
Defined.

Extraction Implicit bindDM [ 4 5 ].

Definition retDM {E : Type -> Type} {A : Type} (a : A) : @DM E A (Return a).
Proof.
  econstructor; apply PR_ret.
Defined.

Parameter fixDM0 : forall {E : Type -> Type} {A : Type} (f : Program E A -> Program E A) (u : Program E A),
    u = f u ->
    (forall (w : DM u), DM (f u)) ->
    DM u.
Extraction Implicit fixDM0 [ 2 3 4 ].
Extract Inlined Constant fixDM0 => "Data.Function.fix".

Lemma bifrob {E A} (u v : Program E A) :
  frobProgram u = frobProgram v ->
  u = v.
Proof.
  intros H.
  rewrite (frobProgramId u), (frobProgramId v).
  assumption.
Qed.

Ltac autoDM :=
  lazymatch goal with
  | [ |- DM (Return _) ] => apply retDM
  | [ |- DM (Bind _ ?k) ] =>
    let x :=
        match k with
        | (fun x => _) => fresh x
        | _ => fresh "x"
        end in
    eapply bindDM1; [ eauto with relm | intros x; try autoDM ]
  | [ |- DM (bindProgram _ ?k) ] =>
    let x :=
        match k with
        | (fun x => _) => fresh x
        | _ => fresh "x"
        end in
    eapply bindDM; [ | intros x ]; try autoDM
  | [ |- DM (match ?w with _ => _ end) ] => destruct w; try autoDM
  | [ |- DM (cofix u := ?w) ] =>
    let u := fresh "u" in
    apply (fixDM0 (fun u => w));
    [ try (apply bifrob; reflexivity)
    | match goal with
      | [ |- DM ?z -> _ ] =>
        generalize z; intros ? ?; try autoDM
      end ]
  end + eauto with relcon.

End MonadExt.

Notation METH1 := (BMapMethod Key Val) (only parsing).
Notation METH2 := (TransMethod METH1 TS) (only parsing).

(* The monad users of STM see. *)
Module STMContext <: MonadContext.

Parameter M : Type -> Type.
Extract Constant M "a" => "TMLSTM.M a".

Parameter modifying : forall A, (TS -> TMLContext.M (option A * TS)) -> M A.
Arguments modifying {A}.
Extract Inlined Constant modifying => "TMLSTM.modifying".

Definition E0 : Type -> Type := METH1.
Parameter RM : forall (E : Type -> Type) (A : Type), E A -> M A -> Prop.
Arguments RM {E A}.

End STMContext.

Module Import TMLMonadExt := MonadExt TMLContext.
Module STM := MonadExt STMContext.

Definition tml_initLoop : DM (initLoop Val).
Proof.
  unfold initLoop. autoDM.
Defined.
Hint Resolve tml_initLoop : relcon.

Definition tml_impl {R : Type} (m : METH2 R)
  : DM (@impl Val R m).
Proof.
  unfold impl; autoDM.
Defined.

Parameter R_impl : forall R (m : METH2 R), RM m (proj1_sig (tml_impl m)).
Hint Resolve R_impl : relm.

Parameter instrumented
  : forall (A : Type) (p : Program METH1 A) (p' : STM.DM p) (t : TS),
    @DM METH2 (option A * TS) (mapProgram instrumentMethod (instrumentBody A p t)).
Extraction Implicit instrumented [2].
Extract Inlined Constant instrumented => "TMLSTM.runM".

Definition atomically {R : Type} (p : Program METH1 R) (p' : STM.DM p)
  : @DM METH2 _ (instrument R p).
Proof.
  cbv beta delta [ instrument instrumentWrap instrumentWrap' instrumentEnd ].
  autoDM.
  apply instrumented, p'.
Defined.
Extraction Implicit atomically [2].

Definition stmGet : Key -> STM.M (option Val) :=
  fun k => STMContext.modifying (fun s => proj1_sig (tml_impl (liftMethod s (get k)))).

Definition stmPut : Key -> Val -> STM.M unit :=
  fun k v => STMContext.modifying (fun s => proj1_sig (tml_impl (liftMethod s (put k v)))).

Extraction "extraction/ExtractionTMLImpl.hs" atomically stmGet stmPut.

Definition trans1: Program (@BMapMethod nat nat) nat :=
  Bind (put 1 2) (fun _ =>
  Bind (get 1) (fun r =>
  Return match r with
  | Some i => i
  | _ => 0
  end)).

Definition trans2: Program (@BMapMethod nat nat) nat :=
  _ <- put 2 3;
  r <- get 2;
  Return match r with
  | Some i => i
  | _ => 0
  end.

Definition rfun (n: option nat): IO unit :=
  print
    match n with
    | Some i => i
    | _ => 0
    end.

Definition main :=
  ipar
    (tml nat)
    (pairObjTrans natRegObjTrans (bMapObjTrans nat))
    (0, fun _ => None)
    ([(instrument _ trans1, rfun);
      (instrument _ trans2, rfun)]).

Extraction "extraction/ExtractionTML.hs" main.
