From Coq Require Import
  Arith.Peano_dec
  Structures.Equalities.

From C4.Core Require Import
  Program
  ProgramOps
  Pair
  Object.
From C4.App Require Import
  Register
  Maps.

Import Nat.
Import BMap.

Inductive Abortable {M: Type -> Type}: Type -> Type :=
| abortable {R}: M R -> Abortable (option R).

Arguments Abortable : clear implicits.


Inductive Stateful (S : Type) (M : Type -> Type) : Type -> Type :=
| stateful : forall R, M R -> S -> Stateful S M (R * S)%type.

Inductive TransMethod {M TLocal} : Type -> Type :=
| init : TransMethod TLocal
| abort : TLocal -> TransMethod unit
| commit : TLocal -> TransMethod (unit + TLocal)
| liftMethod {R}: TLocal -> M R -> TransMethod (option R * TLocal).

Arguments TransMethod : clear implicits.

Definition instrumentBody {M TLocal} R (p : Program M R) :
  TLocal ->
  Program
    (Stateful TLocal (Abortable M))
    (option R * TLocal) :=
  foldProgram
    (stateful TLocal (Abortable M))
    (optionProgram (@abortable M) p).

Definition instrumentEnd {M TLocal} R
           (rt : option R * TLocal) :
  Program (TransMethod M TLocal) (option R) :=
  match rt with
  | (Some r, t) =>
    ret <- commit t;
      match ret with
      | inl _ =>
        Return (Some r)
      | inr t' =>
        _ <- abort t';
          Return None
      end
  | (None, t) =>
    _ <- abort t;
      Return None
  end.

Definition instrumentMethod {M TLocal} R
           (m : Stateful TLocal (Abortable M) R) :
  TransMethod M TLocal R :=
  let '@stateful _ _ _ m' t := m in
  let '@abortable _ A' m'' := m' in
  liftMethod t m''.

Definition instrumentWrap' {M TLocal} R
           (p : Program (Stateful TLocal (Abortable M)) (option R * TLocal)) :
  Program (TransMethod M TLocal) (option R) :=
  bindProgram (mapProgram instrumentMethod p) (instrumentEnd R).

Definition instrumentWrap {M TLocal} R
           (p : TLocal -> Program (Stateful TLocal (Abortable M)) (option R * TLocal)) :
  Program (TransMethod M TLocal) (option R) :=
  t <- init; instrumentWrap' R (p t).

Definition instrument' {M TLocal} R (p : Program M R) (t : TLocal) :
  Program (TransMethod M TLocal) (option R) :=
  instrumentWrap' R (instrumentBody R p t).

Definition instrument {M TLocal} R (p : Program M R) :
  Program (TransMethod M TLocal) (option R) :=
  instrumentWrap R (instrumentBody R p).



Module TMLFun (E : MiniDecidableType).

  Notation "a =_? b" := (eq_nat_dec a b) (at level 20).
  Notation "a <>_? b" := (negb (a =_? b)) (at level 20).

  Definition Key := E.t.
  Definition eq_key_dec := E.eq_dec.

  Definition TS := nat.

  Module BMap := BMapFun E.
  Import BMap.
  
  Section TMLFun1.

  Context (Val : Type).

  Definition base :=
    pairSpec
      (@atomicReg TS 0)
      (atomicBMap Val).

  Definition BaseState :=
    prod
      (@RegisterState TS)
      (BMapState Val).

  Definition gtState (s: BaseState) :=
    fst s.
  Definition mapState (s: BaseState) :=
    snd s.
        
  (*
  Record BaseState := {
    gtState: TS;
    mapState: BMapState
  }.
  Definition baseState :=
    Build_BaseState.

  Definition initState: BaseState :=
    baseState 0 initBMapState.

  *)  

  Definition BaseMethod :=
    PairInstruction
      (@RegisterMethod TS)
      (BMapMethod Key Val).
  Definition callGT {R} (m: @RegisterMethod nat R): BaseMethod R :=
    inl m.
  Definition callMap {R} (m: BMapMethod Key Val R): BaseMethod R :=
    inr m.

  CoFixpoint initLoop: Program BaseMethod TS :=
    lt <- callGT read;
    if ((lt mod 2) =_? 1) then
       initLoop
    else
      Return lt.

  Definition impl {R} (m: @TransMethod (BMapMethod Key Val) TS R): Program BaseMethod R :=
    match m with
    | init =>
      initLoop

    | abort lt =>
      Return tt

    | commit lt =>
      if ((lt mod 2) =_? 1) then
        _ <- callGT (write (lt + 1));
        Return (inl tt)
      else
        Return (inl tt)
               
    | liftMethod lt m =>
      match m with

      | get k =>
        v <- callMap (get k);
        gt <- callGT read;
        if (gt =_? lt) then
          Return ((Some v), lt)
        else
          Return (None, lt)

      | put k v =>
        if ((lt mod 2) =_? 0) then
          b <- callGT (cas lt (lt+1));
          if b then
            _ <- callMap (put k v);
            Return ((Some tt), lt+1)
          else
            Return (None, lt+1)
        else
          _ <- callMap (put k v);
          Return ((Some tt), lt)
      end
    end.
  
  Definition tml :=
    object base (@impl).

  End TMLFun1.
  
End TMLFun.
