(* C4: Verified Transactional Objects *)

(* Overview of the artifact as it relates to the paper.
   There are a few departures from the paper addressed in the comments here.  *)

Set Implicit Arguments.

From Coq Require Import Equalities.

From C4.Core Require Import
  Abortable Program ProgramOps Object Pair Threads Simulates Serializability HConcSimulates ThreadsFacts Trace TraceFacts.

From C4.App Require Import
  Maps SHashMap CHashMap TML.

From C4.AppExtra Require Import
  HistImpl HistImplLin.

From C4.TPred Require Import
  TP.

(* * Section 3: Concurrent objects *)

(* ** Main definitions: programs, specs, objects *)

(* Programs (aka. interaction trees) *)
(* This project was originally developed independently from the itree library,
   that's why it does not currently depend on coq-itree, even though this type is
   identical (up to cosmetic differences), and a minimal fragment of the equational
   theory was reimplemented here. *)
Print Program.

(* Implementations: what is called [Impl M N] in the paper is denoted [M ~> Program N] with
   a squiggly arrow, desugaring to [forall R, M R -> Program N R]. *)

(* Sequential specifications (state transition system labeled by methods/instructions) *)
Print Spec.

(* Objects (an implementation and a specification of its low-level interface) *)
Print Object.

(** Concurrent trace semantics *)

(* To define the trace semantics, the state transitions (Fig 4 and 5, that we
   call "operational semantics" in prose, but it's admittedly a bit of a misnamer)
   are condensed together in the step relation [ThreadStep]. The parameters of
   type [Thread N] are the current state of one thread (i.e., we're focused on one entry
   of the BigTheta mapping in the paper, associated with a fixed small-theta).
   Most of the time that thread state is a program [p], wrapped in [MkThread p], or
   [NoThread].
   - [LowCallEvent]+[LowReturnEvent] -> [CallStep] in Fig 4
   - [LowSilentEvent] -> [TauStep] in Fig 4
   - [HighCallEvent] -> [Spawn] in Fig 5
   - [HighReturnEvent] -> [Return] in Fig 5
   Relative to the paper, the [Step] rule in Fig 5 disappeared by inlining
   [CallStep]/[TauStep], and [CallStep] was split into a call and a return event to
   make it easier to reason about vertical composition (Lemma 4.5). *)
Print ThreadStep.

(* [ThreadStep] actually defines the trace semantics of an implementation only
   (the [prog] parameter), i.e., the trace predicate [IsConcTraceOf_]. *)
Print IsConcTraceOf_.

(* Separately, [PendingEvalStep] provides the steps for the trace semantics of
   a specification, i.e., the trace predicate [IsConcTraceOfSpec]. *)
Print IsConcTraceOfSpec.

(* They are combined together into [IsConcTraceOf], for which we get
   trace inclusion in [ConcTraceIncl] below. *)
Print IsConcTraceOf.

(** * Section 4: Linearizability and composition *)

(* Simulation: we define the (greatest) simulation relation [Simulates] to use
   as the canonical relation between specs. It is well-known that simulation is
   strictly subsumed by trace inclusion (easily defined with the above), but it
   remains as a historical artifact that I haven't gotten around to replace yet. *)
Print Simulates. (* The user-facing definition ([<=_SP] in the paper) *)
Print Simulates_.  (* The core coinductive type *)

(* Interpreted specification ("interp-as-spec" in the paper) *)
Notation interp_spec := coarsen. (* Again the naming is a mess *)

(* Sequential object simulation ([<=_S] in the paper): we just write
   [Simulates (interp_spec obj) spec] directly, instead of yet another notation *)

(* Concurrent refinement (trace inclusion, where traces represent concurrent executions) *)
(* [<=_C] in the paper *)
Print ConcTraceIncl.

(* Atomic objects ("atomic" in the paper) *)
Notation atomic := serialObject.

(* Linearizability *)
Notation Linearizable obj spec := (ConcTraceIncl obj (atomic spec)).

(* Lemma 4.1 *)
Lemma ConcTraceIncl_serialObject {M} (spec spec' : Spec M)
  : Simulates spec spec' -> ConcTraceIncl (serialObject spec) (serialObject spec').
Proof.  apply ConcTraceIncl_serialObject.  Qed.

(* Lemma 4.2 *)
Lemma lin_sim_lin {M} (o : Object M) (spec spec' : Spec M)
  : Linearizable o spec -> Simulates spec spec' -> Linearizable o spec'.
Proof.  apply lin_sim_lin.  Qed.

(* Lemma 4.3 is literally a specialization of Lemma 4.2 *)
Lemma interp_spec_trans {M} (o o' : Object M) (spec : Spec M)
  : Linearizable o (interp_spec o') -> Simulates (interp_spec o') spec ->
    Linearizable o spec.
Proof.  apply lin_sim_lin.  Qed.

(* Lemma 4.4 *)
Lemma ConcTraceIncl_pairObject {M N} (o1 o1' : Object M) (o2 o2' : Object N)
  : ConcTraceIncl o1 o1' ->
    ConcTraceIncl o2 o2' ->
    ConcTraceIncl (pairObject o1 o2) (pairObject o1' o2').
Proof.  apply ConcTraceIncl_pairObject. Qed.

(* Lemma 4.5 *)
Lemma ConcTraceIncl_flatten {M N} (impl : M ~> Program N) (o o' : Object N)
  : ConcTraceIncl o o' -> ConcTraceIncl (flattenObject o _ impl) (flattenObject o' _ impl).
Proof.  apply ConcTraceIncl_flatten. Qed.

(* Lemma 4.6: [|>] in the paper is [flattenProgram] here, and [interp] in the itree library,
   and all of the equations were already proved over there (as the [Handler] category equations).
   Here are lemmas roughly equivalent to the first three equations of lemma 4.6 *)
About Eutt.flattenId.        (* [f |> id = f] *)
About Eutt.flattenTrigger.   (* [id |> f = f] *)
About Eutt.flatten_flatten.  (* [|>] is associative *)

(** Section 5: Verification of linearizability *)

(* Our proof principle for linearizability (Fig 8) *)
About SimulatesCoIndN.simulatesCoInd.
(* This is used in the TML proof (Section 8) *)

(* There is also a little simplified variant [SimulatesCoInd.simulatesCoInd],
   which is the one used in CHashMap (Section 6), with fewer surprise
   obligations, but where [PPRel] does not relate states. *)

(* Relations:
   - [PDRel] is [R_P] in the paper
   - [DRel] is [R_D]
   - [PPRel] is [R_I] *)

(* Obligations:
   - There are two symmetric assumptions omitted in the paper for dealing with [Noop].
     They can probably be relaxed. Assuming excluded middle and choice, you can
     always normalize programs (remove all finite prefixes of Noops) to satisfy this
     condition. The issue is that a Program might be an infinite
     sequence of [Noop], and whether that is the case is undecidable, while the
     proof needs to decide that. Elsewhere (in the TML proof) a similar issue was
     resolved by admitting excluded middle. Again, this condition remains for
     historical reasons/haven't gotten around to fix it.
       PDRel R (Noop p1) s1 p2 s2 -> False
       PPRel R R' (Noop p1) p2 p1' p2' s1 s2 -> False
   - PPSymObl : PPRel R R' p1 p2 p1' p2' s1 s2 -> PPRel R' R p1' p2' p1 p2 s1 s2
   - InitObl : DRel (initialState (objBase impl1)) (initialState (objBase impl2))
   - CallObl (looks different from the paper because of wrappers for dealing with Noop):
       DRel s1 s2 ->
       SimulatesCoIndN.NPDRel impl1 impl2 PDRel R (objProgram impl1 R m) s1
         (objProgram impl2 R m) s2 /\
       (forall (R' : Type) (p1 : Program (ObjInstruction impl1) R')
          (p2 : Program (ObjInstruction impl2) R'),
        PDRel R' p1 s1 p2 s2 ->
        SimulatesCoIndN.NPPRel impl1 impl2 PPRel R R' (objProgram impl1 R m)
          (objProgram impl2 R m) p1 p2 s1 s2))
   - RetObl : PDRel R (Return r) s1 p2 s2 -> DRel s1 s2 -> p2 = Return r
   - StepObl :
       DRel s1 s2 ->
       PDRel R (x <- m; f x) s1 p2 s2 ->
       EvalInstruction (objBase impl1) m s1 r s1' ->
       exists
         (p2' : Program (ObjInstruction impl2) R) (s2' : State (objBase impl2)),
         ProgramSteps (objBase impl2) R p2 s2 p2' s2' /\
         DRel s1' s2' /\
         SimulatesCoIndN.NPDRel impl1 impl2 PDRel R (f r) s1' p2' s2' /\
         (forall (R0 : Type) (p1'' : Program (ObjInstruction impl1) R0)
            (p2'' : Program (ObjInstruction impl2) R0),
          PDRel R0 p1'' s1 p2'' s2 ->
          PPRel R R0 (x <- m; f x) p2 p1'' p2'' s1 s2 -> PDRel R0 p1'' s1' p2'' s2') /\
         (forall (R0 : Type) (p1'' : Program (ObjInstruction impl1) R0)
            (p2'' : Program (ObjInstruction impl2) R0),
          PDRel R0 p1'' s1 p2'' s2 ->
          PPRel R R0 (x <- m; f x) p2 p1'' p2'' s1 s2 ->
          SimulatesCoIndN.NPPRel impl1 impl2 PPRel R R0 (f r) p2' p1'' p2'' s1' s2') /\
         (forall (R0 : Type) (p11 : Program (ObjInstruction impl1) R0)
            (p21 : Program (ObjInstruction impl2) R0) (R'0 : Type)
            (p12 : Program (ObjInstruction impl1) R'0)
            (p22 : Program (ObjInstruction impl2) R'0),
          PDRel R0 p11 s1 p21 s2 ->
          PDRel R'0 p12 s1 p22 s2 ->
          PPRel R0 R'0 p11 p21 p12 p22 s1 s2 ->
          PPRel R0 R'0 p11 p21 p12 p22 s1' s2')
*)

(** Section 6: Hierarchical Verification of Linearizability *)

(** Linearizable hashmap *)

(** Implementation: [CH.impl]. Theorem: [CHashMapLin]. See below. *)

(* Instantiate parameters *)
Module CH := CHashMapFun SHashMapArgs.

(** Core implementation (Fig 10) *)
Print CH.impl.

(* More parameters (the linearizability theorem is in a module functor in the module CH). *)
Module Import SHashMapSeqRefParams.
  Import CH.

  Module EMap := Maps.EMap.EMapFun SHashMapArgs.
  Import EMap.

  (* Assumption [bmap <=_S map-spec] in Lemmas 5.1 and 5.3 *)
  (* Bucket map object is a sequential Simulates of an abstract map. *)
  Parameter bucketMapImplSeqRefAbsMap :
    forall Val,
      Simulates (interp_spec (bucketMapImpl Val)) (atomicEMap Val).
  (*  [<=_S]^^^^^^^^^^^^^^^^ [bmap]^^^^^^^^^^^^^^ [map-spec]^^^^^^ *)
End SHashMapSeqRefParams.

Module SHF := CH.SHashMap.SHashMapSeqRefFun SHashMapSeqRefParams.
Module CHF := CH.CHashMapLinFun SHashMapSeqRefParams.

(* Lemma 6.1: hash-map <=_S map-spec *)
About SHF.sHashMapSeqRefEMap.
(* atomicEMap, atomicBMap are all [map-spec] in the paper, with small differences
   in the interfaces; the most crucial one is that [EMap.EMapMethod] contains [putIfAbsent],
   for implementing [lazyMap] (see Section 9). *)

(* Lemma 6.2: conc-hash-map <=_L interp-as-spec(hash-map) *)
(* [ConcSimulates] is a variant of [ConcTraceIncl] with simulation instead of trace inclusion
   (and thus implies [ConcTraceIncl]). The paper has an extra [lock <=_L lock-spec] assumption
   which is not here because the implementation [CH.cHashMapImpl] is just defined with the lock
   specification [lock-spec]. One can always refine it with an arbitrary implementation [lock]
   by vertical composition ([ConcTraceIncl_flatten], above). *)
About CHF.CHashMapSHashMap.

(* Lemma 6.2 reformulated with [Linearizable] *)
Lemma CHashMapSHashMap_ (Val : Type) (nL nS : {n : nat | 0 < n})
  : Linearizable (CH.cHashMapImpl Val nL (CH.posMult nL nS))
      (interp_spec (CH.SHashMap.hashMapImpl Val (CHF.nBuckets nL nS))).
Proof.
  apply ConcSimulates_ConcTraceIncl, simulatesHSimulates.
  exact (@CHF.CHashMapSHashMap Val nL nS).
Qed.

(** Linearizability theorem (Corollary 6.3)  *)
Lemma CHashMapLin Val s f
  : Linearizable (CH.cHashMapImpl Val s (CH.posMult s f)) (EMap.atomicEMap Val).
Proof.
  apply ConcSimulates_ConcTraceIncl, simulatesHSimulates.
  exact (CHF.CHashMapLin _ _ _). (* This definition proves a stronger variant of linearizability. *)
Qed.

(** * Section 7: Linearizability of Vertical Compositions *)

Module Section7.

Module E.
Parameter t : Type. (* Some type of keys *)
Parameter eq_dec : forall x y : t, {x = y} + {x <> y}.
End E.

Module HIP : HistImplLinParams E.
Parameter map: Object (EMap.EMapMethod E.t nat).
Parameter mapInit: State (objBase map).
Module Import EMap := EMap.EMapFun E.
Parameter mapLin : Linearizability.Linearizable (atomicEMap nat) map.
End HIP.

Module H := HistImplLinFun E HIP.

(* Figure 11 *)
Print H.HistImpl.impl. (* Object implementing get() (inline) and inc() (calling [incLoop], just below) *)
Print H.HistImpl.incLoop.

(* Lemma 7.1 *)
About H.HistImplLin.
Print Assumptions H.HistImplLin.

End Section7.

(** * Section 8: Transactions *)

(* Program specification *)
About programSpec.

(* Abortable specification *)
Print AbortableEvalInst.  (* Transitions, Fig 12 *)
About abortableSpec.
Print Abortable.

(* Strict serializability of a transaction protocol object [obj] wrt. [spec]:
   linearizability of instrumented [obj] wrt. abortable [spec]. *)
Definition StrictSerializable {TLocal} {Instruction}
           (obj: Object (TransMethod Instruction TLocal))
           (spec : Spec Instruction) : Prop :=
  Linearizable
    (instFlatImpl obj instrument)
    (abortableSpec (programSpec spec)).

(* Instrumentation (Fig 14): interpret a transaction in terms of transactional methods *)
Print instrument.

(* [trans-obj] in the paper. *)
Print instFlatImpl.

(** Serializability of TML *)

Import TPred.TP.Inst.

(** TML core implementation (Fig 15) *)
Print TML.impl.

(** Serializability theorem (Lemma 8.2) *)
Lemma TMLStrictSerializable {V} : StrictSerializable TML.tml (TML.BMap.atomicBMap V).
Proof.
  apply ConcSimulates_ConcTraceIncl, simulatesHSimulates.
  apply TML.TMLSer.TMLStrictSerializable.
Qed.

(** * Section 9: Transactional Predication *)

(* Lazy map/locator implementation (Fig 16) *)
Print AllocatorImpl.lazyMap.

(* [pred-map] in the paper (Fig 17) *)
Print AllocatorImpl.predMap.

(* [intrumentR] in the paper.
   - instrument : Program M ~> Program (TransMethod M TLocal)
   - instrumentRight : Program (M' + M) ~> Program (M' + TransMethod M TLocal) *)
Print SerializabilityAnn.instrumentRight.

(* Final theorem: Strict serializability of a transactionally predicated map. *)
(* The theorem statement below is not very enlightening. *)
(* To follow this proof you'll have to go through the file [TPred/TP.v].
   Some notes:
   - This proof is not in the clean equational style of the paper, instead using
     manual applications of transitivity and congruence lemmas, which we hope can be
     considered "close enough".
     The main obstacles are refactoring the implementation with the set of combinators
     as described in that section (some expressions are associated differently,
     and some definitions are too inlined, making rewriting awkward), and
     setting up the [Proper] instances for generalized rewriting.
   - There are three modules
     + [GenericProof] is where most of the reasoning described in this section happens
     + [LazyMapRefine] handles the refinement of the lazy map (described in the
       Locator paragraph near the beginning of Section 9)
     + [Inst] instantiates all the assumptions (with the concurrent hash map, TML, and predMap)
   - The steps don't match 1-to-1 with the paper, but the key ones are:
     + [TPImplLin__5], matching roughly Step 5, rewriting [instrumentR] to [instrument]
     + [TPImplLin__7], for Step 7, applying the serializability of the TM object (Lemma 8.1)
     + [TPImplLin__11], applying the correctness of [pred-map]
     + [predMapCorrect], that correctness theorem (a non-concurrent simulation)
       used by [TPImplLin__11], which corresponds to Lemmas 9.1 and 9.2 together. *)
About Inst.TPImplLin.

Print Assumptions Inst.TPImplLin.
(* This theorem depends on all previous results, so we get a summary of all
   used axioms in one place. *)

(*
   - Functional extensionality (for dependent functions): [functional_extensionality_dep]
   - LEM: [classic]
   - Type theoretic axiom of choice: [choice]
   - UIP: [eq_rect_eq], [JMeq_eq]; implied by LEM
 *)

(* Axioms:
FunctionalExtensionality.functional_extensionality_dep :
forall (A : Type) (B : A -> Type) (f g : forall x : A, B x),
(forall x : A, f x = g x) -> f = g
Eqdep.Eq_rect_eq.eq_rect_eq : forall (U : Type) (p : U)
                                (Q : U -> Type) (x : Q p)
                                (h : p = p), x = eq_rect p Q x p h
Classical_Prop.classic : forall P : Prop, P \/ ~ P
Util.choice : forall (A B : Type) (R : A -> B -> Prop),
              (forall x : A, exists y : B, R x y) ->
              exists f : A -> B, forall x : A, R x (f x)
JMeq.JMeq_eq : forall (A : Type) (x y : A), JMeq.JMeq x y -> x = y
*)

(* There is some extraction code, but further work is required to obtain
   conclusive benchmark results, so the harness for extraction was omitted. *)
