Set Asymmetric Patterns.
From Coq Require Import
  Arith.Peano_dec
  Sets.Ensembles
  Sets.Image
  Logic.FunctionalExtensionality
  Structures.Equalities
  Setoid
  Morphisms.

From Paco Require Import paco.

From C4.Util Require Import
  FunOverride
  LibTactics
  MTactics
  Util.

From C4.Core Require Import
  Program
  ProgramFacts
  Superposition
  Pair
  PairFacts
  MethodSimulates
  Sig
  ProgramOps
  ProgramOpsFacts
  Simulates
  SimulatesFacts
  Threads
  LinCoInd
  Abortable
  Linearizability
  LinearizabilityTactics
  Serializability
  MapSimulates
  MethodSimulates
  MapState
  Object
  HConcSimulates
  RSimulates
  RSimulatesFacts
  Eutt.

From C4.App Require Import
  SerializabilityAnn
  Maps.

From C4.AppExtra Require Import
  Counter.

From C4.TPred Require Import
  Share
  Allocator.

(** (Void + Method) -> Method. (Unit = Void) *)
Definition removeLeft {Method}
           Ret (m : PairMethod UnitInstruction Method Ret) :
  Method Ret :=
  match m with
  | inl e => Datatypes.Empty_set_rect _ e
  | inr m' => m'
  end.

Lemma removeLeftSimulates {Method} (obj : Object Method) :
  ConcSimulates
    (pairObject unitObject obj)
    (mapObject removeLeft obj).
Proof.
  apply SimulatesCoInd2.simulatesCoInd with
  (fun R p1 p2 => programEq (mapProgram removeLeft p1) p2)
    (fun st1 st2 => st1 = (tt, st2)).
  - reflexivity.
  - intros.
    right.
    subst.
    progFrobSimplInAt H 1.
    inverts H.
    eexists.
    split.
    + reflexivity.
    + assumption.
  - intros.
    simpl.
    destruct m as [[] | ?].
    unfold mapImplProg.
    rewrite mapMapProgram.
    apply mapProgramRefl.
  - intros.
    progFrobSimplInAt H 1.
    inverts H.
    reflexivity.
  - intros.
    subst.
    destruct m as [[] | ?].
    destruct H1.
    progFrobSimplInAt H 1.
    inverts H.
    eexists _, _.
    split; [| split].
    + eapply ProgramStepImplProgramSteps.
      constructor.
      eassumption.
    + eauto.
    + destruct s1'.
      simpl in *.
      subst.
      reflexivity.
Qed.

Definition handlePairMethod {ML MR N}
  (fl : forall R, ML R -> Program N R)
  (fr : forall R, MR R -> Program N R)
  : forall R, PairMethod ML MR R -> Program N R :=
  fun _ m =>
    match m with
    | inl ml => fl _ ml
    | inr mr => fr _ mr
    end.

Definition replaceFunction {ML MR}
           (f : forall Ret, ML Ret -> Ret)
  : forall R, PairMethod ML MR R -> Program MR R :=
  handlePairMethod
    (fun _ ml => Return (f _ ml))
    (fun _ mr => Bind mr Return).

Definition replaceFunctionImpl {Method Method'}
           (f : forall Ret, Method' Ret -> Ret)
           (obj : Object Method) :
  Object (PairMethod Method' Method) :=
  flattenObject obj _ (replaceFunction f).

Section PAIRIMPLFUNCTION.

Context
  {M' M N : Type -> Type}
  (lmap : forall R, M R -> R)
  (o : Spec N)
  (h : forall R, M' R -> Program (PairInstruction M N) R)
  (s : State o).

Let i1 : Object M' :=
  {| ObjInstruction := PairInstruction M N;
     objBase := pairSpec (functionSpec lmap) o;
     objProgram := h |}.

Let i2 : Object M' :=
  {| ObjInstruction := N;
     objBase := o;
     objProgram := flattenImplProg (replaceFunction lmap) h
  |}.

Inductive PIF_eq {A}
  (p : Program (PairInstruction M N) A) : Program N A -> Prop :=
| _PIF_eq_flatten : PIF_eq p (flattenProgram (replaceFunction lmap) p)
.

Inductive zip_rel {M1 M2 : Type -> Type} (RR : IRel M1 M2)
  : list (sigT M1) -> list (sigT M2) -> Prop :=
| zip_rel_nil : zip_rel RR nil nil
| zip_rel_cons T p1 p2 t1 t2 :
  RR T p1 p2 ->
  zip_rel RR t1 t2 ->
  zip_rel RR (existT _ T p1 :: t1) (existT _ T p2 :: t2)
.

Definition _eq_threads := zip_rel (@PIF_eq).

Definition _eq_pair (x : unit * State o) (s : State o) : Prop :=
  snd x = s.

Inductive RRR : ThreadsState i1 -> ThreadsState i2 -> Prop :=
| mkRRR t1 s1 t2 s2 :
  _eq_threads t1 t2 ->
  _eq_pair s1 s2 ->
  RRR (t1, s1) (t2, s2).

Lemma pairImplFunction_ : ConcSimulates i1 i2.
Proof.
  apply (SimulatesCoInd2.simulatesCoInd i1 i2 (@PIF_eq) _eq_pair).
  - constructor.
  - intros ? ? ? [].
    right.
    eexists.
    split.
    + rewrite frobProgramId at 1. cbn.
      reflexivity.
    + constructor.
  - constructor.
  - intros ? ? ? [].
    rewrite frobProgramId at 1. cbn.
    reflexivity.
  - intros * [] SIM.
    cbn.
    destruct m; cbn.
    + intros [[] ?].
      exists (flattenProgram (replaceFunction lmap) (f (lmap _ m))) s2.
      split.
      * rewrite frobProgramId at 1. cbn.
        eapply ProgramStepStepsImplSteps.
        { constructor. }
        rewrite frobProgramId at 1. cbn.
        eapply ProgramStepImplProgramSteps.
        { constructor. }
      * repeat constructor. unfold _eq_pair in *. congruence.
    + intros [? ?].
      exists (flattenProgram (replaceFunction lmap) (f r)) (snd s1').
      split.
      * rewrite frobProgramId at 1. cbn.
        eapply ProgramStepStepsImplSteps.
        { constructor. }
        rewrite frobProgramId at 1. cbn.
        eapply ProgramStepStepsImplSteps.
        { constructor. destruct SIM. eassumption. }
        rewrite frobProgramId at 1. cbn.
        eapply ProgramStepImplProgramSteps.
        { constructor. }
      * repeat constructor.
Qed.

Lemma pairImplFunction : HConcSimulates i1 i2.
Proof.
  apply simulatesHSimulates.
  apply pairImplFunction_.
Qed.

End PAIRIMPLFUNCTION.

Import BMap.

Section FLATTENINSTRUMENT.

Context {TId Loc Val : Type}.
Context {M : Type -> Type}.
Context (f : forall X, M X -> X).

Context {R : Type}.
Let N := BMapMethod Loc Val.

Arguments stateful {_ _}.
Arguments abortable {_}.

Inductive _flattenInstrument_sim
  : Program (TransMethod N TId) (option R) -> Program (TransMethod N TId) (option R) -> Prop :=
| _flattenInstrument_core s p :
    _flattenInstrument_sim
      (flattenProgram (replaceFunction f)
         (mapProgramDep focusRight (instrumentWrap' R (foldProgram stateful (optionProgram abortable p) s))))
      (instrumentWrap' R (foldProgram stateful (optionProgram abortable (flattenProgram (replaceFunction f) p)) s))
.

Ltac step_left :=
  match goal with
  | [ |- _ ?t _ ] => rewrite (frobProgramId t)
  end.

Ltac step_right :=
  match goal with
  | [ |- _ ?u ] => rewrite (frobProgramId u)
  end.

Lemma flattenInstrument
  (fp : Program (PairMethod M (BMapMethod Loc Val)) R)
  : eutt ieq
      (flattenProgram (replaceFunction f)
        (instrumentRight R fp TId))
      (instrument R
        (flattenProgram (replaceFunction f) fp)).
Proof.
  unfold instrumentRight, instrument, instrumentWrap, instrumentBody.
  step_left; cbn. apply eutt_Noop. step_left; cbn.
  pfold. constructor; [ constructor | ].
  intros x. left.
  step_left; cbn; apply eutt_Noop.
  enough (J : forall p q,
             _flattenInstrument_sim p q -> eutt ieq p q).
  { apply J. constructor. }
  clear. pcofix SELF. intros p q. induction 1.
  - step_left; step_right; pfold.
    destruct p as [ ? [ml | mr] k | | ]; cbn.
    + constructor; left; pfold.
      step_left; cbn.
      step_right; cbn.
      constructor; right.
      apply SELF. constructor.
    + constructor; left; pfold.
      step_left; cbn.
      step_right; cbn.
      constructor; [ reflexivity | ].
      intros [ox s'].
      step_left. cbn. destruct ox.
      * step_right; cbn.
        left; pfold; constructor. right; apply SELF; constructor.
      * left. pfold. constructor.
        step_right; cbn.
        step_left; cbn. constructor.
        step_left; cbn. constructor; [ reflexivity | ].
        left. pfold.
        step_left; cbn. constructor.
        step_left; cbn. constructor.
    + constructor.
      step_left; cbn. constructor; [ reflexivity | ].
      intros x.
      left; pfold.
      destruct x; repeat (step_left; cbn; constructor).
      * constructor.
      * left; pfold; repeat (step_left; cbn; constructor).
    + constructor; right; apply SELF; constructor.
Qed.

End FLATTENINSTRUMENT.
