Set Asymmetric Patterns.
From Coq Require Import
  Arith
  Sets.Ensembles
  Sets.Image
  Logic.FunctionalExtensionality
  Structures.Equalities
  Setoid
  Morphisms.

From Paco Require Import paco.

From C4.Util Require Import
  LibTactics
  MTactics
  FunOverride
  Util.

From C4.Core Require Import
  Program
  ProgramFacts
  Object
  Superposition
  Pair
  PairFacts
  MethodSimulates
  Sig
  ProgramOps
  ProgramOpsFacts
  Simulates
  SimulatesFacts
  Threads
  LinCoInd
  Abortable
  Linearizability
  LinearizabilityTactics
  Serializability
  MethodSimulates
  MapState
  MapSimulates
  HConcSimulates
  RSimulates
  RSimulatesFacts
  Eutt
  ImplSimulates
  Trace
  TraceFacts
  ProgramSpec.

From C4.App Require Import
  SerializabilityAnn
  Maps
  TML.

From C4.AppExtra Require Import
  Counter.

From C4.TPred Require Import
  PairImplFunction
  Share
  Allocator
  AllocatorImpl.

Import Map.
Import BMap.
Import EMap.
Import SimulatesCoInd2.

(* Repackage the theorem in terms of the HashMap spec (a pure map) only *)
Module Import GenericProof.
Section GenericProof_Sec.

Context (Method : Type -> Type).
Context (Val : Type).

Context (LocMethod : Type -> Type).

Context (locatorMap : forall R, LocMethod R -> R).
Context (locatorMapInjective : injective _ Loc (locatorMap Loc)).

(* Implementation of the transactional interface [Method]. *)
Context (predMap :
    forall R,
      Method R ->
      Program (PairMethod LocMethod (BMapMethod Loc Val)) R).
Context (topSpec : Spec Method).

(* ASSUMPTION: [predMap] is "correct" wrt. [topSpec] when
   it implements the sequential spec  *)
Context (predMapCorrect :
    forall lmap : forall Ret, LocMethod Ret -> Ret,
      injective _ Loc (lmap Loc) ->
      Simulates
        (mapSpec
           (programSpec (LocBMap.atomicBMap Val))
           (flattenImplProg (replaceFunction lmap) predMap))
        topSpec).

(* An implementation of transactional memory. *)
Context (TId : Type).
Context (tm : Object (@TransMethod (BMapMethod Loc Val) TId)).

(* ASSUMPTION: The transactional memory object [tm] is (strictly) serializable
   with respect to a sequential specification for a map [atomicBMap]. *)
Context (tmSer :
  StrictSerializable
    (LocBMap.atomicBMap Val) tm instrument).

Definition abstractLocator : Object LocMethod :=
        (serialObject (functionSpec locatorMap)).

Definition flattenPStructImplProg :
  forall R : Type,
    Program Method R ->
    Program (PairMethod LocMethod (BMapMethod Loc Val)) R :=
  @flattenProgram _ _ predMap.

Definition instrumentPStructImplProg
           {TId} R (m : Abortable (Program Method) R) :
  Program
    (PairMethod LocMethod (@TransMethod (BMapMethod Loc Val) TId))
    R :=
  let 'abortable R' m' := m in
  instrumentRight _ (flattenPStructImplProg R' m') _.

Definition TPImplOfLocator (locatorObject : Object LocMethod) :
  Object (Abortable (Program Method)) :=

  flattenObject
    (pairObject locatorObject tm)
    _
    instrumentPStructImplProg.

Definition TPImpl :
  Object (Abortable (Program Method)) :=
  TPImplOfLocator abstractLocator.

Definition TPSpec :=
  abortableSpec (programSpec topSpec).

Notation pI := pairObject.

Lemma flattenPairFun {M M' N}
      (lmap : forall R, M R -> R)
      (i : Object N) (h : forall R, M' R -> _ R)
  : ConcTraceIncl
      (fI (pI (serialObject (functionSpec lmap)) i) _ h)
      (fI i _ (flattenImplProg (replaceFunction lmap) h)).
Proof.
  assert (H := pairImplFunction (M' := M') (M := M) (N := ObjInstruction i)).
  cbn in H.
  specialize (H lmap (objBase i) (flattenImplProg
                   (objProgram
                      (pI (serialObject (functionSpec lmap)) i)) h)).
  unfold flattenObject. cbn [ObjInstruction objBase serialObject pI].
  eapply Transitive_ConcTraceIncl.
  { apply ConcSimulates_ConcTraceIncl. apply H. }
  apply ConcSimulates_ConcTraceIncl, Proper_ThreadsBehavior.
  econstructor.
  - apply HRSimulates_ieq.
  - cbn [objProgram pairObject serialObject].
    red. intros ? ? _ [].
    unfold flattenImplProg.
    eapply Transitive_eutt_ieq.
    1:{ eapply flatten_flatten. constructor. }
    eapply Transitive_eutt_ieq.
    2:{ apply Symmetric_eutt_ieq.
        eapply flatten_flatten. constructor. }
    apply (Proper_flattenProgram ieq); [ | apply Reflexive_eutt_ieq ].
    red; intros ? [] _ []; cbn.
    + pfold.
      match goal with
      | [ |- _ ?p1 ?p2 ] =>
        rewrite (frobProgramId p1);
        rewrite (frobProgramId p2); cbn
      end.
      constructor.
      rewrite (frobProgramId (bindFlattenProgram _ _ _)). cbn.
      constructor.
      rewrite (frobProgramId (flattenProgram _ _)). cbn.
      constructor.
    + eapply Transitive_eutt_ieq.
      { apply flattenMap. }
      cbn.
      eapply Transitive_eutt_ieq.
      { apply flattenId. }
      eapply Symmetric_eutt_ieq.
      eapply Transitive_eutt_ieq.
      { apply flattenTrigger. }
      apply Reflexive_eutt_ieq.
Qed.

Arguments flattenObject {_} _ {_} _.

(* Rearrange interpreters (interpret with [locatorObject] then with [tm] instead of in parallel) *)
Lemma TPImplLin2
  : ConcTraceIncl
      (TPImplOfLocator (serialObject (functionSpec locatorMap)))
      (flattenObject
         tm
         (flattenImplProg (replaceFunction locatorMap) instrumentPStructImplProg)).
Proof.
  apply flattenPairFun.
Qed.

(* Step 5 of the final paper proof / application of Lemma 7.3
   - the LHS uses [instrumentRight] in [instrumentPStructImplProg]
   - the RHS uses [instrument] *)
Lemma TPImplLin__5
  : ConcTraceIncl
      (flattenObject
         tm
         (flattenImplProg (replaceFunction locatorMap) instrumentPStructImplProg))
      (mapObject
         (mapAbortable (flattenImplProg (replaceFunction locatorMap) flattenPStructImplProg))
         (instFlatImpl tm instrument)).
Proof.
  apply Transitive_ConcTraceIncl with (o2 :=
    (fI tm
      (fun R (m0 : Abortable (Program Method) R) =>
         let m := mapAbortable (flattenImplProg (replaceFunction locatorMap) flattenPStructImplProg) _ m0 in
         let 'abortable _ m' := m in
         instrument _ m'
      ))).
  { unfold instrumentPStructImplProg.
    unfold flattenPStructImplProg.
    apply ConcSimulates_ConcTraceIncl, HConcSimulatesEq.
    apply Proper_flattenObject.
    - apply Reflexive_HImplSimulates.
    - (* These are bisimilar. *)
      red; intros ? [] _ [].
      apply flattenInstrument. }
  apply ConcSimulates_ConcTraceIncl, HConcSimulatesEq.
  econstructor; cbn.
  - apply HRSimulates_ieq.
  - red; intros ? [] _ [].
    unfold mapImplProg, flattenImplProg. cbn.
    apply Reflexive_eutt_ieq.
Qed.

(* Step 7 of the proof (application of Lemma 8.1, serializability of TML) *)
Lemma TPImplLin__7
  : ConcTraceIncl
      (mapObject
         (mapAbortable (flattenImplProg (replaceFunction locatorMap) flattenPStructImplProg))
         (instFlatImpl tm instrument))
      (mapObject
         (mapAbortable (flattenImplProg (replaceFunction locatorMap) flattenPStructImplProg))
         (serialObject (abortableSpec (programSpec (LocBMap.atomicBMap Val))))).
Proof.
  apply ConcTraceIncl_mapObject.
  apply ConcSimulates_ConcTraceIncl, HConcSimulatesConcSimulates.
  apply tmSer.
Qed.

Lemma TPImplLin3_3
  : ConcTraceIncl
      (mapObject
         (mapAbortable (flattenImplProg (replaceFunction locatorMap) flattenPStructImplProg))
         (serialObject (abortableSpec (programSpec (LocBMap.atomicBMap Val)))))
      (serialObject
         (mapSpec
            (abortableSpec (programSpec (LocBMap.atomicBMap Val)))
            (mapAbortable
               (@flattenProgram _ _
                  (flattenImplProg (replaceFunction locatorMap) predMap))))).
Proof.
  eapply Transitive_ConcTraceIncl.
  { apply ConcSimulates_ConcTraceIncl, mapSerial. }
  unfold flattenImplProg.
  unfold flattenPStructImplProg.
  apply ConcSimulates_ConcTraceIncl, HConcSimulatesEq.
  apply Proper_serialObject.
  apply (@Proper_mapSpec (Abortable (Program Method))
                              (Abortable (Program (BMapMethod _ _)))
                              (eq_Abortable (eutt ieq))).
  - intros.
    apply Proper_mapAbortable.
    apply flatten_flatten.
  - apply Proper_abortableSpec.
    apply Proper_programSpec.
    apply (HRSimulates_ieq (M := BMapMethod _ _)).
Qed.

(* Serializability of the transactional component *)
Lemma TPImplLin3
  : ConcTraceIncl
      (flattenObject
         tm
         (flattenImplProg (replaceFunction locatorMap) instrumentPStructImplProg))
      (serialObject
         (mapSpec
            (abortableSpec (programSpec (LocBMap.atomicBMap Val)))
            (mapAbortable
               (@flattenProgram _ _
                  (flattenImplProg (replaceFunction locatorMap) predMap))))).
Proof.
  eapply Transitive_ConcTraceIncl; [ apply TPImplLin__5 | ].
  eapply Transitive_ConcTraceIncl; [ apply TPImplLin__7 | apply TPImplLin3_3 ].
Qed.

(* Correctness of the [predMap] object (Step 11 of the paper proof, application of Lemma 7.2) *)
Lemma TPImplLin__11
  : ConcTraceIncl
      (serialObject
         (mapSpec
            (abortableSpec (programSpec (LocBMap.atomicBMap Val)))
            (mapAbortable
               (@flattenProgram _ _
                  (flattenImplProg (replaceFunction locatorMap) predMap)))))
      (serialObject TPSpec).
Proof.
  apply ConcSimulates_ConcTraceIncl, HConcSimulatesEq.
  apply Proper_serialObject.
  apply simulatesHSimulates.
  unfold TPSpec.
  eapply simulatesTrans; [ apply mapAbortableSimulates | ].
  apply abortableSimulates.
  eapply simulatesTrans; [ apply flattenProgramSimulates | ].
  apply programSpecSimulates.
  apply predMapCorrect.
  apply locatorMapInjective.
Qed.

Lemma TPImplLin
  : ConcTraceIncl
      TPImpl
      (serialObject TPSpec).
Proof.
  eapply Transitive_ConcTraceIncl.
  { apply TPImplLin2. }
  eapply Transitive_ConcTraceIncl.
  { apply TPImplLin3. }
  apply TPImplLin__11.
Qed.

End GenericProof_Sec.
End GenericProof.

(* Trace inclusion *)
Module LazyMapRefine.

Section LazyMapRefine_Sec.

  Context (Method : Type -> Type).
  Context (Val : Type).
  Context (LocMethod : Type -> Type).

  Context (locator : Object LocMethod).

  (* ASSUMPTION: [locatorImplProg] behaves like a pure function [lmap]. *)
  Context (locatorLinearizable :
    forall (t : ConcTrace LocMethod),
      IsConcTraceOf t locator ->
      exists (locatorMap : forall R, LocMethod R -> R),
        injective _ Loc (locatorMap Loc) /\
        IsConcTraceOf t
          (serialObject (functionSpec locatorMap))).

  (* Implementation of the transactional interface [Method]. *)
  Context (predMap :
      forall R,
        Method R ->
        Program (PairMethod LocMethod (BMapMethod Loc Val)) R).
  Context (topSpec : Spec Method).

  (* ASSUMPTION: [predMap] is "correct" wrt. [topSpec] when
     it implements the sequential spec  *)
  Context (predMapCorrect :
      forall lmap : forall Ret, LocMethod Ret -> Ret,
        injective _ Loc (lmap Loc) ->
        Simulates
          (mapSpec
             (programSpec (LocBMap.atomicBMap Val))
             (flattenImplProg (replaceFunction lmap) predMap))
          topSpec).

  (* An implementation of transactional memory. *)
  Context (TId : Type).
  Context (tm : Object (@TransMethod (BMapMethod Loc Val) TId)).
  Context (tmInit: State (objBase tm)).

  (* ASSUMPTION: The transactional memory object [tm] is (strictly) serializable
     with respect to a sequential specification for a map [atomicBMap]. *)
  Context (tmSer :
    StrictSerializable
      (LocBMap.atomicBMap Val) tm instrument).

  (*
   * Given an object that uses both t-maps,
   * we first flatten the objects of the two t-maps and
   * then share the same tm on right.
   *)

  Definition flattenPStructImplProg :
    forall R : Type,
      Program Method R ->
      Program (PairMethod LocMethod (BMapMethod Loc Val)) R :=
    @flattenProgram _ _ predMap.

  Definition instrumentPStructImplProg
             {TId} R (m : Abortable (Program Method) R) :
    Program
      (PairMethod LocMethod (@TransMethod (BMapMethod Loc Val) TId))
      R :=
    let 'abortable R' m' := m in
    instrumentRight _ (flattenPStructImplProg R' m') _.

  Definition TPImplOfLocator (locatorObject : Object LocMethod) :
    Object (Abortable (Program Method)) :=

    flattenObject
      (pairObject locatorObject tm)
      _
      instrumentPStructImplProg.

  Definition TPImpl : Object (Abortable (Program Method)) :=
    TPImplOfLocator locator.

  Definition TPSpec :=
    abortableSpec (programSpec topSpec).

  Definition extends {Key Val} (f : Key -> Val) (g : Key -> option Val) :=
    forall k,
      match g k with
      | Some v => f k = v
      | None => True
      end.

  Notation pI := pairObject.

  Lemma flattenPairFun {M M' N}
        (lmap : forall R, M R -> R)
        (i : Object N) (h : forall R, M' R -> _ R)
    : ConcTraceIncl
        (fI (pI (serialObject (functionSpec lmap)) i) _ h)
        (fI i _ (flattenImplProg (replaceFunction lmap) h)).
  Proof.
    pose proof (pairImplFunction (M' := M') (M := M) (N := ObjInstruction i)).
    cbn in H. unfold flattenObject. cbn [ObjInstruction objBase serialObject pI].
    specialize (H lmap (objBase i) (flattenImplProg
                     (objProgram
                        (pI (serialObject (functionSpec lmap)) i)) h)).
    eapply Transitive_ConcTraceIncl; [ eapply ConcSimulates_ConcTraceIncl, H | ].
    eapply ConcSimulates_ConcTraceIncl.
    apply Proper_ThreadsBehavior.
    econstructor.
    - apply HRSimulates_ieq.
    - cbn [objProgram pairObject serialObject].
      red. intros ? ? _ [].
      unfold flattenImplProg.
      eapply Transitive_eutt_ieq.
      1:{ eapply flatten_flatten. constructor. }
      eapply Transitive_eutt_ieq.
      2:{ apply Symmetric_eutt_ieq.
          eapply flatten_flatten. constructor. }
      apply (Proper_flattenProgram ieq); [ | apply Reflexive_eutt_ieq ].
      red; intros ? [] _ []; cbn.
      + pfold.
        match goal with
        | [ |- _ ?p1 ?p2 ] =>
          rewrite (frobProgramId p1);
          rewrite (frobProgramId p2); cbn
        end.
        constructor.
        rewrite (frobProgramId (bindFlattenProgram _ _ _)). cbn.
        constructor.
        rewrite (frobProgramId (flattenProgram _ _)). cbn.
        constructor.
      + eapply Transitive_eutt_ieq.
        { apply flattenMap. }
        cbn.
        eapply Transitive_eutt_ieq.
        { apply flattenId. }
        eapply Symmetric_eutt_ieq.
        eapply Transitive_eutt_ieq.
        { apply flattenTrigger. }
        apply Reflexive_eutt_ieq.
  Qed.

  Import FlattenSimulates.

  (* Interpret Locator methods (linearizable concurrent map) *)
  Lemma TPImplLin1 :
    forall t,
      IsConcTraceOf t TPImpl ->
      exists (locatorMap : forall R : Type, LocMethod R -> R),
        injective _ Loc (locatorMap Loc) /\
        IsConcTraceOf t
          (TPImplOfLocator
             (serialObject
                (functionSpec locatorMap))).
  Proof.
    unfold TPImpl, TPImplOfLocator.
    intros t (t' & Tt & Tt').
    unfold fI in Tt.
    cbn [objProgram] in Tt.
    apply ConcTrace_flatten_inv in Tt.
    destruct Tt as (t0 & Tt0 & Tt0').
    apply ConcTrace_pairProgram_inv in Tt0'.
    destruct Tt0' as (tLoc & tLoc' & tTM & tTM' & TtLoc & TtTM & PTLoc & PTTM).
    apply @ConcTraceSpec_pairSpec_inv with (1 := PTTM) in Tt'.
    destruct Tt' as [TtLoc' TtTM'].

    assert (W := locatorLinearizable tLoc).
    forward W.
    { exists tLoc'. split; auto. }
    destruct W as (locatorMap & INJECTIVE & TtLoc_).
    exists locatorMap; split; [ apply INJECTIVE | ].

    apply ConcTrace_flattenObject with (1 := Tt0).
    apply @ConcTrace_pairObject with (1 := PTLoc).
    { auto. }
    { exists tTM'; auto. }
  Qed.

  Lemma TPImplLin:
    ConcTraceIncl TPImpl (serialObject TPSpec).
  Proof.
    intros t Tt.
    apply TPImplLin1 in Tt.
    destruct Tt as (locatorMap & locatorMapInjective & Tt).
    apply (GenericProof.TPImplLin Method Val LocMethod locatorMap locatorMapInjective predMap topSpec predMapCorrect TId tm tmSer).
    apply Tt.
  Qed.

End LazyMapRefine_Sec.
End LazyMapRefine.

Module Import Inst.

Module TML := TML.TMLFun LocTyp.
Section Inst_.
Import LazyMapRefine.

Definition Key0 := nat.
Context (Val : Type).

Let Method : Type -> Type := BMapMethod Key0 Val.

Definition LocatorBaseMethod : Type -> Type := MMapM Key0 Loc.
Definition LocMethod : Type -> Type := PureMapM Key0 Loc.
Definition locatorBase : Spec LocatorBaseMethod := mmapSpec (EqDecK := Nat.eq_dec).
Definition locatorImplProg :
    forall Ret,
      LocMethod Ret ->
      Program (PairMethod LocatorBaseMethod (AllocatorMethod Loc)) Ret :=
  fun Ret => lazyMap (R := Ret).

Definition concreteLocator :
  Object LocMethod :=
  object
    (pairSpec locatorBase (atomicAllocator Loc (Full_set _)))
    locatorImplProg.

(* ASSUMPTION: [locatorImplProg]/[concreteLocator] behaves like a pure function [lmap]. *)
Let locatorLinearizable :
  forall (t : ConcTrace LocMethod),
    IsConcTraceOf t concreteLocator ->
    exists (locatorMap : forall R, LocMethod R -> R),
      injective _ Loc (locatorMap Loc) /\
      IsConcTraceOf t
        (serialObject (functionSpec locatorMap)).
Proof.
  intros t H.
  apply (lazyMap_correct (K := Key0) (V := Loc) (f0 := fun x => x)) in H.
  - destruct H as (? & ? & ?).
    eexists; split; eauto using ConcTrace_serialObject.
  - red; auto.
Qed.

(* Implementation of the transactional interface [Method]. *)
Let predMap
  : forall R, Method R ->
      Program (PairMethod LocMethod (BMapMethod Loc Val)) R :=
  AllocatorImpl.predMap.
Let topSpec : Spec Method := AllocatorImpl.topSpec.

(* Lemma 7.1 and 7.2, fused
   - [replaceFunction phi] is [(pure-map phi + id) |> emptyL] in Lemma 7.1.
   - [predMap] is [pred-map]
   - [mapSpec (programSpec spec) impl] is equivalent to [interp-as-spec (impl |> spec)]
   - Thus the LHS is [interp-as-spec (pred-map |> (pure-map phi + id) |> emptyL |> map-spec],
     which matches the RHS of Lemma 7.1
   - [topSpec] (for top-level spec, (essentially) the RHS of the final theorem) is [map-spec] *)
(* ASSUMPTION: [predMap] is "correct" wrt. [topSpec] when
   it implements the sequential spec  *)
Let predMapCorrect
  : forall lmap : forall Ret, LocMethod Ret -> Ret,
      injective _ Loc (lmap Loc) ->
      Simulates
        (mapSpec
           (programSpec (LocBMap.atomicBMap _))
           (flattenImplProg (replaceFunction lmap) predMap))
        topSpec :=
  AllocatorImpl.predMapCorrect.

(* An implementation of transactional memory. *)
Let tm : Object (@TransMethod (BMapMethod Loc Val) _) := TML.tml.

(* Lemma 8.1
   ASSUMPTION: The transactional memory object [tm] is (strictly) serializable
   with respect to a sequential specification for a map [atomicBMap]. *)
Let tmSer : StrictSerializable (LocBMap.atomicBMap Val) tm instrument :=
  TML.TMLSer.TMLStrictSerializable.

Definition flattenPStructImplProg :
  forall R : Type,
    Program Method R ->
    Program (PairMethod LocMethod (BMapMethod Loc Val)) R :=
  @flattenProgram _ _ predMap.

Definition instrumentPStructImplProg
           {TId} R (m : Abortable (Program Method) R) :
  Program
    (PairMethod LocMethod (@TransMethod (BMapMethod Loc Val) TId))
    R :=
  let 'abortable R' m' := m in
  instrumentRight _ (flattenPStructImplProg R' m') _.

Definition TPImplOfLocator (locatorObject : Object LocMethod) :
  Object (Abortable (Program Method)) :=

  flattenObject
    (pairObject locatorObject tm)
    _
    instrumentPStructImplProg.

Definition TPImpl :
  Object (Abortable (Program Method)) :=
  TPImplOfLocator
    concreteLocator.

Definition TPSpec :=
  abortableSpec (programSpec topSpec).

Theorem TPImplLin
  : ConcTraceIncl
      TPImpl
      (serialObject TPSpec).
Proof.
  eapply LazyMapRefine.TPImplLin; eassumption.
Qed.

End Inst_.
End Inst.
Compute (ObjInstruction (TPImpl nat)).
Print Assumptions Inst.TPImplLin.
