From Coq Require Import List Arith Omega.
From Coq.Logic Require Import
  Classical_Prop
  ClassicalChoice
  EqdepFacts
  FunctionalExtensionality.

From C4.Util Require Import LibTactics.

Set Implicit Arguments.

Notation classic := Classical_Prop.classic.
Axiom choice
     : forall (A B : Type) (R : A -> B -> Prop),
       (forall x : A, exists y, R x y) -> exists f, forall x : A, R x (f x).

Inductive mwitness : Type :=
| Witness (P : Prop) (_ : P)
| NoWitness.

Lemma classicT_inhabited : inhabited (forall (P : Prop), { P } + { ~P }).
Proof.
  destruct (choice (fun (P : Prop) (b : mwitness) =>
    match b return Prop with @Witness Q _ => P = Q | NoWitness => P -> False end)) as [f H].
  { intros P; destruct (classic (inhabited P)) as [[x] | ];
      [exists (Witness x) | exists NoWitness]; auto. }
  constructor. intros P; specialize (H P); destruct (f P); [subst | ]; auto.
Qed.

Fixpoint zip {X Y} (lx : list X) (ly : list Y): list (X*Y) :=
  match lx with
  | nil =>
    nil
  | x::tx =>
    match ly with
    | nil => nil
    | y::ty => (x,y) :: (zip tx ty)
    end
  end.

Fixpoint del_nth {A} (n : nat) (xs : list A) : list A :=
  match xs, n with
  | x :: xs, O => xs
  | x :: xs, S n => x :: del_nth n xs
  | nil, _ => nil
  end.

Fixpoint last_error {A} (l:list A) : option A :=
  match l with
  | nil => None
  | a::nil => Some a
  | a :: l => last_error l
  end.

Fixpoint mapfilter {A B} (f : A -> option B) (xs : list A) : list B :=
  match xs with
  | nil => nil
  | x :: xs =>
    match f x with
    | None => mapfilter f xs
    | Some y => y :: mapfilter f xs
    end
  end.

Definition eq_hd {A} (xs ys : list A) : Prop :=
  match xs, ys with
  | x :: _, y :: _ => x = y
  | _, _ => False
  end.

Definition unwrap {A} (a : A) (o : option A) : A :=
  match o with
  | None => a
  | Some a => a
  end.

Definition isSome {A} (o : option A) : Prop :=
  match o with
  | None => False
  | Some _ => True
  end.

Definition option_bind {A B} (oa : option A) (g : A -> option B) : option B :=
  match oa with
  | None => None
  | Some a => g a
  end.

(* [>=>] in Haskell *)
Definition option_fish {A B C} (f : A -> option B) (g : B -> option C) (a : A) : option C :=
  option_bind (f a) g.

Definition HList (Rs : list Type): Type :=
  fold_right prod unit Rs.

Definition hnil: HList nil :=
  tt.

Lemma Distinct_not_In {A} (x : A) xs xs' : NoDup (xs ++ x :: xs') -> In x (xs ++ xs') -> False.
Proof.
  induction xs; cbn; rewrite NoDup_cons_iff; firstorder.
  clear H1 H2; subst; induction xs; cbn in H; firstorder.
Qed.

Lemma Distinct_del {A} (x : A) xs xs' : NoDup (xs ++ x :: xs') -> NoDup (xs ++ xs').
Proof.
  induction xs; cbn; rewrite NoDup_cons_iff; [ firstorder | constructor; firstorder ].
  clear H0 H1; induction xs; cbn in H; firstorder.
Qed.

Lemma not_In_app_inv {A} x (xs xs' : list A) : ~ In x (xs ++ xs') -> ~In x xs /\ ~In x xs'.
Proof.
  induction xs; cbn; firstorder.
Qed.

Theorem list_ne_cons2list A (x y : A) zs : zs <> cons x (cons y zs).
Proof.
  intro H.
  apply (f_equal (@length A)) in H.
  simpl in H.
  omega.
Qed.

Lemma Forall2_impl_ {A B} (R R' : A -> B -> Prop) xs ys
  : (forall x y, In x xs -> R x y -> R' x y) -> Forall2 R xs ys -> Forall2 R' xs ys.
Proof.
  intros H1 H2; induction H2; cbn in H1; auto.
Qed.

Lemma Forall2_length {A B} (R : A -> B -> Prop) xs ys
  : Forall2 R xs ys -> length xs = length ys.
Proof.
  induction 1; cbn; congruence.
Qed.

Lemma nth_In_1 {A} (n : nat) (x x0 : A) (xs : list A)
  : x = nth n xs x0 -> n < length xs -> In x xs.
Proof.
  intros ->; apply nth_In.
Qed.

Lemma nth_not_In {A} (n : nat) (x x0 : A) (xs : list A)
  : ~ In x xs -> n < length xs -> x <> nth n xs x0.
Proof.
  intros X Y Z; apply X; eauto using nth_In_1.
Qed.

Lemma mapfilter_nil_inv {A B} (f : A -> option B) (xs : list A)
  : mapfilter f xs = nil -> List.Forall (fun x => f x = None) xs.
Proof.
  induction xs; cbn.
  - constructor.
  - destruct (f a) as [ | ] eqn:Ef.
    + discriminate.
    + constructor; auto.
Qed.

Lemma mapfilter_mapfilter {A B C} (f : A -> option B) (g : B -> option C) (t : list A)
  : mapfilter g (mapfilter f t) = mapfilter (option_fish f g) t.
Proof.
  induction t; cbn.
  - reflexivity.
  - unfold option_fish, option_bind.
    destruct (f a); cbn; auto.
    destruct (g b); cbn; auto.
    f_equal; auto.
Qed.

Lemma Proper_mapfilter {A B} (f g : A -> option B) (t u : list A)
  : t = u -> (forall a, f a = g a) -> mapfilter f t = mapfilter g u.
Proof.
  intros []; induction t; cbn; intros E.
  - reflexivity.
  - rewrite E. destruct (g a).
    + f_equal; auto.
    + auto.
Qed.

Lemma mapfilter_app {A B} (f : A -> option B) (xs ys : list A)
  : mapfilter f (xs ++ ys) = mapfilter f xs ++ mapfilter f ys.
Proof.
  induction xs; cbn; auto.
  destruct f; cbn; f_equal; auto.
Qed.

Lemma Forall_inv2 A (P : A -> Prop) (x : A) (xs : list A)
  : Forall P (x :: xs) -> P x /\ Forall P xs.
Proof.
  inversion 1; subst; auto.
Qed.

Lemma mapfilter_nil {A B} (f : A -> option B) (xs : list A)
  : List.Forall (fun x => f x = None) xs -> mapfilter f xs = nil.
Proof.
  induction xs; cbn.
  - constructor.
  - intros H. apply Forall_inv2 in H. destruct H as [X Y].
    rewrite X.
    auto.
Qed.

Lemma Forall_map {A B} (f : A -> B) (P : B -> Prop) (xs : list A)
  : Forall (fun x => P (f x)) xs -> Forall P (map f xs).
Proof.
  induction xs; cbn; auto.
  intros H; apply Forall_inv2 in H; destruct H; auto.
Qed.

Lemma Forall2_update {A B} (R R' : A -> B -> Prop) x x' (xs xs' : list A) y y' (ys ys' : list B)
  : (length xs = length ys) ->
    (R x y -> R' x' y') ->
    (forall i y, i <> x -> R i y -> R' i y) ->
    NoDup (xs ++ x :: xs') ->
    Forall2 R (xs ++ x :: xs') (ys ++ y :: ys') ->
    Forall2 R' (xs ++ x' :: xs') (ys ++ y' :: ys').
Proof.
  revert ys; induction xs as [ | x_ xs IH ]; cbn; intros ys Elength RR CONSTANT NODUP FORALL;
    destruct ys as [ | y_ ys ]; inversion Elength; cbn.
  - inversion NODUP; inversion FORALL; subst; constructor; auto.
    revert H8; apply Forall2_impl_.
    intros x0 y0 IN; apply CONSTANT.
    { intros ->. contradiction. }
  - inversion NODUP. inversion FORALL; subst; constructor; auto.
    revert H7; apply CONSTANT.
    { intros ->. apply H2. apply in_or_app; right; cbn; auto. }
Qed.

Lemma Forall2_del {A B} (R R' : A -> B -> Prop) xs x0 xs' (ys : list B)
  : (forall x y, In x (xs ++ xs') -> R x y -> R' x y) ->
    Forall2 R (xs ++ x0 :: xs') ys -> Forall2 R' (xs ++ xs') (del_nth (length xs) ys).
Proof.
  revert ys; induction xs; intros ys H1 H2; inversion H2; cbn; auto.
  - eapply Forall2_impl_; eauto.
  - constructor.
    + eapply H1; cbn; eauto.
    + eapply IHxs; eauto.
      intros. eapply H1; cbn; eauto.
Qed.

Class EqDec (A : Type) : Type :=
  eq_dec : forall x y : A, { x = y } + { x <> y }.

Arguments eq_dec : clear implicits.
Arguments eq_dec {A _}.

Instance EqDec_nat : EqDec nat :=
  Nat.eq_dec.

Instance EqDec_pair {A B} `{EqDec A} `{EqDec B} : EqDec (A * B).
Proof.
  red. intros p1 p2.
  destruct (eq_dec (fst p1) (fst p2)); [ | right; congruence ].
  destruct (eq_dec (snd p1) (snd p2)); [ | right; congruence ].
  left; destruct p1, p2; f_equal; auto.
Defined.

Section SET.
Context {A B : Type} `{EQDEC : EqDec A}.

Definition set (a0 : A) (b : B) (f : A -> B) : A -> B :=
  fun a => if eq_dec a0 a then b else f a.

Lemma set_eq {x : B} {f : A -> B} {n} : set n x f n = x.
Proof.
  unfold set. destruct eq_dec; auto; contradiction.
Qed.

Lemma set_neq {x : B} {f : A -> B} {n m}
  : m <> n -> set n x f m = f m.
Proof.
  intros H. unfold set. destruct eq_dec; auto. symmetry in e; contradiction.
Qed.

Lemma set_neq_flip {x : B} {f : A -> B} {n m}
  : m <> n -> f m = set n x f m.
Proof.
  exact (fun H => eq_sym (set_neq H)).
Qed.

End SET.

Lemma set_or_not n m
  : { n =  m /\ forall X (x : X) (f : nat -> X), set n x f m = x }
  + { n <> m /\ forall X (x : X) (f : nat -> X), set n x f m = f m }.
Proof.
  unfold set. destruct eq_dec; [left | right]; split; auto.
Qed.

Ltac forward H :=
  match type of H with
  | (?Y -> _) =>
    let tmp := fresh "_tmp_" in
    assert (tmp : Y); [ | specialize (H tmp); clear tmp ]
  end.

Ltac split_conj :=
  lazymatch goal with
  | [ |- _ /\ _ ] => split; try split_conj
  end.

Ltac my_decompose H :=
  try lazymatch type of H with
  | False => destruct H
  | (_ /\ _) =>
    let H' := fresh H in
    destruct H as [H' H]; my_decompose H'; my_decompose H
  | (exists x, _) =>
    let x := fresh x in
    destruct H as [x H]; my_decompose H
  | (_ \/ _) =>
    destruct H as [H | H]; my_decompose H
  | (_ = ?i) => subst i
  | (?i <> _) =>
    let NEi := fresh "NE" i in
    rename H into NEi
  | _ => idtac
  end.

(** Indexed equality *)

Notation IRel M N := (forall X, M X -> N X -> Prop) (only parsing).

Definition ieq {M : Type -> Type} R (e1 e2 : M R) : Prop := e1 = e2.
Hint Unfold ieq : core.

Definition irespectful {M1 M2 N1 N2 : Type -> Type}
           (RM : IRel M1 M2)
           (RN : IRel N1 N2)
           (f1 : forall T, M1 T -> N1 T)
           (f2 : forall T, M2 T -> N2 T)
  : Prop :=
  forall T e1 e2, RM T e1 e2 -> RN T (f1 T e1) (f2 T e2).

Delimit Scope isig_scope with isig.
Infix "==>" := irespectful (right associativity, at level 55) : isig_scope.

Definition irel_fun {M N} (f : forall R : Type, M R -> N R) : IRel N M :=
  fun _ e1 e2 => e1 = f _ e2.
